/* Generated by Cython 3.0a5 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            ".",
            "/tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/core/include"
        ],
        "name": "cherab.solps.solps_2d_functions",
        "sources": [
            "cherab/solps/solps_2d_functions.pyx"
        ]
    },
    "module_name": "cherab.solps.solps_2d_functions"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#define CYTHON_ABI "3_0a5"
#define CYTHON_HEX_VERSION 0x030000A5
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP487_INIT_SUBCLASS
  #define CYTHON_PEP487_INIT_SUBCLASS 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(CYTHON_LIMITED_API)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030800B2
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#elif PY_VERSION_HEX >= 0x030800A4
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS(n)
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((n) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET)
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  (n)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static inline void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #if defined(_PyType_Name)
    #define __Pyx_PyType_Name(tp)     (_PyType_Name((PyTypeObject *)tp))
  #else
    #define __Pyx_PyType_Name(tp)     (((PyTypeObject *)tp)->tp_name)
  #endif
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
#else
  #define __Pyx_PyType_Name(tp)       (((PyTypeObject *)tp)->tp_name)
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__cherab__solps__solps_2d_functions
#define __PYX_HAVE_API__cherab__solps__solps_2d_functions
/* Early includes */
#include <string.h>
#include <stdio.h>

    /* Using NumPy API declarations from "numpy/__init__.cython-30.pxd" */
    
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"
#include <math.h>
#include <stdint.h>
#include "pythread.h"
#include <stdlib.h>
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const wchar_t *u)
{
    const wchar_t *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "cherab/solps/solps_2d_functions.pyx",
  "__init__.cython-30.pxd",
  "affinematrix.pxd",
  "vector.pxd",
  "point.pxd",
  "normal.pxd",
  "quaternion.pxd",
  "ray.pxd",
  "intersection.pxd",
  "boundingbox.pxd",
  "base.pxd",
  "base.pxd",
  "base.pxd",
  "base.pxd",
  "base.pxd",
  "base.pxd",
  "stringsource",
  "type.pxd",
  "_vec3.pxd",
  "_mat4.pxd",
  "statsarray.pxd",
  "solidangle.pxd",
  "surface3d.pxd",
  "targetted.pxd",
  "material.pxd",
  "_nodebase.pxd",
  "node.pxd",
  "signal.pxd",
  "primitive.pxd",
  "accelerator.pxd",
  "world.pxd",
  "observer.pxd",
  "utility.pxd",
  "containers.pxd",
  "boundingsphere.pxd",
  "kdtree2d.pxd",
  "common.pxd",
  "base.pxd",
  "base.pxd",
  "constant.pxd",
  "blend.pxd",
  "autowrap.pxd",
  "arg.pxd",
  "cmath.pxd",
  "interpolate.pxd",
  "constant.pxd",
  "blend.pxd",
  "autowrap.pxd",
  "interpolator2dmesh.pxd",
  "discrete2dmesh.pxd",
  "interpolator2darray.pxd",
  "arg.pxd",
  "cmath.pxd",
  "constant.pxd",
  "blend.pxd",
  "autowrap.pxd",
  "kdtree3d.pxd",
  "common.pxd",
  "discrete3dmesh.pxd",
  "interpolator3darray.pxd",
  "arg.pxd",
  "cmath.pxd",
  "base.pxd",
  "constant.pxd",
  "blend.pxd",
  "autowrap.pxd",
  "utility.pxd",
  "constant.pxd",
  "blend.pxd",
  "autowrap.pxd",
  "utility.pxd",
  "constant.pxd",
  "blend.pxd",
  "autowrap.pxd",
  "utility.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && __GNUC__ >= 4 && (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL >= 2)) &&\
                    !defined(__i386__)
    #define __pyx_atomic_incr_aligned(value, lock) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value, lock) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && 0
    #include <Windows.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type LONG
    #define __pyx_atomic_incr_aligned(value, lock) InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#elif CYTHON_ATOMICS && (defined(__ICC) || defined(__INTEL_COMPILER)) && 0
    #define __pyx_atomic_incr_aligned(value, lock) _InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) _InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using Intel atomics"
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* #### Code section: numeric_typedefs ### */

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":732
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":733
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":734
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":735
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":739
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":740
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":741
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":742
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":746
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":747
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":756
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":757
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":758
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":760
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":761
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":762
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":764
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":765
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":767
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":768
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":769
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* #### Code section: complex_type_declarations ### */
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3;
struct __pyx_obj_7raysect_4core_4math_5_mat4__Mat4;
struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D;
struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D;
struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D;
struct __pyx_obj_7raysect_4core_4math_5point_Point3D;
struct __pyx_obj_7raysect_4core_4math_5point_Point2D;
struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D;
struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion;
struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsBin;
struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D;
struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D;
struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D;
struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler;
struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SphereSampler;
struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler;
struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler;
struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler;
struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D;
struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D;
struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D;
struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D;
struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler;
struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler;
struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler;
struct __pyx_obj_7raysect_4core_3ray_Ray;
struct __pyx_obj_7raysect_4core_8material_Material;
struct __pyx_obj_7raysect_4core_10scenegraph_9_nodebase__NodeBase;
struct __pyx_obj_7raysect_4core_10scenegraph_4node_Node;
struct __pyx_obj_7raysect_4core_10scenegraph_6signal_ChangeSignal;
struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive;
struct __pyx_obj_7raysect_4core_12intersection_Intersection;
struct __pyx_obj_7raysect_4core_12acceleration_11accelerator_Accelerator;
struct __pyx_obj_7raysect_4core_10scenegraph_5world_World;
struct __pyx_obj_7raysect_4core_10scenegraph_8observer_Observer;
struct __pyx_obj_7raysect_4core_10scenegraph_7utility_BridgeNode;
struct __pyx_obj_7raysect_4core_10containers__Item;
struct __pyx_obj_7raysect_4core_10containers_LinkedList;
struct __pyx_obj_7raysect_4core_10containers_Stack;
struct __pyx_obj_7raysect_4core_10containers_Queue;
struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D;
struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D;
struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D;
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_Item2D;
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore;
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D;
struct __pyx_obj_7raysect_4core_4math_8function_4base_Function;
struct __pyx_obj_7raysect_4core_4math_8function_5float_4base_FloatFunction;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D;
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_Item3D;
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore;
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D;
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D;
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D;
struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D;
struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":771
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":772
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":773
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":775
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_7raysect_4core_4math_5_mat4_5_Mat4_is_identity;
struct __pyx_opt_args_7raysect_4core_4math_5_mat4_5_Mat4_is_close;

/* "raysect/core/math/_mat4.pxd":40
 *     cdef void set_element(self, int row, int column, double v)
 * 
 *     cpdef bint is_identity(self, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_close(self, _Mat4 other, double tolerance=*)
 */
struct __pyx_opt_args_7raysect_4core_4math_5_mat4_5_Mat4_is_identity {
  int __pyx_n;
  double tolerance;
};

/* "raysect/core/math/_mat4.pxd":42
 *     cpdef bint is_identity(self, double tolerance=*)
 * 
 *     cpdef bint is_close(self, _Mat4 other, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 */
struct __pyx_opt_args_7raysect_4core_4math_5_mat4_5_Mat4_is_close {
  int __pyx_n;
  double tolerance;
};
struct __pyx_opt_args_7raysect_4core_4math_10quaternion_10Quaternion_is_unit;

/* "raysect/core/math/quaternion.pxd":48
 *     cpdef Quaternion normalise(self)
 * 
 *     cpdef bint is_unit(self, double tolerance=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef Quaternion transform(self, AffineMatrix3D m)
 */
struct __pyx_opt_args_7raysect_4core_4math_10quaternion_10Quaternion_is_unit {
  int __pyx_n;
  double tolerance;
};
struct __pyx_opt_args_7raysect_4core_4math_9transform_extract_rotation;

/* "raysect/core/math/transform.pxd":55
 * cpdef Point3D from_cylindrical(double r, double z, double phi)
 * 
 * cpdef tuple extract_rotation(AffineMatrix3D m, str z_axis=*)             # <<<<<<<<<<<<<<
 * 
 * cpdef tuple extract_translation(AffineMatrix3D m)
 */
struct __pyx_opt_args_7raysect_4core_4math_9transform_extract_rotation {
  int __pyx_n;
  PyObject *z_axis;
};
struct __pyx_opt_args_7raysect_4core_4math_6random_seed;

/* "raysect/core/math/random.pxd":35
 * from raysect.core.math.point cimport Point2D, Point3D
 * 
 * cpdef seed(object d=*)             # <<<<<<<<<<<<<<
 * 
 * cpdef double uniform()
 */
struct __pyx_opt_args_7raysect_4core_4math_6random_seed {
  int __pyx_n;
  PyObject *d;
};
struct __pyx_opt_args_7raysect_4core_3ray_3Ray_copy;

/* "raysect/core/ray.pxd":43
 *     cpdef Point3D point_on(self, double t)
 * 
 *     cpdef Ray copy(self, Point3D origin=*, Vector3D direction=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_7raysect_4core_3ray_3Ray_copy {
  int __pyx_n;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *origin;
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *direction;
};
struct __pyx_opt_args_7raysect_4core_10scenegraph_9primitive_9Primitive_instance;

/* "raysect/core/scenegraph/primitive.pxd":58
 *     cpdef BoundingSphere3D bounding_sphere(self)
 * 
 *     cpdef object instance(self, object parent=*, AffineMatrix3D transform=*, Material material=*, str name=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object notify_geometry_change(self)
 */
struct __pyx_opt_args_7raysect_4core_10scenegraph_9primitive_9Primitive_instance {
  int __pyx_n;
  PyObject *parent;
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *transform;
  struct __pyx_obj_7raysect_4core_8material_Material *material;
  PyObject *name;
};
struct __pyx_opt_args_7raysect_4core_10scenegraph_5world_5World_build_accelerator;

/* "raysect/core/scenegraph/world.pxd":51
 *     cpdef list contains(self, Point3D point)
 * 
 *     cpdef build_accelerator(self, bint force=*)             # <<<<<<<<<<<<<<
 * 
 */
struct __pyx_opt_args_7raysect_4core_10scenegraph_5world_5World_build_accelerator {
  int __pyx_n;
  int force;
};
struct __pyx_opt_args_7raysect_4core_14boundingsphere_16BoundingSphere3D_extend;

/* "raysect/core/boundingsphere.pxd":51
 *     cpdef object union(self, BoundingSphere3D sphere)
 * 
 *     cpdef object extend(self, Point3D point, double padding=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double surface_area(self)
 */
struct __pyx_opt_args_7raysect_4core_14boundingsphere_16BoundingSphere3D_extend {
  int __pyx_n;
  double padding;
};
struct __pyx_opt_args_7raysect_4core_11boundingbox_13BoundingBox3D_extend;
struct __pyx_opt_args_7raysect_4core_11boundingbox_13BoundingBox2D_extend;

/* "raysect/core/boundingbox.pxd":56
 *     cpdef object union(self, BoundingBox3D box)
 * 
 *     cpdef object extend(self, Point3D point, double padding=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double surface_area(self)
 */
struct __pyx_opt_args_7raysect_4core_11boundingbox_13BoundingBox3D_extend {
  int __pyx_n;
  double padding;
};

/* "raysect/core/boundingbox.pxd":101
 *     cpdef object union(self, BoundingBox2D box)
 * 
 *     cpdef object extend(self, Point2D point, double padding=*)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double surface_area(self)
 */
struct __pyx_opt_args_7raysect_4core_11boundingbox_13BoundingBox2D_extend {
  int __pyx_n;
  double padding;
};
struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree2d_kdnode;
struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree2d_edge;
struct __pyx_opt_args_7raysect_4core_4math_7spatial_8kdtree2d_12KDTree2DCore__build;

/* "raysect/core/math/spatial/kdtree2d.pxd":38
 * 
 * # c-structure that represent a kd-tree node
 * cdef struct kdnode:             # <<<<<<<<<<<<<<
 * 
 *     int32_t type        # LEAF, X_AXIS, Y_AXIS
 */
struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree2d_kdnode {
  int32_t type;
  double split;
  int32_t count;
  int32_t *items;
};

/* "raysect/core/math/spatial/kdtree2d.pxd":46
 * 
 * 
 * cdef struct edge:             # <<<<<<<<<<<<<<
 * 
 *     bint is_upper_edge
 */
struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree2d_edge {
  int is_upper_edge;
  double value;
};

/* "raysect/core/math/spatial/kdtree2d.pxd":71
 *         double _empty_bonus
 * 
 *     cdef int32_t _build(self, list items, BoundingBox2D bounds, int32_t depth=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef tuple _split(self, list items, BoundingBox2D bounds)
 */
struct __pyx_opt_args_7raysect_4core_4math_7spatial_8kdtree2d_12KDTree2DCore__build {
  int __pyx_n;
  int32_t depth;
};

/* "raysect/core/math/function/float/function2d/arg.pxd":34
 * from raysect.core.math.function.float.function2d.base cimport Function2D
 * 
 * cdef enum ArgLabel:             # <<<<<<<<<<<<<<
 *     X, Y
 * 
 */
enum __pyx_t_7raysect_4core_4math_8function_5float_10function2d_3arg_ArgLabel {
  __pyx_e_7raysect_4core_4math_8function_5float_10function2d_3arg_X,
  __pyx_e_7raysect_4core_4math_8function_5float_10function2d_3arg_Y
};
struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree3d_kdnode;
struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree3d_edge;
struct __pyx_opt_args_7raysect_4core_4math_7spatial_8kdtree3d_12KDTree3DCore__build;

/* "raysect/core/math/spatial/kdtree3d.pxd":38
 * 
 * # c-structure that represent a kd-tree node
 * cdef struct kdnode:             # <<<<<<<<<<<<<<
 * 
 *     int32_t type        # LEAF, X_AXIS, Y_AXIS, Z_AXIS
 */
struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree3d_kdnode {
  int32_t type;
  double split;
  int32_t count;
  int32_t *items;
};

/* "raysect/core/math/spatial/kdtree3d.pxd":46
 * 
 * 
 * cdef struct edge:             # <<<<<<<<<<<<<<
 * 
 *     bint is_upper_edge
 */
struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree3d_edge {
  int is_upper_edge;
  double value;
};

/* "raysect/core/math/spatial/kdtree3d.pxd":71
 *         double _empty_bonus
 * 
 *     cdef int32_t _build(self, list items, BoundingBox3D bounds, int32_t depth=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef tuple _split(self, list items, BoundingBox3D bounds)
 */
struct __pyx_opt_args_7raysect_4core_4math_7spatial_8kdtree3d_12KDTree3DCore__build {
  int __pyx_n;
  int32_t depth;
};

/* "raysect/core/math/function/float/function3d/arg.pxd":34
 * from raysect.core.math.function.float.function3d.base cimport Function3D
 * 
 * cdef enum ArgLabel:             # <<<<<<<<<<<<<<
 *     X, Y, Z
 * 
 */
enum __pyx_t_7raysect_4core_4math_8function_5float_10function3d_3arg_ArgLabel {
  __pyx_e_7raysect_4core_4math_8function_5float_10function3d_3arg_X,
  __pyx_e_7raysect_4core_4math_8function_5float_10function3d_3arg_Y,
  __pyx_e_7raysect_4core_4math_8function_5float_10function3d_3arg_Z
};

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/cherab/core/math/function/__init__.pxd":34
 * 
 * # Alias deprecated names for FloatToVector3DFunctionxD
 * ctypedef FloatToVector3DFunction2D ScalarToVectorFunction2D             # <<<<<<<<<<<<<<
 * ctypedef FloatToVector3DFunction3D ScalarToVectorFunction3D
 */
typedef struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D *__pyx_t_6cherab_4core_4math_8function_ScalarToVectorFunction2D;

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/cherab/core/math/function/__init__.pxd":35
 * # Alias deprecated names for FloatToVector3DFunctionxD
 * ctypedef FloatToVector3DFunction2D ScalarToVectorFunction2D
 * ctypedef FloatToVector3DFunction3D ScalarToVectorFunction3D             # <<<<<<<<<<<<<<
 */
typedef struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D *__pyx_t_6cherab_4core_4math_8function_ScalarToVectorFunction3D;

/* "raysect/core/math/_vec3.pxd":32
 * # POSSIBILITY OF SUCH DAMAGE.
 * 
 * cdef class _Vec3:             # <<<<<<<<<<<<<<
 * 
 *     cdef public double x, y, z
 */
struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_5_vec3__Vec3 *__pyx_vtab;
  double x;
  double y;
  double z;
};


/* "raysect/core/math/_mat4.pxd":32
 * # POSSIBILITY OF SUCH DAMAGE.
 * 
 * cdef class _Mat4:             # <<<<<<<<<<<<<<
 * 
 *     cdef double m[4][4]
 */
struct __pyx_obj_7raysect_4core_4math_5_mat4__Mat4 {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_5_mat4__Mat4 *__pyx_vtab;
  double m[4][4];
};


/* "raysect/core/math/affinematrix.pxd":35
 * 
 * 
 * cdef class AffineMatrix3D(_Mat4):             # <<<<<<<<<<<<<<
 * 
 *     cpdef AffineMatrix3D inverse(self)
 */
struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D {
  struct __pyx_obj_7raysect_4core_4math_5_mat4__Mat4 __pyx_base;
};


/* "raysect/core/math/vector.pxd":35
 * from raysect.core.math.affinematrix cimport AffineMatrix3D
 * 
 * cdef class Vector3D(_Vec3):             # <<<<<<<<<<<<<<
 * 
 *     cpdef Vector3D cross(self, _Vec3 v)
 */
struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D {
  struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 __pyx_base;
};


/* "raysect/core/math/vector.pxd":78
 * 
 * 
 * cdef class Vector2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef public double x, y
 */
struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_6vector_Vector2D *__pyx_vtab;
  double x;
  double y;
};


/* "raysect/core/math/point.pxd":36
 * from raysect.core.math.affinematrix cimport AffineMatrix3D
 * 
 * cdef class Point3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef public double x, y, z
 */
struct __pyx_obj_7raysect_4core_4math_5point_Point3D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_5point_Point3D *__pyx_vtab;
  double x;
  double y;
  double z;
};


/* "raysect/core/math/point.pxd":73
 * 
 * 
 * cdef class Point2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef public double x, y
 */
struct __pyx_obj_7raysect_4core_4math_5point_Point2D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_5point_Point2D *__pyx_vtab;
  double x;
  double y;
};


/* "raysect/core/math/normal.pxd":36
 * from raysect.core.math.affinematrix cimport AffineMatrix3D
 * 
 * cdef class Normal3D(_Vec3):             # <<<<<<<<<<<<<<
 * 
 *     cpdef Vector3D cross(self, _Vec3 v)
 */
struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D {
  struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 __pyx_base;
};


/* "raysect/core/math/quaternion.pxd":36
 * 
 * 
 * cdef class Quaternion:             # <<<<<<<<<<<<<<
 * 
 *     cdef public double x, y, z, s
 */
struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_10quaternion_Quaternion *__pyx_vtab;
  double x;
  double y;
  double z;
  double s;
};


/* "raysect/core/math/statsarray.pxd":35
 * 
 * 
 * cdef class StatsBin:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsBin {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsBin *__pyx_vtab;
  double mean;
  double variance;
  int samples;
};


/* "raysect/core/math/statsarray.pxd":53
 * 
 * 
 * cdef class StatsArray1D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray1D *__pyx_vtab;
  int length;
  PyArrayObject *mean;
  PyArrayObject *variance;
  PyArrayObject *samples;
  __Pyx_memviewslice mean_mv;
  __Pyx_memviewslice variance_mv;
  __Pyx_memviewslice samples_mv;
};


/* "raysect/core/math/statsarray.pxd":81
 * 
 * 
 * cdef class StatsArray2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray2D *__pyx_vtab;
  int nx;
  int ny;
  PyArrayObject *mean;
  PyArrayObject *variance;
  PyArrayObject *samples;
  __Pyx_memviewslice mean_mv;
  __Pyx_memviewslice variance_mv;
  __Pyx_memviewslice samples_mv;
};


/* "raysect/core/math/statsarray.pxd":109
 * 
 * 
 * cdef class StatsArray3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray3D *__pyx_vtab;
  int nx;
  int ny;
  int nz;
  PyArrayObject *mean;
  PyArrayObject *variance;
  PyArrayObject *samples;
  __Pyx_memviewslice mean_mv;
  __Pyx_memviewslice variance_mv;
  __Pyx_memviewslice samples_mv;
};


/* "raysect/core/math/sampler/solidangle.pxd":41
 * 
 * 
 * cdef class SolidAngleSampler:             # <<<<<<<<<<<<<<
 * 
 *     cpdef double pdf(self, Vector3D sample)
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler *__pyx_vtab;
};


/* "raysect/core/math/sampler/solidangle.pxd":54
 * 
 * 
 * cdef class SphereSampler(SolidAngleSampler):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SphereSampler {
  struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler __pyx_base;
};


/* "raysect/core/math/sampler/solidangle.pxd":58
 * 
 * 
 * cdef class HemisphereUniformSampler(SolidAngleSampler):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler {
  struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler __pyx_base;
};


/* "raysect/core/math/sampler/solidangle.pxd":62
 * 
 * 
 * cdef class HemisphereCosineSampler(SolidAngleSampler):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler {
  struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler __pyx_base;
};


/* "raysect/core/math/sampler/solidangle.pxd":66
 * 
 * 
 * cdef class ConeUniformSampler(SolidAngleSampler):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler {
  struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler __pyx_base;
  double angle;
  double _angle_radians;
  double _angle_cosine;
  double _solid_angle;
  double _solid_angle_inv;
};


/* "raysect/core/math/sampler/surface3d.pxd":35
 * 
 * 
 * cdef class SurfaceSampler3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef Point3D sample(self)
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D *__pyx_vtab;
};


/* "raysect/core/math/sampler/surface3d.pxd":46
 * 
 * 
 * cdef class DiskSampler3D(SurfaceSampler3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D {
  struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D __pyx_base;
  double radius;
  double area;
  double _area_inv;
};


/* "raysect/core/math/sampler/surface3d.pxd":53
 * 
 * 
 * cdef class RectangleSampler3D(SurfaceSampler3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D {
  struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D __pyx_base;
  double width;
  double height;
  double area;
  double _area_inv;
  double _width_offset;
  double _height_offset;
};


/* "raysect/core/math/sampler/surface3d.pxd":60
 * 
 * 
 * cdef class TriangleSampler3D(SurfaceSampler3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D {
  struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *_v1;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *_v2;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *_v3;
  double area;
  double _area_inv;
};


/* "raysect/core/math/sampler/targetted.pxd":35
 * from raysect.core.math cimport Point3D, Vector3D
 * 
 * cdef class _TargettedSampler:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted__TargettedSampler *__pyx_vtab;
  double _total_weight;
  PyObject *_targets;
  PyArrayObject *_cdf;
  __Pyx_memviewslice _cdf_mv;
};


/* "raysect/core/math/sampler/targetted.pxd":60
 * 
 * 
 * cdef class TargettedHemisphereSampler(_TargettedSampler):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler {
  struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler __pyx_base;
};


/* "raysect/core/math/sampler/targetted.pxd":63
 *     pass
 * 
 * cdef class TargettedSphereSampler(_TargettedSampler):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler {
  struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler __pyx_base;
};


/* "raysect/core/ray.pxd":35
 * 
 * 
 * cdef class Ray:             # <<<<<<<<<<<<<<
 * 
 *     cdef public Point3D origin
 */
struct __pyx_obj_7raysect_4core_3ray_Ray {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_3ray_Ray *__pyx_vtab;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *origin;
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *direction;
  double max_distance;
};


/* "raysect/core/material.pxd":32
 * # POSSIBILITY OF SUCH DAMAGE.
 * 
 * cdef class Material:             # <<<<<<<<<<<<<<
 * 
 *     cdef public list primitives
 */
struct __pyx_obj_7raysect_4core_8material_Material {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_8material_Material *__pyx_vtab;
  PyObject *primitives;
};


/* "raysect/core/scenegraph/_nodebase.pxd":34
 * from raysect.core.math.affinematrix cimport AffineMatrix3D
 * 
 * cdef class _NodeBase:             # <<<<<<<<<<<<<<
 * 
 *     cdef str _name
 */
struct __pyx_obj_7raysect_4core_10scenegraph_9_nodebase__NodeBase {
  PyObject_HEAD
  PyObject *_name;
  PyObject *_parent;
  PyObject *children;
  struct __pyx_obj_7raysect_4core_10scenegraph_9_nodebase__NodeBase *root;
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *_transform;
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *_root_transform;
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *_root_transform_inverse;
  int _track_modifications;
  PyObject *meta;
};


/* "raysect/core/scenegraph/node.pxd":35
 * from raysect.core.math.affinematrix cimport AffineMatrix3D
 * 
 * cdef class Node(_NodeBase):             # <<<<<<<<<<<<<<
 * 
 *     cpdef AffineMatrix3D to(self, _NodeBase node)
 */
struct __pyx_obj_7raysect_4core_10scenegraph_4node_Node {
  struct __pyx_obj_7raysect_4core_10scenegraph_9_nodebase__NodeBase __pyx_base;
  struct __pyx_vtabstruct_7raysect_4core_10scenegraph_4node_Node *__pyx_vtab;
};


/* "raysect/core/scenegraph/signal.pxd":33
 * 
 * 
 * cdef class ChangeSignal:             # <<<<<<<<<<<<<<
 * 
 *     cdef str _name
 */
struct __pyx_obj_7raysect_4core_10scenegraph_6signal_ChangeSignal {
  PyObject_HEAD
  PyObject *_name;
};


/* "raysect/core/scenegraph/primitive.pxd":42
 * 
 * 
 * cdef class Primitive(Node):             # <<<<<<<<<<<<<<
 * 
 *     cdef Material _material
 */
struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive {
  struct __pyx_obj_7raysect_4core_10scenegraph_4node_Node __pyx_base;
  struct __pyx_obj_7raysect_4core_8material_Material *_material;
};


/* "raysect/core/intersection.pxd":37
 * 
 * 
 * cdef class Intersection:             # <<<<<<<<<<<<<<
 * 
 *     cdef public Ray ray
 */
struct __pyx_obj_7raysect_4core_12intersection_Intersection {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_12intersection_Intersection *__pyx_vtab;
  struct __pyx_obj_7raysect_4core_3ray_Ray *ray;
  double ray_distance;
  int exiting;
  struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *primitive;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *hit_point;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *inside_point;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *outside_point;
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *normal;
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *world_to_primitive;
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *primitive_to_world;
};


/* "raysect/core/acceleration/accelerator.pxd":36
 * from raysect.core.intersection cimport Intersection
 * 
 * cdef class Accelerator:             # <<<<<<<<<<<<<<
 * 
 *     cpdef build(self, list primitives)
 */
struct __pyx_obj_7raysect_4core_12acceleration_11accelerator_Accelerator {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_12acceleration_11accelerator_Accelerator *__pyx_vtab;
};


/* "raysect/core/scenegraph/world.pxd":38
 * from raysect.core.scenegraph._nodebase cimport _NodeBase
 * 
 * cdef class World(_NodeBase):             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _rebuild_accelerator
 */
struct __pyx_obj_7raysect_4core_10scenegraph_5world_World {
  struct __pyx_obj_7raysect_4core_10scenegraph_9_nodebase__NodeBase __pyx_base;
  struct __pyx_vtabstruct_7raysect_4core_10scenegraph_5world_World *__pyx_vtab;
  int _rebuild_accelerator;
  struct __pyx_obj_7raysect_4core_12acceleration_11accelerator_Accelerator *_accelerator;
  PyObject *_primitives;
  PyObject *_observers;
};


/* "raysect/core/scenegraph/observer.pxd":34
 * from raysect.core.scenegraph.node cimport Node
 * 
 * cdef class Observer(Node):             # <<<<<<<<<<<<<<
 * 
 *     cpdef observe(self)
 */
struct __pyx_obj_7raysect_4core_10scenegraph_8observer_Observer {
  struct __pyx_obj_7raysect_4core_10scenegraph_4node_Node __pyx_base;
};


/* "raysect/core/scenegraph/utility.pxd":36
 * 
 * 
 * cdef class BridgeNode(Node):             # <<<<<<<<<<<<<<
 * 
 *     cdef _NodeBase destination
 */
struct __pyx_obj_7raysect_4core_10scenegraph_7utility_BridgeNode {
  struct __pyx_obj_7raysect_4core_10scenegraph_4node_Node __pyx_base;
  struct __pyx_obj_7raysect_4core_10scenegraph_9_nodebase__NodeBase *destination;
};


/* "raysect/core/containers.pxd":33
 * 
 * 
 * cdef class _Item:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_10containers__Item {
  PyObject_HEAD
  struct __pyx_obj_7raysect_4core_10containers__Item *previous;
  struct __pyx_obj_7raysect_4core_10containers__Item *next;
  PyObject *value;
};


/* "raysect/core/containers.pxd":41
 * 
 * 
 * cdef class LinkedList:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_10containers_LinkedList {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_10containers_LinkedList *__pyx_vtab;
  int length;
  struct __pyx_obj_7raysect_4core_10containers__Item *first;
  struct __pyx_obj_7raysect_4core_10containers__Item *last;
};


/* "raysect/core/containers.pxd":61
 * 
 * 
 * cdef class Stack(LinkedList):             # <<<<<<<<<<<<<<
 * 
 *     cpdef push(self, object value)
 */
struct __pyx_obj_7raysect_4core_10containers_Stack {
  struct __pyx_obj_7raysect_4core_10containers_LinkedList __pyx_base;
};


/* "raysect/core/containers.pxd":68
 * 
 * 
 * cdef class Queue(LinkedList):             # <<<<<<<<<<<<<<
 * 
 *     cpdef object next_in_queue(self)
 */
struct __pyx_obj_7raysect_4core_10containers_Queue {
  struct __pyx_obj_7raysect_4core_10containers_LinkedList __pyx_base;
};


/* "raysect/core/boundingsphere.pxd":35
 * 
 * 
 * cdef class BoundingSphere3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_14boundingsphere_BoundingSphere3D *__pyx_vtab;
  double radius;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *centre;
};


/* "raysect/core/boundingbox.pxd":37
 * 
 * 
 * cdef class BoundingBox3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef Point3D lower
 */
struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_11boundingbox_BoundingBox3D *__pyx_vtab;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *lower;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *upper;
};


/* "raysect/core/boundingbox.pxd":92
 * 
 * 
 * cdef class BoundingBox2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef Point2D lower
 */
struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_11boundingbox_BoundingBox2D *__pyx_vtab;
  struct __pyx_obj_7raysect_4core_4math_5point_Point2D *lower;
  struct __pyx_obj_7raysect_4core_4math_5point_Point2D *upper;
};


/* "raysect/core/math/spatial/kdtree2d.pxd":52
 * 
 * 
 * cdef class Item2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_Item2D {
  PyObject_HEAD
  int32_t id;
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *box;
};


/* "raysect/core/math/spatial/kdtree2d.pxd":59
 * 
 * 
 * cdef class KDTree2DCore:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *__pyx_vtab;
  struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree2d_kdnode *_nodes;
  int32_t _allocated_nodes;
  int32_t _next_node;
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *bounds;
  int32_t _max_depth;
  int32_t _min_items;
  double _hit_cost;
  double _empty_bonus;
};


/* "raysect/core/math/spatial/kdtree2d.pxd":116
 * 
 * 
 * cdef class KDTree2D(KDTree2DCore):             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _is_contained_leaf(self, int32_t id, Point2D point)
 */
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D {
  struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore __pyx_base;
};


/* "raysect/core/math/function/float/function2d/interpolate/common.pxd":37
 * 
 * 
 * cdef class MeshKDTree2D(KDTree2DCore):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D {
  struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore __pyx_base;
  PyArrayObject *_vertices;
  PyArrayObject *_triangles;
  __Pyx_memviewslice _vertices_mv;
  __Pyx_memviewslice _triangles_mv;
  __pyx_t_5numpy_int32_t triangle_id;
  __pyx_t_5numpy_int32_t i1;
  __pyx_t_5numpy_int32_t i2;
  __pyx_t_5numpy_int32_t i3;
  double alpha;
  double beta;
  double gamma;
  int _cache_available;
  double _cached_x;
  double _cached_y;
  int _cached_result;
};


/* "raysect/core/math/function/base.pxd":32
 * # POSSIBILITY OF SUCH DAMAGE.
 * 
 * cdef class Function:             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_7raysect_4core_4math_8function_4base_Function {
  PyObject_HEAD
};


/* "raysect/core/math/function/float/base.pxd":35
 * 
 * 
 * cdef class FloatFunction(Function):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_4base_FloatFunction {
  struct __pyx_obj_7raysect_4core_4math_8function_4base_Function __pyx_base;
};


/* "raysect/core/math/function/float/function1d/base.pxd":36
 * 
 * 
 * cdef class Function1D(FloatFunction):             # <<<<<<<<<<<<<<
 *     cdef double evaluate(self, double x) except? -1e999
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_4base_FloatFunction __pyx_base;
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *__pyx_vtab;
};


/* "raysect/core/math/function/float/function1d/base.pxd":40
 * 
 * 
 * cdef class AddFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":44
 * 
 * 
 * cdef class SubtractFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":48
 * 
 * 
 * cdef class MultiplyFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":52
 * 
 * 
 * cdef class DivideFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":56
 * 
 * 
 * cdef class PowFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":60
 * 
 * 
 * cdef class ModuloFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":64
 * 
 * 
 * cdef class AbsFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":68
 * 
 * 
 * cdef class EqualsFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":72
 * 
 * 
 * cdef class NotEqualsFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":76
 * 
 * 
 * cdef class LessThanFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":80
 * 
 * 
 * cdef class GreaterThanFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":84
 * 
 * 
 * cdef class LessEqualsFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":88
 * 
 * 
 * cdef class GreaterEqualsFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/float/function1d/base.pxd":92
 * 
 * 
 * cdef class AddScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":97
 * 
 * 
 * cdef class SubtractScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":102
 * 
 * 
 * cdef class MultiplyScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":107
 * 
 * 
 * cdef class DivideScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":112
 * 
 * 
 * cdef class ModuloScalarFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":117
 * 
 * 
 * cdef class ModuloFunctionScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":122
 * 
 * 
 * cdef class PowScalarFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":127
 * 
 * 
 * cdef class PowFunctionScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":132
 * 
 * 
 * cdef class EqualsScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":137
 * 
 * 
 * cdef class NotEqualsScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":142
 * 
 * 
 * cdef class LessThanScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":147
 * 
 * 
 * cdef class GreaterThanScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":152
 * 
 * 
 * cdef class LessEqualsScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/base.pxd":157
 * 
 * 
 * cdef class GreaterEqualsScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/constant.pxd":35
 * 
 * 
 * cdef class Constant1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  double _value;
};


/* "raysect/core/math/function/float/function1d/blend.pxd":35
 * 
 * 
 * cdef class Blend1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _f1, _f2, _mask
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_f1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_f2;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_mask;
};


/* "raysect/core/math/function/float/function1d/autowrap.pxd":34
 * from raysect.core.math.function.float.function1d.base cimport Function1D
 * 
 * cdef class PythonFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  PyObject *function;
};


/* "raysect/core/math/function/float/function1d/arg.pxd":35
 * 
 * 
 * cdef class Arg1D(Function1D):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};


/* "raysect/core/math/function/float/function1d/cmath.pxd":35
 * 
 * 
 * cdef class Exp1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/cmath.pxd":39
 * 
 * 
 * cdef class Sin1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/cmath.pxd":43
 * 
 * 
 * cdef class Cos1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/cmath.pxd":47
 * 
 * 
 * cdef class Tan1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/cmath.pxd":51
 * 
 * 
 * cdef class Asin1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/cmath.pxd":55
 * 
 * 
 * cdef class Acos1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/cmath.pxd":59
 * 
 * 
 * cdef class Atan1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/cmath.pxd":63
 * 
 * 
 * cdef class Atan4Q1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _numerator, _denominator
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_numerator;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_denominator;
};


/* "raysect/core/math/function/float/function1d/cmath.pxd":67
 * 
 * 
 * cdef class Sqrt1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/cmath.pxd":71
 * 
 * 
 * cdef class Erf1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/float/function1d/interpolate.pxd":36
 * 
 * 
 * cdef class Interpolator1DArray(Function1D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  PyArrayObject *x;
  PyArrayObject *f;
  __Pyx_memviewslice _x_mv;
  __Pyx_memviewslice _f_mv;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D *_interpolator;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D *_extrapolator;
  int _last_index;
  double _extrapolation_range;
};


/* "raysect/core/math/function/float/function1d/interpolate.pxd":47
 * 
 * 
 * cdef class _Interpolator1D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D *__pyx_vtab;
  __Pyx_memviewslice _x;
  __Pyx_memviewslice _f;
  int _last_index;
};


/* "raysect/core/math/function/float/function1d/interpolate.pxd":57
 * 
 * 
 * cdef class _Interpolator1DLinear(_Interpolator1D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D __pyx_base;
};


/* "raysect/core/math/function/float/function1d/interpolate.pxd":61
 * 
 * 
 * cdef class _Interpolator1DCubic(_Interpolator1D):             # <<<<<<<<<<<<<<
 *     cdef:
 *         np.uint8_t[::1] _calculated
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D __pyx_base;
  __Pyx_memviewslice _calculated;
  __Pyx_memviewslice _a;
  int _n;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D *_array_derivative;
};


/* "raysect/core/math/function/float/function1d/interpolate.pxd":69
 * 
 * 
 * cdef class _Extrapolator1D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D *__pyx_vtab;
  __Pyx_memviewslice _x;
  __Pyx_memviewslice _f;
  int _last_index;
};


/* "raysect/core/math/function/float/function1d/interpolate.pxd":78
 *     # cdef double _analytic_gradient(self, double px, int index, int order)
 * 
 * cdef class _Extrapolator1DNone(_Extrapolator1D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D __pyx_base;
};


/* "raysect/core/math/function/float/function1d/interpolate.pxd":82
 * 
 * 
 * cdef class _Extrapolator1DNearest(_Extrapolator1D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D __pyx_base;
};


/* "raysect/core/math/function/float/function1d/interpolate.pxd":86
 * 
 * 
 * cdef class _Extrapolator1DLinear(_Extrapolator1D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D __pyx_base;
};


/* "raysect/core/math/function/float/function1d/interpolate.pxd":90
 * 
 * 
 * cdef class _Extrapolator1DQuadratic(_Extrapolator1D):             # <<<<<<<<<<<<<<
 * 
 *     cdef double[3] _a_first, _a_last
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D __pyx_base;
  double _a_first[3];
  double _a_last[3];
};


/* "raysect/core/math/function/float/function1d/interpolate.pxd":98
 * 
 * 
 * cdef class _ArrayDerivative1D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D *__pyx_vtab;
  __Pyx_memviewslice _x;
  __Pyx_memviewslice _f;
  int _last_index;
};


/* "raysect/core/math/function/float/function2d/base.pxd":36
 * 
 * 
 * cdef class Function2D(FloatFunction):             # <<<<<<<<<<<<<<
 * 
 *     cdef double evaluate(self, double x, double y) except? -1e999
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_4base_FloatFunction __pyx_base;
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *__pyx_vtab;
};


/* "raysect/core/math/function/float/function2d/base.pxd":41
 * 
 * 
 * cdef class AddFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":45
 * 
 * 
 * cdef class SubtractFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":49
 * 
 * 
 * cdef class MultiplyFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":53
 * 
 * 
 * cdef class DivideFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":57
 * 
 * 
 * cdef class ModuloFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":61
 * 
 * 
 * cdef class PowFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":65
 * 
 * 
 * cdef class AbsFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":69
 * 
 * 
 * cdef class EqualsFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":73
 * 
 * 
 * cdef class NotEqualsFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":77
 * 
 * 
 * cdef class LessThanFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":81
 * 
 * 
 * cdef class GreaterThanFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":85
 * 
 * 
 * cdef class LessEqualsFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":89
 * 
 * 
 * cdef class GreaterEqualsFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/float/function2d/base.pxd":93
 * 
 * 
 * cdef class AddScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":98
 * 
 * 
 * cdef class SubtractScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":103
 * 
 * 
 * cdef class MultiplyScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":108
 * 
 * 
 * cdef class DivideScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":113
 * 
 * 
 * cdef class ModuloScalarFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":118
 * 
 * 
 * cdef class ModuloFunctionScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":123
 * 
 * 
 * cdef class PowScalarFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":128
 * 
 * 
 * cdef class PowFunctionScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":133
 * 
 * 
 * cdef class EqualsScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":138
 * 
 * 
 * cdef class NotEqualsScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":143
 * 
 * 
 * cdef class LessThanScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":148
 * 
 * 
 * cdef class GreaterThanScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":153
 * 
 * 
 * cdef class LessEqualsScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/base.pxd":158
 * 
 * 
 * cdef class GreaterEqualsScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/constant.pxd":35
 * 
 * 
 * cdef class Constant2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  double _value;
};


/* "raysect/core/math/function/float/function2d/blend.pxd":35
 * 
 * 
 * cdef class Blend2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _f1, _f2, _mask
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_f1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_f2;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_mask;
};


/* "raysect/core/math/function/float/function2d/autowrap.pxd":34
 * from raysect.core.math.function.float.function2d.base cimport Function2D
 * 
 * cdef class PythonFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  PyObject *function;
};


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2dmesh.pxd":37
 * 
 * 
 * cdef class Interpolator2DMesh(Function2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  PyArrayObject *_vertex_data;
  __Pyx_memviewslice _vertex_data_mv;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *_kdtree;
  int _limit;
  double _default_value;
};


/* "raysect/core/math/function/float/function2d/interpolate/discrete2dmesh.pxd":37
 * 
 * 
 * cdef class Discrete2DMesh(Function2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  PyArrayObject *_triangle_data;
  __Pyx_memviewslice _triangle_data_mv;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *_kdtree;
  int _limit;
  double _default_value;
};


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":42
 * 
 * 
 * cdef class Interpolator2DArray(Function2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  PyArrayObject *x;
  PyArrayObject *y;
  PyArrayObject *f;
  __Pyx_memviewslice _x_mv;
  __Pyx_memviewslice _y_mv;
  __Pyx_memviewslice _f_mv;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D *_interpolator;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D *_extrapolator;
  int _last_index_x;
  int _last_index_y;
  double _extrapolation_range_x;
  double _extrapolation_range_y;
};


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":56
 * 
 * 
 * cdef class _Interpolator2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D *__pyx_vtab;
  __Pyx_memviewslice _x;
  __Pyx_memviewslice _y;
  __Pyx_memviewslice _f;
  int _last_index_x;
  int _last_index_y;
};


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":67
 * 
 * 
 * cdef class _Interpolator2DLinear(_Interpolator2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef double _calculate_a0(self, int ix, int iy)
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D __pyx_base;
};


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":75
 * 
 * 
 * cdef class _Interpolator2DCubic(_Interpolator2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D __pyx_base;
  __Pyx_memviewslice _a;
  __Pyx_memviewslice _calculated;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *_array_derivative;
};


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":85
 * 
 * 
 * cdef class _Extrapolator2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D *__pyx_vtab;
  __Pyx_memviewslice _x;
  __Pyx_memviewslice _y;
  __Pyx_memviewslice _f;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D *_interpolator;
  int _last_index_x;
  int _last_index_y;
  double _extrapolation_range_x;
  double _extrapolation_range_y;
};


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":100
 * 
 * 
 * cdef class _Extrapolator2DNone(_Extrapolator2D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D __pyx_base;
};


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":104
 * 
 * 
 * cdef class _Extrapolator2DNearest(_Extrapolator2D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D __pyx_base;
};


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":108
 * 
 * 
 * cdef class _Extrapolator2DLinear(_Extrapolator2D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D __pyx_base;
};


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":112
 * 
 * 
 * cdef class _ArrayDerivative2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *__pyx_vtab;
  __Pyx_memviewslice _x;
  __Pyx_memviewslice _y;
  __Pyx_memviewslice _f;
  int _last_index_x;
  int _last_index_y;
};


/* "raysect/core/math/function/float/function2d/arg.pxd":37
 *     X, Y
 * 
 * cdef class Arg2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef ArgLabel _argument
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  enum __pyx_t_7raysect_4core_4math_8function_5float_10function2d_3arg_ArgLabel _argument;
};


/* "raysect/core/math/function/float/function2d/cmath.pxd":35
 * 
 * 
 * cdef class Exp2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/cmath.pxd":39
 * 
 * 
 * cdef class Sin2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/cmath.pxd":43
 * 
 * 
 * cdef class Cos2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/cmath.pxd":47
 * 
 * 
 * cdef class Tan2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/cmath.pxd":51
 * 
 * 
 * cdef class Asin2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/cmath.pxd":55
 * 
 * 
 * cdef class Acos2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/cmath.pxd":59
 * 
 * 
 * cdef class Atan2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/cmath.pxd":63
 * 
 * 
 * cdef class Atan4Q2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _numerator, _denominator
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_numerator;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_denominator;
};


/* "raysect/core/math/function/float/function2d/cmath.pxd":67
 * 
 * 
 * cdef class Sqrt2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function2d/cmath.pxd":71
 * 
 * 
 * cdef class Erf2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":36
 * 
 * 
 * cdef class Function3D(FloatFunction):             # <<<<<<<<<<<<<<
 *     cdef double evaluate(self, double x, double y, double z) except? -1e999
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_4base_FloatFunction __pyx_base;
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *__pyx_vtab;
};


/* "raysect/core/math/function/float/function3d/base.pxd":40
 * 
 * 
 * cdef class AddFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":44
 * 
 * 
 * cdef class SubtractFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":48
 * 
 * 
 * cdef class MultiplyFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":52
 * 
 * 
 * cdef class DivideFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":56
 * 
 * 
 * cdef class ModuloFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":60
 * 
 * 
 * cdef class PowFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":64
 * 
 * 
 * cdef class AbsFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":68
 * 
 * 
 * cdef class EqualsFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":72
 * 
 * 
 * cdef class NotEqualsFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":76
 * 
 * 
 * cdef class LessThanFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":80
 * 
 * 
 * cdef class GreaterThanFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":84
 * 
 * 
 * cdef class LessEqualsFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":88
 * 
 * 
 * cdef class GreaterEqualsFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/float/function3d/base.pxd":92
 * 
 * 
 * cdef class AddScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":97
 * 
 * 
 * cdef class SubtractScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":102
 * 
 * 
 * cdef class MultiplyScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":107
 * 
 * 
 * cdef class DivideScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":112
 * 
 * 
 * cdef class ModuloScalarFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":117
 * 
 * 
 * cdef class ModuloFunctionScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":122
 * 
 * 
 * cdef class PowScalarFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":127
 * 
 * 
 * cdef class PowFunctionScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":132
 * 
 * 
 * cdef class EqualsScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":137
 * 
 * 
 * cdef class NotEqualsScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":142
 * 
 * 
 * cdef class LessThanScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":147
 * 
 * 
 * cdef class GreaterThanScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":152
 * 
 * 
 * cdef class LessEqualsScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/base.pxd":157
 * 
 * 
 * cdef class GreaterEqualsScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/constant.pxd":35
 * 
 * 
 * cdef class Constant3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  double _value;
};


/* "raysect/core/math/function/float/function3d/blend.pxd":35
 * 
 * 
 * cdef class Blend3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _f1, _f2, _mask
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_f1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_f2;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_mask;
};


/* "raysect/core/math/function/float/function3d/autowrap.pxd":34
 * from raysect.core.math.function.float.function3d.base cimport Function3D
 * 
 * cdef class PythonFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  PyObject *function;
};


/* "raysect/core/math/spatial/kdtree3d.pxd":52
 * 
 * 
 * cdef class Item3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_Item3D {
  PyObject_HEAD
  int32_t id;
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *box;
};


/* "raysect/core/math/spatial/kdtree3d.pxd":59
 * 
 * 
 * cdef class KDTree3DCore:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *__pyx_vtab;
  struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree3d_kdnode *_nodes;
  int32_t _allocated_nodes;
  int32_t _next_node;
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *bounds;
  int32_t _max_depth;
  int32_t _min_items;
  double _hit_cost;
  double _empty_bonus;
};


/* "raysect/core/math/spatial/kdtree3d.pxd":126
 * 
 * 
 * cdef class KDTree3D(KDTree3DCore):             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _trace_leaf(self, int32_t id, Ray ray, double max_range)
 */
struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D {
  struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore __pyx_base;
};


/* "raysect/core/math/function/float/function3d/interpolate/common.pxd":37
 * 
 * 
 * cdef class MeshKDTree3D(KDTree3DCore):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D {
  struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore __pyx_base;
  PyArrayObject *_vertices;
  PyArrayObject *_tetrahedra;
  __Pyx_memviewslice _vertices_mv;
  __Pyx_memviewslice _tetrahedra_mv;
  __pyx_t_5numpy_int32_t tetrahedra_id;
  __pyx_t_5numpy_int32_t i1;
  __pyx_t_5numpy_int32_t i2;
  __pyx_t_5numpy_int32_t i3;
  __pyx_t_5numpy_int32_t i4;
  double alpha;
  double beta;
  double gamma;
  double delta;
  int _cache_available;
  double _cached_x;
  double _cached_y;
  double _cached_z;
  int _cached_result;
};


/* "raysect/core/math/function/float/function3d/interpolate/discrete3dmesh.pxd":37
 * 
 * 
 * cdef class Discrete3DMesh(Function3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  PyArrayObject *_tetrahedra_data;
  __Pyx_memviewslice _tetrahedra_data_mv;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D *_kdtree;
  int _limit;
  double _default_value;
};


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":41
 * 
 * 
 * cdef class Interpolator3DArray(Function3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  PyArrayObject *x;
  PyArrayObject *y;
  PyArrayObject *z;
  PyArrayObject *f;
  __Pyx_memviewslice _x_mv;
  __Pyx_memviewslice _y_mv;
  __Pyx_memviewslice _z_mv;
  __Pyx_memviewslice _f_mv;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D *_interpolator;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D *_extrapolator;
  int _last_index_x;
  int _last_index_y;
  int _last_index_z;
  double _extrapolation_range_x;
  double _extrapolation_range_y;
  double _extrapolation_range_z;
};


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":55
 * 
 * 
 * cdef class _Interpolator3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D *__pyx_vtab;
  __Pyx_memviewslice _x;
  __Pyx_memviewslice _y;
  __Pyx_memviewslice _z;
  __Pyx_memviewslice _f;
  int _last_index_x;
  int _last_index_y;
  int _last_index_z;
};


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":67
 * 
 * 
 * cdef class _Interpolator3DLinear(_Interpolator3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef double _calculate_a0(self, int ix, int iy, int iz)
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D __pyx_base;
};


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":79
 * 
 * 
 * cdef class _Interpolator3DCubic(_Interpolator3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D __pyx_base;
  __Pyx_memviewslice _calculated;
  __Pyx_memviewslice _a;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *_array_derivative;
};


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":89
 * 
 * 
 * cdef class _Extrapolator3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D *__pyx_vtab;
  __Pyx_memviewslice _x;
  __Pyx_memviewslice _y;
  __Pyx_memviewslice _z;
  __Pyx_memviewslice _f;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D *_interpolator;
  int _last_index_x;
  int _last_index_y;
  int _last_index_z;
  double _extrapolation_range_x;
  double _extrapolation_range_y;
  double _extrapolation_range_z;
};


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":108
 * 
 * 
 * cdef class _Extrapolator3DNone(_Extrapolator3D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D __pyx_base;
};


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":112
 * 
 * 
 * cdef class _Extrapolator3DLinear(_Extrapolator3D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D __pyx_base;
};


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":116
 * 
 * 
 * cdef class _ArrayDerivative3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D {
  PyObject_HEAD
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *__pyx_vtab;
  __Pyx_memviewslice _x;
  __Pyx_memviewslice _y;
  __Pyx_memviewslice _z;
  __Pyx_memviewslice _f;
  int _last_index_x;
  int _last_index_y;
  int _last_index_z;
};


/* "raysect/core/math/function/float/function3d/arg.pxd":37
 *     X, Y, Z
 * 
 * cdef class Arg3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef ArgLabel _argument
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  enum __pyx_t_7raysect_4core_4math_8function_5float_10function3d_3arg_ArgLabel _argument;
};


/* "raysect/core/math/function/float/function3d/cmath.pxd":35
 * 
 * 
 * cdef class Exp3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/cmath.pxd":39
 * 
 * 
 * cdef class Sin3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/cmath.pxd":43
 * 
 * 
 * cdef class Cos3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/cmath.pxd":47
 * 
 * 
 * cdef class Tan3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/cmath.pxd":51
 * 
 * 
 * cdef class Asin3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/cmath.pxd":55
 * 
 * 
 * cdef class Acos3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/cmath.pxd":59
 * 
 * 
 * cdef class Atan3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/cmath.pxd":63
 * 
 * 
 * cdef class Atan4Q3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _numerator, _denominator
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_numerator;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_denominator;
};


/* "raysect/core/math/function/float/function3d/cmath.pxd":67
 * 
 * 
 * cdef class Sqrt3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/float/function3d/cmath.pxd":71
 * 
 * 
 * cdef class Erf3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 */
struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/vector3d/base.pxd":35
 * 
 * 
 * cdef class Vector3DFunction(Function):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction {
  struct __pyx_obj_7raysect_4core_4math_8function_4base_Function __pyx_base;
};


/* "raysect/core/math/function/vector3d/function1d/base.pxd":38
 * 
 * 
 * cdef class Function1D(Vector3DFunction):             # <<<<<<<<<<<<<<
 *     cdef Vector3D evaluate(self, double x)
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction __pyx_base;
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *__pyx_vtab;
};


/* "raysect/core/math/function/vector3d/function1d/base.pxd":42
 * 
 * 
 * cdef class AddFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/vector3d/function1d/base.pxd":46
 * 
 * 
 * cdef class SubtractFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/vector3d/function1d/base.pxd":50
 * 
 * 
 * cdef class MultiplyFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1
 *     cdef FloatFunction1D _function2
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/vector3d/function1d/base.pxd":55
 * 
 * 
 * cdef class DivideFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1
 *     cdef FloatFunction1D _function2
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/vector3d/function1d/base.pxd":60
 * 
 * 
 * cdef class NegFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_function;
};


/* "raysect/core/math/function/vector3d/function1d/base.pxd":64
 * 
 * 
 * cdef class EqualsFunction1D(FloatFunction1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/vector3d/function1d/base.pxd":68
 * 
 * 
 * cdef class NotEqualsFunction1D(FloatFunction1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_function2;
};


/* "raysect/core/math/function/vector3d/function1d/constant.pxd":36
 * 
 * 
 * cdef class Constant1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Vector3D _value
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *_value;
};


/* "raysect/core/math/function/vector3d/function1d/blend.pxd":36
 * 
 * 
 * cdef class Blend1D(VecFn1D):             # <<<<<<<<<<<<<<
 *     cdef VecFn1D _f1, _f2
 *     cdef FltFn1D _mask
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_f1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *_f2;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_mask;
};


/* "raysect/core/math/function/vector3d/function1d/autowrap.pxd":34
 * from raysect.core.math.function.vector3d.function1d.base cimport Function1D
 * 
 * cdef class PythonFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
  PyObject *function;
};


/* "raysect/core/math/function/vector3d/function1d/utility.pxd":36
 * 
 * 
 * cdef class FloatToVector3DFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef FloatFunction1D _x, _y, _z
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_x;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_y;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *_z;
};


/* "raysect/core/math/function/vector3d/function2d/base.pxd":38
 * 
 * 
 * cdef class Function2D(Vector3DFunction):             # <<<<<<<<<<<<<<
 *     cdef Vector3D evaluate(self, double x, double y)
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction __pyx_base;
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *__pyx_vtab;
};


/* "raysect/core/math/function/vector3d/function2d/base.pxd":42
 * 
 * 
 * cdef class AddFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/vector3d/function2d/base.pxd":46
 * 
 * 
 * cdef class SubtractFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/vector3d/function2d/base.pxd":50
 * 
 * 
 * cdef class MultiplyFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1
 *     cdef FloatFunction2D _function2
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/vector3d/function2d/base.pxd":55
 * 
 * 
 * cdef class DivideFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1
 *     cdef FloatFunction2D _function2
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/vector3d/function2d/base.pxd":60
 * 
 * 
 * cdef class NegFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_function;
};


/* "raysect/core/math/function/vector3d/function2d/base.pxd":64
 * 
 * 
 * cdef class EqualsFunction2D(FloatFunction2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/vector3d/function2d/base.pxd":68
 * 
 * 
 * cdef class NotEqualsFunction2D(FloatFunction2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_function2;
};


/* "raysect/core/math/function/vector3d/function2d/constant.pxd":36
 * 
 * 
 * cdef class Constant2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Vector3D _value
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *_value;
};


/* "raysect/core/math/function/vector3d/function2d/blend.pxd":36
 * 
 * 
 * cdef class Blend2D(VecFn2D):             # <<<<<<<<<<<<<<
 *     cdef VecFn2D _f1, _f2
 *     cdef FltFn2D _mask
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_f1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *_f2;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_mask;
};


/* "raysect/core/math/function/vector3d/function2d/autowrap.pxd":34
 * from raysect.core.math.function.vector3d.function2d.base cimport Function2D
 * 
 * cdef class PythonFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
  PyObject *function;
};


/* "raysect/core/math/function/vector3d/function2d/utility.pxd":36
 * 
 * 
 * cdef class FloatToVector3DFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef FloatFunction2D _x, _y, _z
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_x;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_y;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *_z;
};


/* "raysect/core/math/function/vector3d/function3d/base.pxd":38
 * 
 * 
 * cdef class Function3D(Vector3DFunction):             # <<<<<<<<<<<<<<
 *     cdef Vector3D evaluate(self, double x, double y, double z)
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction __pyx_base;
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *__pyx_vtab;
};


/* "raysect/core/math/function/vector3d/function3d/base.pxd":42
 * 
 * 
 * cdef class AddFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/vector3d/function3d/base.pxd":46
 * 
 * 
 * cdef class SubtractFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/vector3d/function3d/base.pxd":50
 * 
 * 
 * cdef class MultiplyFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1
 *     cdef FloatFunction3D _function2
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/vector3d/function3d/base.pxd":55
 * 
 * 
 * cdef class DivideFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1
 *     cdef FloatFunction3D _function2
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/vector3d/function3d/base.pxd":60
 * 
 * 
 * cdef class NegFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_function;
};


/* "raysect/core/math/function/vector3d/function3d/base.pxd":64
 * 
 * 
 * cdef class EqualsFunction3D(FloatFunction3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/vector3d/function3d/base.pxd":68
 * 
 * 
 * cdef class NotEqualsFunction3D(FloatFunction3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_function1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_function2;
};


/* "raysect/core/math/function/vector3d/function3d/constant.pxd":36
 * 
 * 
 * cdef class Constant3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Vector3D _value
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *_value;
};


/* "raysect/core/math/function/vector3d/function3d/blend.pxd":36
 * 
 * 
 * cdef class Blend3D(VecFn3D):             # <<<<<<<<<<<<<<
 *     cdef VecFn3D _f1, _f2
 *     cdef FltFn3D _mask
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_f1;
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *_f2;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_mask;
};


/* "raysect/core/math/function/vector3d/function3d/autowrap.pxd":34
 * from raysect.core.math.function.vector3d.function3d.base cimport Function3D
 * 
 * cdef class PythonFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
  PyObject *function;
};


/* "raysect/core/math/function/vector3d/function3d/utility.pxd":36
 * 
 * 
 * cdef class FloatToVector3DFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef FloatFunction3D _x, _y, _z
 */
struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_x;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_y;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *_z;
};


/* "cherab/solps/solps_2d_functions.pxd":27
 * 
 * 
 * cdef class SOLPSFunction2D(Function2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *_kdtree;
  PyArrayObject *_grid_data;
  PyArrayObject *_triangle_to_grid_map;
  __Pyx_memviewslice _triangle_to_grid_map_mv;
  __Pyx_memviewslice _grid_data_mv;
};


/* "cherab/solps/solps_2d_functions.pxd":37
 *     cdef double evaluate(self, double x, double y) except? -1e999
 * 
 * cdef class SOLPSVectorFunction2D(VectorFunction2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D {
  struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *_kdtree;
  PyArrayObject *_grid_vectors;
  PyArrayObject *_triangle_to_grid_map;
  __Pyx_memviewslice _triangle_to_grid_map_mv;
  __Pyx_memviewslice _grid_vectors_mv;
};


/* "View.MemoryView":109
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":290
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":341
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview:             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int acquisition_count[2];
  __pyx_atomic_int *acquisition_count_aligned_p;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":975
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "raysect/core/math/_vec3.pxd":32
 * # POSSIBILITY OF SUCH DAMAGE.
 * 
 * cdef class _Vec3:             # <<<<<<<<<<<<<<
 * 
 *     cdef public double x, y, z
 */

struct __pyx_vtabstruct_7raysect_4core_4math_5_vec3__Vec3 {
  double (*dot)(struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *, struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *, int __pyx_skip_dispatch);
  double (*angle)(struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *, struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *, int __pyx_skip_dispatch);
  double (*get_length)(struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *);
  PyObject *(*set_length)(struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *, double);
  double (*get_index)(struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *, int);
  void (*set_index)(struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *, int, double);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_5_vec3__Vec3 *__pyx_vtabptr_7raysect_4core_4math_5_vec3__Vec3;


/* "raysect/core/math/_mat4.pxd":32
 * # POSSIBILITY OF SUCH DAMAGE.
 * 
 * cdef class _Mat4:             # <<<<<<<<<<<<<<
 * 
 *     cdef double m[4][4]
 */

struct __pyx_vtabstruct_7raysect_4core_4math_5_mat4__Mat4 {
  double (*get_element)(struct __pyx_obj_7raysect_4core_4math_5_mat4__Mat4 *, int, int);
  void (*set_element)(struct __pyx_obj_7raysect_4core_4math_5_mat4__Mat4 *, int, int, double);
  int (*is_identity)(struct __pyx_obj_7raysect_4core_4math_5_mat4__Mat4 *, int __pyx_skip_dispatch, struct __pyx_opt_args_7raysect_4core_4math_5_mat4_5_Mat4_is_identity *__pyx_optional_args);
  int (*is_close)(struct __pyx_obj_7raysect_4core_4math_5_mat4__Mat4 *, struct __pyx_obj_7raysect_4core_4math_5_mat4__Mat4 *, int __pyx_skip_dispatch, struct __pyx_opt_args_7raysect_4core_4math_5_mat4_5_Mat4_is_close *__pyx_optional_args);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_5_mat4__Mat4 *__pyx_vtabptr_7raysect_4core_4math_5_mat4__Mat4;


/* "raysect/core/math/affinematrix.pxd":35
 * 
 * 
 * cdef class AffineMatrix3D(_Mat4):             # <<<<<<<<<<<<<<
 * 
 *     cpdef AffineMatrix3D inverse(self)
 */

struct __pyx_vtabstruct_7raysect_4core_4math_12affinematrix_AffineMatrix3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_5_mat4__Mat4 __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*inverse)(struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*mul)(struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_12affinematrix_AffineMatrix3D *__pyx_vtabptr_7raysect_4core_4math_12affinematrix_AffineMatrix3D;


/* "raysect/core/math/vector.pxd":35
 * from raysect.core.math.affinematrix cimport AffineMatrix3D
 * 
 * cdef class Vector3D(_Vec3):             # <<<<<<<<<<<<<<
 * 
 *     cpdef Vector3D cross(self, _Vec3 v)
 */

struct __pyx_vtabstruct_7raysect_4core_4math_6vector_Vector3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_5_vec3__Vec3 __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*cross)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*normalise)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*transform)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*neg)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*add)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*sub)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*mul)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, double);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*div)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, double);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*copy)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*orthogonal)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*lerp)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*slerp)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, double, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_6vector_Vector3D *__pyx_vtabptr_7raysect_4core_4math_6vector_Vector3D;


/* "raysect/core/math/vector.pxd":78
 * 
 * 
 * cdef class Vector2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef public double x, y
 */

struct __pyx_vtabstruct_7raysect_4core_4math_6vector_Vector2D {
  double (*dot)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, int __pyx_skip_dispatch);
  double (*get_length)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *);
  PyObject *(*set_length)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, double);
  double (*get_index)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, int);
  void (*set_index)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, int, double);
  double (*cross)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *(*normalise)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *(*neg)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *(*add)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *(*sub)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *(*mul)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, double);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*div)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, double);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *(*copy)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *(*orthogonal)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_6vector_Vector2D *__pyx_vtabptr_7raysect_4core_4math_6vector_Vector2D;


/* "raysect/core/math/point.pxd":36
 * from raysect.core.math.affinematrix cimport AffineMatrix3D
 * 
 * cdef class Point3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef public double x, y, z
 */

struct __pyx_vtabstruct_7raysect_4core_4math_5point_Point3D {
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*vector_to)(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch);
  double (*distance_to)(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *(*transform)(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *(*add)(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *);
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *(*sub)(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *);
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *(*copy)(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch);
  double (*get_index)(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int);
  void (*set_index)(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int, double);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_5point_Point3D *__pyx_vtabptr_7raysect_4core_4math_5point_Point3D;


/* "raysect/core/math/point.pxd":73
 * 
 * 
 * cdef class Point2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef public double x, y
 */

struct __pyx_vtabstruct_7raysect_4core_4math_5point_Point2D {
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *(*vector_to)(struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, int __pyx_skip_dispatch);
  double (*distance_to)(struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_5point_Point2D *(*add)(struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *);
  struct __pyx_obj_7raysect_4core_4math_5point_Point2D *(*sub)(struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *);
  struct __pyx_obj_7raysect_4core_4math_5point_Point2D *(*copy)(struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, int __pyx_skip_dispatch);
  double (*get_index)(struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, int);
  void (*set_index)(struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, int, double);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_5point_Point2D *__pyx_vtabptr_7raysect_4core_4math_5point_Point2D;


/* "raysect/core/math/normal.pxd":36
 * from raysect.core.math.affinematrix cimport AffineMatrix3D
 * 
 * cdef class Normal3D(_Vec3):             # <<<<<<<<<<<<<<
 * 
 *     cpdef Vector3D cross(self, _Vec3 v)
 */

struct __pyx_vtabstruct_7raysect_4core_4math_6normal_Normal3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_5_vec3__Vec3 __pyx_base;
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*cross)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *(*normalise)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *(*transform)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *(*transform_with_inverse)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *(*neg)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *);
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *(*add)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *);
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *(*sub)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3 *);
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *(*mul)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, double);
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *(*div)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, double);
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *(*copy)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*as_vector)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*orthogonal)(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_6normal_Normal3D *__pyx_vtabptr_7raysect_4core_4math_6normal_Normal3D;


/* "raysect/core/math/quaternion.pxd":36
 * 
 * 
 * cdef class Quaternion:             # <<<<<<<<<<<<<<
 * 
 *     cdef public double x, y, z, s
 */

struct __pyx_vtabstruct_7raysect_4core_4math_10quaternion_Quaternion {
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*copy)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*conjugate)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*inverse)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*normalise)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, int __pyx_skip_dispatch);
  int (*is_unit)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, int __pyx_skip_dispatch, struct __pyx_opt_args_7raysect_4core_4math_10quaternion_10Quaternion_is_unit *__pyx_optional_args);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*transform)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*as_matrix)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*quaternion_to)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*neg)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*add)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*sub)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*mul_quaternion)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*mul_scalar)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, double);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*div_quaternion)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *);
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*div_scalar)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, double);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*get_axis)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *);
  double (*get_angle)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *);
  double (*get_length)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *);
  PyObject *(*set_length)(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *, double);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_10quaternion_Quaternion *__pyx_vtabptr_7raysect_4core_4math_10quaternion_Quaternion;


/* "raysect/core/math/statsarray.pxd":35
 * 
 * 
 * cdef class StatsBin:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsBin {
  PyObject *(*clear)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsBin *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsBin *(*copy)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsBin *, int __pyx_skip_dispatch);
  PyObject *(*add_sample)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsBin *, double, int __pyx_skip_dispatch);
  PyObject *(*combine_samples)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsBin *, double, double, int, int __pyx_skip_dispatch);
  double (*error)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsBin *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsBin *__pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsBin;


/* "raysect/core/math/statsarray.pxd":53
 * 
 * 
 * cdef class StatsArray1D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray1D {
  PyObject *(*clear)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D *(*copy)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D *, int __pyx_skip_dispatch);
  PyObject *(*add_sample)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D *, int, double, int __pyx_skip_dispatch);
  PyObject *(*combine_samples)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D *, int, double, double, int, int __pyx_skip_dispatch);
  double (*error)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D *, int, int __pyx_skip_dispatch);
  PyArrayObject *(*errors)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D *, int __pyx_skip_dispatch);
  void (*_new_buffers)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D *);
  PyObject *(*_bounds_check)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D *, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray1D *__pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsArray1D;


/* "raysect/core/math/statsarray.pxd":81
 * 
 * 
 * cdef class StatsArray2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray2D {
  PyObject *(*clear)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D *(*copy)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D *, int __pyx_skip_dispatch);
  PyObject *(*add_sample)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D *, int, int, double, int __pyx_skip_dispatch);
  PyObject *(*combine_samples)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D *, int, int, double, double, int, int __pyx_skip_dispatch);
  double (*error)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D *, int, int, int __pyx_skip_dispatch);
  PyArrayObject *(*errors)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D *, int __pyx_skip_dispatch);
  void (*_new_buffers)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D *);
  PyObject *(*_bounds_check)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D *, int, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray2D *__pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsArray2D;


/* "raysect/core/math/statsarray.pxd":109
 * 
 * 
 * cdef class StatsArray3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray3D {
  PyObject *(*clear)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D *(*copy)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D *, int __pyx_skip_dispatch);
  PyObject *(*add_sample)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D *, int, int, int, double, int __pyx_skip_dispatch);
  PyObject *(*combine_samples)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D *, int, int, int, double, double, int, int __pyx_skip_dispatch);
  double (*error)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D *, int, int, int, int __pyx_skip_dispatch);
  PyArrayObject *(*errors)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D *, int __pyx_skip_dispatch);
  void (*_new_buffers)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D *);
  PyObject *(*_bounds_check)(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D *, int, int, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray3D *__pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsArray3D;


/* "raysect/core/math/sampler/solidangle.pxd":41
 * 
 * 
 * cdef class SolidAngleSampler:             # <<<<<<<<<<<<<<
 * 
 *     cpdef double pdf(self, Vector3D sample)
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler {
  double (*pdf)(struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*sample)(struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler *);
  PyObject *(*sample_with_pdf)(struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler *);
  PyObject *(*samples)(struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler *, int);
  PyObject *(*samples_with_pdfs)(struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler *, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler *__pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler;


/* "raysect/core/math/sampler/solidangle.pxd":54
 * 
 * 
 * cdef class SphereSampler(SolidAngleSampler):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_SphereSampler {
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_SphereSampler *__pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_SphereSampler;


/* "raysect/core/math/sampler/solidangle.pxd":58
 * 
 * 
 * cdef class HemisphereUniformSampler(SolidAngleSampler):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler {
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler *__pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler;


/* "raysect/core/math/sampler/solidangle.pxd":62
 * 
 * 
 * cdef class HemisphereCosineSampler(SolidAngleSampler):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler {
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler *__pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler;


/* "raysect/core/math/sampler/solidangle.pxd":66
 * 
 * 
 * cdef class ConeUniformSampler(SolidAngleSampler):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler {
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler *__pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler;


/* "raysect/core/math/sampler/surface3d.pxd":35
 * 
 * 
 * cdef class SurfaceSampler3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef Point3D sample(self)
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D {
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *(*sample)(struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D *);
  PyObject *(*sample_with_pdf)(struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D *);
  PyObject *(*samples)(struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D *, int);
  PyObject *(*samples_with_pdfs)(struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D *, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D *__pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D;


/* "raysect/core/math/sampler/surface3d.pxd":46
 * 
 * 
 * cdef class DiskSampler3D(SurfaceSampler3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D *__pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D;


/* "raysect/core/math/sampler/surface3d.pxd":53
 * 
 * 
 * cdef class RectangleSampler3D(SurfaceSampler3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D *__pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D;


/* "raysect/core/math/sampler/surface3d.pxd":60
 * 
 * 
 * cdef class TriangleSampler3D(SurfaceSampler3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D __pyx_base;
  double (*_calculate_area)(struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D *__pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D;


/* "raysect/core/math/sampler/targetted.pxd":35
 * from raysect.core.math cimport Point3D, Vector3D
 * 
 * cdef class _TargettedSampler:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted__TargettedSampler {
  PyObject *(*_validate_targets)(struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler *);
  double (*pdf)(struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*sample)(struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *);
  PyObject *(*sample_with_pdf)(struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *);
  PyObject *(*samples)(struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int);
  PyObject *(*samples_with_pdfs)(struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int);
  PyObject *(*_calculate_cdf)(struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler *);
  PyObject *(*_pick_sphere)(struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler *);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted__TargettedSampler *__pyx_vtabptr_7raysect_4core_4math_7sampler_9targetted__TargettedSampler;


/* "raysect/core/math/sampler/targetted.pxd":60
 * 
 * 
 * cdef class TargettedHemisphereSampler(_TargettedSampler):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler {
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted__TargettedSampler __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler *__pyx_vtabptr_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler;


/* "raysect/core/math/sampler/targetted.pxd":63
 *     pass
 * 
 * cdef class TargettedSphereSampler(_TargettedSampler):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler {
  struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted__TargettedSampler __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler *__pyx_vtabptr_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler;


/* "raysect/core/ray.pxd":35
 * 
 * 
 * cdef class Ray:             # <<<<<<<<<<<<<<
 * 
 *     cdef public Point3D origin
 */

struct __pyx_vtabstruct_7raysect_4core_3ray_Ray {
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *(*point_on)(struct __pyx_obj_7raysect_4core_3ray_Ray *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_3ray_Ray *(*copy)(struct __pyx_obj_7raysect_4core_3ray_Ray *, int __pyx_skip_dispatch, struct __pyx_opt_args_7raysect_4core_3ray_3Ray_copy *__pyx_optional_args);
};
static struct __pyx_vtabstruct_7raysect_4core_3ray_Ray *__pyx_vtabptr_7raysect_4core_3ray_Ray;


/* "raysect/core/material.pxd":32
 * # POSSIBILITY OF SUCH DAMAGE.
 * 
 * cdef class Material:             # <<<<<<<<<<<<<<
 * 
 *     cdef public list primitives
 */

struct __pyx_vtabstruct_7raysect_4core_8material_Material {
  PyObject *(*notify_material_change)(struct __pyx_obj_7raysect_4core_8material_Material *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_8material_Material *__pyx_vtabptr_7raysect_4core_8material_Material;


/* "raysect/core/scenegraph/node.pxd":35
 * from raysect.core.math.affinematrix cimport AffineMatrix3D
 * 
 * cdef class Node(_NodeBase):             # <<<<<<<<<<<<<<
 * 
 *     cpdef AffineMatrix3D to(self, _NodeBase node)
 */

struct __pyx_vtabstruct_7raysect_4core_10scenegraph_4node_Node {
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*to)(struct __pyx_obj_7raysect_4core_10scenegraph_4node_Node *, struct __pyx_obj_7raysect_4core_10scenegraph_9_nodebase__NodeBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*to_local)(struct __pyx_obj_7raysect_4core_10scenegraph_4node_Node *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*to_root)(struct __pyx_obj_7raysect_4core_10scenegraph_4node_Node *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_10scenegraph_4node_Node *__pyx_vtabptr_7raysect_4core_10scenegraph_4node_Node;


/* "raysect/core/scenegraph/primitive.pxd":42
 * 
 * 
 * cdef class Primitive(Node):             # <<<<<<<<<<<<<<
 * 
 *     cdef Material _material
 */

struct __pyx_vtabstruct_7raysect_4core_10scenegraph_9primitive_Primitive {
  struct __pyx_vtabstruct_7raysect_4core_10scenegraph_4node_Node __pyx_base;
  struct __pyx_obj_7raysect_4core_8material_Material *(*get_material)(struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *);
  struct __pyx_obj_7raysect_4core_12intersection_Intersection *(*hit)(struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *, struct __pyx_obj_7raysect_4core_3ray_Ray *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_12intersection_Intersection *(*next_intersection)(struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *, int __pyx_skip_dispatch);
  int (*contains)(struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *(*bounding_box)(struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *(*bounding_sphere)(struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *, int __pyx_skip_dispatch);
  PyObject *(*instance)(struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *, int __pyx_skip_dispatch, struct __pyx_opt_args_7raysect_4core_10scenegraph_9primitive_9Primitive_instance *__pyx_optional_args);
  PyObject *(*notify_geometry_change)(struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *, int __pyx_skip_dispatch);
  PyObject *(*notify_material_change)(struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_10scenegraph_9primitive_Primitive *__pyx_vtabptr_7raysect_4core_10scenegraph_9primitive_Primitive;


/* "raysect/core/intersection.pxd":37
 * 
 * 
 * cdef class Intersection:             # <<<<<<<<<<<<<<
 * 
 *     cdef public Ray ray
 */

struct __pyx_vtabstruct_7raysect_4core_12intersection_Intersection {
  void (*_construct)(struct __pyx_obj_7raysect_4core_12intersection_Intersection *, struct __pyx_obj_7raysect_4core_3ray_Ray *, double, struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *, int, struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *);
};
static struct __pyx_vtabstruct_7raysect_4core_12intersection_Intersection *__pyx_vtabptr_7raysect_4core_12intersection_Intersection;


/* "raysect/core/acceleration/accelerator.pxd":36
 * from raysect.core.intersection cimport Intersection
 * 
 * cdef class Accelerator:             # <<<<<<<<<<<<<<
 * 
 *     cpdef build(self, list primitives)
 */

struct __pyx_vtabstruct_7raysect_4core_12acceleration_11accelerator_Accelerator {
  PyObject *(*build)(struct __pyx_obj_7raysect_4core_12acceleration_11accelerator_Accelerator *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_12intersection_Intersection *(*hit)(struct __pyx_obj_7raysect_4core_12acceleration_11accelerator_Accelerator *, struct __pyx_obj_7raysect_4core_3ray_Ray *, int __pyx_skip_dispatch);
  PyObject *(*contains)(struct __pyx_obj_7raysect_4core_12acceleration_11accelerator_Accelerator *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_12acceleration_11accelerator_Accelerator *__pyx_vtabptr_7raysect_4core_12acceleration_11accelerator_Accelerator;


/* "raysect/core/scenegraph/world.pxd":38
 * from raysect.core.scenegraph._nodebase cimport _NodeBase
 * 
 * cdef class World(_NodeBase):             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _rebuild_accelerator
 */

struct __pyx_vtabstruct_7raysect_4core_10scenegraph_5world_World {
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*to)(struct __pyx_obj_7raysect_4core_10scenegraph_5world_World *, struct __pyx_obj_7raysect_4core_10scenegraph_9_nodebase__NodeBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_12intersection_Intersection *(*hit)(struct __pyx_obj_7raysect_4core_10scenegraph_5world_World *, struct __pyx_obj_7raysect_4core_3ray_Ray *, int __pyx_skip_dispatch);
  PyObject *(*contains)(struct __pyx_obj_7raysect_4core_10scenegraph_5world_World *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch);
  PyObject *(*build_accelerator)(struct __pyx_obj_7raysect_4core_10scenegraph_5world_World *, int __pyx_skip_dispatch, struct __pyx_opt_args_7raysect_4core_10scenegraph_5world_5World_build_accelerator *__pyx_optional_args);
};
static struct __pyx_vtabstruct_7raysect_4core_10scenegraph_5world_World *__pyx_vtabptr_7raysect_4core_10scenegraph_5world_World;


/* "raysect/core/scenegraph/observer.pxd":34
 * from raysect.core.scenegraph.node cimport Node
 * 
 * cdef class Observer(Node):             # <<<<<<<<<<<<<<
 * 
 *     cpdef observe(self)
 */

struct __pyx_vtabstruct_7raysect_4core_10scenegraph_8observer_Observer {
  struct __pyx_vtabstruct_7raysect_4core_10scenegraph_4node_Node __pyx_base;
  PyObject *(*observe)(struct __pyx_obj_7raysect_4core_10scenegraph_8observer_Observer *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_10scenegraph_8observer_Observer *__pyx_vtabptr_7raysect_4core_10scenegraph_8observer_Observer;


/* "raysect/core/scenegraph/utility.pxd":36
 * 
 * 
 * cdef class BridgeNode(Node):             # <<<<<<<<<<<<<<
 * 
 *     cdef _NodeBase destination
 */

struct __pyx_vtabstruct_7raysect_4core_10scenegraph_7utility_BridgeNode {
  struct __pyx_vtabstruct_7raysect_4core_10scenegraph_4node_Node __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_10scenegraph_7utility_BridgeNode *__pyx_vtabptr_7raysect_4core_10scenegraph_7utility_BridgeNode;


/* "raysect/core/containers.pxd":41
 * 
 * 
 * cdef class LinkedList:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_10containers_LinkedList {
  int (*is_empty)(struct __pyx_obj_7raysect_4core_10containers_LinkedList *, int __pyx_skip_dispatch);
  PyObject *(*add)(struct __pyx_obj_7raysect_4core_10containers_LinkedList *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*add_items)(struct __pyx_obj_7raysect_4core_10containers_LinkedList *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_index)(struct __pyx_obj_7raysect_4core_10containers_LinkedList *, int, int __pyx_skip_dispatch);
  PyObject *(*insert)(struct __pyx_obj_7raysect_4core_10containers_LinkedList *, PyObject *, int, int __pyx_skip_dispatch);
  PyObject *(*remove)(struct __pyx_obj_7raysect_4core_10containers_LinkedList *, int, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_10containers_LinkedList *__pyx_vtabptr_7raysect_4core_10containers_LinkedList;


/* "raysect/core/containers.pxd":61
 * 
 * 
 * cdef class Stack(LinkedList):             # <<<<<<<<<<<<<<
 * 
 *     cpdef push(self, object value)
 */

struct __pyx_vtabstruct_7raysect_4core_10containers_Stack {
  struct __pyx_vtabstruct_7raysect_4core_10containers_LinkedList __pyx_base;
  PyObject *(*push)(struct __pyx_obj_7raysect_4core_10containers_Stack *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_7raysect_4core_10containers_Stack *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_10containers_Stack *__pyx_vtabptr_7raysect_4core_10containers_Stack;


/* "raysect/core/containers.pxd":68
 * 
 * 
 * cdef class Queue(LinkedList):             # <<<<<<<<<<<<<<
 * 
 *     cpdef object next_in_queue(self)
 */

struct __pyx_vtabstruct_7raysect_4core_10containers_Queue {
  struct __pyx_vtabstruct_7raysect_4core_10containers_LinkedList __pyx_base;
  PyObject *(*next_in_queue)(struct __pyx_obj_7raysect_4core_10containers_Queue *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_10containers_Queue *__pyx_vtabptr_7raysect_4core_10containers_Queue;


/* "raysect/core/boundingsphere.pxd":35
 * 
 * 
 * cdef class BoundingSphere3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_14boundingsphere_BoundingSphere3D {
  int (*hit)(struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *, struct __pyx_obj_7raysect_4core_3ray_Ray *, int __pyx_skip_dispatch);
  PyObject *(*full_intersection)(struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *, struct __pyx_obj_7raysect_4core_3ray_Ray *, int __pyx_skip_dispatch);
  int (*intersect)(struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *, struct __pyx_obj_7raysect_4core_3ray_Ray *, double *, double *);
  int (*contains)(struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch);
  PyObject *(*__pyx_union)(struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *, struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *, int __pyx_skip_dispatch);
  PyObject *(*extend)(struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch, struct __pyx_opt_args_7raysect_4core_14boundingsphere_16BoundingSphere3D_extend *__pyx_optional_args);
  double (*surface_area)(struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *, int __pyx_skip_dispatch);
  double (*volume)(struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *, int __pyx_skip_dispatch);
  PyObject *(*pad)(struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *, double, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_14boundingsphere_BoundingSphere3D *__pyx_vtabptr_7raysect_4core_14boundingsphere_BoundingSphere3D;


/* "raysect/core/boundingbox.pxd":37
 * 
 * 
 * cdef class BoundingBox3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef Point3D lower
 */

struct __pyx_vtabstruct_7raysect_4core_11boundingbox_BoundingBox3D {
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *(*get_centre)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *);
  int (*hit)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, struct __pyx_obj_7raysect_4core_3ray_Ray *, int __pyx_skip_dispatch);
  PyObject *(*full_intersection)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, struct __pyx_obj_7raysect_4core_3ray_Ray *, int __pyx_skip_dispatch);
  int (*intersect)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, struct __pyx_obj_7raysect_4core_3ray_Ray *, double *, double *);
  void (*_slab)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, double, double, double, double, double *, double *);
  int (*contains)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch);
  PyObject *(*__pyx_union)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, int __pyx_skip_dispatch);
  PyObject *(*extend)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch, struct __pyx_opt_args_7raysect_4core_11boundingbox_13BoundingBox3D_extend *__pyx_optional_args);
  double (*surface_area)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, int __pyx_skip_dispatch);
  double (*volume)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, int __pyx_skip_dispatch);
  PyObject *(*vertices)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, int __pyx_skip_dispatch);
  double (*extent)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, int, int __pyx_skip_dispatch);
  int (*largest_axis)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, int __pyx_skip_dispatch);
  double (*largest_extent)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, int __pyx_skip_dispatch);
  PyObject *(*pad)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, double, int __pyx_skip_dispatch);
  PyObject *(*pad_axis)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, int, double, int __pyx_skip_dispatch);
  struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D *(*enclosing_sphere)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_11boundingbox_BoundingBox3D *__pyx_vtabptr_7raysect_4core_11boundingbox_BoundingBox3D;


/* "raysect/core/boundingbox.pxd":92
 * 
 * 
 * cdef class BoundingBox2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef Point2D lower
 */

struct __pyx_vtabstruct_7raysect_4core_11boundingbox_BoundingBox2D {
  int (*contains)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, int __pyx_skip_dispatch);
  PyObject *(*__pyx_union)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, int __pyx_skip_dispatch);
  PyObject *(*extend)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, int __pyx_skip_dispatch, struct __pyx_opt_args_7raysect_4core_11boundingbox_13BoundingBox2D_extend *__pyx_optional_args);
  double (*surface_area)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, int __pyx_skip_dispatch);
  PyObject *(*vertices)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, int __pyx_skip_dispatch);
  double (*extent)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, int, int __pyx_skip_dispatch);
  int (*largest_axis)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, int __pyx_skip_dispatch);
  double (*largest_extent)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, int __pyx_skip_dispatch);
  PyObject *(*pad)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, double, int __pyx_skip_dispatch);
  PyObject *(*pad_axis)(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, int, double, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_11boundingbox_BoundingBox2D *__pyx_vtabptr_7raysect_4core_11boundingbox_BoundingBox2D;


/* "raysect/core/math/spatial/kdtree2d.pxd":59
 * 
 * 
 * cdef class KDTree2DCore:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore {
  int32_t (*_build)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, PyObject *, struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, struct __pyx_opt_args_7raysect_4core_4math_7spatial_8kdtree2d_12KDTree2DCore__build *__pyx_optional_args);
  PyObject *(*_split)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, PyObject *, struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *);
  void (*_get_edges)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, PyObject *, int32_t, int32_t *, struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree2d_edge **);
  void (*_free_edges)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree2d_edge **);
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *(*_get_lower_bounds)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, double, int32_t);
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *(*_get_upper_bounds)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *, double, int32_t);
  int32_t (*_new_leaf)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, PyObject *);
  int32_t (*_new_branch)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, PyObject *, int32_t);
  int32_t (*_new_node)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *);
  int (*is_contained)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, int __pyx_skip_dispatch);
  int (*_is_contained)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *);
  int (*_is_contained_node)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *);
  int (*_is_contained_branch)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *);
  int (*_is_contained_leaf)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *);
  PyObject *(*items_containing)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, int __pyx_skip_dispatch);
  PyObject *(*_items_containing)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *);
  PyObject *(*_items_containing_node)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *);
  PyObject *(*_items_containing_branch)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *);
  PyObject *(*_items_containing_leaf)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *);
  void (*_reset)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *);
  double (*_read_double)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, PyObject *);
  int32_t (*_read_int32)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *, PyObject *);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *__pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore;


/* "raysect/core/math/spatial/kdtree2d.pxd":116
 * 
 * 
 * cdef class KDTree2D(KDTree2DCore):             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _is_contained_leaf(self, int32_t id, Point2D point)
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore __pyx_base;
  int (*_is_contained_items)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D *, PyObject *, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, int __pyx_skip_dispatch);
  PyObject *(*_items_containing_items)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D *, PyObject *, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D *__pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D;


/* "raysect/core/math/function/float/function2d/interpolate/common.pxd":37
 * 
 * 
 * cdef class MeshKDTree2D(KDTree2DCore):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore __pyx_base;
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *(*_generate_bounding_box)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *, __pyx_t_5numpy_int32_t);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D;


/* "raysect/core/math/function/float/function1d/base.pxd":36
 * 
 * 
 * cdef class Function1D(FloatFunction):             # <<<<<<<<<<<<<<
 *     cdef double evaluate(self, double x) except? -1e999
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D {
  double (*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *, double);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D;


/* "raysect/core/math/function/float/function1d/base.pxd":40
 * 
 * 
 * cdef class AddFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":44
 * 
 * 
 * cdef class SubtractFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":48
 * 
 * 
 * cdef class MultiplyFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":52
 * 
 * 
 * cdef class DivideFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":56
 * 
 * 
 * cdef class PowFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":60
 * 
 * 
 * cdef class ModuloFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":64
 * 
 * 
 * cdef class AbsFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":68
 * 
 * 
 * cdef class EqualsFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":72
 * 
 * 
 * cdef class NotEqualsFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":76
 * 
 * 
 * cdef class LessThanFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":80
 * 
 * 
 * cdef class GreaterThanFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":84
 * 
 * 
 * cdef class LessEqualsFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":88
 * 
 * 
 * cdef class GreaterEqualsFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":92
 * 
 * 
 * cdef class AddScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D;


/* "raysect/core/math/function/float/function1d/base.pxd":97
 * 
 * 
 * cdef class SubtractScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D;


/* "raysect/core/math/function/float/function1d/base.pxd":102
 * 
 * 
 * cdef class MultiplyScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D;


/* "raysect/core/math/function/float/function1d/base.pxd":107
 * 
 * 
 * cdef class DivideScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D;


/* "raysect/core/math/function/float/function1d/base.pxd":112
 * 
 * 
 * cdef class ModuloScalarFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":117
 * 
 * 
 * cdef class ModuloFunctionScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D;


/* "raysect/core/math/function/float/function1d/base.pxd":122
 * 
 * 
 * cdef class PowScalarFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D;


/* "raysect/core/math/function/float/function1d/base.pxd":127
 * 
 * 
 * cdef class PowFunctionScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D;


/* "raysect/core/math/function/float/function1d/base.pxd":132
 * 
 * 
 * cdef class EqualsScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D;


/* "raysect/core/math/function/float/function1d/base.pxd":137
 * 
 * 
 * cdef class NotEqualsScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D;


/* "raysect/core/math/function/float/function1d/base.pxd":142
 * 
 * 
 * cdef class LessThanScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D;


/* "raysect/core/math/function/float/function1d/base.pxd":147
 * 
 * 
 * cdef class GreaterThanScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D;


/* "raysect/core/math/function/float/function1d/base.pxd":152
 * 
 * 
 * cdef class LessEqualsScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D;


/* "raysect/core/math/function/float/function1d/base.pxd":157
 * 
 * 
 * cdef class GreaterEqualsScalar1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D;


/* "raysect/core/math/function/float/function1d/constant.pxd":35
 * 
 * 
 * cdef class Constant1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D;


/* "raysect/core/math/function/float/function1d/blend.pxd":35
 * 
 * 
 * cdef class Blend1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _f1, _f2, _mask
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D;


/* "raysect/core/math/function/float/function1d/autowrap.pxd":34
 * from raysect.core.math.function.float.function1d.base cimport Function1D
 * 
 * cdef class PythonFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D;


/* "raysect/core/math/function/float/function1d/arg.pxd":35
 * 
 * 
 * cdef class Arg1D(Function1D):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D;


/* "raysect/core/math/function/float/function1d/cmath.pxd":35
 * 
 * 
 * cdef class Exp1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D;


/* "raysect/core/math/function/float/function1d/cmath.pxd":39
 * 
 * 
 * cdef class Sin1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D;


/* "raysect/core/math/function/float/function1d/cmath.pxd":43
 * 
 * 
 * cdef class Cos1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D;


/* "raysect/core/math/function/float/function1d/cmath.pxd":47
 * 
 * 
 * cdef class Tan1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D;


/* "raysect/core/math/function/float/function1d/cmath.pxd":51
 * 
 * 
 * cdef class Asin1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D;


/* "raysect/core/math/function/float/function1d/cmath.pxd":55
 * 
 * 
 * cdef class Acos1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D;


/* "raysect/core/math/function/float/function1d/cmath.pxd":59
 * 
 * 
 * cdef class Atan1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D;


/* "raysect/core/math/function/float/function1d/cmath.pxd":63
 * 
 * 
 * cdef class Atan4Q1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _numerator, _denominator
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D;


/* "raysect/core/math/function/float/function1d/cmath.pxd":67
 * 
 * 
 * cdef class Sqrt1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D;


/* "raysect/core/math/function/float/function1d/cmath.pxd":71
 * 
 * 
 * cdef class Erf1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D;


/* "raysect/core/math/function/float/function1d/interpolate.pxd":36
 * 
 * 
 * cdef class Interpolator1DArray(Function1D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray;


/* "raysect/core/math/function/float/function1d/interpolate.pxd":47
 * 
 * 
 * cdef class _Interpolator1D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D {
  double (*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D *, double, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D;


/* "raysect/core/math/function/float/function1d/interpolate.pxd":57
 * 
 * 
 * cdef class _Interpolator1DLinear(_Interpolator1D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear;


/* "raysect/core/math/function/float/function1d/interpolate.pxd":61
 * 
 * 
 * cdef class _Interpolator1DCubic(_Interpolator1D):             # <<<<<<<<<<<<<<
 *     cdef:
 *         np.uint8_t[::1] _calculated
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic;


/* "raysect/core/math/function/float/function1d/interpolate.pxd":69
 * 
 * 
 * cdef class _Extrapolator1D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D {
  double (*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D *, double, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D;


/* "raysect/core/math/function/float/function1d/interpolate.pxd":78
 *     # cdef double _analytic_gradient(self, double px, int index, int order)
 * 
 * cdef class _Extrapolator1DNone(_Extrapolator1D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone;


/* "raysect/core/math/function/float/function1d/interpolate.pxd":82
 * 
 * 
 * cdef class _Extrapolator1DNearest(_Extrapolator1D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest;


/* "raysect/core/math/function/float/function1d/interpolate.pxd":86
 * 
 * 
 * cdef class _Extrapolator1DLinear(_Extrapolator1D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear;


/* "raysect/core/math/function/float/function1d/interpolate.pxd":90
 * 
 * 
 * cdef class _Extrapolator1DQuadratic(_Extrapolator1D):             # <<<<<<<<<<<<<<
 * 
 *     cdef double[3] _a_first, _a_last
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D __pyx_base;
  void (*_calculate_quadratic_coefficients_start)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic *, double, double, double, double *);
  void (*_calculate_quadratic_coefficients_end)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic *, double, double, double, double *);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic;


/* "raysect/core/math/function/float/function1d/interpolate.pxd":98
 * 
 * 
 * cdef class _ArrayDerivative1D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D {
  double (*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D *, int, int);
  double (*_rescale_lower_normalisation)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D *, double, double, double, double);
  double (*_evaluate_edge_x)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D *, int);
  double (*_evaluate_x)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D *, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D;


/* "raysect/core/math/function/float/function2d/base.pxd":36
 * 
 * 
 * cdef class Function2D(FloatFunction):             # <<<<<<<<<<<<<<
 * 
 *     cdef double evaluate(self, double x, double y) except? -1e999
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D {
  double (*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *, double, double);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D;


/* "raysect/core/math/function/float/function2d/base.pxd":41
 * 
 * 
 * cdef class AddFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":45
 * 
 * 
 * cdef class SubtractFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":49
 * 
 * 
 * cdef class MultiplyFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":53
 * 
 * 
 * cdef class DivideFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":57
 * 
 * 
 * cdef class ModuloFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":61
 * 
 * 
 * cdef class PowFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":65
 * 
 * 
 * cdef class AbsFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":69
 * 
 * 
 * cdef class EqualsFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":73
 * 
 * 
 * cdef class NotEqualsFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":77
 * 
 * 
 * cdef class LessThanFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":81
 * 
 * 
 * cdef class GreaterThanFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":85
 * 
 * 
 * cdef class LessEqualsFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":89
 * 
 * 
 * cdef class GreaterEqualsFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":93
 * 
 * 
 * cdef class AddScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D;


/* "raysect/core/math/function/float/function2d/base.pxd":98
 * 
 * 
 * cdef class SubtractScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D;


/* "raysect/core/math/function/float/function2d/base.pxd":103
 * 
 * 
 * cdef class MultiplyScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D;


/* "raysect/core/math/function/float/function2d/base.pxd":108
 * 
 * 
 * cdef class DivideScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D;


/* "raysect/core/math/function/float/function2d/base.pxd":113
 * 
 * 
 * cdef class ModuloScalarFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":118
 * 
 * 
 * cdef class ModuloFunctionScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D;


/* "raysect/core/math/function/float/function2d/base.pxd":123
 * 
 * 
 * cdef class PowScalarFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D;


/* "raysect/core/math/function/float/function2d/base.pxd":128
 * 
 * 
 * cdef class PowFunctionScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D;


/* "raysect/core/math/function/float/function2d/base.pxd":133
 * 
 * 
 * cdef class EqualsScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D;


/* "raysect/core/math/function/float/function2d/base.pxd":138
 * 
 * 
 * cdef class NotEqualsScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D;


/* "raysect/core/math/function/float/function2d/base.pxd":143
 * 
 * 
 * cdef class LessThanScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D;


/* "raysect/core/math/function/float/function2d/base.pxd":148
 * 
 * 
 * cdef class GreaterThanScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D;


/* "raysect/core/math/function/float/function2d/base.pxd":153
 * 
 * 
 * cdef class LessEqualsScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D;


/* "raysect/core/math/function/float/function2d/base.pxd":158
 * 
 * 
 * cdef class GreaterEqualsScalar2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D;


/* "raysect/core/math/function/float/function2d/constant.pxd":35
 * 
 * 
 * cdef class Constant2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D;


/* "raysect/core/math/function/float/function2d/blend.pxd":35
 * 
 * 
 * cdef class Blend2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _f1, _f2, _mask
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D;


/* "raysect/core/math/function/float/function2d/autowrap.pxd":34
 * from raysect.core.math.function.float.function2d.base cimport Function2D
 * 
 * cdef class PythonFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D;


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2dmesh.pxd":37
 * 
 * 
 * cdef class Interpolator2DMesh(Function2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh;


/* "raysect/core/math/function/float/function2d/interpolate/discrete2dmesh.pxd":37
 * 
 * 
 * cdef class Discrete2DMesh(Function2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh;


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":42
 * 
 * 
 * cdef class Interpolator2DArray(Function2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray;


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":56
 * 
 * 
 * cdef class _Interpolator2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D {
  double (*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D *, double, double, int, int);
  double (*analytic_gradient)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D *, double, double, int, int, int, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D;


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":67
 * 
 * 
 * cdef class _Interpolator2DLinear(_Interpolator2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef double _calculate_a0(self, int ix, int iy)
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D __pyx_base;
  double (*_calculate_a0)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear *, int, int);
  double (*_calculate_a1)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear *, int, int);
  double (*_calculate_a2)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear *, int, int);
  double (*_calculate_a3)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear *, int, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear;


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":75
 * 
 * 
 * cdef class _Interpolator2DCubic(_Interpolator2D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D __pyx_base;
  PyObject *(*_calc_coefficients)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic *, int, int, double (*)[4]);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic;


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":85
 * 
 * 
 * cdef class _Extrapolator2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D {
  double (*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D *, double, double, int, int);
  double (*_evaluate_edge_x)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D *, double, double, int, int, int);
  double (*_evaluate_edge_y)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D *, double, double, int, int, int);
  double (*_evaluate_edge_xy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D *, double, double, int, int, int, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D;


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":100
 * 
 * 
 * cdef class _Extrapolator2DNone(_Extrapolator2D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone;


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":104
 * 
 * 
 * cdef class _Extrapolator2DNearest(_Extrapolator2D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest;


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":108
 * 
 * 
 * cdef class _Extrapolator2DLinear(_Extrapolator2D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear;


/* "raysect/core/math/function/float/function2d/interpolate/interpolator2darray.pxd":112
 * 
 * 
 * cdef class _ArrayDerivative2D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D {
  double (*evaluate_df_dx)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *, int, int, int);
  double (*_derivative_dfdx_edge)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *, int, int);
  double (*_derivative_dfdx)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *, int, int);
  double (*evaluate_df_dy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *, int, int, int);
  double (*_derivative_dfdy_edge)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *, int, int);
  double (*_derivative_dfdy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *, int, int);
  double (*evaluate_d2f_dxdy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *, int, int, int, int);
  double (*_derivative_d2fdxdy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *, int, int);
  double (*_derivative_d2fdxdy_edge_xy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *, int, int);
  double (*_derivative_d2fdxdy_edge_x)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *, int, int);
  double (*_derivative_d2fdxdy_edge_y)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *, int, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D;


/* "raysect/core/math/function/float/function2d/arg.pxd":37
 *     X, Y
 * 
 * cdef class Arg2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef ArgLabel _argument
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D;


/* "raysect/core/math/function/float/function2d/cmath.pxd":35
 * 
 * 
 * cdef class Exp2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D;


/* "raysect/core/math/function/float/function2d/cmath.pxd":39
 * 
 * 
 * cdef class Sin2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D;


/* "raysect/core/math/function/float/function2d/cmath.pxd":43
 * 
 * 
 * cdef class Cos2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D;


/* "raysect/core/math/function/float/function2d/cmath.pxd":47
 * 
 * 
 * cdef class Tan2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D;


/* "raysect/core/math/function/float/function2d/cmath.pxd":51
 * 
 * 
 * cdef class Asin2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D;


/* "raysect/core/math/function/float/function2d/cmath.pxd":55
 * 
 * 
 * cdef class Acos2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D;


/* "raysect/core/math/function/float/function2d/cmath.pxd":59
 * 
 * 
 * cdef class Atan2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D;


/* "raysect/core/math/function/float/function2d/cmath.pxd":63
 * 
 * 
 * cdef class Atan4Q2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _numerator, _denominator
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D;


/* "raysect/core/math/function/float/function2d/cmath.pxd":67
 * 
 * 
 * cdef class Sqrt2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D;


/* "raysect/core/math/function/float/function2d/cmath.pxd":71
 * 
 * 
 * cdef class Erf2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D;


/* "raysect/core/math/function/float/function3d/base.pxd":36
 * 
 * 
 * cdef class Function3D(FloatFunction):             # <<<<<<<<<<<<<<
 *     cdef double evaluate(self, double x, double y, double z) except? -1e999
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D {
  double (*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *, double, double, double);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D;


/* "raysect/core/math/function/float/function3d/base.pxd":40
 * 
 * 
 * cdef class AddFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":44
 * 
 * 
 * cdef class SubtractFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":48
 * 
 * 
 * cdef class MultiplyFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":52
 * 
 * 
 * cdef class DivideFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":56
 * 
 * 
 * cdef class ModuloFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":60
 * 
 * 
 * cdef class PowFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":64
 * 
 * 
 * cdef class AbsFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":68
 * 
 * 
 * cdef class EqualsFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":72
 * 
 * 
 * cdef class NotEqualsFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":76
 * 
 * 
 * cdef class LessThanFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":80
 * 
 * 
 * cdef class GreaterThanFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":84
 * 
 * 
 * cdef class LessEqualsFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":88
 * 
 * 
 * cdef class GreaterEqualsFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":92
 * 
 * 
 * cdef class AddScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D;


/* "raysect/core/math/function/float/function3d/base.pxd":97
 * 
 * 
 * cdef class SubtractScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D;


/* "raysect/core/math/function/float/function3d/base.pxd":102
 * 
 * 
 * cdef class MultiplyScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D;


/* "raysect/core/math/function/float/function3d/base.pxd":107
 * 
 * 
 * cdef class DivideScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D;


/* "raysect/core/math/function/float/function3d/base.pxd":112
 * 
 * 
 * cdef class ModuloScalarFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":117
 * 
 * 
 * cdef class ModuloFunctionScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D;


/* "raysect/core/math/function/float/function3d/base.pxd":122
 * 
 * 
 * cdef class PowScalarFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D;


/* "raysect/core/math/function/float/function3d/base.pxd":127
 * 
 * 
 * cdef class PowFunctionScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D;


/* "raysect/core/math/function/float/function3d/base.pxd":132
 * 
 * 
 * cdef class EqualsScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D;


/* "raysect/core/math/function/float/function3d/base.pxd":137
 * 
 * 
 * cdef class NotEqualsScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D;


/* "raysect/core/math/function/float/function3d/base.pxd":142
 * 
 * 
 * cdef class LessThanScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D;


/* "raysect/core/math/function/float/function3d/base.pxd":147
 * 
 * 
 * cdef class GreaterThanScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D;


/* "raysect/core/math/function/float/function3d/base.pxd":152
 * 
 * 
 * cdef class LessEqualsScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D;


/* "raysect/core/math/function/float/function3d/base.pxd":157
 * 
 * 
 * cdef class GreaterEqualsScalar3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D;


/* "raysect/core/math/function/float/function3d/constant.pxd":35
 * 
 * 
 * cdef class Constant3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef double _value
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D;


/* "raysect/core/math/function/float/function3d/blend.pxd":35
 * 
 * 
 * cdef class Blend3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _f1, _f2, _mask
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D;


/* "raysect/core/math/function/float/function3d/autowrap.pxd":34
 * from raysect.core.math.function.float.function3d.base cimport Function3D
 * 
 * cdef class PythonFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D;


/* "raysect/core/math/spatial/kdtree3d.pxd":59
 * 
 * 
 * cdef class KDTree3DCore:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore {
  int32_t (*_build)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, PyObject *, struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, struct __pyx_opt_args_7raysect_4core_4math_7spatial_8kdtree3d_12KDTree3DCore__build *__pyx_optional_args);
  PyObject *(*_split)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, PyObject *, struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *);
  void (*_get_edges)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, PyObject *, int32_t, int32_t *, struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree3d_edge **);
  void (*_free_edges)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, struct __pyx_t_7raysect_4core_4math_7spatial_8kdtree3d_edge **);
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *(*_get_lower_bounds)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, double, int32_t);
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *(*_get_upper_bounds)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *, double, int32_t);
  int32_t (*_new_leaf)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, PyObject *);
  int32_t (*_new_branch)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, PyObject *, int32_t);
  int32_t (*_new_node)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *);
  int (*is_contained)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch);
  int (*_is_contained)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *);
  int (*_is_contained_node)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *);
  int (*_is_contained_branch)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *);
  int (*_is_contained_leaf)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *);
  int (*trace)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, struct __pyx_obj_7raysect_4core_3ray_Ray *, int __pyx_skip_dispatch);
  int (*_trace)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, struct __pyx_obj_7raysect_4core_3ray_Ray *);
  int (*_trace_node)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, int32_t, struct __pyx_obj_7raysect_4core_3ray_Ray *, double, double);
  int (*_trace_branch)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, int32_t, struct __pyx_obj_7raysect_4core_3ray_Ray *, double, double);
  int (*_trace_leaf)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, int32_t, struct __pyx_obj_7raysect_4core_3ray_Ray *, double);
  PyObject *(*items_containing)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch);
  PyObject *(*_items_containing)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *);
  PyObject *(*_items_containing_node)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *);
  PyObject *(*_items_containing_branch)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *);
  PyObject *(*_items_containing_leaf)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, int32_t, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *);
  void (*_reset)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *);
  double (*_read_double)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, PyObject *);
  int32_t (*_read_int32)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *, PyObject *);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore *__pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore;


/* "raysect/core/math/spatial/kdtree3d.pxd":126
 * 
 * 
 * cdef class KDTree3D(KDTree3DCore):             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _trace_leaf(self, int32_t id, Ray ray, double max_range)
 */

struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore __pyx_base;
  int (*_trace_items)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D *, PyObject *, struct __pyx_obj_7raysect_4core_3ray_Ray *, double, int __pyx_skip_dispatch);
  PyObject *(*_items_containing_items)(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D *, PyObject *, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D *__pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D;


/* "raysect/core/math/function/float/function3d/interpolate/common.pxd":37
 * 
 * 
 * cdef class MeshKDTree3D(KDTree3DCore):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore __pyx_base;
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *(*_generate_bounding_box)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D *, __pyx_t_5numpy_int32_t);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D;


/* "raysect/core/math/function/float/function3d/interpolate/discrete3dmesh.pxd":37
 * 
 * 
 * cdef class Discrete3DMesh(Function3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh;


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":41
 * 
 * 
 * cdef class Interpolator3DArray(Function3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray;


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":55
 * 
 * 
 * cdef class _Interpolator3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D {
  double (*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D *, double, double, double, int, int, int);
  double (*analytic_gradient)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D *, double, double, double, int, int, int, int, int, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D;


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":67
 * 
 * 
 * cdef class _Interpolator3DLinear(_Interpolator3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef double _calculate_a0(self, int ix, int iy, int iz)
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D __pyx_base;
  double (*_calculate_a0)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear *, int, int, int);
  double (*_calculate_a1)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear *, int, int, int);
  double (*_calculate_a2)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear *, int, int, int);
  double (*_calculate_a3)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear *, int, int, int);
  double (*_calculate_a4)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear *, int, int, int);
  double (*_calculate_a5)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear *, int, int, int);
  double (*_calculate_a6)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear *, int, int, int);
  double (*_calculate_a7)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear *, int, int, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear;


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":79
 * 
 * 
 * cdef class _Interpolator3DCubic(_Interpolator3D):             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D __pyx_base;
  PyObject *(*_cache_coefficients)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic *, int, int, int, double (*)[4][4]);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic;


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":89
 * 
 * 
 * cdef class _Extrapolator3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D {
  double (*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D *, double, double, double, int, int, int);
  double (*_evaluate_edge_x)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D *, double, double, double, int, int, int, int, int, int);
  double (*_evaluate_edge_y)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D *, double, double, double, int, int, int, int, int, int);
  double (*_evaluate_edge_z)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D *, double, double, double, int, int, int, int, int, int);
  double (*_evaluate_edge_xy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D *, double, double, double, int, int, int, int, int, int);
  double (*_evaluate_edge_xz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D *, double, double, double, int, int, int, int, int, int);
  double (*_evaluate_edge_yz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D *, double, double, double, int, int, int, int, int, int);
  double (*_evaluate_edge_xyz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D *, double, double, double, int, int, int, int, int, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D;


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":108
 * 
 * 
 * cdef class _Extrapolator3DNone(_Extrapolator3D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone;


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":112
 * 
 * 
 * cdef class _Extrapolator3DLinear(_Extrapolator3D):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear;


/* "raysect/core/math/function/float/function3d/interpolate/interpolator3darray.pxd":116
 * 
 * 
 * cdef class _ArrayDerivative3D:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D {
  double (*evaluate_df_dx)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int, int);
  double (*_derivative_dfdx)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_dfdx_edge)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*evaluate_df_dy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int, int);
  double (*_derivative_dfdy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_dfdy_edge)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*evaluate_df_dz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int, int);
  double (*_derivative_dfdz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_dfdz_edge)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*evaluate_d2fdxdy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int, int, int);
  double (*_derivative_d2fdxdy_edge_xy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d2fdxdy_edge_x)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d2fdxdy_edge_y)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d2fdxdy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*evaluate_d2fdxdz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int, int, int);
  double (*_derivative_d2fdxdz_edge_xz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d2fdxdz_edge_x)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d2fdxdz_edge_z)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d2fdxdz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*evaluate_d2fdydz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int, int, int);
  double (*_derivative_d2fdydz_edge_yz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d2fdydz_edge_y)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d2fdydz_edge_z)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d2fdydz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*evaluate_d3fdxdydz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int, int, int, int);
  double (*_derivative_d3fdxdydz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d3fdxdydz_edge_x)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d3fdxdydz_edge_y)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d3fdxdydz_edge_z)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d3fdxdydz_edge_xy)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d3fdxdydz_edge_xz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d3fdxdydz_edge_yz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
  double (*_derivative_d3fdxdydz_edge_xyz)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *, int, int, int);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D;


/* "raysect/core/math/function/float/function3d/arg.pxd":37
 *     X, Y, Z
 * 
 * cdef class Arg3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef ArgLabel _argument
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D;


/* "raysect/core/math/function/float/function3d/cmath.pxd":35
 * 
 * 
 * cdef class Exp3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D;


/* "raysect/core/math/function/float/function3d/cmath.pxd":39
 * 
 * 
 * cdef class Sin3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D;


/* "raysect/core/math/function/float/function3d/cmath.pxd":43
 * 
 * 
 * cdef class Cos3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D;


/* "raysect/core/math/function/float/function3d/cmath.pxd":47
 * 
 * 
 * cdef class Tan3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D;


/* "raysect/core/math/function/float/function3d/cmath.pxd":51
 * 
 * 
 * cdef class Asin3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D;


/* "raysect/core/math/function/float/function3d/cmath.pxd":55
 * 
 * 
 * cdef class Acos3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D;


/* "raysect/core/math/function/float/function3d/cmath.pxd":59
 * 
 * 
 * cdef class Atan3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D;


/* "raysect/core/math/function/float/function3d/cmath.pxd":63
 * 
 * 
 * cdef class Atan4Q3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _numerator, _denominator
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D;


/* "raysect/core/math/function/float/function3d/cmath.pxd":67
 * 
 * 
 * cdef class Sqrt3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D;


/* "raysect/core/math/function/float/function3d/cmath.pxd":71
 * 
 * 
 * cdef class Erf3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D;


/* "raysect/core/math/function/vector3d/function1d/base.pxd":38
 * 
 * 
 * cdef class Function1D(Vector3DFunction):             # <<<<<<<<<<<<<<
 *     cdef Vector3D evaluate(self, double x)
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D {
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *, double);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D;


/* "raysect/core/math/function/vector3d/function1d/base.pxd":42
 * 
 * 
 * cdef class AddFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D;


/* "raysect/core/math/function/vector3d/function1d/base.pxd":46
 * 
 * 
 * cdef class SubtractFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D;


/* "raysect/core/math/function/vector3d/function1d/base.pxd":50
 * 
 * 
 * cdef class MultiplyFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1
 *     cdef FloatFunction1D _function2
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D;


/* "raysect/core/math/function/vector3d/function1d/base.pxd":55
 * 
 * 
 * cdef class DivideFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1
 *     cdef FloatFunction1D _function2
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D;


/* "raysect/core/math/function/vector3d/function1d/base.pxd":60
 * 
 * 
 * cdef class NegFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D;


/* "raysect/core/math/function/vector3d/function1d/base.pxd":64
 * 
 * 
 * cdef class EqualsFunction1D(FloatFunction1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D;


/* "raysect/core/math/function/vector3d/function1d/base.pxd":68
 * 
 * 
 * cdef class NotEqualsFunction1D(FloatFunction1D):             # <<<<<<<<<<<<<<
 *     cdef Function1D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D;


/* "raysect/core/math/function/vector3d/function1d/constant.pxd":36
 * 
 * 
 * cdef class Constant1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef Vector3D _value
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D;


/* "raysect/core/math/function/vector3d/function1d/blend.pxd":36
 * 
 * 
 * cdef class Blend1D(VecFn1D):             # <<<<<<<<<<<<<<
 *     cdef VecFn1D _f1, _f2
 *     cdef FltFn1D _mask
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D;


/* "raysect/core/math/function/vector3d/function1d/autowrap.pxd":34
 * from raysect.core.math.function.vector3d.function1d.base cimport Function1D
 * 
 * cdef class PythonFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D;


/* "raysect/core/math/function/vector3d/function1d/utility.pxd":36
 * 
 * 
 * cdef class FloatToVector3DFunction1D(Function1D):             # <<<<<<<<<<<<<<
 *     cdef FloatFunction1D _x, _y, _z
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D;


/* "raysect/core/math/function/vector3d/function2d/base.pxd":38
 * 
 * 
 * cdef class Function2D(Vector3DFunction):             # <<<<<<<<<<<<<<
 *     cdef Vector3D evaluate(self, double x, double y)
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D {
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *, double, double);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D;


/* "raysect/core/math/function/vector3d/function2d/base.pxd":42
 * 
 * 
 * cdef class AddFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D;


/* "raysect/core/math/function/vector3d/function2d/base.pxd":46
 * 
 * 
 * cdef class SubtractFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D;


/* "raysect/core/math/function/vector3d/function2d/base.pxd":50
 * 
 * 
 * cdef class MultiplyFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1
 *     cdef FloatFunction2D _function2
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D;


/* "raysect/core/math/function/vector3d/function2d/base.pxd":55
 * 
 * 
 * cdef class DivideFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1
 *     cdef FloatFunction2D _function2
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D;


/* "raysect/core/math/function/vector3d/function2d/base.pxd":60
 * 
 * 
 * cdef class NegFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D;


/* "raysect/core/math/function/vector3d/function2d/base.pxd":64
 * 
 * 
 * cdef class EqualsFunction2D(FloatFunction2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D;


/* "raysect/core/math/function/vector3d/function2d/base.pxd":68
 * 
 * 
 * cdef class NotEqualsFunction2D(FloatFunction2D):             # <<<<<<<<<<<<<<
 *     cdef Function2D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D;


/* "raysect/core/math/function/vector3d/function2d/constant.pxd":36
 * 
 * 
 * cdef class Constant2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef Vector3D _value
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D;


/* "raysect/core/math/function/vector3d/function2d/blend.pxd":36
 * 
 * 
 * cdef class Blend2D(VecFn2D):             # <<<<<<<<<<<<<<
 *     cdef VecFn2D _f1, _f2
 *     cdef FltFn2D _mask
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D;


/* "raysect/core/math/function/vector3d/function2d/autowrap.pxd":34
 * from raysect.core.math.function.vector3d.function2d.base cimport Function2D
 * 
 * cdef class PythonFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D;


/* "raysect/core/math/function/vector3d/function2d/utility.pxd":36
 * 
 * 
 * cdef class FloatToVector3DFunction2D(Function2D):             # <<<<<<<<<<<<<<
 *     cdef FloatFunction2D _x, _y, _z
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D;


/* "raysect/core/math/function/vector3d/function3d/base.pxd":38
 * 
 * 
 * cdef class Function3D(Vector3DFunction):             # <<<<<<<<<<<<<<
 *     cdef Vector3D evaluate(self, double x, double y, double z)
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D {
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*evaluate)(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *, double, double, double);
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D;


/* "raysect/core/math/function/vector3d/function3d/base.pxd":42
 * 
 * 
 * cdef class AddFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D;


/* "raysect/core/math/function/vector3d/function3d/base.pxd":46
 * 
 * 
 * cdef class SubtractFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D;


/* "raysect/core/math/function/vector3d/function3d/base.pxd":50
 * 
 * 
 * cdef class MultiplyFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1
 *     cdef FloatFunction3D _function2
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D;


/* "raysect/core/math/function/vector3d/function3d/base.pxd":55
 * 
 * 
 * cdef class DivideFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1
 *     cdef FloatFunction3D _function2
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D;


/* "raysect/core/math/function/vector3d/function3d/base.pxd":60
 * 
 * 
 * cdef class NegFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D;


/* "raysect/core/math/function/vector3d/function3d/base.pxd":64
 * 
 * 
 * cdef class EqualsFunction3D(FloatFunction3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D;


/* "raysect/core/math/function/vector3d/function3d/base.pxd":68
 * 
 * 
 * cdef class NotEqualsFunction3D(FloatFunction3D):             # <<<<<<<<<<<<<<
 *     cdef Function3D _function1, _function2
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D;


/* "raysect/core/math/function/vector3d/function3d/constant.pxd":36
 * 
 * 
 * cdef class Constant3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef Vector3D _value
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D;


/* "raysect/core/math/function/vector3d/function3d/blend.pxd":36
 * 
 * 
 * cdef class Blend3D(VecFn3D):             # <<<<<<<<<<<<<<
 *     cdef VecFn3D _f1, _f2
 *     cdef FltFn3D _mask
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D;


/* "raysect/core/math/function/vector3d/function3d/autowrap.pxd":34
 * from raysect.core.math.function.vector3d.function3d.base cimport Function3D
 * 
 * cdef class PythonFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef public object function
 * 
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D;


/* "raysect/core/math/function/vector3d/function3d/utility.pxd":36
 * 
 * 
 * cdef class FloatToVector3DFunction3D(Function3D):             # <<<<<<<<<<<<<<
 *     cdef FloatFunction3D _x, _y, _z
 */

struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_base;
};
static struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D;


/* "cherab/solps/solps_2d_functions.pyx":32
 * 
 * 
 * cdef class SOLPSFunction2D(Function2D):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, object vertex_coords not None, object triangles not None, object triangle_to_grid_map not None, object grid_data not None):
 */

struct __pyx_vtabstruct_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_vtabptr_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D;


/* "cherab/solps/solps_2d_functions.pyx":117
 *         return 0.0
 * 
 * cdef class SOLPSVectorFunction2D(VectorFunction2D):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, object vertex_coords not None, object triangles not None, object triangle_to_grid_map not None, object grid_vectors not None):
 */

struct __pyx_vtabstruct_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D {
  struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_base;
};
static struct __pyx_vtabstruct_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_vtabptr_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D;


/* "View.MemoryView":109
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "View.MemoryView":341
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview:             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
  PyObject *(*_get_base)(struct __pyx_memoryview_obj *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":975
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) (&args[nargs])
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, nargs, NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* tp_new.proto */
#define __Pyx_tp_new(type_obj, args) __Pyx_tp_new_kwargs(type_obj, args, NULL)
static CYTHON_INLINE PyObject* __Pyx_tp_new_kwargs(PyObject* type_obj, PyObject* args, PyObject* kwargs) {
    return (PyObject*) (((PyTypeObject*)type_obj)->tp_new((PyTypeObject*)type_obj, args, kwargs));
}

/* CallableCheck.proto */
#if CYTHON_USE_TYPE_SLOTS && PY_MAJOR_VERSION >= 3
#define __Pyx_PyCallable_Check(obj)   ((obj)->ob_type->tp_call != NULL)
#else
#define __Pyx_PyCallable_Check(obj)   PyCallable_Check(obj)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* None.proto */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t, Py_ssize_t);

/* UnaryNegOverflows.proto */
#define __Pyx_UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);

/* ssize_strlen.proto */
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_Occurred(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* None.proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#if CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, CYTHON_UNUSED PyObject *typ) {
    if (!self) {
        Py_INCREF(func);
        return func;
    }
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, Py_ssize_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_CyFunctionType = 0;
#endif
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func.vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* ClassMethod.proto */
#include "descrobject.h"
static CYTHON_UNUSED PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5numpy_int32_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_double(PyObject *, int writable_flag);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str);
#else
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);
#endif

/* #### Code section: module_declarations ### */
static double __pyx_f_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_evaluate(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_self, double __pyx_v_x, double __pyx_v_y); /* proto*/
static struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *__pyx_f_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_evaluate(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_self, double __pyx_v_x, double __pyx_v_y); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self); /* proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview__get_base(struct __pyx_memoryview_obj *__pyx_v_self); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice__get_base(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto*/

/* Module declarations from "libc.string" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "libc.stdio" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "__builtin__" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "cpython.type" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;
#endif

/* Module declarations from "cpython" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "cpython.object" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "cpython.ref" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "numpy" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "numpy" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
#endif

/* Module declarations from "raysect.core.math._vec3" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_5_vec3__Vec3 = 0;
#endif

/* Module declarations from "raysect.core.math._mat4" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_5_mat4__Mat4 = 0;
#endif

/* Module declarations from "raysect.core.math.affinematrix" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_12affinematrix_AffineMatrix3D = 0;
#endif
static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *__pyx_f_7raysect_4core_4math_12affinematrix_new_affinematrix3d(double, double, double, double, double, double, double, double, double, double, double, double, double, double, double, double); /*proto*/

/* Module declarations from "raysect.core.math.vector" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_6vector_Vector3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_6vector_Vector2D = 0;
#endif
static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *__pyx_f_7raysect_4core_4math_6vector_new_vector3d(double, double, double); /*proto*/
static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *__pyx_f_7raysect_4core_4math_6vector_new_vector2d(double, double); /*proto*/

/* Module declarations from "raysect.core.math.point" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_5point_Point3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_5point_Point2D = 0;
#endif
static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_5point_Point3D *__pyx_f_7raysect_4core_4math_5point_new_point3d(double, double, double); /*proto*/
static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_5point_Point2D *__pyx_f_7raysect_4core_4math_5point_new_point2d(double, double); /*proto*/

/* Module declarations from "raysect.core.math.normal" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_6normal_Normal3D = 0;
#endif
static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *__pyx_f_7raysect_4core_4math_6normal_new_normal3d(double, double, double); /*proto*/

/* Module declarations from "raysect.core.math.quaternion" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_10quaternion_Quaternion = 0;
#endif
static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *__pyx_f_7raysect_4core_4math_10quaternion_new_quaternion(double, double, double, double); /*proto*/
static struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*__pyx_f_7raysect_4core_4math_10quaternion_new_quaternion_from_matrix)(struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *); /*proto*/
static struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(*__pyx_f_7raysect_4core_4math_10quaternion_new_quaternion_from_axis_angle)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, double); /*proto*/

/* Module declarations from "raysect.core.math.transform" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif
static struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*__pyx_f_7raysect_4core_4math_9transform_translate)(double, double, double, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*__pyx_f_7raysect_4core_4math_9transform_rotate_x)(double, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*__pyx_f_7raysect_4core_4math_9transform_rotate_y)(double, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*__pyx_f_7raysect_4core_4math_9transform_rotate_z)(double, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*__pyx_f_7raysect_4core_4math_9transform_rotate_vector)(double, struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*__pyx_f_7raysect_4core_4math_9transform_rotate)(double, double, double, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(*__pyx_f_7raysect_4core_4math_9transform_rotate_basis)(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, int __pyx_skip_dispatch); /*proto*/
static PyObject *(*__pyx_f_7raysect_4core_4math_9transform_to_cylindrical)(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_7raysect_4core_4math_5point_Point3D *(*__pyx_f_7raysect_4core_4math_9transform_from_cylindrical)(double, double, double, int __pyx_skip_dispatch); /*proto*/
static PyObject *(*__pyx_f_7raysect_4core_4math_9transform_extract_rotation)(struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, int __pyx_skip_dispatch, struct __pyx_opt_args_7raysect_4core_4math_9transform_extract_rotation *__pyx_optional_args); /*proto*/
static PyObject *(*__pyx_f_7raysect_4core_4math_9transform_extract_translation)(struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, int __pyx_skip_dispatch); /*proto*/

/* Module declarations from "raysect.core.math.units" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.statsarray" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_10statsarray_StatsBin = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray3D = 0;
#endif

/* Module declarations from "libc.math" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.random" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif
static double (*__pyx_f_7raysect_4core_4math_6random_uniform)(int __pyx_skip_dispatch); /*proto*/

/* Module declarations from "cython.view" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "cython" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.cython.utility" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.cython.transform" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.cython.triangle" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif
static void (*__pyx_f_7raysect_4core_4math_6cython_8triangle_barycentric_coords)(double, double, double, double, double, double, double, double, double *, double *, double *); /*proto*/
static double (*__pyx_f_7raysect_4core_4math_6cython_8triangle_barycentric_interpolation)(double, double, double, double, double, double); /*proto*/

/* Module declarations from "raysect.core.math.cython.tetrahedra" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.cython" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.sampler.solidangle" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SphereSampler = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler = 0;
#endif

/* Module declarations from "raysect.core.math.sampler.surface3d" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D = 0;
#endif

/* Module declarations from "raysect.core.math.sampler.targetted" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9targetted__TargettedSampler = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler = 0;
#endif

/* Module declarations from "raysect.core.math.sampler" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.polygon" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif
static PyArrayObject *(*__pyx_f_7raysect_4core_4math_7polygon_triangulate2d)(PyArrayObject *, int __pyx_skip_dispatch); /*proto*/

/* Module declarations from "raysect.core.math" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.ray" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_3ray_Ray = 0;
#endif

/* Module declarations from "raysect.core.material" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_8material_Material = 0;
#endif

/* Module declarations from "raysect.core.scenegraph._nodebase" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_9_nodebase__NodeBase = 0;
#endif

/* Module declarations from "raysect.core.scenegraph.node" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_4node_Node = 0;
#endif

/* Module declarations from "raysect.core.scenegraph.signal" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_6signal_ChangeSignal = 0;
#endif

/* Module declarations from "raysect.core.scenegraph.primitive" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_9primitive_Primitive = 0;
#endif

/* Module declarations from "raysect.core.intersection" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_12intersection_Intersection = 0;
#endif

/* Module declarations from "raysect.core.acceleration.accelerator" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_12acceleration_11accelerator_Accelerator = 0;
#endif

/* Module declarations from "raysect.core.scenegraph.world" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_5world_World = 0;
#endif

/* Module declarations from "raysect.core.scenegraph.observer" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_8observer_Observer = 0;
#endif

/* Module declarations from "raysect.core.scenegraph.utility" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_7utility_BridgeNode = 0;
#endif

/* Module declarations from "raysect.core.scenegraph" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.containers" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_10containers__Item = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_10containers_LinkedList = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_10containers_Stack = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_10containers_Queue = 0;
#endif

/* Module declarations from "raysect.core" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.boundingsphere" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_14boundingsphere_BoundingSphere3D = 0;
#endif

/* Module declarations from "raysect.core.boundingbox" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox2D = 0;
#endif

/* Module declarations from "libc.stdint" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.spatial.kdtree2d" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_Item2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function2d.interpolate.common" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.base" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_4base_Function = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.base" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_4base_FloatFunction = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function1d.base" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function1d.constant" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function1d.blend" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function1d.autowrap" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D = 0;
#endif
static struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *(*__pyx_f_7raysect_4core_4math_8function_5float_10function1d_8autowrap_autowrap_function1d)(PyObject *); /*proto*/

/* Module declarations from "raysect.core.math.function.float.function1d.arg" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function1d.cmath" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function1d.interpolate" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function1d" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.function.float.function2d.base" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function2d.constant" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function2d.blend" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function2d.autowrap" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D = 0;
#endif
static struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *(*__pyx_f_7raysect_4core_4math_8function_5float_10function2d_8autowrap_autowrap_function2d)(PyObject *); /*proto*/

/* Module declarations from "raysect.core.math.function.float.function2d.interpolate.interpolator2dmesh" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function2d.interpolate.discrete2dmesh" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function2d.interpolate.interpolator2darray" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function2d.interpolate" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.function.float.function2d.arg" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function2d.cmath" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function2d" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.function.float.function3d.base" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function3d.constant" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function3d.blend" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function3d.autowrap" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D = 0;
#endif
static struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *(*__pyx_f_7raysect_4core_4math_8function_5float_10function3d_8autowrap_autowrap_function3d)(PyObject *); /*proto*/

/* Module declarations from "raysect.core.math.spatial.kdtree3d" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_Item3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function3d.interpolate.common" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function3d.interpolate.discrete3dmesh" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function3d.interpolate.interpolator3darray" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function3d.interpolate" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.function.float.function3d.arg" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function3d.cmath" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.float.function3d" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.function.float" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.function.vector3d.base" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function1d.base" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function1d.constant" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function1d.blend" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function1d.autowrap" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D = 0;
#endif
static struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *(*__pyx_f_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_autowrap_function1d)(PyObject *); /*proto*/

/* Module declarations from "raysect.core.math.function.vector3d.function1d.utility" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function1d" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function2d.base" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function2d.constant" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function2d.blend" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function2d.autowrap" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D = 0;
#endif
static struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *(*__pyx_f_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_autowrap_function2d)(PyObject *); /*proto*/

/* Module declarations from "raysect.core.math.function.vector3d.function2d.utility" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function2d" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function3d.base" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D = 0;
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function3d.constant" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function3d.blend" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function3d.autowrap" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D = 0;
#endif
static struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *(*__pyx_f_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_autowrap_function3d)(PyObject *); /*proto*/

/* Module declarations from "raysect.core.math.function.vector3d.function3d.utility" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D = 0;
#endif

/* Module declarations from "raysect.core.math.function.vector3d.function3d" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "raysect.core.math.function.vector3d" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "cherab.core.math.function" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "cherab.solps.solps_2d_functions" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D = 0;
static PyTypeObject *__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D = 0;
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
#endif
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static int __pyx_array_allocate_buffer(struct __pyx_array_obj *); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, PyObject *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, PyObject *); /*proto*/
static int __pyx_memoryview_err_no_memory(void); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t = { "int32_t", NULL, sizeof(__pyx_t_5numpy_int32_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int32_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int32_t), 0 };
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "cherab.solps.solps_2d_functions"
extern int __pyx_module_is_main_cherab__solps__solps_2d_functions;
int __pyx_module_is_main_cherab__solps__solps_2d_functions = 0;

/* Implementation of "cherab.solps.solps_2d_functions" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_IndexError;
/* #### Code section: string_decls ### */
static const char __pyx_k_O[] = "O";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_m[] = "m";
static const char __pyx_k__3[] = ": ";
static const char __pyx_k__4[] = ".";
static const char __pyx_k__5[] = "*";
static const char __pyx_k__8[] = "'";
static const char __pyx_k__9[] = ")";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_and[] = " and ";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_got[] = " (got ";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_instance[] = "instance";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_grid_data[] = "grid_data";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_triangles[] = "triangles";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_grid_vectors[] = "grid_vectors";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_vertex_coords[] = "vertex_coords";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_SOLPSFunction2D[] = "SOLPSFunction2D";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_triangle_to_grid_map[] = "triangle_to_grid_map";
static const char __pyx_k_Invalid_shape_in_axis[] = "Invalid shape in axis ";
static const char __pyx_k_SOLPSVectorFunction2D[] = "SOLPSVectorFunction2D";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_Cannot_index_with_type[] = "Cannot index with type '";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_SOLPSFunction2D___reduce[] = "SOLPSFunction2D.__reduce__";
static const char __pyx_k_SOLPSFunction2D_instance[] = "SOLPSFunction2D.instance";
static const char __pyx_k_Dimension_d_is_not_direct[] = "Dimension %d is not direct";
static const char __pyx_k_Index_out_of_bounds_axis_d[] = "Index out of bounds (axis %d)";
static const char __pyx_k_SOLPSFunction2D___getstate[] = "SOLPSFunction2D.__getstate__";
static const char __pyx_k_SOLPSFunction2D___setstate[] = "SOLPSFunction2D.__setstate__";
static const char __pyx_k_Step_may_not_be_zero_axis_d[] = "Step may not be zero (axis %d)";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_SOLPSVectorFunction2D___reduce[] = "SOLPSVectorFunction2D.__reduce__";
static const char __pyx_k_SOLPSVectorFunction2D_instance[] = "SOLPSVectorFunction2D.instance";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_cherab_solps_solps_2d_functions[] = "cherab/solps/solps_2d_functions.pyx";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_All_dimensions_preceding_dimensi[] = "All dimensions preceding dimension %d must be indexed and not sliced";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Cannot_transpose_memoryview_with[] = "Cannot transpose memoryview with indirect dimensions";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Incompatible_checksums_s_vs_0x6a[] = "Incompatible checksums (%s vs 0x6ae9995 = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got ";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis ";
static const char __pyx_k_SOLPSVectorFunction2D___getstate[] = "SOLPSVectorFunction2D.__getstate__";
static const char __pyx_k_SOLPSVectorFunction2D___setstate[] = "SOLPSVectorFunction2D.__setstate__";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension ";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static const char __pyx_k_cherab_solps_solps_2d_functions_2[] = "cherab.solps.solps_2d_functions";
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_kp_s_All_dimensions_preceding_dimensi;
static PyObject *__pyx_n_s_AssertionError;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_u_Cannot_index_with_type;
static PyObject *__pyx_kp_s_Cannot_transpose_memoryview_with;
static PyObject *__pyx_kp_s_Dimension_d_is_not_direct;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x6a;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_s_Index_out_of_bounds_axis_d;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_kp_u_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_u_Invalid_shape_in_axis;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_kp_u_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_SOLPSFunction2D;
static PyObject *__pyx_n_s_SOLPSFunction2D___getstate;
static PyObject *__pyx_n_s_SOLPSFunction2D___reduce;
static PyObject *__pyx_n_s_SOLPSFunction2D___setstate;
static PyObject *__pyx_n_s_SOLPSFunction2D_instance;
static PyObject *__pyx_n_s_SOLPSVectorFunction2D;
static PyObject *__pyx_n_s_SOLPSVectorFunction2D___getstate;
static PyObject *__pyx_n_s_SOLPSVectorFunction2D___reduce;
static PyObject *__pyx_n_s_SOLPSVectorFunction2D___setstate;
static PyObject *__pyx_n_s_SOLPSVectorFunction2D_instance;
static PyObject *__pyx_kp_s_Step_may_not_be_zero_axis_d;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_kp_u__3;
static PyObject *__pyx_kp_u__4;
static PyObject *__pyx_n_s__5;
static PyObject *__pyx_kp_u__8;
static PyObject *__pyx_kp_u__9;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_kp_u_and;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_kp_s_cherab_solps_solps_2d_functions;
static PyObject *__pyx_n_s_cherab_solps_solps_2d_functions_2;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_class_getitem;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_cls;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_u_got;
static PyObject *__pyx_kp_u_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_grid_data;
static PyObject *__pyx_n_s_grid_vectors;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_initializing;
static PyObject *__pyx_n_s_instance;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_m;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_spec;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_triangle_to_grid_map;
static PyObject *__pyx_n_s_triangles;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_vertex_coords;
#endif
/* #### Code section: decls ### */
static int __pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D___init__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_self, PyObject *__pyx_v_vertex_coords, PyObject *__pyx_v_triangles, PyObject *__pyx_v_triangle_to_grid_map, PyObject *__pyx_v_grid_data); /* proto */
static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_2__getstate__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_4__setstate__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_6__reduce__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_8instance(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_instance, PyObject *__pyx_v_grid_data); /* proto */
static int __pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D___init__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_self, PyObject *__pyx_v_vertex_coords, PyObject *__pyx_v_triangles, PyObject *__pyx_v_triangle_to_grid_map, PyObject *__pyx_v_grid_vectors); /* proto */
static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_2__getstate__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_4__setstate__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_6__reduce__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_8instance(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_instance, PyObject *__pyx_v_grid_vectors); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_112105877;
static PyObject *__pyx_int_neg_1;
#endif
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__pyx_tuple_;
static PyObject *__pyx_slice__7;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_codeobj__11;
static PyObject *__pyx_codeobj__13;
static PyObject *__pyx_codeobj__15;
static PyObject *__pyx_codeobj__17;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__32;
#endif
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyTypeObject *__pyx_ptype_5numpy_dtype;
  PyTypeObject *__pyx_ptype_5numpy_flatiter;
  PyTypeObject *__pyx_ptype_5numpy_broadcast;
  PyTypeObject *__pyx_ptype_5numpy_ndarray;
  PyTypeObject *__pyx_ptype_5numpy_ufunc;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_5_vec3__Vec3;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_5_mat4__Mat4;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_12affinematrix_AffineMatrix3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_6vector_Vector3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_6vector_Vector2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_5point_Point3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_5point_Point2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_6normal_Normal3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_10quaternion_Quaternion;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_10statsarray_StatsBin;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SphereSampler;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9targetted__TargettedSampler;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler;
  PyTypeObject *__pyx_ptype_7raysect_4core_3ray_Ray;
  PyTypeObject *__pyx_ptype_7raysect_4core_8material_Material;
  PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_9_nodebase__NodeBase;
  PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_4node_Node;
  PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_6signal_ChangeSignal;
  PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_9primitive_Primitive;
  PyTypeObject *__pyx_ptype_7raysect_4core_12intersection_Intersection;
  PyTypeObject *__pyx_ptype_7raysect_4core_12acceleration_11accelerator_Accelerator;
  PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_5world_World;
  PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_8observer_Observer;
  PyTypeObject *__pyx_ptype_7raysect_4core_10scenegraph_7utility_BridgeNode;
  PyTypeObject *__pyx_ptype_7raysect_4core_10containers__Item;
  PyTypeObject *__pyx_ptype_7raysect_4core_10containers_LinkedList;
  PyTypeObject *__pyx_ptype_7raysect_4core_10containers_Stack;
  PyTypeObject *__pyx_ptype_7raysect_4core_10containers_Queue;
  PyTypeObject *__pyx_ptype_7raysect_4core_14boundingsphere_BoundingSphere3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_Item2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_4base_Function;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_4base_FloatFunction;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_Item3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D;
  PyTypeObject *__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D;
  PyTypeObject *__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D;
  PyObject *__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D;
  PyTypeObject *__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D;
  PyObject *__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D;
  PyTypeObject *__pyx_array_type;
  PyObject *__pyx_type___pyx_array;
  PyTypeObject *__pyx_MemviewEnum_type;
  PyObject *__pyx_type___pyx_MemviewEnum;
  PyTypeObject *__pyx_memoryview_type;
  PyObject *__pyx_type___pyx_memoryview;
  PyTypeObject *__pyx_memoryviewslice_type;
  PyObject *__pyx_type___pyx_memoryviewslice;
  PyObject *__pyx_n_s_ASCII;
  PyObject *__pyx_kp_s_All_dimensions_preceding_dimensi;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
  PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
  PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
  PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
  PyObject *__pyx_kp_u_Cannot_index_with_type;
  PyObject *__pyx_kp_s_Cannot_transpose_memoryview_with;
  PyObject *__pyx_kp_s_Dimension_d_is_not_direct;
  PyObject *__pyx_n_s_Ellipsis;
  PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
  PyObject *__pyx_n_s_ImportError;
  PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x6a;
  PyObject *__pyx_n_s_IndexError;
  PyObject *__pyx_kp_s_Index_out_of_bounds_axis_d;
  PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
  PyObject *__pyx_kp_u_Invalid_mode_expected_c_or_fortr;
  PyObject *__pyx_kp_u_Invalid_shape_in_axis;
  PyObject *__pyx_n_s_MemoryError;
  PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
  PyObject *__pyx_kp_s_MemoryView_of_r_object;
  PyObject *__pyx_n_b_O;
  PyObject *__pyx_kp_u_Out_of_bounds_on_buffer_access_a;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_n_s_SOLPSFunction2D;
  PyObject *__pyx_n_s_SOLPSFunction2D___getstate;
  PyObject *__pyx_n_s_SOLPSFunction2D___reduce;
  PyObject *__pyx_n_s_SOLPSFunction2D___setstate;
  PyObject *__pyx_n_s_SOLPSFunction2D_instance;
  PyObject *__pyx_n_s_SOLPSVectorFunction2D;
  PyObject *__pyx_n_s_SOLPSVectorFunction2D___getstate;
  PyObject *__pyx_n_s_SOLPSVectorFunction2D___reduce;
  PyObject *__pyx_n_s_SOLPSVectorFunction2D___setstate;
  PyObject *__pyx_n_s_SOLPSVectorFunction2D_instance;
  PyObject *__pyx_kp_s_Step_may_not_be_zero_axis_d;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_n_s_View_MemoryView;
  PyObject *__pyx_kp_u__3;
  PyObject *__pyx_kp_u__4;
  PyObject *__pyx_n_s__5;
  PyObject *__pyx_kp_u__8;
  PyObject *__pyx_kp_u__9;
  PyObject *__pyx_n_s_allocate_buffer;
  PyObject *__pyx_kp_u_and;
  PyObject *__pyx_n_s_array;
  PyObject *__pyx_n_s_base;
  PyObject *__pyx_n_s_c;
  PyObject *__pyx_n_u_c;
  PyObject *__pyx_kp_s_cherab_solps_solps_2d_functions;
  PyObject *__pyx_n_s_cherab_solps_solps_2d_functions_2;
  PyObject *__pyx_n_s_class;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_cls;
  PyObject *__pyx_kp_s_contiguous_and_direct;
  PyObject *__pyx_kp_s_contiguous_and_indirect;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dtype;
  PyObject *__pyx_n_s_dtype_is_object;
  PyObject *__pyx_n_s_encode;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_error;
  PyObject *__pyx_n_s_flags;
  PyObject *__pyx_n_s_float64;
  PyObject *__pyx_n_s_format;
  PyObject *__pyx_n_s_fortran;
  PyObject *__pyx_n_u_fortran;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_kp_u_got;
  PyObject *__pyx_kp_u_got_differing_extents_in_dimensi;
  PyObject *__pyx_n_s_grid_data;
  PyObject *__pyx_n_s_grid_vectors;
  PyObject *__pyx_n_s_id;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_instance;
  PyObject *__pyx_n_s_int32;
  PyObject *__pyx_n_s_itemsize;
  PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
  PyObject *__pyx_n_s_m;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_memview;
  PyObject *__pyx_n_s_mode;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_ndim;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
  PyObject *__pyx_n_s_np;
  PyObject *__pyx_n_s_numpy;
  PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
  PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
  PyObject *__pyx_n_s_obj;
  PyObject *__pyx_n_s_pack;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_getbuffer;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_Enum;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_shape;
  PyObject *__pyx_n_s_size;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_start;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_step;
  PyObject *__pyx_n_s_stop;
  PyObject *__pyx_kp_s_strided_and_direct;
  PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
  PyObject *__pyx_kp_s_strided_and_indirect;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_struct;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_triangle_to_grid_map;
  PyObject *__pyx_n_s_triangles;
  PyObject *__pyx_kp_s_unable_to_allocate_array_data;
  PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
  PyObject *__pyx_n_s_unpack;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_vertex_coords;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_112105877;
  PyObject *__pyx_int_neg_1;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_slice__7;
  PyObject *__pyx_tuple__2;
  PyObject *__pyx_tuple__6;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__16;
  PyObject *__pyx_tuple__18;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__27;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__29;
  PyObject *__pyx_tuple__30;
  PyObject *__pyx_tuple__31;
  PyObject *__pyx_codeobj__11;
  PyObject *__pyx_codeobj__13;
  PyObject *__pyx_codeobj__15;
  PyObject *__pyx_codeobj__17;
  PyObject *__pyx_codeobj__19;
  PyObject *__pyx_codeobj__21;
  PyObject *__pyx_codeobj__23;
  PyObject *__pyx_codeobj__25;
  PyObject *__pyx_codeobj__32;
} __pyx_mstate;

#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_dtype);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flatiter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_broadcast);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ndarray);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ufunc);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_5_vec3__Vec3);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_5_mat4__Mat4);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_12affinematrix_AffineMatrix3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_6vector_Vector3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_6vector_Vector2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_5point_Point3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_5point_Point2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_6normal_Normal3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_10quaternion_Quaternion);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsBin);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SphereSampler);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9targetted__TargettedSampler);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_3ray_Ray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_8material_Material);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_10scenegraph_9_nodebase__NodeBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_10scenegraph_4node_Node);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_10scenegraph_6signal_ChangeSignal);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_10scenegraph_9primitive_Primitive);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_12intersection_Intersection);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_12acceleration_11accelerator_Accelerator);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_10scenegraph_5world_World);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_10scenegraph_8observer_Observer);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_10scenegraph_7utility_BridgeNode);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_10containers__Item);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_10containers_LinkedList);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_10containers_Stack);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_10containers_Queue);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_14boundingsphere_BoundingSphere3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_Item2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_4base_Function);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_4base_FloatFunction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_Item3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D);
  Py_CLEAR(clear_module_state->__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D);
  Py_CLEAR(clear_module_state->__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D);
  Py_CLEAR(clear_module_state->__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D);
  Py_CLEAR(clear_module_state->__pyx_array_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_array);
  Py_CLEAR(clear_module_state->__pyx_MemviewEnum_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_MemviewEnum);
  Py_CLEAR(clear_module_state->__pyx_memoryview_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_memoryview);
  Py_CLEAR(clear_module_state->__pyx_memoryviewslice_type);
  Py_CLEAR(clear_module_state->__pyx_type___pyx_memoryviewslice);
  Py_CLEAR(clear_module_state->__pyx_n_s_ASCII);
  Py_CLEAR(clear_module_state->__pyx_kp_s_All_dimensions_preceding_dimensi);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Buffer_view_does_not_expose_stri);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Can_only_create_a_buffer_that_is);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_assign_to_read_only_memor);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_create_writable_memory_vi);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_index_with_type);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Cannot_transpose_memoryview_with);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Dimension_d_is_not_direct);
  Py_CLEAR(clear_module_state->__pyx_n_s_Ellipsis);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Empty_shape_tuple_for_cython_arr);
  Py_CLEAR(clear_module_state->__pyx_n_s_ImportError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_s_vs_0x6a);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Index_out_of_bounds_axis_d);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Indirect_dimensions_not_supporte);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_mode_expected_c_or_fortr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_shape_in_axis);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_MemoryView_of_r_at_0x_x);
  Py_CLEAR(clear_module_state->__pyx_kp_s_MemoryView_of_r_object);
  Py_CLEAR(clear_module_state->__pyx_n_b_O);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_SOLPSFunction2D);
  Py_CLEAR(clear_module_state->__pyx_n_s_SOLPSFunction2D___getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_SOLPSFunction2D___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_SOLPSFunction2D___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_SOLPSFunction2D_instance);
  Py_CLEAR(clear_module_state->__pyx_n_s_SOLPSVectorFunction2D);
  Py_CLEAR(clear_module_state->__pyx_n_s_SOLPSVectorFunction2D___getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_SOLPSVectorFunction2D___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_SOLPSVectorFunction2D___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_SOLPSVectorFunction2D_instance);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Step_may_not_be_zero_axis_d);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Unable_to_convert_item_to_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_n_s_View_MemoryView);
  Py_CLEAR(clear_module_state->__pyx_kp_u__3);
  Py_CLEAR(clear_module_state->__pyx_kp_u__4);
  Py_CLEAR(clear_module_state->__pyx_n_s__5);
  Py_CLEAR(clear_module_state->__pyx_kp_u__8);
  Py_CLEAR(clear_module_state->__pyx_kp_u__9);
  Py_CLEAR(clear_module_state->__pyx_n_s_allocate_buffer);
  Py_CLEAR(clear_module_state->__pyx_kp_u_and);
  Py_CLEAR(clear_module_state->__pyx_n_s_array);
  Py_CLEAR(clear_module_state->__pyx_n_s_base);
  Py_CLEAR(clear_module_state->__pyx_n_s_c);
  Py_CLEAR(clear_module_state->__pyx_n_u_c);
  Py_CLEAR(clear_module_state->__pyx_kp_s_cherab_solps_solps_2d_functions);
  Py_CLEAR(clear_module_state->__pyx_n_s_cherab_solps_solps_2d_functions_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_cls);
  Py_CLEAR(clear_module_state->__pyx_kp_s_contiguous_and_direct);
  Py_CLEAR(clear_module_state->__pyx_kp_s_contiguous_and_indirect);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dtype);
  Py_CLEAR(clear_module_state->__pyx_n_s_dtype_is_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_error);
  Py_CLEAR(clear_module_state->__pyx_n_s_flags);
  Py_CLEAR(clear_module_state->__pyx_n_s_float64);
  Py_CLEAR(clear_module_state->__pyx_n_s_format);
  Py_CLEAR(clear_module_state->__pyx_n_s_fortran);
  Py_CLEAR(clear_module_state->__pyx_n_u_fortran);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_kp_u_got);
  Py_CLEAR(clear_module_state->__pyx_kp_u_got_differing_extents_in_dimensi);
  Py_CLEAR(clear_module_state->__pyx_n_s_grid_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_grid_vectors);
  Py_CLEAR(clear_module_state->__pyx_n_s_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_instance);
  Py_CLEAR(clear_module_state->__pyx_n_s_int32);
  Py_CLEAR(clear_module_state->__pyx_n_s_itemsize);
  Py_CLEAR(clear_module_state->__pyx_kp_s_itemsize_0_for_cython_array);
  Py_CLEAR(clear_module_state->__pyx_n_s_m);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_memview);
  Py_CLEAR(clear_module_state->__pyx_n_s_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_ndim);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_CLEAR(clear_module_state->__pyx_n_s_np);
  Py_CLEAR(clear_module_state->__pyx_n_s_numpy);
  Py_CLEAR(clear_module_state->__pyx_kp_u_numpy_core_multiarray_failed_to);
  Py_CLEAR(clear_module_state->__pyx_kp_u_numpy_core_umath_failed_to_impor);
  Py_CLEAR(clear_module_state->__pyx_n_s_obj);
  Py_CLEAR(clear_module_state->__pyx_n_s_pack);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_getbuffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Enum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_shape);
  Py_CLEAR(clear_module_state->__pyx_n_s_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_start);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_step);
  Py_CLEAR(clear_module_state->__pyx_n_s_stop);
  Py_CLEAR(clear_module_state->__pyx_kp_s_strided_and_direct);
  Py_CLEAR(clear_module_state->__pyx_kp_s_strided_and_direct_or_indirect);
  Py_CLEAR(clear_module_state->__pyx_kp_s_strided_and_indirect);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_struct);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_triangle_to_grid_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_triangles);
  Py_CLEAR(clear_module_state->__pyx_kp_s_unable_to_allocate_array_data);
  Py_CLEAR(clear_module_state->__pyx_kp_s_unable_to_allocate_shape_and_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_unpack);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_vertex_coords);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_112105877);
  Py_CLEAR(clear_module_state->__pyx_int_neg_1);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_slice__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__2);
  Py_CLEAR(clear_module_state->__pyx_tuple__6);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__16);
  Py_CLEAR(clear_module_state->__pyx_tuple__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__27);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__29);
  Py_CLEAR(clear_module_state->__pyx_tuple__30);
  Py_CLEAR(clear_module_state->__pyx_tuple__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__11);
  Py_CLEAR(clear_module_state->__pyx_codeobj__13);
  Py_CLEAR(clear_module_state->__pyx_codeobj__15);
  Py_CLEAR(clear_module_state->__pyx_codeobj__17);
  Py_CLEAR(clear_module_state->__pyx_codeobj__19);
  Py_CLEAR(clear_module_state->__pyx_codeobj__21);
  Py_CLEAR(clear_module_state->__pyx_codeobj__23);
  Py_CLEAR(clear_module_state->__pyx_codeobj__25);
  Py_CLEAR(clear_module_state->__pyx_codeobj__32);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_dtype);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flatiter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_broadcast);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ndarray);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ufunc);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_5_vec3__Vec3);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_5_mat4__Mat4);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_12affinematrix_AffineMatrix3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_6vector_Vector3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_6vector_Vector2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_5point_Point3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_5point_Point2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_6normal_Normal3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_10quaternion_Quaternion);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsBin);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SphereSampler);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9targetted__TargettedSampler);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_3ray_Ray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_8material_Material);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_10scenegraph_9_nodebase__NodeBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_10scenegraph_4node_Node);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_10scenegraph_6signal_ChangeSignal);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_10scenegraph_9primitive_Primitive);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_12intersection_Intersection);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_12acceleration_11accelerator_Accelerator);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_10scenegraph_5world_World);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_10scenegraph_8observer_Observer);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_10scenegraph_7utility_BridgeNode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_10containers__Item);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_10containers_LinkedList);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_10containers_Stack);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_10containers_Queue);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_14boundingsphere_BoundingSphere3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_Item2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_4base_Function);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_4base_FloatFunction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_Item3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D);
  Py_VISIT(traverse_module_state->__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D);
  Py_VISIT(traverse_module_state->__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D);
  Py_VISIT(traverse_module_state->__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D);
  Py_VISIT(traverse_module_state->__pyx_array_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_array);
  Py_VISIT(traverse_module_state->__pyx_MemviewEnum_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_MemviewEnum);
  Py_VISIT(traverse_module_state->__pyx_memoryview_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_memoryview);
  Py_VISIT(traverse_module_state->__pyx_memoryviewslice_type);
  Py_VISIT(traverse_module_state->__pyx_type___pyx_memoryviewslice);
  Py_VISIT(traverse_module_state->__pyx_n_s_ASCII);
  Py_VISIT(traverse_module_state->__pyx_kp_s_All_dimensions_preceding_dimensi);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Buffer_view_does_not_expose_stri);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Can_only_create_a_buffer_that_is);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_assign_to_read_only_memor);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_create_writable_memory_vi);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_index_with_type);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Cannot_transpose_memoryview_with);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Dimension_d_is_not_direct);
  Py_VISIT(traverse_module_state->__pyx_n_s_Ellipsis);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Empty_shape_tuple_for_cython_arr);
  Py_VISIT(traverse_module_state->__pyx_n_s_ImportError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_s_vs_0x6a);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Index_out_of_bounds_axis_d);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Indirect_dimensions_not_supporte);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_mode_expected_c_or_fortr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_shape_in_axis);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_MemoryView_of_r_at_0x_x);
  Py_VISIT(traverse_module_state->__pyx_kp_s_MemoryView_of_r_object);
  Py_VISIT(traverse_module_state->__pyx_n_b_O);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_SOLPSFunction2D);
  Py_VISIT(traverse_module_state->__pyx_n_s_SOLPSFunction2D___getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_SOLPSFunction2D___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_SOLPSFunction2D___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_SOLPSFunction2D_instance);
  Py_VISIT(traverse_module_state->__pyx_n_s_SOLPSVectorFunction2D);
  Py_VISIT(traverse_module_state->__pyx_n_s_SOLPSVectorFunction2D___getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_SOLPSVectorFunction2D___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_SOLPSVectorFunction2D___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_SOLPSVectorFunction2D_instance);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Step_may_not_be_zero_axis_d);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Unable_to_convert_item_to_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_n_s_View_MemoryView);
  Py_VISIT(traverse_module_state->__pyx_kp_u__3);
  Py_VISIT(traverse_module_state->__pyx_kp_u__4);
  Py_VISIT(traverse_module_state->__pyx_n_s__5);
  Py_VISIT(traverse_module_state->__pyx_kp_u__8);
  Py_VISIT(traverse_module_state->__pyx_kp_u__9);
  Py_VISIT(traverse_module_state->__pyx_n_s_allocate_buffer);
  Py_VISIT(traverse_module_state->__pyx_kp_u_and);
  Py_VISIT(traverse_module_state->__pyx_n_s_array);
  Py_VISIT(traverse_module_state->__pyx_n_s_base);
  Py_VISIT(traverse_module_state->__pyx_n_s_c);
  Py_VISIT(traverse_module_state->__pyx_n_u_c);
  Py_VISIT(traverse_module_state->__pyx_kp_s_cherab_solps_solps_2d_functions);
  Py_VISIT(traverse_module_state->__pyx_n_s_cherab_solps_solps_2d_functions_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_cls);
  Py_VISIT(traverse_module_state->__pyx_kp_s_contiguous_and_direct);
  Py_VISIT(traverse_module_state->__pyx_kp_s_contiguous_and_indirect);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dtype);
  Py_VISIT(traverse_module_state->__pyx_n_s_dtype_is_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_error);
  Py_VISIT(traverse_module_state->__pyx_n_s_flags);
  Py_VISIT(traverse_module_state->__pyx_n_s_float64);
  Py_VISIT(traverse_module_state->__pyx_n_s_format);
  Py_VISIT(traverse_module_state->__pyx_n_s_fortran);
  Py_VISIT(traverse_module_state->__pyx_n_u_fortran);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_kp_u_got);
  Py_VISIT(traverse_module_state->__pyx_kp_u_got_differing_extents_in_dimensi);
  Py_VISIT(traverse_module_state->__pyx_n_s_grid_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_grid_vectors);
  Py_VISIT(traverse_module_state->__pyx_n_s_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_instance);
  Py_VISIT(traverse_module_state->__pyx_n_s_int32);
  Py_VISIT(traverse_module_state->__pyx_n_s_itemsize);
  Py_VISIT(traverse_module_state->__pyx_kp_s_itemsize_0_for_cython_array);
  Py_VISIT(traverse_module_state->__pyx_n_s_m);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_memview);
  Py_VISIT(traverse_module_state->__pyx_n_s_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_ndim);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_VISIT(traverse_module_state->__pyx_n_s_np);
  Py_VISIT(traverse_module_state->__pyx_n_s_numpy);
  Py_VISIT(traverse_module_state->__pyx_kp_u_numpy_core_multiarray_failed_to);
  Py_VISIT(traverse_module_state->__pyx_kp_u_numpy_core_umath_failed_to_impor);
  Py_VISIT(traverse_module_state->__pyx_n_s_obj);
  Py_VISIT(traverse_module_state->__pyx_n_s_pack);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_getbuffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Enum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_shape);
  Py_VISIT(traverse_module_state->__pyx_n_s_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_start);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_step);
  Py_VISIT(traverse_module_state->__pyx_n_s_stop);
  Py_VISIT(traverse_module_state->__pyx_kp_s_strided_and_direct);
  Py_VISIT(traverse_module_state->__pyx_kp_s_strided_and_direct_or_indirect);
  Py_VISIT(traverse_module_state->__pyx_kp_s_strided_and_indirect);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_struct);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_triangle_to_grid_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_triangles);
  Py_VISIT(traverse_module_state->__pyx_kp_s_unable_to_allocate_array_data);
  Py_VISIT(traverse_module_state->__pyx_kp_s_unable_to_allocate_shape_and_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_unpack);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_vertex_coords);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_112105877);
  Py_VISIT(traverse_module_state->__pyx_int_neg_1);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_slice__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__2);
  Py_VISIT(traverse_module_state->__pyx_tuple__6);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__16);
  Py_VISIT(traverse_module_state->__pyx_tuple__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__27);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__29);
  Py_VISIT(traverse_module_state->__pyx_tuple__30);
  Py_VISIT(traverse_module_state->__pyx_tuple__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__11);
  Py_VISIT(traverse_module_state->__pyx_codeobj__13);
  Py_VISIT(traverse_module_state->__pyx_codeobj__15);
  Py_VISIT(traverse_module_state->__pyx_codeobj__17);
  Py_VISIT(traverse_module_state->__pyx_codeobj__19);
  Py_VISIT(traverse_module_state->__pyx_codeobj__21);
  Py_VISIT(traverse_module_state->__pyx_codeobj__23);
  Py_VISIT(traverse_module_state->__pyx_codeobj__25);
  Py_VISIT(traverse_module_state->__pyx_codeobj__32);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#if CYTHON_COMPILING_IN_LIMITED_API
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#define __pyx_ptype_5numpy_dtype __pyx_mstate_global->__pyx_ptype_5numpy_dtype
#define __pyx_ptype_5numpy_flatiter __pyx_mstate_global->__pyx_ptype_5numpy_flatiter
#define __pyx_ptype_5numpy_broadcast __pyx_mstate_global->__pyx_ptype_5numpy_broadcast
#define __pyx_ptype_5numpy_ndarray __pyx_mstate_global->__pyx_ptype_5numpy_ndarray
#define __pyx_ptype_5numpy_ufunc __pyx_mstate_global->__pyx_ptype_5numpy_ufunc
#define __pyx_ptype_7raysect_4core_4math_5_vec3__Vec3 __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_5_vec3__Vec3
#define __pyx_ptype_7raysect_4core_4math_5_mat4__Mat4 __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_5_mat4__Mat4
#define __pyx_ptype_7raysect_4core_4math_12affinematrix_AffineMatrix3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_12affinematrix_AffineMatrix3D
#define __pyx_ptype_7raysect_4core_4math_6vector_Vector3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_6vector_Vector3D
#define __pyx_ptype_7raysect_4core_4math_6vector_Vector2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_6vector_Vector2D
#define __pyx_ptype_7raysect_4core_4math_5point_Point3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_5point_Point3D
#define __pyx_ptype_7raysect_4core_4math_5point_Point2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_5point_Point2D
#define __pyx_ptype_7raysect_4core_4math_6normal_Normal3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_6normal_Normal3D
#define __pyx_ptype_7raysect_4core_4math_10quaternion_Quaternion __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_10quaternion_Quaternion
#define __pyx_ptype_7raysect_4core_4math_10statsarray_StatsBin __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsBin
#define __pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray1D
#define __pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray2D
#define __pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray3D
#define __pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler
#define __pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SphereSampler __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SphereSampler
#define __pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler
#define __pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler
#define __pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler
#define __pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D
#define __pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D
#define __pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D
#define __pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D
#define __pyx_ptype_7raysect_4core_4math_7sampler_9targetted__TargettedSampler __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_9targetted__TargettedSampler
#define __pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler
#define __pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler
#define __pyx_ptype_7raysect_4core_3ray_Ray __pyx_mstate_global->__pyx_ptype_7raysect_4core_3ray_Ray
#define __pyx_ptype_7raysect_4core_8material_Material __pyx_mstate_global->__pyx_ptype_7raysect_4core_8material_Material
#define __pyx_ptype_7raysect_4core_10scenegraph_9_nodebase__NodeBase __pyx_mstate_global->__pyx_ptype_7raysect_4core_10scenegraph_9_nodebase__NodeBase
#define __pyx_ptype_7raysect_4core_10scenegraph_4node_Node __pyx_mstate_global->__pyx_ptype_7raysect_4core_10scenegraph_4node_Node
#define __pyx_ptype_7raysect_4core_10scenegraph_6signal_ChangeSignal __pyx_mstate_global->__pyx_ptype_7raysect_4core_10scenegraph_6signal_ChangeSignal
#define __pyx_ptype_7raysect_4core_10scenegraph_9primitive_Primitive __pyx_mstate_global->__pyx_ptype_7raysect_4core_10scenegraph_9primitive_Primitive
#define __pyx_ptype_7raysect_4core_12intersection_Intersection __pyx_mstate_global->__pyx_ptype_7raysect_4core_12intersection_Intersection
#define __pyx_ptype_7raysect_4core_12acceleration_11accelerator_Accelerator __pyx_mstate_global->__pyx_ptype_7raysect_4core_12acceleration_11accelerator_Accelerator
#define __pyx_ptype_7raysect_4core_10scenegraph_5world_World __pyx_mstate_global->__pyx_ptype_7raysect_4core_10scenegraph_5world_World
#define __pyx_ptype_7raysect_4core_10scenegraph_8observer_Observer __pyx_mstate_global->__pyx_ptype_7raysect_4core_10scenegraph_8observer_Observer
#define __pyx_ptype_7raysect_4core_10scenegraph_7utility_BridgeNode __pyx_mstate_global->__pyx_ptype_7raysect_4core_10scenegraph_7utility_BridgeNode
#define __pyx_ptype_7raysect_4core_10containers__Item __pyx_mstate_global->__pyx_ptype_7raysect_4core_10containers__Item
#define __pyx_ptype_7raysect_4core_10containers_LinkedList __pyx_mstate_global->__pyx_ptype_7raysect_4core_10containers_LinkedList
#define __pyx_ptype_7raysect_4core_10containers_Stack __pyx_mstate_global->__pyx_ptype_7raysect_4core_10containers_Stack
#define __pyx_ptype_7raysect_4core_10containers_Queue __pyx_mstate_global->__pyx_ptype_7raysect_4core_10containers_Queue
#define __pyx_ptype_7raysect_4core_14boundingsphere_BoundingSphere3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_14boundingsphere_BoundingSphere3D
#define __pyx_ptype_7raysect_4core_11boundingbox_BoundingBox3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox3D
#define __pyx_ptype_7raysect_4core_11boundingbox_BoundingBox2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox2D
#define __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_Item2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_Item2D
#define __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore
#define __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D
#define __pyx_ptype_7raysect_4core_4math_8function_4base_Function __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_4base_Function
#define __pyx_ptype_7raysect_4core_4math_8function_5float_4base_FloatFunction __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_4base_FloatFunction
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D
#define __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_Item3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_Item3D
#define __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore
#define __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D
#define __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D
#define __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D __pyx_mstate_global->__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D
#define __pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D __pyx_mstate_global->__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D
#define __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D __pyx_mstate_global->__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D
#define __pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D __pyx_mstate_global->__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D
#define __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D __pyx_mstate_global->__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D
#define __pyx_array_type __pyx_mstate_global->__pyx_array_type
#define __pyx_type___pyx_array __pyx_mstate_global->__pyx_type___pyx_array
#define __pyx_MemviewEnum_type __pyx_mstate_global->__pyx_MemviewEnum_type
#define __pyx_type___pyx_MemviewEnum __pyx_mstate_global->__pyx_type___pyx_MemviewEnum
#define __pyx_memoryview_type __pyx_mstate_global->__pyx_memoryview_type
#define __pyx_type___pyx_memoryview __pyx_mstate_global->__pyx_type___pyx_memoryview
#define __pyx_memoryviewslice_type __pyx_mstate_global->__pyx_memoryviewslice_type
#define __pyx_type___pyx_memoryviewslice __pyx_mstate_global->__pyx_type___pyx_memoryviewslice
#define __pyx_n_s_ASCII __pyx_mstate_global->__pyx_n_s_ASCII
#define __pyx_kp_s_All_dimensions_preceding_dimensi __pyx_mstate_global->__pyx_kp_s_All_dimensions_preceding_dimensi
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_kp_s_Buffer_view_does_not_expose_stri __pyx_mstate_global->__pyx_kp_s_Buffer_view_does_not_expose_stri
#define __pyx_kp_s_Can_only_create_a_buffer_that_is __pyx_mstate_global->__pyx_kp_s_Can_only_create_a_buffer_that_is
#define __pyx_kp_s_Cannot_assign_to_read_only_memor __pyx_mstate_global->__pyx_kp_s_Cannot_assign_to_read_only_memor
#define __pyx_kp_s_Cannot_create_writable_memory_vi __pyx_mstate_global->__pyx_kp_s_Cannot_create_writable_memory_vi
#define __pyx_kp_u_Cannot_index_with_type __pyx_mstate_global->__pyx_kp_u_Cannot_index_with_type
#define __pyx_kp_s_Cannot_transpose_memoryview_with __pyx_mstate_global->__pyx_kp_s_Cannot_transpose_memoryview_with
#define __pyx_kp_s_Dimension_d_is_not_direct __pyx_mstate_global->__pyx_kp_s_Dimension_d_is_not_direct
#define __pyx_n_s_Ellipsis __pyx_mstate_global->__pyx_n_s_Ellipsis
#define __pyx_kp_s_Empty_shape_tuple_for_cython_arr __pyx_mstate_global->__pyx_kp_s_Empty_shape_tuple_for_cython_arr
#define __pyx_n_s_ImportError __pyx_mstate_global->__pyx_n_s_ImportError
#define __pyx_kp_s_Incompatible_checksums_s_vs_0x6a __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_s_vs_0x6a
#define __pyx_n_s_IndexError __pyx_mstate_global->__pyx_n_s_IndexError
#define __pyx_kp_s_Index_out_of_bounds_axis_d __pyx_mstate_global->__pyx_kp_s_Index_out_of_bounds_axis_d
#define __pyx_kp_s_Indirect_dimensions_not_supporte __pyx_mstate_global->__pyx_kp_s_Indirect_dimensions_not_supporte
#define __pyx_kp_u_Invalid_mode_expected_c_or_fortr __pyx_mstate_global->__pyx_kp_u_Invalid_mode_expected_c_or_fortr
#define __pyx_kp_u_Invalid_shape_in_axis __pyx_mstate_global->__pyx_kp_u_Invalid_shape_in_axis
#define __pyx_n_s_MemoryError __pyx_mstate_global->__pyx_n_s_MemoryError
#define __pyx_kp_s_MemoryView_of_r_at_0x_x __pyx_mstate_global->__pyx_kp_s_MemoryView_of_r_at_0x_x
#define __pyx_kp_s_MemoryView_of_r_object __pyx_mstate_global->__pyx_kp_s_MemoryView_of_r_object
#define __pyx_n_b_O __pyx_mstate_global->__pyx_n_b_O
#define __pyx_kp_u_Out_of_bounds_on_buffer_access_a __pyx_mstate_global->__pyx_kp_u_Out_of_bounds_on_buffer_access_a
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_n_s_SOLPSFunction2D __pyx_mstate_global->__pyx_n_s_SOLPSFunction2D
#define __pyx_n_s_SOLPSFunction2D___getstate __pyx_mstate_global->__pyx_n_s_SOLPSFunction2D___getstate
#define __pyx_n_s_SOLPSFunction2D___reduce __pyx_mstate_global->__pyx_n_s_SOLPSFunction2D___reduce
#define __pyx_n_s_SOLPSFunction2D___setstate __pyx_mstate_global->__pyx_n_s_SOLPSFunction2D___setstate
#define __pyx_n_s_SOLPSFunction2D_instance __pyx_mstate_global->__pyx_n_s_SOLPSFunction2D_instance
#define __pyx_n_s_SOLPSVectorFunction2D __pyx_mstate_global->__pyx_n_s_SOLPSVectorFunction2D
#define __pyx_n_s_SOLPSVectorFunction2D___getstate __pyx_mstate_global->__pyx_n_s_SOLPSVectorFunction2D___getstate
#define __pyx_n_s_SOLPSVectorFunction2D___reduce __pyx_mstate_global->__pyx_n_s_SOLPSVectorFunction2D___reduce
#define __pyx_n_s_SOLPSVectorFunction2D___setstate __pyx_mstate_global->__pyx_n_s_SOLPSVectorFunction2D___setstate
#define __pyx_n_s_SOLPSVectorFunction2D_instance __pyx_mstate_global->__pyx_n_s_SOLPSVectorFunction2D_instance
#define __pyx_kp_s_Step_may_not_be_zero_axis_d __pyx_mstate_global->__pyx_kp_s_Step_may_not_be_zero_axis_d
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_kp_s_Unable_to_convert_item_to_object __pyx_mstate_global->__pyx_kp_s_Unable_to_convert_item_to_object
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_n_s_View_MemoryView __pyx_mstate_global->__pyx_n_s_View_MemoryView
#define __pyx_kp_u__3 __pyx_mstate_global->__pyx_kp_u__3
#define __pyx_kp_u__4 __pyx_mstate_global->__pyx_kp_u__4
#define __pyx_n_s__5 __pyx_mstate_global->__pyx_n_s__5
#define __pyx_kp_u__8 __pyx_mstate_global->__pyx_kp_u__8
#define __pyx_kp_u__9 __pyx_mstate_global->__pyx_kp_u__9
#define __pyx_n_s_allocate_buffer __pyx_mstate_global->__pyx_n_s_allocate_buffer
#define __pyx_kp_u_and __pyx_mstate_global->__pyx_kp_u_and
#define __pyx_n_s_array __pyx_mstate_global->__pyx_n_s_array
#define __pyx_n_s_base __pyx_mstate_global->__pyx_n_s_base
#define __pyx_n_s_c __pyx_mstate_global->__pyx_n_s_c
#define __pyx_n_u_c __pyx_mstate_global->__pyx_n_u_c
#define __pyx_kp_s_cherab_solps_solps_2d_functions __pyx_mstate_global->__pyx_kp_s_cherab_solps_solps_2d_functions
#define __pyx_n_s_cherab_solps_solps_2d_functions_2 __pyx_mstate_global->__pyx_n_s_cherab_solps_solps_2d_functions_2
#define __pyx_n_s_class __pyx_mstate_global->__pyx_n_s_class
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_cls __pyx_mstate_global->__pyx_n_s_cls
#define __pyx_kp_s_contiguous_and_direct __pyx_mstate_global->__pyx_kp_s_contiguous_and_direct
#define __pyx_kp_s_contiguous_and_indirect __pyx_mstate_global->__pyx_kp_s_contiguous_and_indirect
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dtype __pyx_mstate_global->__pyx_n_s_dtype
#define __pyx_n_s_dtype_is_object __pyx_mstate_global->__pyx_n_s_dtype_is_object
#define __pyx_n_s_encode __pyx_mstate_global->__pyx_n_s_encode
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_error __pyx_mstate_global->__pyx_n_s_error
#define __pyx_n_s_flags __pyx_mstate_global->__pyx_n_s_flags
#define __pyx_n_s_float64 __pyx_mstate_global->__pyx_n_s_float64
#define __pyx_n_s_format __pyx_mstate_global->__pyx_n_s_format
#define __pyx_n_s_fortran __pyx_mstate_global->__pyx_n_s_fortran
#define __pyx_n_u_fortran __pyx_mstate_global->__pyx_n_u_fortran
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_kp_u_got __pyx_mstate_global->__pyx_kp_u_got
#define __pyx_kp_u_got_differing_extents_in_dimensi __pyx_mstate_global->__pyx_kp_u_got_differing_extents_in_dimensi
#define __pyx_n_s_grid_data __pyx_mstate_global->__pyx_n_s_grid_data
#define __pyx_n_s_grid_vectors __pyx_mstate_global->__pyx_n_s_grid_vectors
#define __pyx_n_s_id __pyx_mstate_global->__pyx_n_s_id
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_instance __pyx_mstate_global->__pyx_n_s_instance
#define __pyx_n_s_int32 __pyx_mstate_global->__pyx_n_s_int32
#define __pyx_n_s_itemsize __pyx_mstate_global->__pyx_n_s_itemsize
#define __pyx_kp_s_itemsize_0_for_cython_array __pyx_mstate_global->__pyx_kp_s_itemsize_0_for_cython_array
#define __pyx_n_s_m __pyx_mstate_global->__pyx_n_s_m
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_memview __pyx_mstate_global->__pyx_n_s_memview
#define __pyx_n_s_mode __pyx_mstate_global->__pyx_n_s_mode
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_ndim __pyx_mstate_global->__pyx_n_s_ndim
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_kp_s_no_default___reduce___due_to_non __pyx_mstate_global->__pyx_kp_s_no_default___reduce___due_to_non
#define __pyx_n_s_np __pyx_mstate_global->__pyx_n_s_np
#define __pyx_n_s_numpy __pyx_mstate_global->__pyx_n_s_numpy
#define __pyx_kp_u_numpy_core_multiarray_failed_to __pyx_mstate_global->__pyx_kp_u_numpy_core_multiarray_failed_to
#define __pyx_kp_u_numpy_core_umath_failed_to_impor __pyx_mstate_global->__pyx_kp_u_numpy_core_umath_failed_to_impor
#define __pyx_n_s_obj __pyx_mstate_global->__pyx_n_s_obj
#define __pyx_n_s_pack __pyx_mstate_global->__pyx_n_s_pack
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_getbuffer __pyx_mstate_global->__pyx_n_s_pyx_getbuffer
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_Enum __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Enum
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_shape __pyx_mstate_global->__pyx_n_s_shape
#define __pyx_n_s_size __pyx_mstate_global->__pyx_n_s_size
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_start __pyx_mstate_global->__pyx_n_s_start
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_step __pyx_mstate_global->__pyx_n_s_step
#define __pyx_n_s_stop __pyx_mstate_global->__pyx_n_s_stop
#define __pyx_kp_s_strided_and_direct __pyx_mstate_global->__pyx_kp_s_strided_and_direct
#define __pyx_kp_s_strided_and_direct_or_indirect __pyx_mstate_global->__pyx_kp_s_strided_and_direct_or_indirect
#define __pyx_kp_s_strided_and_indirect __pyx_mstate_global->__pyx_kp_s_strided_and_indirect
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_struct __pyx_mstate_global->__pyx_n_s_struct
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_triangle_to_grid_map __pyx_mstate_global->__pyx_n_s_triangle_to_grid_map
#define __pyx_n_s_triangles __pyx_mstate_global->__pyx_n_s_triangles
#define __pyx_kp_s_unable_to_allocate_array_data __pyx_mstate_global->__pyx_kp_s_unable_to_allocate_array_data
#define __pyx_kp_s_unable_to_allocate_shape_and_str __pyx_mstate_global->__pyx_kp_s_unable_to_allocate_shape_and_str
#define __pyx_n_s_unpack __pyx_mstate_global->__pyx_n_s_unpack
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_vertex_coords __pyx_mstate_global->__pyx_n_s_vertex_coords
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_112105877 __pyx_mstate_global->__pyx_int_112105877
#define __pyx_int_neg_1 __pyx_mstate_global->__pyx_int_neg_1
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_slice__7 __pyx_mstate_global->__pyx_slice__7
#define __pyx_tuple__2 __pyx_mstate_global->__pyx_tuple__2
#define __pyx_tuple__6 __pyx_mstate_global->__pyx_tuple__6
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__16 __pyx_mstate_global->__pyx_tuple__16
#define __pyx_tuple__18 __pyx_mstate_global->__pyx_tuple__18
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__27 __pyx_mstate_global->__pyx_tuple__27
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__29 __pyx_mstate_global->__pyx_tuple__29
#define __pyx_tuple__30 __pyx_mstate_global->__pyx_tuple__30
#define __pyx_tuple__31 __pyx_mstate_global->__pyx_tuple__31
#define __pyx_codeobj__11 __pyx_mstate_global->__pyx_codeobj__11
#define __pyx_codeobj__13 __pyx_mstate_global->__pyx_codeobj__13
#define __pyx_codeobj__15 __pyx_mstate_global->__pyx_codeobj__15
#define __pyx_codeobj__17 __pyx_mstate_global->__pyx_codeobj__17
#define __pyx_codeobj__19 __pyx_mstate_global->__pyx_codeobj__19
#define __pyx_codeobj__21 __pyx_mstate_global->__pyx_codeobj__21
#define __pyx_codeobj__23 __pyx_mstate_global->__pyx_codeobj__23
#define __pyx_codeobj__25 __pyx_mstate_global->__pyx_codeobj__25
#define __pyx_codeobj__32 __pyx_mstate_global->__pyx_codeobj__32
#endif
/* #### Code section: module_code ### */

/* "cherab/solps/solps_2d_functions.pyx":34
 * cdef class SOLPSFunction2D(Function2D):
 * 
 *     def __init__(self, object vertex_coords not None, object triangles not None, object triangle_to_grid_map not None, object grid_data not None):             # <<<<<<<<<<<<<<
 * 
 *         # use numpy arrays to store data internally
 */

/* Python wrapper */
static int __pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_vertex_coords = 0;
  PyObject *__pyx_v_triangles = 0;
  PyObject *__pyx_v_triangle_to_grid_map = 0;
  PyObject *__pyx_v_grid_data = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_vertex_coords,&__pyx_n_s_triangles,&__pyx_n_s_triangle_to_grid_map,&__pyx_n_s_grid_data,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_vertex_coords,&__pyx_n_s_triangles,&__pyx_n_s_triangle_to_grid_map,&__pyx_n_s_grid_data,0};
    #endif
    PyObject* values[4] = {0,0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_vertex_coords)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 34, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_triangles)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 34, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(0, 34, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_triangle_to_grid_map)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 34, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(0, 34, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_grid_data)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 34, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(0, 34, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 34, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
    }
    __pyx_v_vertex_coords = values[0];
    __pyx_v_triangles = values[1];
    __pyx_v_triangle_to_grid_map = values[2];
    __pyx_v_grid_data = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 34, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSFunction2D.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_vertex_coords) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "vertex_coords"); __PYX_ERR(0, 34, __pyx_L1_error)
  }
  if (unlikely(((PyObject *)__pyx_v_triangles) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "triangles"); __PYX_ERR(0, 34, __pyx_L1_error)
  }
  if (unlikely(((PyObject *)__pyx_v_triangle_to_grid_map) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "triangle_to_grid_map"); __PYX_ERR(0, 34, __pyx_L1_error)
  }
  if (unlikely(((PyObject *)__pyx_v_grid_data) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "grid_data"); __PYX_ERR(0, 34, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D___init__(((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *)__pyx_v_self), __pyx_v_vertex_coords, __pyx_v_triangles, __pyx_v_triangle_to_grid_map, __pyx_v_grid_data);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D___init__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_self, PyObject *__pyx_v_vertex_coords, PyObject *__pyx_v_triangles, PyObject *__pyx_v_triangle_to_grid_map, PyObject *__pyx_v_grid_data) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_vertex_coords);
  __Pyx_INCREF(__pyx_v_triangles);
  __Pyx_INCREF(__pyx_v_triangle_to_grid_map);

  /* "cherab/solps/solps_2d_functions.pyx":37
 * 
 *         # use numpy arrays to store data internally
 *         vertex_coords = np.array(vertex_coords, dtype=np.float64)             # <<<<<<<<<<<<<<
 *         triangles = np.array(triangles, dtype=np.int32)
 *         triangle_to_grid_map = np.array(triangle_to_grid_map, dtype=np.int32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_vertex_coords);
  __Pyx_GIVEREF(__pyx_v_vertex_coords);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_vertex_coords);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_vertex_coords, __pyx_t_5);
  __pyx_t_5 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":38
 *         # use numpy arrays to store data internally
 *         vertex_coords = np.array(vertex_coords, dtype=np.float64)
 *         triangles = np.array(triangles, dtype=np.int32)             # <<<<<<<<<<<<<<
 *         triangle_to_grid_map = np.array(triangle_to_grid_map, dtype=np.int32)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_triangles);
  __Pyx_GIVEREF(__pyx_v_triangles);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_triangles);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int32); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_triangles, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":39
 *         vertex_coords = np.array(vertex_coords, dtype=np.float64)
 *         triangles = np.array(triangles, dtype=np.int32)
 *         triangle_to_grid_map = np.array(triangle_to_grid_map, dtype=np.int32)             # <<<<<<<<<<<<<<
 * 
 *         # Attention!!! Do not copy grid_data! Attribute self._grid_data must point to the original data array,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_triangle_to_grid_map);
  __Pyx_GIVEREF(__pyx_v_triangle_to_grid_map);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_triangle_to_grid_map);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF_SET(__pyx_v_triangle_to_grid_map, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":45
 * 
 *         # build kdtree
 *         self._kdtree = MeshKDTree2D(vertex_coords, triangles)             # <<<<<<<<<<<<<<
 * 
 *         # populate internal attributes
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_vertex_coords);
  __Pyx_GIVEREF(__pyx_v_vertex_coords);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_vertex_coords);
  __Pyx_INCREF(__pyx_v_triangles);
  __Pyx_GIVEREF(__pyx_v_triangles);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_triangles);
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D), __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_kdtree);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_kdtree);
  __pyx_v_self->_kdtree = ((struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":48
 * 
 *         # populate internal attributes
 *         self._grid_data = grid_data             # <<<<<<<<<<<<<<
 *         self._triangle_to_grid_map = triangle_to_grid_map
 * 
 */
  if (!(likely(__Pyx_TypeTest(__pyx_v_grid_data, __pyx_ptype_5numpy_ndarray)))) __PYX_ERR(0, 48, __pyx_L1_error)
  __pyx_t_5 = __pyx_v_grid_data;
  __Pyx_INCREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_grid_data);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_grid_data);
  __pyx_v_self->_grid_data = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":49
 *         # populate internal attributes
 *         self._grid_data = grid_data
 *         self._triangle_to_grid_map = triangle_to_grid_map             # <<<<<<<<<<<<<<
 * 
 *         self._grid_data_mv = self._grid_data
 */
  if (!(likely(((__pyx_v_triangle_to_grid_map) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_triangle_to_grid_map, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 49, __pyx_L1_error)
  __pyx_t_5 = __pyx_v_triangle_to_grid_map;
  __Pyx_INCREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  __pyx_v_self->_triangle_to_grid_map = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":51
 *         self._triangle_to_grid_map = triangle_to_grid_map
 * 
 *         self._grid_data_mv = self._grid_data             # <<<<<<<<<<<<<<
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(((PyObject *)__pyx_v_self->_grid_data), PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 51, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_grid_data_mv, 0);
  __pyx_v_self->_grid_data_mv = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":52
 * 
 *         self._grid_data_mv = self._grid_data
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_self->_triangle_to_grid_map), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 52, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_triangle_to_grid_map_mv, 0);
  __pyx_v_self->_triangle_to_grid_map_mv = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":34
 * cdef class SOLPSFunction2D(Function2D):
 * 
 *     def __init__(self, object vertex_coords not None, object triangles not None, object triangle_to_grid_map not None, object grid_data not None):             # <<<<<<<<<<<<<<
 * 
 *         # use numpy arrays to store data internally
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSFunction2D.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vertex_coords);
  __Pyx_XDECREF(__pyx_v_triangles);
  __Pyx_XDECREF(__pyx_v_triangle_to_grid_map);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cherab/solps/solps_2d_functions.pyx":54
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self._grid_data, self._triangle_to_grid_map, self._kdtree
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_3__getstate__ = {"__getstate__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_3__getstate__, METH_NOARGS, 0};
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_2__getstate__(((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_2__getstate__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "cherab/solps/solps_2d_functions.pyx":55
 * 
 *     def __getstate__(self):
 *         return self._grid_data, self._triangle_to_grid_map, self._kdtree             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_grid_data);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_grid_data);
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->_grid_data));
  __Pyx_INCREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->_triangle_to_grid_map));
  __Pyx_INCREF((PyObject *)__pyx_v_self->_kdtree);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_kdtree);
  PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->_kdtree));
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cherab/solps/solps_2d_functions.pyx":54
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self._grid_data, self._triangle_to_grid_map, self._kdtree
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSFunction2D.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cherab/solps/solps_2d_functions.pyx":57
 *         return self._grid_data, self._triangle_to_grid_map, self._kdtree
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._grid_data, self._triangle_to_grid_map, self._kdtree = state
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 */

/* Python wrapper */
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static PyMethodDef __pyx_mdef_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_5__setstate__ = {"__setstate__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_5__setstate__, METH_O, 0};
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_4__setstate__(((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_4__setstate__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "cherab/solps/solps_2d_functions.pyx":58
 * 
 *     def __setstate__(self, state):
 *         self._grid_data, self._triangle_to_grid_map, self._kdtree = state             # <<<<<<<<<<<<<<
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 *         self._grid_data_mv = self._grid_data
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {
    PyObject* sequence = __pyx_v_state;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 58, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 2; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 3) < 0) __PYX_ERR(0, 58, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 58, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 58, __pyx_L1_error)
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 58, __pyx_L1_error)
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D))))) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_grid_data);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_grid_data);
  __pyx_v_self->_grid_data = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  __pyx_v_self->_triangle_to_grid_map = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_kdtree);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_kdtree);
  __pyx_v_self->_kdtree = ((struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":59
 *     def __setstate__(self, state):
 *         self._grid_data, self._triangle_to_grid_map, self._kdtree = state
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map             # <<<<<<<<<<<<<<
 *         self._grid_data_mv = self._grid_data
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_self->_triangle_to_grid_map), PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 59, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_triangle_to_grid_map_mv, 0);
  __pyx_v_self->_triangle_to_grid_map_mv = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":60
 *         self._grid_data, self._triangle_to_grid_map, self._kdtree = state
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 *         self._grid_data_mv = self._grid_data             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(((PyObject *)__pyx_v_self->_grid_data), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 60, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_grid_data_mv, 0);
  __pyx_v_self->_grid_data_mv = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":57
 *         return self._grid_data, self._triangle_to_grid_map, self._kdtree
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._grid_data, self._triangle_to_grid_map, self._kdtree = state
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSFunction2D.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cherab/solps/solps_2d_functions.pyx":62
 *         self._grid_data_mv = self._grid_data
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_7__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_7__reduce__ = {"__reduce__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_7__reduce__, METH_NOARGS, 0};
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_7__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_6__reduce__(((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_6__reduce__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "cherab/solps/solps_2d_functions.pyx":63
 * 
 *     def __reduce__(self):
 *         return self.__new__, (self.__class__, ), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_5, };
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cherab/solps/solps_2d_functions.pyx":62
 *         self._grid_data_mv = self._grid_data
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSFunction2D.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cherab/solps/solps_2d_functions.pyx":65
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def instance(cls, SOLPSFunction2D instance not None, object grid_data=None):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_9instance(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_8instance, "\n        Creates a new interpolator instance from an existing interpolator instance.\n        The new interpolator instance will share the same internal acceleration\n        data as the original interpolator. The grid_data of the new instance can\n        be redefined.\n        This method should be used if the user has multiple sets of grid_data\n        that lie on the same mesh geometry. Using this methods avoids the\n        repeated rebuilding of the mesh acceleration structures by sharing the\n        geometry data between multiple interpolator objects.\n        :param SOLPSFunction2D instance: SOLPSFunction2D object.\n        :param ndarray grid_data: An array containing data on SOLPS grid.\n        :return: A SOLPSFunction2D object.\n        :rtype: SOLPSFunction2D\n        ");
static PyMethodDef __pyx_mdef_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_9instance = {"instance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_9instance, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_8instance};
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_9instance(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_instance = 0;
  PyObject *__pyx_v_grid_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("instance (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_instance,&__pyx_n_s_grid_data,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_instance,&__pyx_n_s_grid_data,0};
    #endif
    PyObject* values[2] = {0,0};

    /* "cherab/solps/solps_2d_functions.pyx":66
 * 
 *     @classmethod
 *     def instance(cls, SOLPSFunction2D instance not None, object grid_data=None):             # <<<<<<<<<<<<<<
 *         """
 *         Creates a new interpolator instance from an existing interpolator instance.
 */
    values[1] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_instance)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 65, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_grid_data);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 65, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "instance") < 0)) __PYX_ERR(0, 65, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_instance = ((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *)values[0]);
    __pyx_v_grid_data = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("instance", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 65, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSFunction2D.instance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_instance), __pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D, 0, "instance", 0))) __PYX_ERR(0, 66, __pyx_L1_error)
  __pyx_r = __pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_8instance(((PyTypeObject*)__pyx_v_cls), __pyx_v_instance, __pyx_v_grid_data);

  /* "cherab/solps/solps_2d_functions.pyx":65
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def instance(cls, SOLPSFunction2D instance not None, object grid_data=None):
 *         """
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_8instance(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_instance, PyObject *__pyx_v_grid_data) {
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_m = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("instance", 0);

  /* "cherab/solps/solps_2d_functions.pyx":85
 * 
 *         # copy source data
 *         m = SOLPSFunction2D.__new__(SOLPSFunction2D)             # <<<<<<<<<<<<<<
 *         m._kdtree = instance._kdtree
 *         m._triangle_to_grid_map = instance._triangle_to_grid_map
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D(((PyTypeObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_m = ((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":86
 *         # copy source data
 *         m = SOLPSFunction2D.__new__(SOLPSFunction2D)
 *         m._kdtree = instance._kdtree             # <<<<<<<<<<<<<<
 *         m._triangle_to_grid_map = instance._triangle_to_grid_map
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_instance->_kdtree);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_m->_kdtree);
  __Pyx_DECREF((PyObject *)__pyx_v_m->_kdtree);
  __pyx_v_m->_kdtree = ((struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":87
 *         m = SOLPSFunction2D.__new__(SOLPSFunction2D)
 *         m._kdtree = instance._kdtree
 *         m._triangle_to_grid_map = instance._triangle_to_grid_map             # <<<<<<<<<<<<<<
 * 
 *         # do we have replacement triangle data?
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_instance->_triangle_to_grid_map);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_m->_triangle_to_grid_map);
  __Pyx_DECREF((PyObject *)__pyx_v_m->_triangle_to_grid_map);
  __pyx_v_m->_triangle_to_grid_map = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":90
 * 
 *         # do we have replacement triangle data?
 *         if grid_data is None:             # <<<<<<<<<<<<<<
 *             m._grid_data = instance._grid_data
 *         else:
 */
  __pyx_t_2 = (__pyx_v_grid_data == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "cherab/solps/solps_2d_functions.pyx":91
 *         # do we have replacement triangle data?
 *         if grid_data is None:
 *             m._grid_data = instance._grid_data             # <<<<<<<<<<<<<<
 *         else:
 *             m._grid_data = grid_data
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_instance->_grid_data);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_m->_grid_data);
    __Pyx_DECREF((PyObject *)__pyx_v_m->_grid_data);
    __pyx_v_m->_grid_data = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "cherab/solps/solps_2d_functions.pyx":90
 * 
 *         # do we have replacement triangle data?
 *         if grid_data is None:             # <<<<<<<<<<<<<<
 *             m._grid_data = instance._grid_data
 *         else:
 */
    goto __pyx_L3;
  }

  /* "cherab/solps/solps_2d_functions.pyx":93
 *             m._grid_data = instance._grid_data
 *         else:
 *             m._grid_data = grid_data             # <<<<<<<<<<<<<<
 * 
 *         m._triangle_to_grid_map_mv = m._triangle_to_grid_map
 */
  /*else*/ {
    if (!(likely(((__pyx_v_grid_data) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_grid_data, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 93, __pyx_L1_error)
    __pyx_t_1 = __pyx_v_grid_data;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_m->_grid_data);
    __Pyx_DECREF((PyObject *)__pyx_v_m->_grid_data);
    __pyx_v_m->_grid_data = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "cherab/solps/solps_2d_functions.pyx":95
 *             m._grid_data = grid_data
 * 
 *         m._triangle_to_grid_map_mv = m._triangle_to_grid_map             # <<<<<<<<<<<<<<
 *         m._grid_data_mv = m._grid_data
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_m->_triangle_to_grid_map), PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 95, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_m->_triangle_to_grid_map_mv, 0);
  __pyx_v_m->_triangle_to_grid_map_mv = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":96
 * 
 *         m._triangle_to_grid_map_mv = m._triangle_to_grid_map
 *         m._grid_data_mv = m._grid_data             # <<<<<<<<<<<<<<
 * 
 *         return m
 */
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(((PyObject *)__pyx_v_m->_grid_data), PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 96, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_m->_grid_data_mv, 0);
  __pyx_v_m->_grid_data_mv = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":98
 *         m._grid_data_mv = m._grid_data
 * 
 *         return m             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_m);
  __pyx_r = ((PyObject *)__pyx_v_m);
  goto __pyx_L0;

  /* "cherab/solps/solps_2d_functions.pyx":65
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def instance(cls, SOLPSFunction2D instance not None, object grid_data=None):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSFunction2D.instance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_m);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cherab/solps/solps_2d_functions.pyx":103
 *     @cython.wraparound(False)
 *     @cython.initializedcheck(False)
 *     cdef double evaluate(self, double x, double y) except? -1e999:             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */

static double __pyx_f_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_evaluate(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *__pyx_v_self, double __pyx_v_x, double __pyx_v_y) {
  __pyx_t_5numpy_int32_t __pyx_v_triangle_id;
  __pyx_t_5numpy_int32_t __pyx_v_ix;
  __pyx_t_5numpy_int32_t __pyx_v_iy;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __pyx_t_5numpy_int32_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("evaluate", 0);

  /* "cherab/solps/solps_2d_functions.pyx":108
 *             np.int32_t triangle_id, ix, iy
 * 
 *         if self._kdtree.is_contained(new_point2d(x, y)):             # <<<<<<<<<<<<<<
 * 
 *             triangle_id = self._kdtree.triangle_id
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7raysect_4core_4math_5point_new_point2d(__pyx_v_x, __pyx_v_y)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (((struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)__pyx_v_self->_kdtree->__pyx_base.__pyx_vtab)->__pyx_base.is_contained(((struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *)__pyx_v_self->_kdtree), ((struct __pyx_obj_7raysect_4core_4math_5point_Point2D *)__pyx_t_1), 0) != 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "cherab/solps/solps_2d_functions.pyx":110
 *         if self._kdtree.is_contained(new_point2d(x, y)):
 * 
 *             triangle_id = self._kdtree.triangle_id             # <<<<<<<<<<<<<<
 *             ix = self._triangle_to_grid_map_mv[triangle_id, 0]
 *             iy = self._triangle_to_grid_map_mv[triangle_id, 1]
 */
    __pyx_t_3 = __pyx_v_self->_kdtree->triangle_id;
    __pyx_v_triangle_id = __pyx_t_3;

    /* "cherab/solps/solps_2d_functions.pyx":111
 * 
 *             triangle_id = self._kdtree.triangle_id
 *             ix = self._triangle_to_grid_map_mv[triangle_id, 0]             # <<<<<<<<<<<<<<
 *             iy = self._triangle_to_grid_map_mv[triangle_id, 1]
 *             return self._grid_data_mv[ix, iy]
 */
    __pyx_t_4 = __pyx_v_triangle_id;
    __pyx_t_5 = 0;
    __pyx_v_ix = (*((__pyx_t_5numpy_int32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->_triangle_to_grid_map_mv.data + __pyx_t_4 * __pyx_v_self->_triangle_to_grid_map_mv.strides[0]) )) + __pyx_t_5)) )));

    /* "cherab/solps/solps_2d_functions.pyx":112
 *             triangle_id = self._kdtree.triangle_id
 *             ix = self._triangle_to_grid_map_mv[triangle_id, 0]
 *             iy = self._triangle_to_grid_map_mv[triangle_id, 1]             # <<<<<<<<<<<<<<
 *             return self._grid_data_mv[ix, iy]
 * 
 */
    __pyx_t_5 = __pyx_v_triangle_id;
    __pyx_t_4 = 1;
    __pyx_v_iy = (*((__pyx_t_5numpy_int32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->_triangle_to_grid_map_mv.data + __pyx_t_5 * __pyx_v_self->_triangle_to_grid_map_mv.strides[0]) )) + __pyx_t_4)) )));

    /* "cherab/solps/solps_2d_functions.pyx":113
 *             ix = self._triangle_to_grid_map_mv[triangle_id, 0]
 *             iy = self._triangle_to_grid_map_mv[triangle_id, 1]
 *             return self._grid_data_mv[ix, iy]             # <<<<<<<<<<<<<<
 * 
 *         return 0.0
 */
    __pyx_t_4 = __pyx_v_ix;
    __pyx_t_5 = __pyx_v_iy;
    __pyx_r = (*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v_self->_grid_data_mv.data + __pyx_t_4 * __pyx_v_self->_grid_data_mv.strides[0]) )) + __pyx_t_5)) )));
    goto __pyx_L0;

    /* "cherab/solps/solps_2d_functions.pyx":108
 *             np.int32_t triangle_id, ix, iy
 * 
 *         if self._kdtree.is_contained(new_point2d(x, y)):             # <<<<<<<<<<<<<<
 * 
 *             triangle_id = self._kdtree.triangle_id
 */
  }

  /* "cherab/solps/solps_2d_functions.pyx":115
 *             return self._grid_data_mv[ix, iy]
 * 
 *         return 0.0             # <<<<<<<<<<<<<<
 * 
 * cdef class SOLPSVectorFunction2D(VectorFunction2D):
 */
  __pyx_r = 0.0;
  goto __pyx_L0;

  /* "cherab/solps/solps_2d_functions.pyx":103
 *     @cython.wraparound(False)
 *     @cython.initializedcheck(False)
 *     cdef double evaluate(self, double x, double y) except? -1e999:             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSFunction2D.evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = (-Py_HUGE_VAL);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cherab/solps/solps_2d_functions.pyx":119
 * cdef class SOLPSVectorFunction2D(VectorFunction2D):
 * 
 *     def __init__(self, object vertex_coords not None, object triangles not None, object triangle_to_grid_map not None, object grid_vectors not None):             # <<<<<<<<<<<<<<
 * 
 *         # use numpy arrays to store data internally
 */

/* Python wrapper */
static int __pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_vertex_coords = 0;
  PyObject *__pyx_v_triangles = 0;
  PyObject *__pyx_v_triangle_to_grid_map = 0;
  PyObject *__pyx_v_grid_vectors = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_vertex_coords,&__pyx_n_s_triangles,&__pyx_n_s_triangle_to_grid_map,&__pyx_n_s_grid_vectors,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_vertex_coords,&__pyx_n_s_triangles,&__pyx_n_s_triangle_to_grid_map,&__pyx_n_s_grid_vectors,0};
    #endif
    PyObject* values[4] = {0,0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_vertex_coords)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 119, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_triangles)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 119, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(0, 119, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_triangle_to_grid_map)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 119, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(0, 119, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_grid_vectors)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 119, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(0, 119, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 119, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
    }
    __pyx_v_vertex_coords = values[0];
    __pyx_v_triangles = values[1];
    __pyx_v_triangle_to_grid_map = values[2];
    __pyx_v_grid_vectors = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 119, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSVectorFunction2D.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_vertex_coords) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "vertex_coords"); __PYX_ERR(0, 119, __pyx_L1_error)
  }
  if (unlikely(((PyObject *)__pyx_v_triangles) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "triangles"); __PYX_ERR(0, 119, __pyx_L1_error)
  }
  if (unlikely(((PyObject *)__pyx_v_triangle_to_grid_map) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "triangle_to_grid_map"); __PYX_ERR(0, 119, __pyx_L1_error)
  }
  if (unlikely(((PyObject *)__pyx_v_grid_vectors) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "grid_vectors"); __PYX_ERR(0, 119, __pyx_L1_error)
  }
  __pyx_r = __pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D___init__(((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *)__pyx_v_self), __pyx_v_vertex_coords, __pyx_v_triangles, __pyx_v_triangle_to_grid_map, __pyx_v_grid_vectors);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D___init__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_self, PyObject *__pyx_v_vertex_coords, PyObject *__pyx_v_triangles, PyObject *__pyx_v_triangle_to_grid_map, PyObject *__pyx_v_grid_vectors) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_vertex_coords);
  __Pyx_INCREF(__pyx_v_triangles);
  __Pyx_INCREF(__pyx_v_triangle_to_grid_map);

  /* "cherab/solps/solps_2d_functions.pyx":122
 * 
 *         # use numpy arrays to store data internally
 *         vertex_coords = np.array(vertex_coords, dtype=np.float64)             # <<<<<<<<<<<<<<
 *         triangles = np.array(triangles, dtype=np.int32)
 *         triangle_to_grid_map = np.array(triangle_to_grid_map, dtype=np.int32)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_vertex_coords);
  __Pyx_GIVEREF(__pyx_v_vertex_coords);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_vertex_coords);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_vertex_coords, __pyx_t_5);
  __pyx_t_5 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":123
 *         # use numpy arrays to store data internally
 *         vertex_coords = np.array(vertex_coords, dtype=np.float64)
 *         triangles = np.array(triangles, dtype=np.int32)             # <<<<<<<<<<<<<<
 *         triangle_to_grid_map = np.array(triangle_to_grid_map, dtype=np.int32)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_triangles);
  __Pyx_GIVEREF(__pyx_v_triangles);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_triangles);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int32); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_triangles, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":124
 *         vertex_coords = np.array(vertex_coords, dtype=np.float64)
 *         triangles = np.array(triangles, dtype=np.int32)
 *         triangle_to_grid_map = np.array(triangle_to_grid_map, dtype=np.int32)             # <<<<<<<<<<<<<<
 * 
 *         # Attention!!! Do not copy grid_vectors! Attribute self._grid_vectors must point to the original data array,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_triangle_to_grid_map);
  __Pyx_GIVEREF(__pyx_v_triangle_to_grid_map);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_triangle_to_grid_map);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_int32); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF_SET(__pyx_v_triangle_to_grid_map, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":130
 * 
 *         # build kdtree
 *         self._kdtree = MeshKDTree2D(vertex_coords, triangles)             # <<<<<<<<<<<<<<
 * 
 *         # populate internal attributes
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_vertex_coords);
  __Pyx_GIVEREF(__pyx_v_vertex_coords);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_vertex_coords);
  __Pyx_INCREF(__pyx_v_triangles);
  __Pyx_GIVEREF(__pyx_v_triangles);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_triangles);
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D), __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_kdtree);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_kdtree);
  __pyx_v_self->_kdtree = ((struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":133
 * 
 *         # populate internal attributes
 *         self._grid_vectors = grid_vectors             # <<<<<<<<<<<<<<
 *         self._triangle_to_grid_map = triangle_to_grid_map
 *         self._grid_vectors_mv = self._grid_vectors
 */
  if (!(likely(__Pyx_TypeTest(__pyx_v_grid_vectors, __pyx_ptype_5numpy_ndarray)))) __PYX_ERR(0, 133, __pyx_L1_error)
  __pyx_t_5 = __pyx_v_grid_vectors;
  __Pyx_INCREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_grid_vectors);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_grid_vectors);
  __pyx_v_self->_grid_vectors = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":134
 *         # populate internal attributes
 *         self._grid_vectors = grid_vectors
 *         self._triangle_to_grid_map = triangle_to_grid_map             # <<<<<<<<<<<<<<
 *         self._grid_vectors_mv = self._grid_vectors
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 */
  if (!(likely(((__pyx_v_triangle_to_grid_map) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_triangle_to_grid_map, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 134, __pyx_L1_error)
  __pyx_t_5 = __pyx_v_triangle_to_grid_map;
  __Pyx_INCREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  __pyx_v_self->_triangle_to_grid_map = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":135
 *         self._grid_vectors = grid_vectors
 *         self._triangle_to_grid_map = triangle_to_grid_map
 *         self._grid_vectors_mv = self._grid_vectors             # <<<<<<<<<<<<<<
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_double(((PyObject *)__pyx_v_self->_grid_vectors), PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 135, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_grid_vectors_mv, 0);
  __pyx_v_self->_grid_vectors_mv = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":136
 *         self._triangle_to_grid_map = triangle_to_grid_map
 *         self._grid_vectors_mv = self._grid_vectors
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_self->_triangle_to_grid_map), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 136, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_triangle_to_grid_map_mv, 0);
  __pyx_v_self->_triangle_to_grid_map_mv = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":119
 * cdef class SOLPSVectorFunction2D(VectorFunction2D):
 * 
 *     def __init__(self, object vertex_coords not None, object triangles not None, object triangle_to_grid_map not None, object grid_vectors not None):             # <<<<<<<<<<<<<<
 * 
 *         # use numpy arrays to store data internally
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSVectorFunction2D.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vertex_coords);
  __Pyx_XDECREF(__pyx_v_triangles);
  __Pyx_XDECREF(__pyx_v_triangle_to_grid_map);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cherab/solps/solps_2d_functions.pyx":138
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self._grid_vectors, self._triangle_to_grid_map, self._kdtree
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_3__getstate__ = {"__getstate__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_3__getstate__, METH_NOARGS, 0};
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_2__getstate__(((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_2__getstate__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "cherab/solps/solps_2d_functions.pyx":139
 * 
 *     def __getstate__(self):
 *         return self._grid_vectors, self._triangle_to_grid_map, self._kdtree             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_grid_vectors);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_grid_vectors);
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->_grid_vectors));
  __Pyx_INCREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->_triangle_to_grid_map));
  __Pyx_INCREF((PyObject *)__pyx_v_self->_kdtree);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_kdtree);
  PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->_kdtree));
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cherab/solps/solps_2d_functions.pyx":138
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self._grid_vectors, self._triangle_to_grid_map, self._kdtree
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSVectorFunction2D.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cherab/solps/solps_2d_functions.pyx":141
 *         return self._grid_vectors, self._triangle_to_grid_map, self._kdtree
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._grid_vectors, self._triangle_to_grid_map, self._kdtree = state
 *         self._grid_vectors_mv = self._grid_vectors
 */

/* Python wrapper */
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static PyMethodDef __pyx_mdef_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_5__setstate__ = {"__setstate__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_5__setstate__, METH_O, 0};
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_4__setstate__(((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_4__setstate__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "cherab/solps/solps_2d_functions.pyx":142
 * 
 *     def __setstate__(self, state):
 *         self._grid_vectors, self._triangle_to_grid_map, self._kdtree = state             # <<<<<<<<<<<<<<
 *         self._grid_vectors_mv = self._grid_vectors
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {
    PyObject* sequence = __pyx_v_state;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 142, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 2; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 3) < 0) __PYX_ERR(0, 142, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 142, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 142, __pyx_L1_error)
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 142, __pyx_L1_error)
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D))))) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_grid_vectors);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_grid_vectors);
  __pyx_v_self->_grid_vectors = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_triangle_to_grid_map);
  __pyx_v_self->_triangle_to_grid_map = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_kdtree);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_kdtree);
  __pyx_v_self->_kdtree = ((struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":143
 *     def __setstate__(self, state):
 *         self._grid_vectors, self._triangle_to_grid_map, self._kdtree = state
 *         self._grid_vectors_mv = self._grid_vectors             # <<<<<<<<<<<<<<
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_double(((PyObject *)__pyx_v_self->_grid_vectors), PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 143, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_grid_vectors_mv, 0);
  __pyx_v_self->_grid_vectors_mv = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":144
 *         self._grid_vectors, self._triangle_to_grid_map, self._kdtree = state
 *         self._grid_vectors_mv = self._grid_vectors
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_self->_triangle_to_grid_map), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 144, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_triangle_to_grid_map_mv, 0);
  __pyx_v_self->_triangle_to_grid_map_mv = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":141
 *         return self._grid_vectors, self._triangle_to_grid_map, self._kdtree
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._grid_vectors, self._triangle_to_grid_map, self._kdtree = state
 *         self._grid_vectors_mv = self._grid_vectors
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSVectorFunction2D.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cherab/solps/solps_2d_functions.pyx":146
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_7__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_7__reduce__ = {"__reduce__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_7__reduce__, METH_NOARGS, 0};
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_7__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_6__reduce__(((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_6__reduce__(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "cherab/solps/solps_2d_functions.pyx":147
 * 
 *     def __reduce__(self):
 *         return self.__new__, (self.__class__, ), self.__getstate__()             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getstate); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_5, };
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cherab/solps/solps_2d_functions.pyx":146
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSVectorFunction2D.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cherab/solps/solps_2d_functions.pyx":149
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def instance(cls, SOLPSVectorFunction2D instance not None, object grid_vectors=None):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_9instance(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_8instance, "\n        Creates a new interpolator instance from an existing interpolator instance.\n        The new interpolator instance will share the same internal acceleration\n        data as the original interpolator. The grid_data of the new instance can\n        be redefined.\n        This method should be used if the user has multiple sets of grid_data\n        that lie on the same mesh geometry. Using this methods avoids the\n        repeated rebuilding of the mesh acceleration structures by sharing the\n        geometry data between multiple interpolator objects.\n        :param SOLPSVectorFunction2D instance: SOLPSVectorFunction2D object.\n        :param ndarray grid_vectors: An array containing vector data on SOLPS grid.\n        :return: A SOLPSVectorFunction2D object.\n        :rtype: SOLPSVectorFunction2D\n        ");
static PyMethodDef __pyx_mdef_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_9instance = {"instance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_9instance, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_8instance};
static PyObject *__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_9instance(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_instance = 0;
  PyObject *__pyx_v_grid_vectors = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("instance (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_instance,&__pyx_n_s_grid_vectors,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_instance,&__pyx_n_s_grid_vectors,0};
    #endif
    PyObject* values[2] = {0,0};

    /* "cherab/solps/solps_2d_functions.pyx":150
 * 
 *     @classmethod
 *     def instance(cls, SOLPSVectorFunction2D instance not None, object grid_vectors=None):             # <<<<<<<<<<<<<<
 *         """
 *         Creates a new interpolator instance from an existing interpolator instance.
 */
    values[1] = ((PyObject *)Py_None);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_instance)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 149, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_grid_vectors);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 149, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "instance") < 0)) __PYX_ERR(0, 149, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_instance = ((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *)values[0]);
    __pyx_v_grid_vectors = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("instance", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 149, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSVectorFunction2D.instance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_instance), __pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D, 0, "instance", 0))) __PYX_ERR(0, 150, __pyx_L1_error)
  __pyx_r = __pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_8instance(((PyTypeObject*)__pyx_v_cls), __pyx_v_instance, __pyx_v_grid_vectors);

  /* "cherab/solps/solps_2d_functions.pyx":149
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def instance(cls, SOLPSVectorFunction2D instance not None, object grid_vectors=None):
 *         """
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_8instance(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_instance, PyObject *__pyx_v_grid_vectors) {
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_m = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("instance", 0);

  /* "cherab/solps/solps_2d_functions.pyx":169
 * 
 *         # copy source data
 *         m = SOLPSVectorFunction2D.__new__(SOLPSVectorFunction2D)             # <<<<<<<<<<<<<<
 *         m._kdtree = instance._kdtree
 *         m._triangle_to_grid_map = instance._triangle_to_grid_map
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D(((PyTypeObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_m = ((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":170
 *         # copy source data
 *         m = SOLPSVectorFunction2D.__new__(SOLPSVectorFunction2D)
 *         m._kdtree = instance._kdtree             # <<<<<<<<<<<<<<
 *         m._triangle_to_grid_map = instance._triangle_to_grid_map
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_instance->_kdtree);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_m->_kdtree);
  __Pyx_DECREF((PyObject *)__pyx_v_m->_kdtree);
  __pyx_v_m->_kdtree = ((struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":171
 *         m = SOLPSVectorFunction2D.__new__(SOLPSVectorFunction2D)
 *         m._kdtree = instance._kdtree
 *         m._triangle_to_grid_map = instance._triangle_to_grid_map             # <<<<<<<<<<<<<<
 * 
 *         # do we have replacement triangle data?
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_instance->_triangle_to_grid_map);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_m->_triangle_to_grid_map);
  __Pyx_DECREF((PyObject *)__pyx_v_m->_triangle_to_grid_map);
  __pyx_v_m->_triangle_to_grid_map = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cherab/solps/solps_2d_functions.pyx":174
 * 
 *         # do we have replacement triangle data?
 *         if grid_vectors is None:             # <<<<<<<<<<<<<<
 *             m._grid_vectors = instance._grid_vectors
 *         else:
 */
  __pyx_t_2 = (__pyx_v_grid_vectors == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "cherab/solps/solps_2d_functions.pyx":175
 *         # do we have replacement triangle data?
 *         if grid_vectors is None:
 *             m._grid_vectors = instance._grid_vectors             # <<<<<<<<<<<<<<
 *         else:
 *             m._grid_vectors = grid_vectors
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_instance->_grid_vectors);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_m->_grid_vectors);
    __Pyx_DECREF((PyObject *)__pyx_v_m->_grid_vectors);
    __pyx_v_m->_grid_vectors = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "cherab/solps/solps_2d_functions.pyx":174
 * 
 *         # do we have replacement triangle data?
 *         if grid_vectors is None:             # <<<<<<<<<<<<<<
 *             m._grid_vectors = instance._grid_vectors
 *         else:
 */
    goto __pyx_L3;
  }

  /* "cherab/solps/solps_2d_functions.pyx":177
 *             m._grid_vectors = instance._grid_vectors
 *         else:
 *             m._grid_vectors = grid_vectors             # <<<<<<<<<<<<<<
 * 
 *         m._triangle_to_grid_map_mv = m._triangle_to_grid_map
 */
  /*else*/ {
    if (!(likely(((__pyx_v_grid_vectors) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_grid_vectors, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 177, __pyx_L1_error)
    __pyx_t_1 = __pyx_v_grid_vectors;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_m->_grid_vectors);
    __Pyx_DECREF((PyObject *)__pyx_v_m->_grid_vectors);
    __pyx_v_m->_grid_vectors = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "cherab/solps/solps_2d_functions.pyx":179
 *             m._grid_vectors = grid_vectors
 * 
 *         m._triangle_to_grid_map_mv = m._triangle_to_grid_map             # <<<<<<<<<<<<<<
 *         m._grid_vectors_mv = m._grid_vectors
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_m->_triangle_to_grid_map), PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 179, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_m->_triangle_to_grid_map_mv, 0);
  __pyx_v_m->_triangle_to_grid_map_mv = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":180
 * 
 *         m._triangle_to_grid_map_mv = m._triangle_to_grid_map
 *         m._grid_vectors_mv = m._grid_vectors             # <<<<<<<<<<<<<<
 * 
 *         return m
 */
  __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_double(((PyObject *)__pyx_v_m->_grid_vectors), PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 180, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_m->_grid_vectors_mv, 0);
  __pyx_v_m->_grid_vectors_mv = __pyx_t_5;
  __pyx_t_5.memview = NULL;
  __pyx_t_5.data = NULL;

  /* "cherab/solps/solps_2d_functions.pyx":182
 *         m._grid_vectors_mv = m._grid_vectors
 * 
 *         return m             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_m);
  __pyx_r = ((PyObject *)__pyx_v_m);
  goto __pyx_L0;

  /* "cherab/solps/solps_2d_functions.pyx":149
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def instance(cls, SOLPSVectorFunction2D instance not None, object grid_vectors=None):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 1);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSVectorFunction2D.instance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_m);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cherab/solps/solps_2d_functions.pyx":187
 *     @cython.wraparound(False)
 *     @cython.initializedcheck(False)
 *     cdef Vector3D evaluate(self, double x, double y):             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */

static struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *__pyx_f_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_evaluate(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *__pyx_v_self, double __pyx_v_x, double __pyx_v_y) {
  __pyx_t_5numpy_int32_t __pyx_v_triangle_id;
  __pyx_t_5numpy_int32_t __pyx_v_ix;
  __pyx_t_5numpy_int32_t __pyx_v_iy;
  double __pyx_v_vx;
  double __pyx_v_vy;
  double __pyx_v_vz;
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __pyx_t_5numpy_int32_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("evaluate", 0);

  /* "cherab/solps/solps_2d_functions.pyx":193
 *             double vx, vy, vz
 * 
 *         if self._kdtree.is_contained(new_point2d(x, y)):             # <<<<<<<<<<<<<<
 * 
 *             triangle_id = self._kdtree.triangle_id
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_7raysect_4core_4math_5point_new_point2d(__pyx_v_x, __pyx_v_y)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (((struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)__pyx_v_self->_kdtree->__pyx_base.__pyx_vtab)->__pyx_base.is_contained(((struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore *)__pyx_v_self->_kdtree), ((struct __pyx_obj_7raysect_4core_4math_5point_Point2D *)__pyx_t_1), 0) != 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "cherab/solps/solps_2d_functions.pyx":195
 *         if self._kdtree.is_contained(new_point2d(x, y)):
 * 
 *             triangle_id = self._kdtree.triangle_id             # <<<<<<<<<<<<<<
 *             ix = self._triangle_to_grid_map_mv[triangle_id, 0]
 *             iy = self._triangle_to_grid_map_mv[triangle_id, 1]
 */
    __pyx_t_3 = __pyx_v_self->_kdtree->triangle_id;
    __pyx_v_triangle_id = __pyx_t_3;

    /* "cherab/solps/solps_2d_functions.pyx":196
 * 
 *             triangle_id = self._kdtree.triangle_id
 *             ix = self._triangle_to_grid_map_mv[triangle_id, 0]             # <<<<<<<<<<<<<<
 *             iy = self._triangle_to_grid_map_mv[triangle_id, 1]
 *             vx = self._grid_vectors_mv[0, ix, iy]
 */
    __pyx_t_4 = __pyx_v_triangle_id;
    __pyx_t_5 = 0;
    __pyx_v_ix = (*((__pyx_t_5numpy_int32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->_triangle_to_grid_map_mv.data + __pyx_t_4 * __pyx_v_self->_triangle_to_grid_map_mv.strides[0]) )) + __pyx_t_5)) )));

    /* "cherab/solps/solps_2d_functions.pyx":197
 *             triangle_id = self._kdtree.triangle_id
 *             ix = self._triangle_to_grid_map_mv[triangle_id, 0]
 *             iy = self._triangle_to_grid_map_mv[triangle_id, 1]             # <<<<<<<<<<<<<<
 *             vx = self._grid_vectors_mv[0, ix, iy]
 *             vy = self._grid_vectors_mv[1, ix, iy]
 */
    __pyx_t_5 = __pyx_v_triangle_id;
    __pyx_t_4 = 1;
    __pyx_v_iy = (*((__pyx_t_5numpy_int32_t *) ( /* dim=1 */ ((char *) (((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->_triangle_to_grid_map_mv.data + __pyx_t_5 * __pyx_v_self->_triangle_to_grid_map_mv.strides[0]) )) + __pyx_t_4)) )));

    /* "cherab/solps/solps_2d_functions.pyx":198
 *             ix = self._triangle_to_grid_map_mv[triangle_id, 0]
 *             iy = self._triangle_to_grid_map_mv[triangle_id, 1]
 *             vx = self._grid_vectors_mv[0, ix, iy]             # <<<<<<<<<<<<<<
 *             vy = self._grid_vectors_mv[1, ix, iy]
 *             vz = self._grid_vectors_mv[2, ix, iy]
 */
    __pyx_t_4 = 0;
    __pyx_t_5 = __pyx_v_ix;
    __pyx_t_6 = __pyx_v_iy;
    __pyx_v_vx = (*((double *) ( /* dim=2 */ ((char *) (((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_self->_grid_vectors_mv.data + __pyx_t_4 * __pyx_v_self->_grid_vectors_mv.strides[0]) ) + __pyx_t_5 * __pyx_v_self->_grid_vectors_mv.strides[1]) )) + __pyx_t_6)) )));

    /* "cherab/solps/solps_2d_functions.pyx":199
 *             iy = self._triangle_to_grid_map_mv[triangle_id, 1]
 *             vx = self._grid_vectors_mv[0, ix, iy]
 *             vy = self._grid_vectors_mv[1, ix, iy]             # <<<<<<<<<<<<<<
 *             vz = self._grid_vectors_mv[2, ix, iy]
 * 
 */
    __pyx_t_6 = 1;
    __pyx_t_5 = __pyx_v_ix;
    __pyx_t_4 = __pyx_v_iy;
    __pyx_v_vy = (*((double *) ( /* dim=2 */ ((char *) (((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_self->_grid_vectors_mv.data + __pyx_t_6 * __pyx_v_self->_grid_vectors_mv.strides[0]) ) + __pyx_t_5 * __pyx_v_self->_grid_vectors_mv.strides[1]) )) + __pyx_t_4)) )));

    /* "cherab/solps/solps_2d_functions.pyx":200
 *             vx = self._grid_vectors_mv[0, ix, iy]
 *             vy = self._grid_vectors_mv[1, ix, iy]
 *             vz = self._grid_vectors_mv[2, ix, iy]             # <<<<<<<<<<<<<<
 * 
 *             return new_vector3d(vx, vy, vz)
 */
    __pyx_t_4 = 2;
    __pyx_t_5 = __pyx_v_ix;
    __pyx_t_6 = __pyx_v_iy;
    __pyx_v_vz = (*((double *) ( /* dim=2 */ ((char *) (((double *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_self->_grid_vectors_mv.data + __pyx_t_4 * __pyx_v_self->_grid_vectors_mv.strides[0]) ) + __pyx_t_5 * __pyx_v_self->_grid_vectors_mv.strides[1]) )) + __pyx_t_6)) )));

    /* "cherab/solps/solps_2d_functions.pyx":202
 *             vz = self._grid_vectors_mv[2, ix, iy]
 * 
 *             return new_vector3d(vx, vy, vz)             # <<<<<<<<<<<<<<
 * 
 *         return new_vector3d(0, 0, 0)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_t_1 = ((PyObject *)__pyx_f_7raysect_4core_4math_6vector_new_vector3d(__pyx_v_vx, __pyx_v_vy, __pyx_v_vz)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = ((struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "cherab/solps/solps_2d_functions.pyx":193
 *             double vx, vy, vz
 * 
 *         if self._kdtree.is_contained(new_point2d(x, y)):             # <<<<<<<<<<<<<<
 * 
 *             triangle_id = self._kdtree.triangle_id
 */
  }

  /* "cherab/solps/solps_2d_functions.pyx":204
 *             return new_vector3d(vx, vy, vz)
 * 
 *         return new_vector3d(0, 0, 0)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_7raysect_4core_4math_6vector_new_vector3d(0.0, 0.0, 0.0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cherab/solps/solps_2d_functions.pyx":187
 *     @cython.wraparound(False)
 *     @cython.initializedcheck(False)
 *     cdef Vector3D evaluate(self, double x, double y):             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cherab.solps.solps_2d_functions.SOLPSVectorFunction2D.evaluate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":246
 * 
 *         @property
 *         cdef inline PyObject* base(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self) {
  PyObject *__pyx_r;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":249
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 *             return PyArray_BASE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_BASE(__pyx_v_self);
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":246
 * 
 *         @property
 *         cdef inline PyObject* base(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":252
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self) {
  PyArray_Descr *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyArray_Descr *__pyx_t_1;
  __Pyx_RefNannySetupContext("descr", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":255
 *             """Returns an owned reference to the dtype of the array.
 *             """
 *             return <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyArray_DESCR(__pyx_v_self);
  __Pyx_INCREF((PyObject *)((PyArray_Descr *)__pyx_t_1));
  __pyx_r = ((PyArray_Descr *)__pyx_t_1);
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":252
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":258
 * 
 *         @property
 *         cdef inline int ndim(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self) {
  int __pyx_r;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":261
 *             """Returns the number of dimensions in the array.
 *             """
 *             return PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":258
 * 
 *         @property
 *         cdef inline int ndim(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":264
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":269
 *             Can return NULL for 0-dimensional arrays.
 *             """
 *             return PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":264
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":272
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":276
 *             The number of elements matches the number of dimensions of the array (ndim).
 *             """
 *             return PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_STRIDES(__pyx_v_self);
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":272
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":279
 * 
 *         @property
 *         cdef inline npy_intp size(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":282
 *             """Returns the total size (in number of elements) of the array.
 *             """
 *             return PyArray_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":279
 * 
 *         @property
 *         cdef inline npy_intp size(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":285
 * 
 *         @property
 *         cdef inline char* data(self) nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self) {
  char *__pyx_r;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":291
 *             of `PyArray_DATA()` instead, which returns a 'void*'.
 *             """
 *             return PyArray_BYTES(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef unsigned char      npy_bool
 */
  __pyx_r = PyArray_BYTES(__pyx_v_self);
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":285
 * 
 *         @property
 *         cdef inline char* data(self) nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":777
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":778
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":777
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":780
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":781
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":780
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":783
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":784
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":783
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":786
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":787
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":786
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":789
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":790
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":789
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":792
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":793
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":794
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":793
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":796
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":792
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":950
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":951
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":952
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":950
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":954
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":955
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":956
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":957
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":956
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":958
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":954
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":962
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":963
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":964
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 964, __pyx_L3_error)

      /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":963
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":965
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 965, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":966
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 966, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 966, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":963
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":962
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":968
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":969
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":970
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 970, __pyx_L3_error)

      /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":969
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":971
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 971, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":972
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 972, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 972, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":969
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":968
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":974
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":975
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":976
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 976, __pyx_L3_error)

      /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":975
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":977
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 977, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":978
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 978, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 978, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":975
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":974
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":981
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":993
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":981
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":996
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1008
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":996
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1011
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1018
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1011
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1021
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1025
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1021
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1028
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1032
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":1028
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "raysect/core/math/affinematrix.pxd":42
 * 
 * 
 * cdef inline AffineMatrix3D new_affinematrix3d(double m00, double m01, double m02, double m03,             # <<<<<<<<<<<<<<
 *                                               double m10, double m11, double m12, double m13,
 *                                               double m20, double m21, double m22, double m23,
 */

static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *__pyx_f_7raysect_4core_4math_12affinematrix_new_affinematrix3d(double __pyx_v_m00, double __pyx_v_m01, double __pyx_v_m02, double __pyx_v_m03, double __pyx_v_m10, double __pyx_v_m11, double __pyx_v_m12, double __pyx_v_m13, double __pyx_v_m20, double __pyx_v_m21, double __pyx_v_m22, double __pyx_v_m23, double __pyx_v_m30, double __pyx_v_m31, double __pyx_v_m32, double __pyx_v_m33) {
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *__pyx_v_v = 0;
  struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_affinematrix3d", 0);

  /* "raysect/core/math/affinematrix.pxd":54
 * 
 *     cdef AffineMatrix3D v
 *     v = AffineMatrix3D.__new__(AffineMatrix3D)             # <<<<<<<<<<<<<<
 *     v.m[0][0] = m00
 *     v.m[0][1] = m01
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7raysect_4core_4math_12affinematrix_AffineMatrix3D), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7raysect_4core_4math_12affinematrix_AffineMatrix3D)))) __PYX_ERR(2, 54, __pyx_L1_error)
  __pyx_v_v = ((struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "raysect/core/math/affinematrix.pxd":55
 *     cdef AffineMatrix3D v
 *     v = AffineMatrix3D.__new__(AffineMatrix3D)
 *     v.m[0][0] = m00             # <<<<<<<<<<<<<<
 *     v.m[0][1] = m01
 *     v.m[0][2] = m02
 */
  ((__pyx_v_v->__pyx_base.m[0])[0]) = __pyx_v_m00;

  /* "raysect/core/math/affinematrix.pxd":56
 *     v = AffineMatrix3D.__new__(AffineMatrix3D)
 *     v.m[0][0] = m00
 *     v.m[0][1] = m01             # <<<<<<<<<<<<<<
 *     v.m[0][2] = m02
 *     v.m[0][3] = m03
 */
  ((__pyx_v_v->__pyx_base.m[0])[1]) = __pyx_v_m01;

  /* "raysect/core/math/affinematrix.pxd":57
 *     v.m[0][0] = m00
 *     v.m[0][1] = m01
 *     v.m[0][2] = m02             # <<<<<<<<<<<<<<
 *     v.m[0][3] = m03
 *     v.m[1][0] = m10
 */
  ((__pyx_v_v->__pyx_base.m[0])[2]) = __pyx_v_m02;

  /* "raysect/core/math/affinematrix.pxd":58
 *     v.m[0][1] = m01
 *     v.m[0][2] = m02
 *     v.m[0][3] = m03             # <<<<<<<<<<<<<<
 *     v.m[1][0] = m10
 *     v.m[1][1] = m11
 */
  ((__pyx_v_v->__pyx_base.m[0])[3]) = __pyx_v_m03;

  /* "raysect/core/math/affinematrix.pxd":59
 *     v.m[0][2] = m02
 *     v.m[0][3] = m03
 *     v.m[1][0] = m10             # <<<<<<<<<<<<<<
 *     v.m[1][1] = m11
 *     v.m[1][2] = m12
 */
  ((__pyx_v_v->__pyx_base.m[1])[0]) = __pyx_v_m10;

  /* "raysect/core/math/affinematrix.pxd":60
 *     v.m[0][3] = m03
 *     v.m[1][0] = m10
 *     v.m[1][1] = m11             # <<<<<<<<<<<<<<
 *     v.m[1][2] = m12
 *     v.m[1][3] = m13
 */
  ((__pyx_v_v->__pyx_base.m[1])[1]) = __pyx_v_m11;

  /* "raysect/core/math/affinematrix.pxd":61
 *     v.m[1][0] = m10
 *     v.m[1][1] = m11
 *     v.m[1][2] = m12             # <<<<<<<<<<<<<<
 *     v.m[1][3] = m13
 *     v.m[2][0] = m20
 */
  ((__pyx_v_v->__pyx_base.m[1])[2]) = __pyx_v_m12;

  /* "raysect/core/math/affinematrix.pxd":62
 *     v.m[1][1] = m11
 *     v.m[1][2] = m12
 *     v.m[1][3] = m13             # <<<<<<<<<<<<<<
 *     v.m[2][0] = m20
 *     v.m[2][1] = m21
 */
  ((__pyx_v_v->__pyx_base.m[1])[3]) = __pyx_v_m13;

  /* "raysect/core/math/affinematrix.pxd":63
 *     v.m[1][2] = m12
 *     v.m[1][3] = m13
 *     v.m[2][0] = m20             # <<<<<<<<<<<<<<
 *     v.m[2][1] = m21
 *     v.m[2][2] = m22
 */
  ((__pyx_v_v->__pyx_base.m[2])[0]) = __pyx_v_m20;

  /* "raysect/core/math/affinematrix.pxd":64
 *     v.m[1][3] = m13
 *     v.m[2][0] = m20
 *     v.m[2][1] = m21             # <<<<<<<<<<<<<<
 *     v.m[2][2] = m22
 *     v.m[2][3] = m23
 */
  ((__pyx_v_v->__pyx_base.m[2])[1]) = __pyx_v_m21;

  /* "raysect/core/math/affinematrix.pxd":65
 *     v.m[2][0] = m20
 *     v.m[2][1] = m21
 *     v.m[2][2] = m22             # <<<<<<<<<<<<<<
 *     v.m[2][3] = m23
 *     v.m[3][0] = m30
 */
  ((__pyx_v_v->__pyx_base.m[2])[2]) = __pyx_v_m22;

  /* "raysect/core/math/affinematrix.pxd":66
 *     v.m[2][1] = m21
 *     v.m[2][2] = m22
 *     v.m[2][3] = m23             # <<<<<<<<<<<<<<
 *     v.m[3][0] = m30
 *     v.m[3][1] = m31
 */
  ((__pyx_v_v->__pyx_base.m[2])[3]) = __pyx_v_m23;

  /* "raysect/core/math/affinematrix.pxd":67
 *     v.m[2][2] = m22
 *     v.m[2][3] = m23
 *     v.m[3][0] = m30             # <<<<<<<<<<<<<<
 *     v.m[3][1] = m31
 *     v.m[3][2] = m32
 */
  ((__pyx_v_v->__pyx_base.m[3])[0]) = __pyx_v_m30;

  /* "raysect/core/math/affinematrix.pxd":68
 *     v.m[2][3] = m23
 *     v.m[3][0] = m30
 *     v.m[3][1] = m31             # <<<<<<<<<<<<<<
 *     v.m[3][2] = m32
 *     v.m[3][3] = m33
 */
  ((__pyx_v_v->__pyx_base.m[3])[1]) = __pyx_v_m31;

  /* "raysect/core/math/affinematrix.pxd":69
 *     v.m[3][0] = m30
 *     v.m[3][1] = m31
 *     v.m[3][2] = m32             # <<<<<<<<<<<<<<
 *     v.m[3][3] = m33
 *     return v
 */
  ((__pyx_v_v->__pyx_base.m[3])[2]) = __pyx_v_m32;

  /* "raysect/core/math/affinematrix.pxd":70
 *     v.m[3][1] = m31
 *     v.m[3][2] = m32
 *     v.m[3][3] = m33             # <<<<<<<<<<<<<<
 *     return v
 */
  ((__pyx_v_v->__pyx_base.m[3])[3]) = __pyx_v_m33;

  /* "raysect/core/math/affinematrix.pxd":71
 *     v.m[3][2] = m32
 *     v.m[3][3] = m33
 *     return v             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "raysect/core/math/affinematrix.pxd":42
 * 
 * 
 * cdef inline AffineMatrix3D new_affinematrix3d(double m00, double m01, double m02, double m03,             # <<<<<<<<<<<<<<
 *                                               double m10, double m11, double m12, double m13,
 *                                               double m20, double m21, double m22, double m23,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("raysect.core.math.affinematrix.new_affinematrix3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_v);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/vector.pxd":62
 * 
 * 
 * cdef inline Vector3D new_vector3d(double x, double y, double z):             # <<<<<<<<<<<<<<
 *     """
 *     Vector3D factory function.
 */

static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *__pyx_f_7raysect_4core_4math_6vector_new_vector3d(double __pyx_v_x, double __pyx_v_y, double __pyx_v_z) {
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *__pyx_v_v = 0;
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_vector3d", 0);

  /* "raysect/core/math/vector.pxd":71
 * 
 *     cdef Vector3D v
 *     v = Vector3D.__new__(Vector3D)             # <<<<<<<<<<<<<<
 *     v.x = x
 *     v.y = y
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7raysect_4core_4math_6vector_Vector3D), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7raysect_4core_4math_6vector_Vector3D)))) __PYX_ERR(3, 71, __pyx_L1_error)
  __pyx_v_v = ((struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "raysect/core/math/vector.pxd":72
 *     cdef Vector3D v
 *     v = Vector3D.__new__(Vector3D)
 *     v.x = x             # <<<<<<<<<<<<<<
 *     v.y = y
 *     v.z = z
 */
  __pyx_v_v->__pyx_base.x = __pyx_v_x;

  /* "raysect/core/math/vector.pxd":73
 *     v = Vector3D.__new__(Vector3D)
 *     v.x = x
 *     v.y = y             # <<<<<<<<<<<<<<
 *     v.z = z
 *     return v
 */
  __pyx_v_v->__pyx_base.y = __pyx_v_y;

  /* "raysect/core/math/vector.pxd":74
 *     v.x = x
 *     v.y = y
 *     v.z = z             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
  __pyx_v_v->__pyx_base.z = __pyx_v_z;

  /* "raysect/core/math/vector.pxd":75
 *     v.y = y
 *     v.z = z
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "raysect/core/math/vector.pxd":62
 * 
 * 
 * cdef inline Vector3D new_vector3d(double x, double y, double z):             # <<<<<<<<<<<<<<
 *     """
 *     Vector3D factory function.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("raysect.core.math.vector.new_vector3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_v);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/vector.pxd":113
 * 
 * 
 * cdef inline Vector2D new_vector2d(double x, double y):             # <<<<<<<<<<<<<<
 *     """
 *     Vector2D factory function.
 */

static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *__pyx_f_7raysect_4core_4math_6vector_new_vector2d(double __pyx_v_x, double __pyx_v_y) {
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *__pyx_v_v = 0;
  struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_vector2d", 0);

  /* "raysect/core/math/vector.pxd":122
 * 
 *     cdef Vector2D v
 *     v = Vector2D.__new__(Vector2D)             # <<<<<<<<<<<<<<
 *     v.x = x
 *     v.y = y
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7raysect_4core_4math_6vector_Vector2D), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7raysect_4core_4math_6vector_Vector2D)))) __PYX_ERR(3, 122, __pyx_L1_error)
  __pyx_v_v = ((struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "raysect/core/math/vector.pxd":123
 *     cdef Vector2D v
 *     v = Vector2D.__new__(Vector2D)
 *     v.x = x             # <<<<<<<<<<<<<<
 *     v.y = y
 *     return v
 */
  __pyx_v_v->x = __pyx_v_x;

  /* "raysect/core/math/vector.pxd":124
 *     v = Vector2D.__new__(Vector2D)
 *     v.x = x
 *     v.y = y             # <<<<<<<<<<<<<<
 *     return v
 */
  __pyx_v_v->y = __pyx_v_y;

  /* "raysect/core/math/vector.pxd":125
 *     v.x = x
 *     v.y = y
 *     return v             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "raysect/core/math/vector.pxd":113
 * 
 * 
 * cdef inline Vector2D new_vector2d(double x, double y):             # <<<<<<<<<<<<<<
 *     """
 *     Vector2D factory function.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("raysect.core.math.vector.new_vector2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_v);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/point.pxd":57
 * 
 * 
 * cdef inline Point3D new_point3d(double x, double y, double z):             # <<<<<<<<<<<<<<
 *     """
 *     Point3D factory function.
 */

static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_5point_Point3D *__pyx_f_7raysect_4core_4math_5point_new_point3d(double __pyx_v_x, double __pyx_v_y, double __pyx_v_z) {
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *__pyx_v_v = 0;
  struct __pyx_obj_7raysect_4core_4math_5point_Point3D *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_point3d", 0);

  /* "raysect/core/math/point.pxd":66
 * 
 *     cdef Point3D v
 *     v = Point3D.__new__(Point3D)             # <<<<<<<<<<<<<<
 *     v.x = x
 *     v.y = y
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7raysect_4core_4math_5point_Point3D), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7raysect_4core_4math_5point_Point3D)))) __PYX_ERR(4, 66, __pyx_L1_error)
  __pyx_v_v = ((struct __pyx_obj_7raysect_4core_4math_5point_Point3D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "raysect/core/math/point.pxd":67
 *     cdef Point3D v
 *     v = Point3D.__new__(Point3D)
 *     v.x = x             # <<<<<<<<<<<<<<
 *     v.y = y
 *     v.z = z
 */
  __pyx_v_v->x = __pyx_v_x;

  /* "raysect/core/math/point.pxd":68
 *     v = Point3D.__new__(Point3D)
 *     v.x = x
 *     v.y = y             # <<<<<<<<<<<<<<
 *     v.z = z
 *     return v
 */
  __pyx_v_v->y = __pyx_v_y;

  /* "raysect/core/math/point.pxd":69
 *     v.x = x
 *     v.y = y
 *     v.z = z             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
  __pyx_v_v->z = __pyx_v_z;

  /* "raysect/core/math/point.pxd":70
 *     v.y = y
 *     v.z = z
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "raysect/core/math/point.pxd":57
 * 
 * 
 * cdef inline Point3D new_point3d(double x, double y, double z):             # <<<<<<<<<<<<<<
 *     """
 *     Point3D factory function.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("raysect.core.math.point.new_point3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_v);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/point.pxd":94
 * 
 * 
 * cdef inline Point2D new_point2d(double x, double y):             # <<<<<<<<<<<<<<
 *     """
 *     Point2D factory function.
 */

static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_5point_Point2D *__pyx_f_7raysect_4core_4math_5point_new_point2d(double __pyx_v_x, double __pyx_v_y) {
  struct __pyx_obj_7raysect_4core_4math_5point_Point2D *__pyx_v_a = 0;
  struct __pyx_obj_7raysect_4core_4math_5point_Point2D *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_point2d", 0);

  /* "raysect/core/math/point.pxd":103
 * 
 *     cdef Point2D a
 *     a = Point2D.__new__(Point2D)             # <<<<<<<<<<<<<<
 *     a.x = x
 *     a.y = y
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7raysect_4core_4math_5point_Point2D), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7raysect_4core_4math_5point_Point2D)))) __PYX_ERR(4, 103, __pyx_L1_error)
  __pyx_v_a = ((struct __pyx_obj_7raysect_4core_4math_5point_Point2D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "raysect/core/math/point.pxd":104
 *     cdef Point2D a
 *     a = Point2D.__new__(Point2D)
 *     a.x = x             # <<<<<<<<<<<<<<
 *     a.y = y
 *     return a
 */
  __pyx_v_a->x = __pyx_v_x;

  /* "raysect/core/math/point.pxd":105
 *     a = Point2D.__new__(Point2D)
 *     a.x = x
 *     a.y = y             # <<<<<<<<<<<<<<
 *     return a
 */
  __pyx_v_a->y = __pyx_v_y;

  /* "raysect/core/math/point.pxd":106
 *     a.x = x
 *     a.y = y
 *     return a             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_a);
  __pyx_r = __pyx_v_a;
  goto __pyx_L0;

  /* "raysect/core/math/point.pxd":94
 * 
 * 
 * cdef inline Point2D new_point2d(double x, double y):             # <<<<<<<<<<<<<<
 *     """
 *     Point2D factory function.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("raysect.core.math.point.new_point2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_a);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/normal.pxd":63
 * 
 * 
 * cdef inline Normal3D new_normal3d(double x, double y, double z):             # <<<<<<<<<<<<<<
 *     """
 *     Normal3D factory function.
 */

static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *__pyx_f_7raysect_4core_4math_6normal_new_normal3d(double __pyx_v_x, double __pyx_v_y, double __pyx_v_z) {
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *__pyx_v_v = 0;
  struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_normal3d", 0);

  /* "raysect/core/math/normal.pxd":72
 * 
 *     cdef Normal3D v
 *     v = Normal3D.__new__(Normal3D)             # <<<<<<<<<<<<<<
 *     v.x = x
 *     v.y = y
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7raysect_4core_4math_6normal_Normal3D), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7raysect_4core_4math_6normal_Normal3D)))) __PYX_ERR(5, 72, __pyx_L1_error)
  __pyx_v_v = ((struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "raysect/core/math/normal.pxd":73
 *     cdef Normal3D v
 *     v = Normal3D.__new__(Normal3D)
 *     v.x = x             # <<<<<<<<<<<<<<
 *     v.y = y
 *     v.z = z
 */
  __pyx_v_v->__pyx_base.x = __pyx_v_x;

  /* "raysect/core/math/normal.pxd":74
 *     v = Normal3D.__new__(Normal3D)
 *     v.x = x
 *     v.y = y             # <<<<<<<<<<<<<<
 *     v.z = z
 *     return v
 */
  __pyx_v_v->__pyx_base.y = __pyx_v_y;

  /* "raysect/core/math/normal.pxd":75
 *     v.x = x
 *     v.y = y
 *     v.z = z             # <<<<<<<<<<<<<<
 *     return v
 */
  __pyx_v_v->__pyx_base.z = __pyx_v_z;

  /* "raysect/core/math/normal.pxd":76
 *     v.y = y
 *     v.z = z
 *     return v             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "raysect/core/math/normal.pxd":63
 * 
 * 
 * cdef inline Normal3D new_normal3d(double x, double y, double z):             # <<<<<<<<<<<<<<
 *     """
 *     Normal3D factory function.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("raysect.core.math.normal.new_normal3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_v);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/quaternion.pxd":79
 * 
 * 
 * cdef inline Quaternion new_quaternion(double x, double y, double z, double s):             # <<<<<<<<<<<<<<
 *     """
 *     Quaternion factory function.
 */

static CYTHON_INLINE struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *__pyx_f_7raysect_4core_4math_10quaternion_new_quaternion(double __pyx_v_x, double __pyx_v_y, double __pyx_v_z, double __pyx_v_s) {
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *__pyx_v_q = 0;
  struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_quaternion", 0);

  /* "raysect/core/math/quaternion.pxd":88
 * 
 *     cdef Quaternion q
 *     q = Quaternion.__new__(Quaternion)             # <<<<<<<<<<<<<<
 *     q.x = x
 *     q.y = y
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7raysect_4core_4math_10quaternion_Quaternion), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7raysect_4core_4math_10quaternion_Quaternion)))) __PYX_ERR(6, 88, __pyx_L1_error)
  __pyx_v_q = ((struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "raysect/core/math/quaternion.pxd":89
 *     cdef Quaternion q
 *     q = Quaternion.__new__(Quaternion)
 *     q.x = x             # <<<<<<<<<<<<<<
 *     q.y = y
 *     q.z = z
 */
  __pyx_v_q->x = __pyx_v_x;

  /* "raysect/core/math/quaternion.pxd":90
 *     q = Quaternion.__new__(Quaternion)
 *     q.x = x
 *     q.y = y             # <<<<<<<<<<<<<<
 *     q.z = z
 *     q.s = s
 */
  __pyx_v_q->y = __pyx_v_y;

  /* "raysect/core/math/quaternion.pxd":91
 *     q.x = x
 *     q.y = y
 *     q.z = z             # <<<<<<<<<<<<<<
 *     q.s = s
 *     return q
 */
  __pyx_v_q->z = __pyx_v_z;

  /* "raysect/core/math/quaternion.pxd":92
 *     q.y = y
 *     q.z = z
 *     q.s = s             # <<<<<<<<<<<<<<
 *     return q
 * 
 */
  __pyx_v_q->s = __pyx_v_s;

  /* "raysect/core/math/quaternion.pxd":93
 *     q.z = z
 *     q.s = s
 *     return q             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_q);
  __pyx_r = __pyx_v_q;
  goto __pyx_L0;

  /* "raysect/core/math/quaternion.pxd":79
 * 
 * 
 * cdef inline Quaternion new_quaternion(double x, double y, double z, double s):             # <<<<<<<<<<<<<<
 *     """
 *     Quaternion factory function.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("raysect.core.math.quaternion.new_quaternion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_q);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/cython/utility.pxd":48
 * cdef double peak_to_peak(double[:] data) nogil
 * 
 * cdef inline double clamp(double v, double minimum, double maximum) nogil:             # <<<<<<<<<<<<<<
 *     if v < minimum:
 *         return minimum
 */

static CYTHON_INLINE double __pyx_f_7raysect_4core_4math_6cython_7utility_clamp(double __pyx_v_v, double __pyx_v_minimum, double __pyx_v_maximum) {
  double __pyx_r;
  int __pyx_t_1;

  /* "raysect/core/math/cython/utility.pxd":49
 * 
 * cdef inline double clamp(double v, double minimum, double maximum) nogil:
 *     if v < minimum:             # <<<<<<<<<<<<<<
 *         return minimum
 *     if v > maximum:
 */
  __pyx_t_1 = ((__pyx_v_v < __pyx_v_minimum) != 0);
  if (__pyx_t_1) {

    /* "raysect/core/math/cython/utility.pxd":50
 * cdef inline double clamp(double v, double minimum, double maximum) nogil:
 *     if v < minimum:
 *         return minimum             # <<<<<<<<<<<<<<
 *     if v > maximum:
 *         return maximum
 */
    __pyx_r = __pyx_v_minimum;
    goto __pyx_L0;

    /* "raysect/core/math/cython/utility.pxd":49
 * 
 * cdef inline double clamp(double v, double minimum, double maximum) nogil:
 *     if v < minimum:             # <<<<<<<<<<<<<<
 *         return minimum
 *     if v > maximum:
 */
  }

  /* "raysect/core/math/cython/utility.pxd":51
 *     if v < minimum:
 *         return minimum
 *     if v > maximum:             # <<<<<<<<<<<<<<
 *         return maximum
 *     return v
 */
  __pyx_t_1 = ((__pyx_v_v > __pyx_v_maximum) != 0);
  if (__pyx_t_1) {

    /* "raysect/core/math/cython/utility.pxd":52
 *         return minimum
 *     if v > maximum:
 *         return maximum             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
    __pyx_r = __pyx_v_maximum;
    goto __pyx_L0;

    /* "raysect/core/math/cython/utility.pxd":51
 *     if v < minimum:
 *         return minimum
 *     if v > maximum:             # <<<<<<<<<<<<<<
 *         return maximum
 *     return v
 */
  }

  /* "raysect/core/math/cython/utility.pxd":53
 *     if v > maximum:
 *         return maximum
 *     return v             # <<<<<<<<<<<<<<
 * 
 * cdef inline void swap_double(double *a, double *b) nogil:
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "raysect/core/math/cython/utility.pxd":48
 * cdef double peak_to_peak(double[:] data) nogil
 * 
 * cdef inline double clamp(double v, double minimum, double maximum) nogil:             # <<<<<<<<<<<<<<
 *     if v < minimum:
 *         return minimum
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "raysect/core/math/cython/utility.pxd":55
 *     return v
 * 
 * cdef inline void swap_double(double *a, double *b) nogil:             # <<<<<<<<<<<<<<
 *     cdef double temp
 *     temp = a[0]
 */

static CYTHON_INLINE void __pyx_f_7raysect_4core_4math_6cython_7utility_swap_double(double *__pyx_v_a, double *__pyx_v_b) {
  double __pyx_v_temp;

  /* "raysect/core/math/cython/utility.pxd":57
 * cdef inline void swap_double(double *a, double *b) nogil:
 *     cdef double temp
 *     temp = a[0]             # <<<<<<<<<<<<<<
 *     a[0] = b[0]
 *     b[0] = temp
 */
  __pyx_v_temp = (__pyx_v_a[0]);

  /* "raysect/core/math/cython/utility.pxd":58
 *     cdef double temp
 *     temp = a[0]
 *     a[0] = b[0]             # <<<<<<<<<<<<<<
 *     b[0] = temp
 * 
 */
  (__pyx_v_a[0]) = (__pyx_v_b[0]);

  /* "raysect/core/math/cython/utility.pxd":59
 *     temp = a[0]
 *     a[0] = b[0]
 *     b[0] = temp             # <<<<<<<<<<<<<<
 * 
 * cdef inline void swap_int(int *a, int *b) nogil:
 */
  (__pyx_v_b[0]) = __pyx_v_temp;

  /* "raysect/core/math/cython/utility.pxd":55
 *     return v
 * 
 * cdef inline void swap_double(double *a, double *b) nogil:             # <<<<<<<<<<<<<<
 *     cdef double temp
 *     temp = a[0]
 */

  /* function exit code */
}

/* "raysect/core/math/cython/utility.pxd":61
 *     b[0] = temp
 * 
 * cdef inline void swap_int(int *a, int *b) nogil:             # <<<<<<<<<<<<<<
 *     cdef int temp
 *     temp = a[0]
 */

static CYTHON_INLINE void __pyx_f_7raysect_4core_4math_6cython_7utility_swap_int(int *__pyx_v_a, int *__pyx_v_b) {
  int __pyx_v_temp;

  /* "raysect/core/math/cython/utility.pxd":63
 * cdef inline void swap_int(int *a, int *b) nogil:
 *     cdef int temp
 *     temp = a[0]             # <<<<<<<<<<<<<<
 *     a[0] = b[0]
 *     b[0] = temp
 */
  __pyx_v_temp = (__pyx_v_a[0]);

  /* "raysect/core/math/cython/utility.pxd":64
 *     cdef int temp
 *     temp = a[0]
 *     a[0] = b[0]             # <<<<<<<<<<<<<<
 *     b[0] = temp
 * 
 */
  (__pyx_v_a[0]) = (__pyx_v_b[0]);

  /* "raysect/core/math/cython/utility.pxd":65
 *     temp = a[0]
 *     a[0] = b[0]
 *     b[0] = temp             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  (__pyx_v_b[0]) = __pyx_v_temp;

  /* "raysect/core/math/cython/utility.pxd":61
 *     b[0] = temp
 * 
 * cdef inline void swap_int(int *a, int *b) nogil:             # <<<<<<<<<<<<<<
 *     cdef int temp
 *     temp = a[0]
 */

  /* function exit code */
}

/* "raysect/core/math/cython/utility.pxd":68
 * 
 * @cython.cdivision(True)
 * cdef inline double lerp(double x0, double x1, double y0, double y1, double x) nogil:             # <<<<<<<<<<<<<<
 *     return ((y1 - y0) / (x1 - x0)) * (x - x0) + y0
 * 
 */

static CYTHON_INLINE double __pyx_f_7raysect_4core_4math_6cython_7utility_lerp(double __pyx_v_x0, double __pyx_v_x1, double __pyx_v_y0, double __pyx_v_y1, double __pyx_v_x) {
  double __pyx_r;

  /* "raysect/core/math/cython/utility.pxd":69
 * @cython.cdivision(True)
 * cdef inline double lerp(double x0, double x1, double y0, double y1, double x) nogil:
 *     return ((y1 - y0) / (x1 - x0)) * (x - x0) + y0             # <<<<<<<<<<<<<<
 * 
 * cdef bint solve_quadratic(double a, double b, double c, double *t0, double *t1) nogil
 */
  __pyx_r = ((((__pyx_v_y1 - __pyx_v_y0) / (__pyx_v_x1 - __pyx_v_x0)) * (__pyx_v_x - __pyx_v_x0)) + __pyx_v_y0);
  goto __pyx_L0;

  /* "raysect/core/math/cython/utility.pxd":68
 * 
 * @cython.cdivision(True)
 * cdef inline double lerp(double x0, double x1, double y0, double y1, double x) nogil:             # <<<<<<<<<<<<<<
 *     return ((y1 - y0) / (x1 - x0)) * (x - x0) + y0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "raysect/core/ray.pxd":46
 * 
 * 
 * cdef inline new_ray(Point3D origin, Vector3D direction, double max_distance):             # <<<<<<<<<<<<<<
 * 
 *     cdef Ray ray
 */

static CYTHON_INLINE PyObject *__pyx_f_7raysect_4core_3ray_new_ray(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *__pyx_v_origin, struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *__pyx_v_direction, double __pyx_v_max_distance) {
  struct __pyx_obj_7raysect_4core_3ray_Ray *__pyx_v_ray = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_ray", 0);

  /* "raysect/core/ray.pxd":50
 *     cdef Ray ray
 * 
 *     ray = Ray.__new__(Ray)             # <<<<<<<<<<<<<<
 *     ray.origin = origin
 *     ray.direction = direction
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7raysect_4core_3ray_Ray), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7raysect_4core_3ray_Ray)))) __PYX_ERR(7, 50, __pyx_L1_error)
  __pyx_v_ray = ((struct __pyx_obj_7raysect_4core_3ray_Ray *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "raysect/core/ray.pxd":51
 * 
 *     ray = Ray.__new__(Ray)
 *     ray.origin = origin             # <<<<<<<<<<<<<<
 *     ray.direction = direction
 *     ray.max_distance = max_distance
 */
  __Pyx_INCREF((PyObject *)__pyx_v_origin);
  __Pyx_GIVEREF((PyObject *)__pyx_v_origin);
  __Pyx_GOTREF((PyObject *)__pyx_v_ray->origin);
  __Pyx_DECREF((PyObject *)__pyx_v_ray->origin);
  __pyx_v_ray->origin = __pyx_v_origin;

  /* "raysect/core/ray.pxd":52
 *     ray = Ray.__new__(Ray)
 *     ray.origin = origin
 *     ray.direction = direction             # <<<<<<<<<<<<<<
 *     ray.max_distance = max_distance
 *     return ray
 */
  __Pyx_INCREF((PyObject *)__pyx_v_direction);
  __Pyx_GIVEREF((PyObject *)__pyx_v_direction);
  __Pyx_GOTREF((PyObject *)__pyx_v_ray->direction);
  __Pyx_DECREF((PyObject *)__pyx_v_ray->direction);
  __pyx_v_ray->direction = __pyx_v_direction;

  /* "raysect/core/ray.pxd":53
 *     ray.origin = origin
 *     ray.direction = direction
 *     ray.max_distance = max_distance             # <<<<<<<<<<<<<<
 *     return ray
 */
  __pyx_v_ray->max_distance = __pyx_v_max_distance;

  /* "raysect/core/ray.pxd":54
 *     ray.direction = direction
 *     ray.max_distance = max_distance
 *     return ray             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_ray);
  __pyx_r = ((PyObject *)__pyx_v_ray);
  goto __pyx_L0;

  /* "raysect/core/ray.pxd":46
 * 
 * 
 * cdef inline new_ray(Point3D origin, Vector3D direction, double max_distance):             # <<<<<<<<<<<<<<
 * 
 *     cdef Ray ray
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("raysect.core.ray.new_ray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ray);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/intersection.pxd":57
 * 
 * 
 * cdef inline Intersection new_intersection(Ray ray, double ray_distance, Primitive primitive,             # <<<<<<<<<<<<<<
 *                                           Point3D hit_point, Point3D inside_point, Point3D outside_point,
 *                                           Normal3D normal, bint exiting, AffineMatrix3D world_to_primitive,
 */

static CYTHON_INLINE struct __pyx_obj_7raysect_4core_12intersection_Intersection *__pyx_f_7raysect_4core_12intersection_new_intersection(struct __pyx_obj_7raysect_4core_3ray_Ray *__pyx_v_ray, double __pyx_v_ray_distance, struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive *__pyx_v_primitive, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *__pyx_v_hit_point, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *__pyx_v_inside_point, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *__pyx_v_outside_point, struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D *__pyx_v_normal, int __pyx_v_exiting, struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *__pyx_v_world_to_primitive, struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *__pyx_v_primitive_to_world) {
  struct __pyx_obj_7raysect_4core_12intersection_Intersection *__pyx_v_intersection = 0;
  struct __pyx_obj_7raysect_4core_12intersection_Intersection *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_intersection", 0);

  /* "raysect/core/intersection.pxd":63
 * 
 *     cdef Intersection intersection
 *     intersection = Intersection.__new__(Intersection)             # <<<<<<<<<<<<<<
 *     intersection._construct(ray, ray_distance, primitive, hit_point, inside_point, outside_point, normal, exiting, world_to_primitive, primitive_to_world)
 *     return intersection
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7raysect_4core_12intersection_Intersection), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7raysect_4core_12intersection_Intersection)))) __PYX_ERR(8, 63, __pyx_L1_error)
  __pyx_v_intersection = ((struct __pyx_obj_7raysect_4core_12intersection_Intersection *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "raysect/core/intersection.pxd":64
 *     cdef Intersection intersection
 *     intersection = Intersection.__new__(Intersection)
 *     intersection._construct(ray, ray_distance, primitive, hit_point, inside_point, outside_point, normal, exiting, world_to_primitive, primitive_to_world)             # <<<<<<<<<<<<<<
 *     return intersection
 */
  ((struct __pyx_vtabstruct_7raysect_4core_12intersection_Intersection *)__pyx_v_intersection->__pyx_vtab)->_construct(__pyx_v_intersection, __pyx_v_ray, __pyx_v_ray_distance, __pyx_v_primitive, __pyx_v_hit_point, __pyx_v_inside_point, __pyx_v_outside_point, __pyx_v_normal, __pyx_v_exiting, __pyx_v_world_to_primitive, __pyx_v_primitive_to_world);

  /* "raysect/core/intersection.pxd":65
 *     intersection = Intersection.__new__(Intersection)
 *     intersection._construct(ray, ray_distance, primitive, hit_point, inside_point, outside_point, normal, exiting, world_to_primitive, primitive_to_world)
 *     return intersection             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_intersection);
  __pyx_r = __pyx_v_intersection;
  goto __pyx_L0;

  /* "raysect/core/intersection.pxd":57
 * 
 * 
 * cdef inline Intersection new_intersection(Ray ray, double ray_distance, Primitive primitive,             # <<<<<<<<<<<<<<
 *                                           Point3D hit_point, Point3D inside_point, Point3D outside_point,
 *                                           Normal3D normal, bint exiting, AffineMatrix3D world_to_primitive,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("raysect.core.intersection.new_intersection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_intersection);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/boundingbox.pxd":77
 * 
 * 
 * cdef inline BoundingBox3D new_boundingbox3d(Point3D lower, Point3D upper):             # <<<<<<<<<<<<<<
 *     """
 *     BoundingBox3D factory function.
 */

static CYTHON_INLINE struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *__pyx_f_7raysect_4core_11boundingbox_new_boundingbox3d(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *__pyx_v_lower, struct __pyx_obj_7raysect_4core_4math_5point_Point3D *__pyx_v_upper) {
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *__pyx_v_v = 0;
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_boundingbox3d", 0);

  /* "raysect/core/boundingbox.pxd":86
 * 
 *     cdef BoundingBox3D v
 *     v = BoundingBox3D.__new__(BoundingBox3D)             # <<<<<<<<<<<<<<
 *     v.lower = lower
 *     v.upper = upper
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox3D), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7raysect_4core_11boundingbox_BoundingBox3D)))) __PYX_ERR(9, 86, __pyx_L1_error)
  __pyx_v_v = ((struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "raysect/core/boundingbox.pxd":87
 *     cdef BoundingBox3D v
 *     v = BoundingBox3D.__new__(BoundingBox3D)
 *     v.lower = lower             # <<<<<<<<<<<<<<
 *     v.upper = upper
 *     return v
 */
  __Pyx_INCREF((PyObject *)__pyx_v_lower);
  __Pyx_GIVEREF((PyObject *)__pyx_v_lower);
  __Pyx_GOTREF((PyObject *)__pyx_v_v->lower);
  __Pyx_DECREF((PyObject *)__pyx_v_v->lower);
  __pyx_v_v->lower = __pyx_v_lower;

  /* "raysect/core/boundingbox.pxd":88
 *     v = BoundingBox3D.__new__(BoundingBox3D)
 *     v.lower = lower
 *     v.upper = upper             # <<<<<<<<<<<<<<
 *     return v
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_upper);
  __Pyx_GIVEREF((PyObject *)__pyx_v_upper);
  __Pyx_GOTREF((PyObject *)__pyx_v_v->upper);
  __Pyx_DECREF((PyObject *)__pyx_v_v->upper);
  __pyx_v_v->upper = __pyx_v_upper;

  /* "raysect/core/boundingbox.pxd":89
 *     v.lower = lower
 *     v.upper = upper
 *     return v             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "raysect/core/boundingbox.pxd":77
 * 
 * 
 * cdef inline BoundingBox3D new_boundingbox3d(Point3D lower, Point3D upper):             # <<<<<<<<<<<<<<
 *     """
 *     BoundingBox3D factory function.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("raysect.core.boundingbox.new_boundingbox3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_v);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/boundingbox.pxd":118
 * 
 * 
 * cdef inline BoundingBox2D new_boundingbox2d(Point2D lower, Point2D upper):             # <<<<<<<<<<<<<<
 *     """
 *     BoundingBox2D factory function.
 */

static CYTHON_INLINE struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *__pyx_f_7raysect_4core_11boundingbox_new_boundingbox2d(struct __pyx_obj_7raysect_4core_4math_5point_Point2D *__pyx_v_lower, struct __pyx_obj_7raysect_4core_4math_5point_Point2D *__pyx_v_upper) {
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *__pyx_v_b = 0;
  struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_boundingbox2d", 0);

  /* "raysect/core/boundingbox.pxd":127
 * 
 *     cdef BoundingBox2D b
 *     b = BoundingBox2D.__new__(BoundingBox2D)             # <<<<<<<<<<<<<<
 *     b.lower = lower
 *     b.upper = upper
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox2D), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7raysect_4core_11boundingbox_BoundingBox2D)))) __PYX_ERR(9, 127, __pyx_L1_error)
  __pyx_v_b = ((struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "raysect/core/boundingbox.pxd":128
 *     cdef BoundingBox2D b
 *     b = BoundingBox2D.__new__(BoundingBox2D)
 *     b.lower = lower             # <<<<<<<<<<<<<<
 *     b.upper = upper
 *     return b
 */
  __Pyx_INCREF((PyObject *)__pyx_v_lower);
  __Pyx_GIVEREF((PyObject *)__pyx_v_lower);
  __Pyx_GOTREF((PyObject *)__pyx_v_b->lower);
  __Pyx_DECREF((PyObject *)__pyx_v_b->lower);
  __pyx_v_b->lower = __pyx_v_lower;

  /* "raysect/core/boundingbox.pxd":129
 *     b = BoundingBox2D.__new__(BoundingBox2D)
 *     b.lower = lower
 *     b.upper = upper             # <<<<<<<<<<<<<<
 *     return b
 */
  __Pyx_INCREF((PyObject *)__pyx_v_upper);
  __Pyx_GIVEREF((PyObject *)__pyx_v_upper);
  __Pyx_GOTREF((PyObject *)__pyx_v_b->upper);
  __Pyx_DECREF((PyObject *)__pyx_v_b->upper);
  __pyx_v_b->upper = __pyx_v_upper;

  /* "raysect/core/boundingbox.pxd":130
 *     b.lower = lower
 *     b.upper = upper
 *     return b             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_b);
  __pyx_r = __pyx_v_b;
  goto __pyx_L0;

  /* "raysect/core/boundingbox.pxd":118
 * 
 * 
 * cdef inline BoundingBox2D new_boundingbox2d(Point2D lower, Point2D upper):             # <<<<<<<<<<<<<<
 *     """
 *     BoundingBox2D factory function.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("raysect.core.boundingbox.new_boundingbox2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_b);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/function/float/function1d/base.pxd":162
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a Function1D object.
 */

static CYTHON_INLINE int __pyx_f_7raysect_4core_4math_8function_5float_10function1d_4base_is_callable(PyObject *__pyx_v_f) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_callable", 0);

  /* "raysect/core/math/function/float/function1d/base.pxd":169
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function1D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "raysect/core/math/function/float/function1d/base.pxd":170
 *     """
 *     if isinstance(f, Function1D):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     # other function classes are incompatible
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "raysect/core/math/function/float/function1d/base.pxd":169
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function1D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  }

  /* "raysect/core/math/function/float/function1d/base.pxd":173
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_4base_Function); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "raysect/core/math/function/float/function1d/base.pxd":174
 *     # other function classes are incompatible
 *     if isinstance(f, Function):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     return callable(f)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "raysect/core/math/function/float/function1d/base.pxd":173
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "raysect/core/math/function/float/function1d/base.pxd":176
 *         return False
 * 
 *     return callable(f)             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyCallable_Check(__pyx_v_f); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(10, 176, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "raysect/core/math/function/float/function1d/base.pxd":162
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a Function1D object.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("raysect.core.math.function.float.function1d.base.is_callable", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/function/float/function2d/base.pxd":163
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a Function2D object.
 */

static CYTHON_INLINE int __pyx_f_7raysect_4core_4math_8function_5float_10function2d_4base_is_callable(PyObject *__pyx_v_f) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_callable", 0);

  /* "raysect/core/math/function/float/function2d/base.pxd":170
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function2D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "raysect/core/math/function/float/function2d/base.pxd":171
 *     """
 *     if isinstance(f, Function2D):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     # other function classes are incompatible
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "raysect/core/math/function/float/function2d/base.pxd":170
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function2D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  }

  /* "raysect/core/math/function/float/function2d/base.pxd":174
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_4base_Function); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "raysect/core/math/function/float/function2d/base.pxd":175
 *     # other function classes are incompatible
 *     if isinstance(f, Function):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     return callable(f)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "raysect/core/math/function/float/function2d/base.pxd":174
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "raysect/core/math/function/float/function2d/base.pxd":177
 *         return False
 * 
 *     return callable(f)             # <<<<<<<<<<<<<<
 * 
 */
  __pyx_t_1 = __Pyx_PyCallable_Check(__pyx_v_f); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(11, 177, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "raysect/core/math/function/float/function2d/base.pxd":163
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a Function2D object.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("raysect.core.math.function.float.function2d.base.is_callable", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/function/float/function3d/base.pxd":162
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a Function3D object.
 */

static CYTHON_INLINE int __pyx_f_7raysect_4core_4math_8function_5float_10function3d_4base_is_callable(PyObject *__pyx_v_f) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_callable", 0);

  /* "raysect/core/math/function/float/function3d/base.pxd":169
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function3D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "raysect/core/math/function/float/function3d/base.pxd":170
 *     """
 *     if isinstance(f, Function3D):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     # other function classes are incompatible
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "raysect/core/math/function/float/function3d/base.pxd":169
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function3D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  }

  /* "raysect/core/math/function/float/function3d/base.pxd":173
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_4base_Function); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "raysect/core/math/function/float/function3d/base.pxd":174
 *     # other function classes are incompatible
 *     if isinstance(f, Function):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     return callable(f)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "raysect/core/math/function/float/function3d/base.pxd":173
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "raysect/core/math/function/float/function3d/base.pxd":176
 *         return False
 * 
 *     return callable(f)             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyCallable_Check(__pyx_v_f); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(12, 176, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "raysect/core/math/function/float/function3d/base.pxd":162
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a Function3D object.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("raysect.core.math.function.float.function3d.base.is_callable", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/function/vector3d/function1d/base.pxd":72
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a vector3d.Function1D object.
 */

static CYTHON_INLINE int __pyx_f_7raysect_4core_4math_8function_8vector3d_10function1d_4base_is_callable(PyObject *__pyx_v_f) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_callable", 0);

  /* "raysect/core/math/function/vector3d/function1d/base.pxd":79
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function1D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "raysect/core/math/function/vector3d/function1d/base.pxd":80
 *     """
 *     if isinstance(f, Function1D):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     # other function classes are incompatible
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "raysect/core/math/function/vector3d/function1d/base.pxd":79
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function1D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  }

  /* "raysect/core/math/function/vector3d/function1d/base.pxd":83
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_4base_Function); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "raysect/core/math/function/vector3d/function1d/base.pxd":84
 *     # other function classes are incompatible
 *     if isinstance(f, Function):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     return callable(f)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "raysect/core/math/function/vector3d/function1d/base.pxd":83
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "raysect/core/math/function/vector3d/function1d/base.pxd":86
 *         return False
 * 
 *     return callable(f)             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyCallable_Check(__pyx_v_f); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(13, 86, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "raysect/core/math/function/vector3d/function1d/base.pxd":72
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a vector3d.Function1D object.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("raysect.core.math.function.vector3d.function1d.base.is_callable", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/function/vector3d/function2d/base.pxd":72
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a vector3d.Function2D object.
 */

static CYTHON_INLINE int __pyx_f_7raysect_4core_4math_8function_8vector3d_10function2d_4base_is_callable(PyObject *__pyx_v_f) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_callable", 0);

  /* "raysect/core/math/function/vector3d/function2d/base.pxd":79
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function2D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "raysect/core/math/function/vector3d/function2d/base.pxd":80
 *     """
 *     if isinstance(f, Function2D):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     # other function classes are incompatible
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "raysect/core/math/function/vector3d/function2d/base.pxd":79
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function2D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  }

  /* "raysect/core/math/function/vector3d/function2d/base.pxd":83
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_4base_Function); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "raysect/core/math/function/vector3d/function2d/base.pxd":84
 *     # other function classes are incompatible
 *     if isinstance(f, Function):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     return callable(f)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "raysect/core/math/function/vector3d/function2d/base.pxd":83
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "raysect/core/math/function/vector3d/function2d/base.pxd":86
 *         return False
 * 
 *     return callable(f)             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyCallable_Check(__pyx_v_f); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(14, 86, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "raysect/core/math/function/vector3d/function2d/base.pxd":72
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a vector3d.Function2D object.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("raysect.core.math.function.vector3d.function2d.base.is_callable", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "raysect/core/math/function/vector3d/function3d/base.pxd":72
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a vector3d.Function3D object.
 */

static CYTHON_INLINE int __pyx_f_7raysect_4core_4math_8function_8vector3d_10function3d_4base_is_callable(PyObject *__pyx_v_f) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_callable", 0);

  /* "raysect/core/math/function/vector3d/function3d/base.pxd":79
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function3D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "raysect/core/math/function/vector3d/function3d/base.pxd":80
 *     """
 *     if isinstance(f, Function3D):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     # other function classes are incompatible
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "raysect/core/math/function/vector3d/function3d/base.pxd":79
 *     :return: True if callable, False otherwise.
 *     """
 *     if isinstance(f, Function3D):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  }

  /* "raysect/core/math/function/vector3d/function3d/base.pxd":83
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_f, __pyx_ptype_7raysect_4core_4math_8function_4base_Function); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "raysect/core/math/function/vector3d/function3d/base.pxd":84
 *     # other function classes are incompatible
 *     if isinstance(f, Function):
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     return callable(f)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "raysect/core/math/function/vector3d/function3d/base.pxd":83
 * 
 *     # other function classes are incompatible
 *     if isinstance(f, Function):             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
  }

  /* "raysect/core/math/function/vector3d/function3d/base.pxd":86
 *         return False
 * 
 *     return callable(f)             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyCallable_Check(__pyx_v_f); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(15, 86, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "raysect/core/math/function/vector3d/function3d/base.pxd":72
 * 
 * 
 * cdef inline bint is_callable(object f):             # <<<<<<<<<<<<<<
 *     """
 *     Tests if an object is a python callable or a vector3d.Function3D object.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("raysect.core.math.function.vector3d.function3d.base.is_callable", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":126
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    #endif
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_n_s_c);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_shape)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 126, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_itemsize)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 126, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(16, 126, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_format)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 126, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(16, 126, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_mode);
          if (value) { values[3] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 126, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 126, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(16, 126, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(16, 126, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(16, 127, __pyx_L3_error)
    } else {

      /* "View.MemoryView":127
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, __pyx_nargs); __PYX_ERR(16, 126, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(16, 126, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(16, 126, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":126
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_dim;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  char *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":132
 *         cdef Py_ssize_t dim
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(16, 132, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(16, 132, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":133
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":135
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":136
 * 
 *         if not self.ndim:
 *             raise ValueError, "Empty shape tuple for cython.array"             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Empty_shape_tuple_for_cython_arr, 0, 0);
    __PYX_ERR(16, 136, __pyx_L1_error)

    /* "View.MemoryView":135
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 */
  }

  /* "View.MemoryView":138
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 */
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":139
 * 
 *         if itemsize <= 0:
 *             raise ValueError, "itemsize <= 0 for cython.array"             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_itemsize_0_for_cython_array, 0, 0);
    __PYX_ERR(16, 139, __pyx_L1_error)

    /* "View.MemoryView":138
 *             raise ValueError, "Empty shape tuple for cython.array"
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 */
  }

  /* "View.MemoryView":141
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":142
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_n_s_ASCII};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 142, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":141
 *             raise ValueError, "itemsize <= 0 for cython.array"
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":143
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", __Pyx_PyType_Name(Py_TYPE(__pyx_v_format))), 0))) __PYX_ERR(16, 143, __pyx_L1_error)
  __pyx_t_4 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":144
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(16, 144, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(16, 144, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_8;

  /* "View.MemoryView":147
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":148
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":150
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError, "unable to allocate shape and strides."
 * 
 */
  __pyx_t_3 = ((!(__pyx_v_self->_shape != 0)) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "View.MemoryView":151
 * 
 *         if not self._shape:
 *             raise MemoryError, "unable to allocate shape and strides."             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_Raise(__pyx_builtin_MemoryError, __pyx_kp_s_unable_to_allocate_shape_and_str, 0, 0);
    __PYX_ERR(16, 151, __pyx_L1_error)

    /* "View.MemoryView":150
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError, "unable to allocate shape and strides."
 * 
 */
  }

  /* "View.MemoryView":154
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 */
  __pyx_t_7 = 0;
  __pyx_t_4 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_4); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(16, 154, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(16, 154, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_7;
    __pyx_t_7 = (__pyx_t_7 + 1);

    /* "View.MemoryView":155
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 *             self._shape[idx] = dim
 */
    __pyx_t_3 = ((__pyx_v_dim <= 0) != 0);
    if (unlikely(__pyx_t_3)) {

      /* "View.MemoryView":156
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __Pyx_INCREF(__pyx_kp_u_Invalid_shape_in_axis);
      __pyx_t_9 += 22;
      __Pyx_GIVEREF(__pyx_kp_u_Invalid_shape_in_axis);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Invalid_shape_in_axis);
      __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_idx, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u__3);
      __pyx_t_9 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__3);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u__3);
      __pyx_t_6 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_dim, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u__4);
      __pyx_t_9 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__4);
      PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_kp_u__4);
      __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_5, 5, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_builtin_ValueError, __pyx_t_6, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(16, 156, __pyx_L1_error)

      /* "View.MemoryView":155
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":157
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":154
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError, f"Invalid shape in axis {idx}: {dim}."
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "View.MemoryView":160
 * 
 *         cdef char order
 *         if mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(16, 160, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "View.MemoryView":161
 *         cdef char order
 *         if mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         elif mode == 'fortran':
 */
    __pyx_v_order = 'C';

    /* "View.MemoryView":162
 *         if mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         elif mode == 'fortran':
 *             order = b'F'
 */
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":160
 * 
 *         cdef char order
 *         if mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":163
 *             order = b'C'
 *             self.mode = u'c'
 *         elif mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(16, 163, __pyx_L1_error)
  if (likely(__pyx_t_3)) {

    /* "View.MemoryView":164
 *             self.mode = u'c'
 *         elif mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         else:
 */
    __pyx_v_order = 'F';

    /* "View.MemoryView":165
 *         elif mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError, f"Invalid mode, expected 'c' or 'fortran', got {mode}"
 */
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":163
 *             order = b'C'
 *             self.mode = u'c'
 *         elif mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":167
 *             self.mode = u'fortran'
 *         else:
 *             raise ValueError, f"Invalid mode, expected 'c' or 'fortran', got {mode}"             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides, itemsize, self.ndim, order)
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_mode, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Invalid_mode_expected_c_or_fortr, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_t_6, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(16, 167, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "View.MemoryView":169
 *             raise ValueError, f"Invalid mode, expected 'c' or 'fortran', got {mode}"
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides, itemsize, self.ndim, order)             # <<<<<<<<<<<<<<
 * 
 *         self.free_data = allocate_buffer
 */
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":171
 *         self.len = fill_contig_strides_array(self._shape, self._strides, itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 * 
 */
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":172
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 * 
 *         if allocate_buffer:
 */
  __pyx_t_6 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 172, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(16, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_3;

  /* "View.MemoryView":174
 *         self.dtype_is_object = format == b'O'
 * 
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 *             _allocate_buffer(self)
 * 
 */
  __pyx_t_3 = (__pyx_v_allocate_buffer != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":175
 * 
 *         if allocate_buffer:
 *             _allocate_buffer(self)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_t_7 = __pyx_array_allocate_buffer(__pyx_v_self); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(16, 175, __pyx_L1_error)

    /* "View.MemoryView":174
 *         self.dtype_is_object = format == b'O'
 * 
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 *             _allocate_buffer(self)
 * 
 */
  }

  /* "View.MemoryView":126
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":177
 *             _allocate_buffer(self)
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  char *__pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (unlikely(__pyx_v_info == NULL)) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":179
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":180
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(16, 180, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":181
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":180
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":182
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(16, 182, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":183
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError, "Can only create a buffer that is contiguous in memory."
 */
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":182
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":184
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data
 */
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":185
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError, "Can only create a buffer that is contiguous in memory."             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Can_only_create_a_buffer_that_is, 0, 0);
    __PYX_ERR(16, 185, __pyx_L1_error)

    /* "View.MemoryView":184
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data
 */
  }

  /* "View.MemoryView":186
 *         if not (flags & bufmode):
 *             raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */
  __pyx_t_3 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_3;

  /* "View.MemoryView":187
 *             raise ValueError, "Can only create a buffer that is contiguous in memory."
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */
  __pyx_t_4 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_4;

  /* "View.MemoryView":188
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */
  __pyx_t_5 = __pyx_v_self->ndim;
  __pyx_v_info->ndim = __pyx_t_5;

  /* "View.MemoryView":189
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */
  __pyx_t_6 = __pyx_v_self->_shape;
  __pyx_v_info->shape = __pyx_t_6;

  /* "View.MemoryView":190
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */
  __pyx_t_6 = __pyx_v_self->_strides;
  __pyx_v_info->strides = __pyx_t_6;

  /* "View.MemoryView":191
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":192
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */
  __pyx_t_4 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_4;

  /* "View.MemoryView":193
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":195
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":196
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_3 = __pyx_v_self->format;
    __pyx_v_info->format = __pyx_t_3;

    /* "View.MemoryView":195
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
    goto __pyx_L5;
  }

  /* "View.MemoryView":198
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L5:;

  /* "View.MemoryView":200
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":177
 *             _allocate_buffer(self)
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":204
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":205
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":206
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:
 */
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":205
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":207
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 */
  __pyx_t_2 = (__pyx_v_self->free_data != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_self->data != NULL) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":208
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)
 */
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":209
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)             # <<<<<<<<<<<<<<
 *             free(self.data)
 *         PyObject_Free(self._shape)
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":208
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)
 */
    }

    /* "View.MemoryView":210
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    free(__pyx_v_self->data);

    /* "View.MemoryView":207
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data and self.data is not NULL:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":211
 *                 refcount_objects_in_slice(self.data, self._shape, self._strides, self.ndim, inc=False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":204
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":213
 *         PyObject_Free(self._shape)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def memview(self):
 *         return self.get_memview()
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":215
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":213
 *         PyObject_Free(self._shape)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def memview(self):
 *         return self.get_memview()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":218
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":219
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":220
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":218
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":222
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":223
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":222
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":225
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":226
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":225
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":228
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":229
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":228
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":231
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":232
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(16, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":231
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(16, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(16, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":236
 * 
 * @cname("__pyx_array_allocate_buffer")
 * cdef int _allocate_buffer(array self) except -1:             # <<<<<<<<<<<<<<
 * 
 * 
 */

static int __pyx_array_allocate_buffer(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_i;
  PyObject **__pyx_v_p;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_allocate_buffer", 0);

  /* "View.MemoryView":242
 *     cdef PyObject **p
 * 
 *     self.free_data = True             # <<<<<<<<<<<<<<
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:
 */
  __pyx_v_self->free_data = 1;

  /* "View.MemoryView":243
 * 
 *     self.free_data = True
 *     self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *     if not self.data:
 *         raise MemoryError, "unable to allocate array data."
 */
  __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

  /* "View.MemoryView":244
 *     self.free_data = True
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:             # <<<<<<<<<<<<<<
 *         raise MemoryError, "unable to allocate array data."
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->data != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":245
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:
 *         raise MemoryError, "unable to allocate array data."             # <<<<<<<<<<<<<<
 * 
 *     if self.dtype_is_object:
 */
    __Pyx_Raise(__pyx_builtin_MemoryError, __pyx_kp_s_unable_to_allocate_array_data, 0, 0);
    __PYX_ERR(16, 245, __pyx_L1_error)

    /* "View.MemoryView":244
 *     self.free_data = True
 *     self.data = <char *>malloc(self.len)
 *     if not self.data:             # <<<<<<<<<<<<<<
 *         raise MemoryError, "unable to allocate array data."
 * 
 */
  }

  /* "View.MemoryView":247
 *         raise MemoryError, "unable to allocate array data."
 * 
 *     if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):
 */
  __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":248
 * 
 *     if self.dtype_is_object:
 *         p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *         for i in range(self.len // self.itemsize):
 *             p[i] = Py_None
 */
    __pyx_v_p = ((PyObject **)__pyx_v_self->data);

    /* "View.MemoryView":249
 *     if self.dtype_is_object:
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):             # <<<<<<<<<<<<<<
 *             p[i] = Py_None
 *             Py_INCREF(Py_None)
 */
    if (unlikely(__pyx_v_self->itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(16, 249, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_self->itemsize == (Py_ssize_t)-1)  && unlikely(__Pyx_UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(16, 249, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_div_Py_ssize_t(__pyx_v_self->len, __pyx_v_self->itemsize);
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":250
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):
 *             p[i] = Py_None             # <<<<<<<<<<<<<<
 *             Py_INCREF(Py_None)
 *     return 0
 */
      (__pyx_v_p[__pyx_v_i]) = Py_None;

      /* "View.MemoryView":251
 *         for i in range(self.len // self.itemsize):
 *             p[i] = Py_None
 *             Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
      Py_INCREF(Py_None);
    }

    /* "View.MemoryView":247
 *         raise MemoryError, "unable to allocate array data."
 * 
 *     if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *         p = <PyObject **> self.data
 *         for i in range(self.len // self.itemsize):
 */
  }

  /* "View.MemoryView":252
 *             p[i] = Py_None
 *             Py_INCREF(Py_None)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":236
 * 
 * @cname("__pyx_array_allocate_buffer")
 * cdef int _allocate_buffer(array self) except -1:             # <<<<<<<<<<<<<<
 * 
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._allocate_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":256
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format, char *c_mode, char *buf):             # <<<<<<<<<<<<<<
 *     cdef array result
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_c_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  PyObject *__pyx_v_mode = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":258
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format, char *c_mode, char *buf):
 *     cdef array result
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.             # <<<<<<<<<<<<<<
 * 
 *     if buf is NULL:
 */
  if ((((__pyx_v_c_mode[0]) == 'f') != 0)) {
    __Pyx_INCREF(__pyx_n_s_fortran);
    __pyx_t_1 = __pyx_n_s_fortran;
  } else {
    __Pyx_INCREF(__pyx_n_s_c);
    __pyx_t_1 = __pyx_n_s_c;
  }
  __pyx_v_mode = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":260
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 * 
 *     if buf is NULL:             # <<<<<<<<<<<<<<
 *         result = array.__new__(array, shape, itemsize, format, mode)
 *     else:
 */
  __pyx_t_2 = ((__pyx_v_buf == NULL) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":261
 * 
 *     if buf is NULL:
 *         result = array.__new__(array, shape, itemsize, format, mode)             # <<<<<<<<<<<<<<
 *     else:
 *         result = array.__new__(array, shape, itemsize, format, mode, allocate_buffer=False)
 */
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_mode);
    __Pyx_GIVEREF(__pyx_v_mode);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_mode);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = ((PyObject *)__pyx_tp_new_array(((PyTypeObject *)__pyx_array_type), __pyx_t_4, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 261, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":260
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 * 
 *     if buf is NULL:             # <<<<<<<<<<<<<<
 *         result = array.__new__(array, shape, itemsize, format, mode)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":263
 *         result = array.__new__(array, shape, itemsize, format, mode)
 *     else:
 *         result = array.__new__(array, shape, itemsize, format, mode, allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
  /*else*/ {
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_mode);
    __Pyx_GIVEREF(__pyx_v_mode);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_mode);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(16, 263, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_tp_new_array(((PyTypeObject *)__pyx_array_type), __pyx_t_1, __pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 263, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":264
 *     else:
 *         result = array.__new__(array, shape, itemsize, format, mode, allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":266
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":256
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format, char *c_mode, char *buf):             # <<<<<<<<<<<<<<
 *     cdef array result
 *     cdef str mode = "fortran" if c_mode[0] == b'f' else "c"  # this often comes from a constant C string.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_mode);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":292
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    #endif
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 292, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(16, 292, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(16, 292, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":293
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":292
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":294
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":295
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":294
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0x6ae9995, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0x6ae9995, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0x6ae9995, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x6ae9995, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_112105877);
    __Pyx_GIVEREF(__pyx_int_112105877);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_112105877);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0x6ae9995, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0x6ae9995, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x6ae9995, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_112105877);
    __Pyx_GIVEREF(__pyx_int_112105877);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_112105877);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x6ae9995, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0x6ae9995, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", __Pyx_PyType_Name(Py_TYPE(__pyx_v___pyx_state))), 0))) __PYX_ERR(16, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0x6ae9995, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":309
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p;
  size_t __pyx_v_offset;
  void *__pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":311
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":315
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":317
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":318
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":317
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  }

  /* "View.MemoryView":320
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((void *)__pyx_v_aligned_p);
  goto __pyx_L0;

  /* "View.MemoryView":309
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":356
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_obj)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 356, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_flags)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 356, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(16, 356, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 356, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(16, 356, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(16, 356, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(16, 356, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, __pyx_nargs); __PYX_ERR(16, 356, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":357
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":358
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":359
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_obj != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":360
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(16, 360, __pyx_L1_error)

    /* "View.MemoryView":361
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":362
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":363
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         global __pyx_memoryview_thread_locks_used
 */
      Py_INCREF(Py_None);

      /* "View.MemoryView":361
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":359
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":366
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":367
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 */
    __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

    /* "View.MemoryView":368
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 */
    __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

    /* "View.MemoryView":366
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  }

  /* "View.MemoryView":369
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":370
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 raise MemoryError
 */
    __pyx_v_self->lock = PyThread_allocate_lock();

    /* "View.MemoryView":371
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":372
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
      PyErr_NoMemory(); __PYX_ERR(16, 372, __pyx_L1_error)

      /* "View.MemoryView":371
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    }

    /* "View.MemoryView":369
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  }

  /* "View.MemoryView":374
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":375
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":374
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":377
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L10:;

  /* "View.MemoryView":379
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":381
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":356
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":383
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyThread_type_lock __pyx_t_6;
  PyThread_type_lock __pyx_t_7;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":384
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":385
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":384
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":386
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":388
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":389
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    Py_DECREF(Py_None);

    /* "View.MemoryView":386
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":393
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":394
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":395
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":396
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":397
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":399
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":398
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":397
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":400
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        goto __pyx_L6_break;

        /* "View.MemoryView":395
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":402
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":393
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":383
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":404
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":406
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":408
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 408, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(16, 408, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 408, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(16, 408, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 408, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(16, 408, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":409
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(16, 409, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(16, 409, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":408
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":411
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":404
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":414
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":415
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":416
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":415
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":418
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(16, 418, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(16, 418, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_indices = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "View.MemoryView":421
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(16, 421, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "View.MemoryView":422
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":421
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":424
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  /*else*/ {
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(16, 424, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":425
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":414
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":427
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError, "Cannot assign to read-only memoryview"
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":428
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError, "Cannot assign to read-only memoryview"
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":429
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError, "Cannot assign to read-only memoryview"             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_Cannot_assign_to_read_only_memor, 0, 0);
    __PYX_ERR(16, 429, __pyx_L1_error)

    /* "View.MemoryView":428
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError, "Cannot assign to read-only memoryview"
 * 
 */
  }

  /* "View.MemoryView":431
 *             raise TypeError, "Cannot assign to read-only memoryview"
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(16, 431, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(16, 431, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":433
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(16, 433, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":434
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "View.MemoryView":435
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(16, 435, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":436
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 436, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 436, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":435
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":438
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(16, 438, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":433
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":440
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":427
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError, "Cannot assign to read-only memoryview"
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":442
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":443
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":444
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":445
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 445, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":446
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(16, 446, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":445
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(16, 445, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(16, 445, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":444
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":447
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(16, 447, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":448
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "View.MemoryView":444
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":443
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":450
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":442
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":452
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  __Pyx_memviewslice __pyx_v_msrc;
  __Pyx_memviewslice __pyx_v_mdst;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":455
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 *         cdef __Pyx_memviewslice msrc = get_slice_from_memview(src, &src_slice)[0]             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mdst = get_slice_from_memview(dst, &dst_slice)[0]
 * 
 */
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(16, 455, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(16, 455, __pyx_L1_error)
  __pyx_v_msrc = (__pyx_t_1[0]);

  /* "View.MemoryView":456
 *         cdef __Pyx_memviewslice src_slice
 *         cdef __Pyx_memviewslice msrc = get_slice_from_memview(src, &src_slice)[0]
 *         cdef __Pyx_memviewslice mdst = get_slice_from_memview(dst, &dst_slice)[0]             # <<<<<<<<<<<<<<
 * 
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)
 */
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(16, 456, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(16, 456, __pyx_L1_error)
  __pyx_v_mdst = (__pyx_t_1[0]);

  /* "View.MemoryView":458
 *         cdef __Pyx_memviewslice mdst = get_slice_from_memview(dst, &dst_slice)[0]
 * 
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(16, 458, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(16, 458, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __pyx_memoryview_copy_contents(__pyx_v_msrc, __pyx_v_mdst, __pyx_t_3, __pyx_t_4, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(16, 458, __pyx_L1_error)

  /* "View.MemoryView":452
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":460
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":462
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":467
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(16, 467, __pyx_L1_error)
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":469
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __pyx_t_2 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":470
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":471
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":472
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      PyErr_NoMemory(); __PYX_ERR(16, 472, __pyx_L1_error)

      /* "View.MemoryView":471
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":473
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":469
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":475
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":477
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  /*try:*/ {

    /* "View.MemoryView":478
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    __pyx_t_2 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":479
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":478
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":481
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 481, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":485
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __pyx_t_2 = ((__pyx_v_self->view.suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":486
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __pyx_t_3 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 486, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":485
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":487
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":490
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":460
 *         memoryview_copy_contents(msrc, mdst, src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":492
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":493
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(16, 493, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":494
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":492
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":496
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":499
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":502
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":503
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":504
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError, "Unable to convert item to object"
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 504, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 504, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 504, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":503
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":508
 *             raise ValueError, "Unable to convert item to object"
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    /*else:*/ {
      __pyx_t_9 = __Pyx_ssize_strlen(__pyx_v_self->view.format); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(16, 508, __pyx_L5_except_error)
      __pyx_t_10 = ((__pyx_t_9 == 1) != 0);
      if (__pyx_t_10) {

        /* "View.MemoryView":509
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 509, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":508
 *             raise ValueError, "Unable to convert item to object"
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":510
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "View.MemoryView":505
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Unable to convert item to object"
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_7)) __PYX_ERR(16, 505, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_6);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(16, 505, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":506
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError, "Unable to convert item to object"             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Unable_to_convert_item_to_object, 0, 0);
      __PYX_ERR(16, 506, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "View.MemoryView":503
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":496
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":512
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  char *__pyx_t_10;
  char *__pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":515
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_struct, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":520
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":521
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", __Pyx_PyType_Name(Py_TYPE(__pyx_t_4))), 0))) __PYX_ERR(16, 521, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":520
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":523
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 523, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", __Pyx_PyType_Name(Py_TYPE(__pyx_t_4))), 0))) __PYX_ERR(16, 523, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":525
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __pyx_t_8 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(16, 525, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_9 = __pyx_v_bytesvalue;
  __pyx_t_11 = PyBytes_AS_STRING(__pyx_t_9);
  __pyx_t_12 = (__pyx_t_11 + PyBytes_GET_SIZE(__pyx_t_9));
  for (__pyx_t_13 = __pyx_t_11; __pyx_t_13 < __pyx_t_12; __pyx_t_13++) {
    __pyx_t_10 = __pyx_t_13;
    __pyx_v_c = (__pyx_t_10[0]);

    /* "View.MemoryView":526
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_v_i = __pyx_t_8;

    /* "View.MemoryView":525
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":526
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "View.MemoryView":512
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":528
 *             itemp[i] = c
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  char *__pyx_t_4;
  void *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (unlikely(__pyx_v_info == NULL)) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":530
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 */
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":531
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Cannot_create_writable_memory_vi, 0, 0);
    __PYX_ERR(16, 531, __pyx_L1_error)

    /* "View.MemoryView":530
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 */
  }

  /* "View.MemoryView":533
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":534
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __pyx_t_3 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_3;

    /* "View.MemoryView":533
 *             raise ValueError, "Cannot create writable memory view from read-only memoryview"
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":536
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":538
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":539
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __pyx_t_3 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_3;

    /* "View.MemoryView":538
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":541
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":543
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":544
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __pyx_t_3 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_3;

    /* "View.MemoryView":543
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":546
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":548
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":549
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":548
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":551
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":553
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __pyx_t_5 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_5;

  /* "View.MemoryView":554
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __pyx_t_6 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":555
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __pyx_t_7 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_7;

  /* "View.MemoryView":556
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __pyx_t_7 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_7;

  /* "View.MemoryView":557
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":558
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":528
 *             itemp[i] = c
 * 
 *     @cname('getbuffer')             # <<<<<<<<<<<<<<
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":563
 * 
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":565
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(16, 565, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":566
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(16, 566, __pyx_L1_error)

  /* "View.MemoryView":567
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":563
 * 
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":569
 *         return result
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base(self):
 *         return self._get_base()
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":571
 *     @property
 *     def base(self):
 *         return self._get_base()             # <<<<<<<<<<<<<<
 * 
 *     cdef _get_base(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->_get_base(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":569
 *         return result
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base(self):
 *         return self._get_base()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.base.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":573
 *         return self._get_base()
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

static PyObject *__pyx_memoryview__get_base(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_base", 0);

  /* "View.MemoryView":574
 * 
 *     cdef _get_base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":573
 *         return self._get_base()
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":576
 *         return self.obj
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_7genexpr__pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":578
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
      __pyx_t_2 = __pyx_t_4;
      __pyx_7genexpr__pyx_v_length = (__pyx_t_2[0]);
      __pyx_t_5 = PyInt_FromSsize_t(__pyx_7genexpr__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 578, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(16, 578, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":576
 *         return self.obj
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":580
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def strides(self):
 *         if self.view.strides == NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_8genexpr1__pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":582
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError, "Buffer view does not expose strides"
 */
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":584
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError, "Buffer view does not expose strides"             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Buffer_view_does_not_expose_stri, 0, 0);
    __PYX_ERR(16, 584, __pyx_L1_error)

    /* "View.MemoryView":582
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError, "Buffer view does not expose strides"
 */
  }

  /* "View.MemoryView":586
 *             raise ValueError, "Buffer view does not expose strides"
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_8genexpr1__pyx_v_stride = (__pyx_t_3[0]);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_8genexpr1__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(16, 586, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  } /* exit inner scope */
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":580
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def strides(self):
 *         if self.view.strides == NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":588
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_8genexpr2__pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":590
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":591
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__6, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":590
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":593
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 593, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
    for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {
      __pyx_t_4 = __pyx_t_6;
      __pyx_8genexpr2__pyx_v_suboffset = (__pyx_t_4[0]);
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_8genexpr2__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 593, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(16, 593, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  } /* exit inner scope */
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":588
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":595
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ndim(self):
 *         return self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":597
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":595
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ndim(self):
 *         return self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":599
 *         return self.view.ndim
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def itemsize(self):
 *         return self.view.itemsize
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":601
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":599
 *         return self.view.ndim
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def itemsize(self):
 *         return self.view.itemsize
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":603
 *         return self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def nbytes(self):
 *         return self.size * self.view.itemsize
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":605
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":603
 *         return self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def nbytes(self):
 *         return self.size * self.view.itemsize
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":607
 *         return self.size * self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def size(self):
 *         if self._size is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":609
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":610
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":612
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 612, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "View.MemoryView":613
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 613, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);
      __pyx_t_6 = 0;
    }

    /* "View.MemoryView":615
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":609
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":617
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":607
 *         return self.size * self.view.itemsize
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def size(self):
 *         if self._size is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":619
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":620
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":621
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":620
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":623
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":619
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":625
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":626
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":627
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":626
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":625
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":629
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":630
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":629
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":633
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":636
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(16, 636, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":637
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":633
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":639
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":642
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(16, 642, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":643
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":639
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":645
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":647
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":649
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":650
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 650, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":655
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":645
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":657
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":659
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":661
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":662
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 662, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":667
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":657
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(16, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(16, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":671
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":672
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":673
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":674
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":671
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":677
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":678
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":677
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":680
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_idx;
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":686
 *     """
 *     cdef Py_ssize_t idx
 *     tup = <tuple>index if isinstance(index, tuple) else (index,)             # <<<<<<<<<<<<<<
 * 
 *     result = [slice(None)] * ndim
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_index); 
  if ((__pyx_t_2 != 0)) {
    __Pyx_INCREF(((PyObject*)__pyx_v_index));
    __pyx_t_1 = __pyx_v_index;
  } else {
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 686, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_v_tup = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":688
 *     tup = <tuple>index if isinstance(index, tuple) else (index,)
 * 
 *     result = [slice(None)] * ndim             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_t_1 = PyList_New(1 * ((__pyx_v_ndim<0) ? 0:__pyx_v_ndim)); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_v_ndim; __pyx_temp++) {
      __Pyx_INCREF(__pyx_slice__7);
      __Pyx_GIVEREF(__pyx_slice__7);
      PyList_SET_ITEM(__pyx_t_1, __pyx_temp, __pyx_slice__7);
    }
  }
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":689
 * 
 *     result = [slice(None)] * ndim
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     idx = 0
 */
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":690
 *     result = [slice(None)] * ndim
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     idx = 0
 *     for item in tup:
 */
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":691
 *     have_slices = False
 *     seen_ellipsis = False
 *     idx = 0             # <<<<<<<<<<<<<<
 *     for item in tup:
 *         if item is Ellipsis:
 */
  __pyx_v_idx = 0;

  /* "View.MemoryView":692
 *     seen_ellipsis = False
 *     idx = 0
 *     for item in tup:             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  if (unlikely(__pyx_v_tup == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(16, 692, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
  for (;;) {
    if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(16, 692, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":693
 *     idx = 0
 *     for item in tup:
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)
 */
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (__pyx_t_5) {

      /* "View.MemoryView":694
 *     for item in tup:
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True
 */
      __pyx_t_5 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);
      if (__pyx_t_5) {

        /* "View.MemoryView":695
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             have_slices = True
 */
        if (unlikely(__pyx_v_tup == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
          __PYX_ERR(16, 695, __pyx_L1_error)
        }
        __pyx_t_6 = PyTuple_GET_SIZE(__pyx_v_tup); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(16, 695, __pyx_L1_error)
        __pyx_v_idx = (__pyx_v_idx + (__pyx_v_ndim - __pyx_t_6));

        /* "View.MemoryView":696
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":694
 *     for item in tup:
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True
 */
      }

      /* "View.MemoryView":697
 *                 idx += ndim - len(tup)
 *                 seen_ellipsis = True
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if isinstance(item, slice):
 */
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":693
 *     idx = 0
 *     for item in tup:
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 idx += ndim - len(tup)
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":699
 *             have_slices = True
 *         else:
 *             if isinstance(item, slice):             # <<<<<<<<<<<<<<
 *                 have_slices = True
 *             elif not PyIndex_Check(item):
 */
    /*else*/ {
      __pyx_t_5 = PySlice_Check(__pyx_v_item); 
      __pyx_t_2 = (__pyx_t_5 != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":700
 *         else:
 *             if isinstance(item, slice):
 *                 have_slices = True             # <<<<<<<<<<<<<<
 *             elif not PyIndex_Check(item):
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 */
        __pyx_v_have_slices = 1;

        /* "View.MemoryView":699
 *             have_slices = True
 *         else:
 *             if isinstance(item, slice):             # <<<<<<<<<<<<<<
 *                 have_slices = True
 *             elif not PyIndex_Check(item):
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":701
 *             if isinstance(item, slice):
 *                 have_slices = True
 *             elif not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item
 */
      __pyx_t_2 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);
      if (unlikely(__pyx_t_2)) {

        /* "View.MemoryView":702
 *                 have_slices = True
 *             elif not PyIndex_Check(item):
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"             # <<<<<<<<<<<<<<
 *             result[idx] = item
 *         idx += 1
 */
        __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 702, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = 0;
        __pyx_t_7 = 127;
        __Pyx_INCREF(__pyx_kp_u_Cannot_index_with_type);
        __pyx_t_6 += 24;
        __Pyx_GIVEREF(__pyx_kp_u_Cannot_index_with_type);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Cannot_index_with_type);
        __pyx_t_8 = __Pyx_PyObject_FormatSimple(((PyObject *)Py_TYPE(__pyx_v_item)), __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(16, 702, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_7;
        __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_8);
        __pyx_t_8 = 0;
        __Pyx_INCREF(__pyx_kp_u__8);
        __pyx_t_6 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__8);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__8);
        __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(16, 702, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_builtin_TypeError, __pyx_t_8, 0, 0);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(16, 702, __pyx_L1_error)

        /* "View.MemoryView":701
 *             if isinstance(item, slice):
 *                 have_slices = True
 *             elif not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item
 */
      }
      __pyx_L7:;

      /* "View.MemoryView":703
 *             elif not PyIndex_Check(item):
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item             # <<<<<<<<<<<<<<
 *         idx += 1
 * 
 */
      if (unlikely(__Pyx_SetItemInt(__pyx_v_result, __pyx_v_idx, __pyx_v_item, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1) < 0)) __PYX_ERR(16, 703, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "View.MemoryView":704
 *                 raise TypeError, f"Cannot index with type '{type(item)}'"
 *             result[idx] = item
 *         idx += 1             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - idx
 */
    __pyx_v_idx = (__pyx_v_idx + 1);

    /* "View.MemoryView":692
 *     seen_ellipsis = False
 *     idx = 0
 *     for item in tup:             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":706
 *         idx += 1
 * 
 *     nslices = ndim - idx             # <<<<<<<<<<<<<<
 *     return have_slices or nslices, tuple(result)
 * 
 */
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_v_idx);

  /* "View.MemoryView":707
 * 
 *     nslices = ndim - idx
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_8 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_8)) __PYX_ERR(16, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_8 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_8)) __PYX_ERR(16, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = __pyx_t_8;
  __pyx_t_8 = 0;
  __pyx_L8_bool_binop_done:;
  __pyx_t_8 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_8)) __PYX_ERR(16, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_8);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":680
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":709
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":710
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError, "Indirect dimensions not supported"
 */
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":711
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Indirect dimensions not supported"
 * 
 */
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":712
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError, "Indirect dimensions not supported"             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_Raise(__pyx_builtin_ValueError, __pyx_kp_s_Indirect_dimensions_not_supporte, 0, 0);
      __PYX_ERR(16, 712, __pyx_L1_error)

      /* "View.MemoryView":711
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError, "Indirect dimensions not supported"
 * 
 */
    }
  }

  /* "View.MemoryView":709
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":719
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  Py_ssize_t __pyx_v_cindex;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":720
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":727
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":731
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = ((__pyx_v_memview->view.ndim > 0) != 0);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(16, 731, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(16, 731, __pyx_L1_error)
  #endif

  /* "View.MemoryView":733
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":734
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(16, 734, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":735
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":733
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":737
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":738
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":744
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __pyx_t_4 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":745
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":750
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step, cindex
 */
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":751
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step, cindex
 *     cdef bint have_start, have_stop, have_step
 */
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":755
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             cindex = index
 */
  __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(16, 755, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(16, 755, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(16, 755, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(16, 755, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(16, 755, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(16, 755, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_v_dim = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":756
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             cindex = index
 *             slice_memviewslice(
 */
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":757
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             cindex = index             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(16, 757, __pyx_L1_error)
      __pyx_v_cindex = __pyx_t_10;

      /* "View.MemoryView":758
 *         if PyIndex_Check(index):
 *             cindex = index
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_cindex, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(16, 758, __pyx_L1_error)

      /* "View.MemoryView":756
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             cindex = index
 *             slice_memviewslice(
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":764
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __pyx_t_2 = (__pyx_v_index == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":765
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":766
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":767
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":768
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":764
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":770
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(16, 770, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(16, 770, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(16, 770, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":771
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(16, 771, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(16, 771, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(16, 771, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":772
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(16, 772, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(16, 772, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(16, 772, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":774
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(16, 774, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":775
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(16, 775, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":776
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(16, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":778
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(16, 778, __pyx_L1_error)

      /* "View.MemoryView":784
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":755
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             cindex = index
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":786
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":787
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);

    /* "View.MemoryView":788
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(16, 788, __pyx_L1_error) }

    /* "View.MemoryView":789
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(16, 789, __pyx_L1_error) }

    /* "View.MemoryView":787
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(16, 787, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":786
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":792
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF((PyObject *)__pyx_r);

    /* "View.MemoryView":793
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":792
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(16, 792, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":719
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":817
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":837
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":839
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":840
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)
 */
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":839
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":841
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":842
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(PyExc_IndexError, __pyx_kp_s_Index_out_of_bounds_axis_d, __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(16, 842, __pyx_L1_error)

      /* "View.MemoryView":841
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":837
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":845
 *     else:
 * 
 *         if have_step:             # <<<<<<<<<<<<<<
 *             negative_step = step < 0
 *             if step == 0:
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_have_step != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":846
 * 
 *         if have_step:
 *             negative_step = step < 0             # <<<<<<<<<<<<<<
 *             if step == 0:
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 */
      __pyx_v_negative_step = (__pyx_v_step < 0);

      /* "View.MemoryView":847
 *         if have_step:
 *             negative_step = step < 0
 *             if step == 0:             # <<<<<<<<<<<<<<
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_step == 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":848
 *             negative_step = step < 0
 *             if step == 0:
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             negative_step = False
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(PyExc_ValueError, __pyx_kp_s_Step_may_not_be_zero_axis_d, __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(16, 848, __pyx_L1_error)

        /* "View.MemoryView":847
 *         if have_step:
 *             negative_step = step < 0
 *             if step == 0:             # <<<<<<<<<<<<<<
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 *         else:
 */
      }

      /* "View.MemoryView":845
 *     else:
 * 
 *         if have_step:             # <<<<<<<<<<<<<<
 *             negative_step = step < 0
 *             if step == 0:
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":850
 *                 _err_dim(PyExc_ValueError, "Step may not be zero (axis %d)", dim)
 *         else:
 *             negative_step = False             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    /*else*/ {
      __pyx_v_negative_step = 0;

      /* "View.MemoryView":851
 *         else:
 *             negative_step = False
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_step = 1;
    }
    __pyx_L6:;

    /* "View.MemoryView":854
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":855
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":856
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":857
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":858
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __pyx_v_start = 0;

          /* "View.MemoryView":857
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":855
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L9;
      }

      /* "View.MemoryView":859
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":860
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        __pyx_t_2 = (__pyx_v_negative_step != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":861
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":860
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L11;
        }

        /* "View.MemoryView":863
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L11:;

        /* "View.MemoryView":859
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L9:;

      /* "View.MemoryView":854
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":865
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":866
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":865
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":868
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L12:;
    }
    __pyx_L8:;

    /* "View.MemoryView":870
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":871
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":872
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":873
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":874
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __pyx_v_stop = 0;

          /* "View.MemoryView":873
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":871
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L14;
      }

      /* "View.MemoryView":875
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":876
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":875
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L14:;

      /* "View.MemoryView":870
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L13;
    }

    /* "View.MemoryView":878
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":879
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __pyx_v_stop = -1L;

        /* "View.MemoryView":878
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L16;
      }

      /* "View.MemoryView":881
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 * 
 */
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L16:;
    }
    __pyx_L13:;

    /* "View.MemoryView":885
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":887
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":888
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":887
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":890
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":891
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":890
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":894
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":895
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":896
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":899
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":900
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":899
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L19;
  }

  /* "View.MemoryView":902
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L19:;

  /* "View.MemoryView":904
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":905
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":906
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":907
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(PyExc_IndexError, "All dimensions preceding dimension %d "
 */
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":906
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L22;
      }

      /* "View.MemoryView":909
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(PyExc_IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      /*else*/ {

        /* "View.MemoryView":910
 *             else:
 *                 _err_dim(PyExc_IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(PyExc_IndexError, __pyx_kp_s_All_dimensions_preceding_dimensi, __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(16, 909, __pyx_L1_error)
      }
      __pyx_L22:;

      /* "View.MemoryView":905
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L21;
    }

    /* "View.MemoryView":912
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L21:;

    /* "View.MemoryView":904
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":914
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":817
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":920
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":922
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":923
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":926
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len // itemsize
 *         stride = itemsize
 */
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":927
 * 
 *     if view.ndim == 0:
 *         shape = view.len // itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(16, 927, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(__Pyx_UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(16, 927, __pyx_L1_error)
    }
    __pyx_v_shape = __Pyx_div_Py_ssize_t(__pyx_v_view->len, __pyx_v_itemsize);

    /* "View.MemoryView":928
 *     if view.ndim == 0:
 *         shape = view.len // itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":926
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len // itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":930
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":931
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":932
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":933
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":932
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":935
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":936
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 */
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":937
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":938
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = 0;
      __pyx_t_4 = 127;
      __Pyx_INCREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
      __pyx_t_1 += 37;
      __Pyx_GIVEREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Out_of_bounds_on_buffer_access_a);
      __pyx_t_5 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_dim, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u__9);
      __pyx_t_1 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__9);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__9);
      __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_builtin_IndexError, __pyx_t_5, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(16, 938, __pyx_L1_error)

      /* "View.MemoryView":937
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
    }

    /* "View.MemoryView":935
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":940
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":941
 * 
 *     if index >= shape:
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
    __pyx_t_1 += 37;
    __Pyx_GIVEREF(__pyx_kp_u_Out_of_bounds_on_buffer_access_a);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Out_of_bounds_on_buffer_access_a);
    __pyx_t_3 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_dim, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u__9);
    __pyx_t_1 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__9);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u__9);
    __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_builtin_IndexError, __pyx_t_3, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(16, 941, __pyx_L1_error)

    /* "View.MemoryView":940
 *             raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 */
  }

  /* "View.MemoryView":943
 *         raise IndexError, f"Out of bounds on buffer access (axis {dim})"
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":944
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":945
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":944
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":947
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":920
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":953
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except -1:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":954
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except -1:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":956
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":957
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":961
 * 
 *     cdef int i, j
 *     for i in range(ndim // 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __pyx_t_3 = __Pyx_div_long(__pyx_v_ndim, 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":962
 *     cdef int i, j
 *     for i in range(ndim // 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":963
 *     for i in range(ndim // 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":964
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":966
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":967
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
      __pyx_t_9 = __pyx_memoryview_err(PyExc_ValueError, __pyx_kp_s_Cannot_transpose_memoryview_with); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(16, 967, __pyx_L1_error)

      /* "View.MemoryView":966
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":969
 *             _err(PyExc_ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":953
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except -1:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":986
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":987
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":986
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":989
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":990
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":991
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 991, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":990
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":993
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":989
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":995
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":996
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":997
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(16, 997, __pyx_L1_error)

    /* "View.MemoryView":996
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":999
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef _get_base(self):
 */
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":995
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1001
 *             memoryview.assign_item_from_object(self, itemp, value)
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

static PyObject *__pyx_memoryviewslice__get_base(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_base", 0);

  /* "View.MemoryView":1002
 * 
 *     cdef _get_base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":1001
 *             memoryview.assign_item_from_object(self, itemp, value)
 * 
 *     cdef _get_base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(16, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(16, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1008
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1016
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1017
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1016
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1022
 * 
 * 
 *     result = _memoryviewslice.__new__(_memoryviewslice, None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = ((PyObject *)__pyx_tp_new__memoryviewslice(((PyTypeObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 1022, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1024
 *     result = _memoryviewslice.__new__(_memoryviewslice, None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1025
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview)._get_base()
 */
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1027
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview)._get_base()             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->__pyx_vtab)->_get_base(((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1028
 * 
 *     result.from_object = (<memoryview> memviewslice.memview)._get_base()
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1030
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1031
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1032
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1033
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1034
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  Py_INCREF(Py_None);

  /* "View.MemoryView":1036
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1037
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1036
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1039
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1041
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1042
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1045
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1046
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1047
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1048
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1049
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      goto __pyx_L6_break;

      /* "View.MemoryView":1047
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1051
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1052
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 1052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1053
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 1053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 1053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(16, 1053, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1055
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1056
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1058
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result);
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":1008
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1061
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1064
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1065
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(16, 1065, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":1066
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1064
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1068
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1069
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1061
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1072
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1076
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1077
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1078
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1080
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1081
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1083
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1084
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1085
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1086
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    if ((__pyx_v_suboffsets != 0)) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1072
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1092
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1093
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 1093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1096
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *(*__pyx_t_3)(char *);
  int (*__pyx_t_4)(char *, PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1103
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1104
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1105
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1103
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1107
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1108
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1110
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1112
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(16, 1110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1096
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1118
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     return -arg if arg < 0 else arg
 * 
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;

  /* "View.MemoryView":1119
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     return -arg if arg < 0 else arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  if (((__pyx_v_arg < 0) != 0)) {
    __pyx_t_1 = (-__pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":1118
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     return -arg if arg < 0 else arg
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1122
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1127
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1128
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1130
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1131
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1132
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1133
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      goto __pyx_L4_break;

      /* "View.MemoryView":1131
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1135
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1136
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1137
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1138
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      goto __pyx_L7_break;

      /* "View.MemoryView":1136
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1140
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1141
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1140
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1143
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1122
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1146
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1153
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1154
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1155
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1156
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1158
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1159
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1160
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_3 = (__pyx_t_2 != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1159
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1161
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1159
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1163
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */
    /*else*/ {
      __pyx_t_4 = __pyx_v_dst_extent;
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1164
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1165
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1166
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1158
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1168
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_dst_extent;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1169
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1173
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1174
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1146
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1176
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1179
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1176
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1183
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1185
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1187
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1188
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);
  }

  /* "View.MemoryView":1190
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1183
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1193
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1202
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1203
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1204
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1205
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1202
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1207
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1208
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1209
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1211
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1193
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1214
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":1225
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1226
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1228
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err_no_memory()
 */
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1229
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err_no_memory()
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1230
 *     result = malloc(size)
 *     if not result:
 *         _err_no_memory()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_err_no_memory(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(16, 1230, __pyx_L1_error)

    /* "View.MemoryView":1229
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err_no_memory()
 * 
 */
  }

  /* "View.MemoryView":1233
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1234
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1235
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1236
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1237
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize, ndim, order)
 */
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1239
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize, ndim, order)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1242
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1243
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1244
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1243
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1246
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1247
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1246
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1249
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1251
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1214
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1256
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError, f"got differing extents in dimension {i} (got {extent1} and {extent2})"
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1258
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError, f"got differing extents in dimension {i} (got {extent1} and {extent2})"             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_got_differing_extents_in_dimensi);
  __pyx_t_2 += 35;
  __Pyx_GIVEREF(__pyx_kp_u_got_differing_extents_in_dimensi);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_got_differing_extents_in_dimensi);
  __pyx_t_4 = __Pyx_PyUnicode_From_int(__pyx_v_i, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_got);
  __pyx_t_2 += 6;
  __Pyx_GIVEREF(__pyx_kp_u_got);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_got);
  __pyx_t_4 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_extent1, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_and);
  __pyx_t_2 += 5;
  __Pyx_GIVEREF(__pyx_kp_u_and);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_and);
  __pyx_t_4 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_extent2, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__9);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__9);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__9);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(16, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_builtin_ValueError, __pyx_t_4, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(16, 1258, __pyx_L1_error)

  /* "View.MemoryView":1256
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError, f"got differing extents in dimension {i} (got {extent1} and {extent2})"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(PyObject *error, str msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg % dim
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, PyObject *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_INCREF(__pyx_v_msg);

  /* "View.MemoryView":1262
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(PyObject *error, str msg, int dim) except -1 with gil:
 *     raise <object>error, msg % dim             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 1262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyString_FormatSafe(__pyx_v_msg, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 1262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(((PyObject *)__pyx_v_error), __pyx_t_2, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(16, 1262, __pyx_L1_error)

  /* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(PyObject *error, str msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg % dim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1265
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(PyObject *error, str msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg
 * 
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, PyObject *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_INCREF(__pyx_v_msg);

  /* "View.MemoryView":1266
 * @cname('__pyx_memoryview_err')
 * cdef int _err(PyObject *error, str msg) except -1 with gil:
 *     raise <object>error, msg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_no_memory')
 */
  __Pyx_Raise(((PyObject *)__pyx_v_error), __pyx_v_msg, 0, 0);
  __PYX_ERR(16, 1266, __pyx_L1_error)

  /* "View.MemoryView":1265
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(PyObject *error, str msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise <object>error, msg
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1269
 * 
 * @cname('__pyx_memoryview_err_no_memory')
 * cdef int _err_no_memory() except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise MemoryError
 * 
 */

static int __pyx_memoryview_err_no_memory(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_no_memory", 0);

  /* "View.MemoryView":1270
 * @cname('__pyx_memoryview_err_no_memory')
 * cdef int _err_no_memory() except -1 with gil:
 *     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 * 
 */
  PyErr_NoMemory(); __PYX_ERR(16, 1270, __pyx_L1_error)

  /* "View.MemoryView":1269
 * 
 * @cname('__pyx_memoryview_err_no_memory')
 * cdef int _err_no_memory() except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise MemoryError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView._err_no_memory", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1274
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif

  /* "View.MemoryView":1282
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1283
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1285
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1286
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1287
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1290
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1291
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1290
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1292
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1293
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1292
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1295
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1297
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1298
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1299
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":1300
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1301
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1299
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1303
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(16, 1303, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1298
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1305
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 */
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1306
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __pyx_t_6 = __pyx_memoryview_err_dim(PyExc_ValueError, __pyx_kp_s_Dimension_d_is_not_direct, __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(16, 1306, __pyx_L1_error)

      /* "View.MemoryView":1305
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1308
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1310
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1311
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1310
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1313
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(16, 1313, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1314
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1308
 *             _err_dim(PyExc_ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1316
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1319
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1320
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1319
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1321
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1322
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1321
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1324
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 */
    __pyx_t_2 = (__pyx_v_direct_copy != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1326
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1327
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 *             free(tmpdata)
 */
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1328
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1329
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1330
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1324
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 */
    }

    /* "View.MemoryView":1316
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1332
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  __pyx_t_8 = (__pyx_t_2 != 0);
  if (__pyx_t_8) {

    /* "View.MemoryView":1335
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(16, 1335, __pyx_L1_error)

    /* "View.MemoryView":1336
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(16, 1336, __pyx_L1_error)

    /* "View.MemoryView":1332
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1338
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1339
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 * 
 */
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1340
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1342
 *     refcount_copying(&dst, dtype_is_object, ndim, inc=True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1343
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1274
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1346
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":1350
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1352
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1353
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1354
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1355
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1357
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1358
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1359
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1360
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1346
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1368
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) nogil:             # <<<<<<<<<<<<<<
 * 
 *     if dtype_is_object:
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {
  int __pyx_t_1;

  /* "View.MemoryView":1370
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) nogil:
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape, dst.strides, ndim, inc)
 * 
 */
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1371
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape, dst.strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 */
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1370
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) nogil:
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape, dst.strides, ndim, inc)
 * 
 */
  }

  /* "View.MemoryView":1368
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object, int ndim, bint inc) nogil:             # <<<<<<<<<<<<<<
 * 
 *     if dtype_is_object:
 */

  /* function exit code */
}

/* "View.MemoryView":1374
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  __Pyx_RefNannyDeclarations
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1377
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1374
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1380
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1383
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(shape[0]):
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1385
 *     cdef Py_ssize_t stride = strides[0]
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1386
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":1387
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      __pyx_t_4 = (__pyx_v_inc != 0);
      if (__pyx_t_4) {

        /* "View.MemoryView":1388
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1387
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1390
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1, ndim - 1, inc)
 */
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1386
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1392
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1, ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += stride
 */
    /*else*/ {
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1394
 *             refcount_objects_in_slice(data, shape + 1, strides + 1, ndim - 1, inc)
 * 
 *         data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
  }

  /* "View.MemoryView":1380
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1400
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1403
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, inc=False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim, itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, inc=True)
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1404
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, inc=False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim, itemsize, item)             # <<<<<<<<<<<<<<
 *     refcount_copying(dst, dtype_is_object, ndim, inc=True)
 * 
 */
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1405
 *     refcount_copying(dst, dtype_is_object, ndim, inc=False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim, itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, inc=True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1400
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1413
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1414
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1417
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1418
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1419
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1421
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1, ndim - 1, itemsize, item)
 *             data += stride
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1422
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1, ndim - 1, itemsize, item)             # <<<<<<<<<<<<<<
 *             data += stride
 * 
 */
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1423
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1, ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(16, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(16, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(16, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(16, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, __pyx_nargs); __PYX_ERR(16, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x6ae9995:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (%s vs 0x6ae9995 = (name))" % __pyx_checksum
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x6ae9995) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x6ae9995:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (%s vs 0x6ae9995 = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x6ae9995:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (%s vs 0x6ae9995 = (name))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x6a, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_2, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(16, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x6ae9995:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (%s vs 0x6ae9995 = (name))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (%s vs 0x6ae9995 = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (%s vs 0x6ae9995 = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", __Pyx_PyType_Name(Py_TYPE(__pyx_v___pyx_state))), 0))) __PYX_ERR(16, 9, __pyx_L1_error)
    __pyx_t_2 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_2)) __PYX_ERR(16, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (%s vs 0x6ae9995 = (name))" % __pyx_checksum
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(16, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(16, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(16, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(16, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(16, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(16, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D __pyx_vtable_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D;

static PyObject *__pyx_tp_new_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *p;
  #if CYTHON_COMPILING_IN_LIMITED_API
  newfunc new_func = (newfunc)PyType_GetSlot(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D, Py_tp_new);
  PyObject *o = new_func(t, a, k);
  #else
  PyObject *o = __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D->tp_new(t, a, k);
  #endif
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D*)__pyx_vtabptr_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D;
  p->_kdtree = ((struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)Py_None); Py_INCREF(Py_None);
  p->_grid_data = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->_triangle_to_grid_map = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->_triangle_to_grid_map_mv.data = NULL;
  p->_triangle_to_grid_map_mv.memview = NULL;
  p->_grid_data_mv.data = NULL;
  p->_grid_data_mv.memview = NULL;
  return o;
}

#if !CYTHON_COMPILING_IN_LIMITED_API
static void __pyx_tp_dealloc_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D(PyObject *o) {
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *p = (struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_kdtree);
  Py_CLEAR(p->_grid_data);
  Py_CLEAR(p->_triangle_to_grid_map);
  __PYX_XDEC_MEMVIEW(&p->_triangle_to_grid_map_mv, 1);
  p->_triangle_to_grid_map_mv.memview = NULL; p->_triangle_to_grid_map_mv.data = NULL;
  __PYX_XDEC_MEMVIEW(&p->_grid_data_mv, 1);
  p->_grid_data_mv.memview = NULL; p->_grid_data_mv.data = NULL;
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D)) __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D);
}
#endif

static int __pyx_tp_traverse_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *p = (struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *)o;
  e = ((likely(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D)) ? ((__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D->tp_traverse) ? __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D)); if (e) return e;
  if (p->_kdtree) {
    e = (*v)(((PyObject *)p->_kdtree), a); if (e) return e;
  }
  if (p->_grid_data) {
    e = (*v)(((PyObject *)p->_grid_data), a); if (e) return e;
  }
  if (p->_triangle_to_grid_map) {
    e = (*v)(((PyObject *)p->_triangle_to_grid_map), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *p = (struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D *)o;
  if (likely(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D)) { if (__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D->tp_clear) __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D);
  tmp = ((PyObject*)p->_kdtree);
  p->_kdtree = ((struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_grid_data);
  p->_grid_data = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_triangle_to_grid_map);
  p->_triangle_to_grid_map = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D[] = {
  {"__getstate__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_3__getstate__, METH_NOARGS, 0},
  {"__setstate__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_5__setstate__, METH_O, 0},
  {"__reduce__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_7__reduce__, METH_NOARGS, 0},
  {"instance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_9instance, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_8instance},
  {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D_slots[] = {
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D},
  {Py_tp_clear, (void *)__pyx_tp_clear_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D},
  {Py_tp_methods, (void *)__pyx_methods_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D},
  {Py_tp_init, (void *)__pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D},
  {0, 0},
};
static PyType_Spec __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D_spec = {
  "cherab.solps.solps_2d_functions.SOLPSFunction2D",
  sizeof(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D_slots,
};
#else

static PyTypeObject __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D = {
  PyVarObject_HEAD_INIT(0, 0)
  "cherab.solps.solps_2d_functions.""SOLPSFunction2D", /*tp_name*/
  sizeof(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D, /*tp_traverse*/
  __pyx_tp_clear_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D __pyx_vtable_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D;

static PyObject *__pyx_tp_new_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *p;
  #if CYTHON_COMPILING_IN_LIMITED_API
  newfunc new_func = (newfunc)PyType_GetSlot(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D, Py_tp_new);
  PyObject *o = new_func(t, a, k);
  #else
  PyObject *o = __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D->tp_new(t, a, k);
  #endif
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D*)__pyx_vtabptr_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D;
  p->_kdtree = ((struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)Py_None); Py_INCREF(Py_None);
  p->_grid_vectors = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->_triangle_to_grid_map = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->_triangle_to_grid_map_mv.data = NULL;
  p->_triangle_to_grid_map_mv.memview = NULL;
  p->_grid_vectors_mv.data = NULL;
  p->_grid_vectors_mv.memview = NULL;
  return o;
}

#if !CYTHON_COMPILING_IN_LIMITED_API
static void __pyx_tp_dealloc_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D(PyObject *o) {
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *p = (struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_kdtree);
  Py_CLEAR(p->_grid_vectors);
  Py_CLEAR(p->_triangle_to_grid_map);
  __PYX_XDEC_MEMVIEW(&p->_triangle_to_grid_map_mv, 1);
  p->_triangle_to_grid_map_mv.memview = NULL; p->_triangle_to_grid_map_mv.data = NULL;
  __PYX_XDEC_MEMVIEW(&p->_grid_vectors_mv, 1);
  p->_grid_vectors_mv.memview = NULL; p->_grid_vectors_mv.data = NULL;
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D)) __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D);
}
#endif

static int __pyx_tp_traverse_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *p = (struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *)o;
  e = ((likely(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D)) ? ((__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D->tp_traverse) ? __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D)); if (e) return e;
  if (p->_kdtree) {
    e = (*v)(((PyObject *)p->_kdtree), a); if (e) return e;
  }
  if (p->_grid_vectors) {
    e = (*v)(((PyObject *)p->_grid_vectors), a); if (e) return e;
  }
  if (p->_triangle_to_grid_map) {
    e = (*v)(((PyObject *)p->_triangle_to_grid_map), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *p = (struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D *)o;
  if (likely(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D)) { if (__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D->tp_clear) __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D);
  tmp = ((PyObject*)p->_kdtree);
  p->_kdtree = ((struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_grid_vectors);
  p->_grid_vectors = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_triangle_to_grid_map);
  p->_triangle_to_grid_map = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D[] = {
  {"__getstate__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_3__getstate__, METH_NOARGS, 0},
  {"__setstate__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_5__setstate__, METH_O, 0},
  {"__reduce__", (PyCFunction)__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_7__reduce__, METH_NOARGS, 0},
  {"instance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_9instance, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_8instance},
  {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D_slots[] = {
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D},
  {Py_tp_clear, (void *)__pyx_tp_clear_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D},
  {Py_tp_methods, (void *)__pyx_methods_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D},
  {Py_tp_init, (void *)__pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D},
  {0, 0},
};
static PyType_Spec __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D_spec = {
  "cherab.solps.solps_2d_functions.SOLPSVectorFunction2D",
  sizeof(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D_slots,
};
#else

static PyTypeObject __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D = {
  PyVarObject_HEAD_INIT(0, 0)
  "cherab.solps.solps_2d_functions.""SOLPSVectorFunction2D", /*tp_name*/
  sizeof(struct __pyx_obj_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D, /*tp_traverse*/
  __pyx_tp_clear_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_array___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type___pyx_array_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_array},
  {Py_sq_length, (void *)__pyx_array___len__},
  {Py_sq_item, (void *)__pyx_sq_item_array},
  {Py_mp_length, (void *)__pyx_array___len__},
  {Py_mp_subscript, (void *)__pyx_array___getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_array},
  {Py_tp_getattro, (void *)__pyx_tp_getattro_array},
  {Py_tp_methods, (void *)__pyx_methods_array},
  {Py_tp_getset, (void *)__pyx_getsets_array},
  {Py_tp_new, (void *)__pyx_tp_new_array},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_array_spec = {
  "cherab.solps.solps_2d_functions.array",
  sizeof(struct __pyx_array_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type___pyx_array_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "cherab.solps.solps_2d_functions.""array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

#if !CYTHON_COMPILING_IN_LIMITED_API
static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}
#endif

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type___pyx_MemviewEnum_slots[] = {
  {Py_tp_repr, (void *)__pyx_MemviewEnum___repr__},
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_Enum},
  {Py_tp_clear, (void *)__pyx_tp_clear_Enum},
  {Py_tp_methods, (void *)__pyx_methods_Enum},
  {Py_tp_init, (void *)__pyx_MemviewEnum___init__},
  {Py_tp_new, (void *)__pyx_tp_new_Enum},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_MemviewEnum_spec = {
  "cherab.solps.solps_2d_functions.Enum",
  sizeof(struct __pyx_MemviewEnum_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type___pyx_MemviewEnum_slots,
};
#else

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "cherab.solps.solps_2d_functions.""Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_memoryview___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type___pyx_memoryview_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_memoryview},
  {Py_tp_repr, (void *)__pyx_memoryview___repr__},
  {Py_sq_length, (void *)__pyx_memoryview___len__},
  {Py_sq_item, (void *)__pyx_sq_item_memoryview},
  {Py_mp_length, (void *)__pyx_memoryview___len__},
  {Py_mp_subscript, (void *)__pyx_memoryview___getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_memoryview},
  {Py_tp_str, (void *)__pyx_memoryview___str__},
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_memoryview},
  {Py_tp_clear, (void *)__pyx_tp_clear_memoryview},
  {Py_tp_methods, (void *)__pyx_methods_memoryview},
  {Py_tp_getset, (void *)__pyx_getsets_memoryview},
  {Py_tp_new, (void *)__pyx_tp_new_memoryview},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_memoryview_spec = {
  "cherab.solps.solps_2d_functions.memoryview",
  sizeof(struct __pyx_memoryview_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type___pyx_memoryview_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "cherab.solps.solps_2d_functions.""memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  #if CYTHON_COMPILING_IN_LIMITED_API
  newfunc new_func = (newfunc)PyType_GetSlot(__pyx_memoryview_type, Py_tp_new);
  PyObject *o = new_func(t, a, k);
  #else
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  #endif
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_memoryviewslice___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type___pyx_memoryviewslice_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc__memoryviewslice},
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_doc, (void *)PyDoc_STR("Internal class for passing memoryview slices to Python")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse__memoryviewslice},
  {Py_tp_clear, (void *)__pyx_tp_clear__memoryviewslice},
  {Py_tp_methods, (void *)__pyx_methods__memoryviewslice},
  {Py_tp_new, (void *)__pyx_tp_new__memoryviewslice},
  {0, 0},
};
static PyType_Spec __pyx_type___pyx_memoryviewslice_spec = {
  "cherab.solps.solps_2d_functions._memoryviewslice",
  sizeof(struct __pyx_memoryviewslice_obj),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type___pyx_memoryviewslice_slots,
};
#else

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "cherab.solps.solps_2d_functions.""_memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Internal class for passing memoryview slices to Python"), /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  #if CYTHON_COMPILING_IN_LIMITED_API
  {0, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {0, __pyx_k_All_dimensions_preceding_dimensi, sizeof(__pyx_k_All_dimensions_preceding_dimensi), 0, 0, 1, 0},
  {0, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {0, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {0, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {0, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {0, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {0, __pyx_k_Cannot_index_with_type, sizeof(__pyx_k_Cannot_index_with_type), 0, 1, 0, 0},
  {0, __pyx_k_Cannot_transpose_memoryview_with, sizeof(__pyx_k_Cannot_transpose_memoryview_with), 0, 0, 1, 0},
  {0, __pyx_k_Dimension_d_is_not_direct, sizeof(__pyx_k_Dimension_d_is_not_direct), 0, 0, 1, 0},
  {0, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {0, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {0, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {0, __pyx_k_Incompatible_checksums_s_vs_0x6a, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x6a), 0, 0, 1, 0},
  {0, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {0, __pyx_k_Index_out_of_bounds_axis_d, sizeof(__pyx_k_Index_out_of_bounds_axis_d), 0, 0, 1, 0},
  {0, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {0, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 1, 0, 0},
  {0, __pyx_k_Invalid_shape_in_axis, sizeof(__pyx_k_Invalid_shape_in_axis), 0, 1, 0, 0},
  {0, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {0, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {0, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {0, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {0, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 1, 0, 0},
  {0, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {0, __pyx_k_SOLPSFunction2D, sizeof(__pyx_k_SOLPSFunction2D), 0, 0, 1, 1},
  {0, __pyx_k_SOLPSFunction2D___getstate, sizeof(__pyx_k_SOLPSFunction2D___getstate), 0, 0, 1, 1},
  {0, __pyx_k_SOLPSFunction2D___reduce, sizeof(__pyx_k_SOLPSFunction2D___reduce), 0, 0, 1, 1},
  {0, __pyx_k_SOLPSFunction2D___setstate, sizeof(__pyx_k_SOLPSFunction2D___setstate), 0, 0, 1, 1},
  {0, __pyx_k_SOLPSFunction2D_instance, sizeof(__pyx_k_SOLPSFunction2D_instance), 0, 0, 1, 1},
  {0, __pyx_k_SOLPSVectorFunction2D, sizeof(__pyx_k_SOLPSVectorFunction2D), 0, 0, 1, 1},
  {0, __pyx_k_SOLPSVectorFunction2D___getstate, sizeof(__pyx_k_SOLPSVectorFunction2D___getstate), 0, 0, 1, 1},
  {0, __pyx_k_SOLPSVectorFunction2D___reduce, sizeof(__pyx_k_SOLPSVectorFunction2D___reduce), 0, 0, 1, 1},
  {0, __pyx_k_SOLPSVectorFunction2D___setstate, sizeof(__pyx_k_SOLPSVectorFunction2D___setstate), 0, 0, 1, 1},
  {0, __pyx_k_SOLPSVectorFunction2D_instance, sizeof(__pyx_k_SOLPSVectorFunction2D_instance), 0, 0, 1, 1},
  {0, __pyx_k_Step_may_not_be_zero_axis_d, sizeof(__pyx_k_Step_may_not_be_zero_axis_d), 0, 0, 1, 0},
  {0, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {0, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {0, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {0, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {0, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {0, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
  {0, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 1, 1},
  {0, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
  {0, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
  {0, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {0, __pyx_k_and, sizeof(__pyx_k_and), 0, 1, 0, 0},
  {0, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {0, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {0, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {0, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {0, __pyx_k_cherab_solps_solps_2d_functions, sizeof(__pyx_k_cherab_solps_solps_2d_functions), 0, 0, 1, 0},
  {0, __pyx_k_cherab_solps_solps_2d_functions_2, sizeof(__pyx_k_cherab_solps_solps_2d_functions_2), 0, 0, 1, 1},
  {0, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {0, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
  {0, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {0, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {0, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {0, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {0, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {0, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {0, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {0, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {0, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {0, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {0, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {0, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {0, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {0, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {0, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {0, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {0, __pyx_k_got, sizeof(__pyx_k_got), 0, 1, 0, 0},
  {0, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 1, 0, 0},
  {0, __pyx_k_grid_data, sizeof(__pyx_k_grid_data), 0, 0, 1, 1},
  {0, __pyx_k_grid_vectors, sizeof(__pyx_k_grid_vectors), 0, 0, 1, 1},
  {0, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {0, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {0, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
  {0, __pyx_k_instance, sizeof(__pyx_k_instance), 0, 0, 1, 1},
  {0, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {0, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {0, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {0, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
  {0, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {0, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {0, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {0, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {0, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {0, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {0, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {0, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {0, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {0, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {0, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
  {0, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
  {0, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {0, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {0, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {0, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {0, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {0, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {0, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {0, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {0, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {0, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {0, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {0, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {0, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {0, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {0, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {0, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {0, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {0, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {0, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {0, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {0, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
  {0, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {0, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {0, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {0, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {0, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {0, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {0, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {0, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {0, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {0, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, __pyx_k_triangle_to_grid_map, sizeof(__pyx_k_triangle_to_grid_map), 0, 0, 1, 1},
  {0, __pyx_k_triangles, sizeof(__pyx_k_triangles), 0, 0, 1, 1},
  {0, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {0, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {0, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {0, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {0, __pyx_k_vertex_coords, sizeof(__pyx_k_vertex_coords), 0, 0, 1, 1},
  #else
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_kp_s_All_dimensions_preceding_dimensi, __pyx_k_All_dimensions_preceding_dimensi, sizeof(__pyx_k_All_dimensions_preceding_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_u_Cannot_index_with_type, __pyx_k_Cannot_index_with_type, sizeof(__pyx_k_Cannot_index_with_type), 0, 1, 0, 0},
  {&__pyx_kp_s_Cannot_transpose_memoryview_with, __pyx_k_Cannot_transpose_memoryview_with, sizeof(__pyx_k_Cannot_transpose_memoryview_with), 0, 0, 1, 0},
  {&__pyx_kp_s_Dimension_d_is_not_direct, __pyx_k_Dimension_d_is_not_direct, sizeof(__pyx_k_Dimension_d_is_not_direct), 0, 0, 1, 0},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x6a, __pyx_k_Incompatible_checksums_s_vs_0x6a, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x6a), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_s_Index_out_of_bounds_axis_d, __pyx_k_Index_out_of_bounds_axis_d, sizeof(__pyx_k_Index_out_of_bounds_axis_d), 0, 0, 1, 0},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_kp_u_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 1, 0, 0},
  {&__pyx_kp_u_Invalid_shape_in_axis, __pyx_k_Invalid_shape_in_axis, sizeof(__pyx_k_Invalid_shape_in_axis), 0, 1, 0, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_kp_u_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 1, 0, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_SOLPSFunction2D, __pyx_k_SOLPSFunction2D, sizeof(__pyx_k_SOLPSFunction2D), 0, 0, 1, 1},
  {&__pyx_n_s_SOLPSFunction2D___getstate, __pyx_k_SOLPSFunction2D___getstate, sizeof(__pyx_k_SOLPSFunction2D___getstate), 0, 0, 1, 1},
  {&__pyx_n_s_SOLPSFunction2D___reduce, __pyx_k_SOLPSFunction2D___reduce, sizeof(__pyx_k_SOLPSFunction2D___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_SOLPSFunction2D___setstate, __pyx_k_SOLPSFunction2D___setstate, sizeof(__pyx_k_SOLPSFunction2D___setstate), 0, 0, 1, 1},
  {&__pyx_n_s_SOLPSFunction2D_instance, __pyx_k_SOLPSFunction2D_instance, sizeof(__pyx_k_SOLPSFunction2D_instance), 0, 0, 1, 1},
  {&__pyx_n_s_SOLPSVectorFunction2D, __pyx_k_SOLPSVectorFunction2D, sizeof(__pyx_k_SOLPSVectorFunction2D), 0, 0, 1, 1},
  {&__pyx_n_s_SOLPSVectorFunction2D___getstate, __pyx_k_SOLPSVectorFunction2D___getstate, sizeof(__pyx_k_SOLPSVectorFunction2D___getstate), 0, 0, 1, 1},
  {&__pyx_n_s_SOLPSVectorFunction2D___reduce, __pyx_k_SOLPSVectorFunction2D___reduce, sizeof(__pyx_k_SOLPSVectorFunction2D___reduce), 0, 0, 1, 1},
  {&__pyx_n_s_SOLPSVectorFunction2D___setstate, __pyx_k_SOLPSVectorFunction2D___setstate, sizeof(__pyx_k_SOLPSVectorFunction2D___setstate), 0, 0, 1, 1},
  {&__pyx_n_s_SOLPSVectorFunction2D_instance, __pyx_k_SOLPSVectorFunction2D_instance, sizeof(__pyx_k_SOLPSVectorFunction2D_instance), 0, 0, 1, 1},
  {&__pyx_kp_s_Step_may_not_be_zero_axis_d, __pyx_k_Step_may_not_be_zero_axis_d, sizeof(__pyx_k_Step_may_not_be_zero_axis_d), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
  {&__pyx_n_s__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 1, 1},
  {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
  {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_kp_u_and, __pyx_k_and, sizeof(__pyx_k_and), 0, 1, 0, 0},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_kp_s_cherab_solps_solps_2d_functions, __pyx_k_cherab_solps_solps_2d_functions, sizeof(__pyx_k_cherab_solps_solps_2d_functions), 0, 0, 1, 0},
  {&__pyx_n_s_cherab_solps_solps_2d_functions_2, __pyx_k_cherab_solps_solps_2d_functions_2, sizeof(__pyx_k_cherab_solps_solps_2d_functions_2), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_u_got, __pyx_k_got, sizeof(__pyx_k_got), 0, 1, 0, 0},
  {&__pyx_kp_u_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 1, 0, 0},
  {&__pyx_n_s_grid_data, __pyx_k_grid_data, sizeof(__pyx_k_grid_data), 0, 0, 1, 1},
  {&__pyx_n_s_grid_vectors, __pyx_k_grid_vectors, sizeof(__pyx_k_grid_vectors), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
  {&__pyx_n_s_instance, __pyx_k_instance, sizeof(__pyx_k_instance), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
  {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_triangle_to_grid_map, __pyx_k_triangle_to_grid_map, sizeof(__pyx_k_triangle_to_grid_map), 0, 0, 1, 1},
  {&__pyx_n_s_triangles, __pyx_k_triangles, sizeof(__pyx_k_triangles), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_vertex_coords, __pyx_k_vertex_coords, sizeof(__pyx_k_vertex_coords), 0, 0, 1, 1},
  #endif
  {0, 0, 0, 0, 0, 0, 0}
};
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 966, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(16, 136, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(16, 151, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(16, 154, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(16, 2, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(16, 249, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(16, 415, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(16, 627, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(16, 731, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(16, 938, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":966
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "../tmp/build-env-_150tatp/lib/python3.9/site-packages/numpy/__init__.cython-30.pxd":972
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "View.MemoryView":591
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__6 = PyTuple_New(1); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(16, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_tuple__6, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "View.MemoryView":688
 *     tup = <tuple>index if isinstance(index, tuple) else (index,)
 * 
 *     result = [slice(None)] * ndim             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_slice__7 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__7)) __PYX_ERR(16, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__7);
  __Pyx_GIVEREF(__pyx_slice__7);

  /* "cherab/solps/solps_2d_functions.pyx":54
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self._grid_data, self._triangle_to_grid_map, self._kdtree
 * 
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cherab_solps_solps_2d_functions, __pyx_n_s_getstate, 54, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 54, __pyx_L1_error)

  /* "cherab/solps/solps_2d_functions.pyx":57
 *         return self._grid_data, self._triangle_to_grid_map, self._kdtree
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._grid_data, self._triangle_to_grid_map, self._kdtree = state
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 */
  __pyx_tuple__12 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_state); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cherab_solps_solps_2d_functions, __pyx_n_s_setstate, 57, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 57, __pyx_L1_error)

  /* "cherab/solps/solps_2d_functions.pyx":62
 *         self._grid_data_mv = self._grid_data
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cherab_solps_solps_2d_functions, __pyx_n_s_reduce, 62, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(0, 62, __pyx_L1_error)

  /* "cherab/solps/solps_2d_functions.pyx":65
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def instance(cls, SOLPSFunction2D instance not None, object grid_data=None):
 *         """
 */
  __pyx_tuple__16 = PyTuple_Pack(4, __pyx_n_s_cls, __pyx_n_s_instance, __pyx_n_s_grid_data, __pyx_n_s_m); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cherab_solps_solps_2d_functions, __pyx_n_s_instance, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 65, __pyx_L1_error)

  /* "cherab/solps/solps_2d_functions.pyx":138
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self._grid_vectors, self._triangle_to_grid_map, self._kdtree
 * 
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cherab_solps_solps_2d_functions, __pyx_n_s_getstate, 138, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 138, __pyx_L1_error)

  /* "cherab/solps/solps_2d_functions.pyx":141
 *         return self._grid_vectors, self._triangle_to_grid_map, self._kdtree
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._grid_vectors, self._triangle_to_grid_map, self._kdtree = state
 *         self._grid_vectors_mv = self._grid_vectors
 */
  __pyx_tuple__20 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_state); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cherab_solps_solps_2d_functions, __pyx_n_s_setstate, 141, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 141, __pyx_L1_error)

  /* "cherab/solps/solps_2d_functions.pyx":146
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cherab_solps_solps_2d_functions, __pyx_n_s_reduce, 146, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 146, __pyx_L1_error)

  /* "cherab/solps/solps_2d_functions.pyx":149
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def instance(cls, SOLPSVectorFunction2D instance not None, object grid_vectors=None):
 *         """
 */
  __pyx_tuple__24 = PyTuple_Pack(4, __pyx_n_s_cls, __pyx_n_s_instance, __pyx_n_s_grid_vectors, __pyx_n_s_m); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cherab_solps_solps_2d_functions, __pyx_n_s_instance, 149, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 149, __pyx_L1_error)

  /* "View.MemoryView":297
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(16, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "View.MemoryView":298
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(16, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "View.MemoryView":299
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(16, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "View.MemoryView":302
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(16, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "View.MemoryView":303
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(16, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__31 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(16, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(16, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* NumpyImportArray.init */
  /*
 * Cython has automatically inserted a call to _import_array since
 * you didn't include one when you cimported numpy. To disable this
 * add the line
 *   <void>numpy._import_array
 */
#ifdef NPY_NDARRAYOBJECT_H
#if !NO_IMPORT_ARRAY
if (unlikely(_import_array() == -1)) {
    PyErr_SetString(PyExc_ImportError, "numpy.core.multiarray failed to import "
    "(auto-generated because you didn't call 'numpy.import_array()' after cimporting numpy; "
    "use '<void>numpy._import_array' to disable if you are certain you don't need it).");
    __PYX_ERR(0, 22, __pyx_L1_error);
}
#endif
#endif

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  #if CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_InitString(__pyx_string_tab[0], &__pyx_n_s_ASCII) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[1], &__pyx_kp_s_All_dimensions_preceding_dimensi) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[2], &__pyx_n_s_AssertionError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[3], &__pyx_kp_s_Buffer_view_does_not_expose_stri) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[4], &__pyx_kp_s_Can_only_create_a_buffer_that_is) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[5], &__pyx_kp_s_Cannot_assign_to_read_only_memor) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[6], &__pyx_kp_s_Cannot_create_writable_memory_vi) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[7], &__pyx_kp_u_Cannot_index_with_type) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[8], &__pyx_kp_s_Cannot_transpose_memoryview_with) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[9], &__pyx_kp_s_Dimension_d_is_not_direct) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[10], &__pyx_n_s_Ellipsis) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[11], &__pyx_kp_s_Empty_shape_tuple_for_cython_arr) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[12], &__pyx_n_s_ImportError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[13], &__pyx_kp_s_Incompatible_checksums_s_vs_0x6a) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[14], &__pyx_n_s_IndexError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[15], &__pyx_kp_s_Index_out_of_bounds_axis_d) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[16], &__pyx_kp_s_Indirect_dimensions_not_supporte) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[17], &__pyx_kp_u_Invalid_mode_expected_c_or_fortr) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[18], &__pyx_kp_u_Invalid_shape_in_axis) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[19], &__pyx_n_s_MemoryError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[20], &__pyx_kp_s_MemoryView_of_r_at_0x_x) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[21], &__pyx_kp_s_MemoryView_of_r_object) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[22], &__pyx_n_b_O) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[23], &__pyx_kp_u_Out_of_bounds_on_buffer_access_a) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[24], &__pyx_n_s_PickleError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[25], &__pyx_n_s_SOLPSFunction2D) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[26], &__pyx_n_s_SOLPSFunction2D___getstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[27], &__pyx_n_s_SOLPSFunction2D___reduce) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[28], &__pyx_n_s_SOLPSFunction2D___setstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[29], &__pyx_n_s_SOLPSFunction2D_instance) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[30], &__pyx_n_s_SOLPSVectorFunction2D) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[31], &__pyx_n_s_SOLPSVectorFunction2D___getstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[32], &__pyx_n_s_SOLPSVectorFunction2D___reduce) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[33], &__pyx_n_s_SOLPSVectorFunction2D___setstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[34], &__pyx_n_s_SOLPSVectorFunction2D_instance) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[35], &__pyx_kp_s_Step_may_not_be_zero_axis_d) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[36], &__pyx_n_s_TypeError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[37], &__pyx_kp_s_Unable_to_convert_item_to_object) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[38], &__pyx_n_s_ValueError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[39], &__pyx_n_s_View_MemoryView) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[40], &__pyx_kp_u__3) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[41], &__pyx_kp_u__4) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[42], &__pyx_n_s__5) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[43], &__pyx_kp_u__8) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[44], &__pyx_kp_u__9) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[45], &__pyx_n_s_allocate_buffer) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[46], &__pyx_kp_u_and) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[47], &__pyx_n_s_array) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[48], &__pyx_n_s_base) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[49], &__pyx_n_s_c) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[50], &__pyx_n_u_c) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[51], &__pyx_kp_s_cherab_solps_solps_2d_functions) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[52], &__pyx_n_s_cherab_solps_solps_2d_functions_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[53], &__pyx_n_s_class) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[54], &__pyx_n_s_class_getitem) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[55], &__pyx_n_s_cline_in_traceback) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[56], &__pyx_n_s_cls) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[57], &__pyx_kp_s_contiguous_and_direct) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[58], &__pyx_kp_s_contiguous_and_indirect) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[59], &__pyx_n_s_dict) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[60], &__pyx_n_s_dtype) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[61], &__pyx_n_s_dtype_is_object) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[62], &__pyx_n_s_encode) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[63], &__pyx_n_s_enumerate) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[64], &__pyx_n_s_error) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[65], &__pyx_n_s_flags) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[66], &__pyx_n_s_float64) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[67], &__pyx_n_s_format) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[68], &__pyx_n_s_fortran) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[69], &__pyx_n_u_fortran) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[70], &__pyx_n_s_getstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[71], &__pyx_kp_u_got) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[72], &__pyx_kp_u_got_differing_extents_in_dimensi) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[73], &__pyx_n_s_grid_data) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[74], &__pyx_n_s_grid_vectors) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[75], &__pyx_n_s_id) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[76], &__pyx_n_s_import) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[77], &__pyx_n_s_initializing) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[78], &__pyx_n_s_instance) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[79], &__pyx_n_s_int32) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[80], &__pyx_n_s_itemsize) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[81], &__pyx_kp_s_itemsize_0_for_cython_array) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[82], &__pyx_n_s_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[83], &__pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[84], &__pyx_n_s_memview) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[85], &__pyx_n_s_mode) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[86], &__pyx_n_s_name) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[87], &__pyx_n_s_name_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[88], &__pyx_n_s_ndim) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[89], &__pyx_n_s_new) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[90], &__pyx_kp_s_no_default___reduce___due_to_non) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[91], &__pyx_n_s_np) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[92], &__pyx_n_s_numpy) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[93], &__pyx_kp_u_numpy_core_multiarray_failed_to) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[94], &__pyx_kp_u_numpy_core_umath_failed_to_impor) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[95], &__pyx_n_s_obj) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[96], &__pyx_n_s_pack) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[97], &__pyx_n_s_pickle) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[98], &__pyx_n_s_pyx_PickleError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[99], &__pyx_n_s_pyx_checksum) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[100], &__pyx_n_s_pyx_getbuffer) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[101], &__pyx_n_s_pyx_result) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[102], &__pyx_n_s_pyx_state) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[103], &__pyx_n_s_pyx_type) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[104], &__pyx_n_s_pyx_unpickle_Enum) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[105], &__pyx_n_s_pyx_vtable) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[106], &__pyx_n_s_range) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[107], &__pyx_n_s_reduce) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[108], &__pyx_n_s_reduce_cython) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[109], &__pyx_n_s_reduce_ex) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[110], &__pyx_n_s_self) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[111], &__pyx_n_s_setstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[112], &__pyx_n_s_setstate_cython) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[113], &__pyx_n_s_shape) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[114], &__pyx_n_s_size) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[115], &__pyx_n_s_spec) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[116], &__pyx_n_s_start) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[117], &__pyx_n_s_state) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[118], &__pyx_n_s_step) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[119], &__pyx_n_s_stop) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[120], &__pyx_kp_s_strided_and_direct) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[121], &__pyx_kp_s_strided_and_direct_or_indirect) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[122], &__pyx_kp_s_strided_and_indirect) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[123], &__pyx_kp_s_stringsource) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[124], &__pyx_n_s_struct) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[125], &__pyx_n_s_test) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[126], &__pyx_n_s_triangle_to_grid_map) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[127], &__pyx_n_s_triangles) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[128], &__pyx_kp_s_unable_to_allocate_array_data) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[129], &__pyx_kp_s_unable_to_allocate_shape_and_str) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[130], &__pyx_n_s_unpack) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[131], &__pyx_n_s_update) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[132], &__pyx_n_s_vertex_coords) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  #endif
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_112105877 = PyInt_FromLong(112105877L); if (unlikely(!__pyx_int_112105877)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function2d.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "Function2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D = &__pyx_vtable_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D;
  __pyx_vtable_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D.__pyx_base = *__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D;
  __pyx_vtable_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D.__pyx_base.evaluate = (double (*)(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *, double, double))__pyx_f_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_evaluate;
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D = PyType_FromSpecWithBases(&__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D_spec, __pyx_t_2); if (unlikely(!__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  #else
  __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D.tp_base = __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D;
  if (PyType_Ready(&__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D.tp_dictoffset && __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_SetVtable(__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D, __pyx_vtabptr_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  #else
  if (__Pyx_SetVtable(__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D.tp_dict, __pyx_vtabptr_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SOLPSFunction2D, __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  #else
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SOLPSFunction2D, (PyObject *)&__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D = (PyTypeObject *)__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D;
  #else
  __pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D = &__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D;
  #endif
  __pyx_t_2 = PyImport_ImportModule("raysect.core.math.function.vector3d.function2d.base"); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D = __Pyx_ImportType(__pyx_t_2, "raysect.core.math.function.vector3d.function2d.base", "Function2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D = &__pyx_vtable_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D;
  __pyx_vtable_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D.__pyx_base = *__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D;
  __pyx_vtable_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D.__pyx_base.evaluate = (struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *(*)(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *, double, double))__pyx_f_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_evaluate;
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_t_3 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D = PyType_FromSpecWithBases(&__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D_spec, __pyx_t_3); if (unlikely(!__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  #else
  __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D.tp_base = __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D;
  if (PyType_Ready(&__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D.tp_dictoffset && __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_SetVtable(__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D, __pyx_vtabptr_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  #else
  if (__Pyx_SetVtable(__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D.tp_dict, __pyx_vtabptr_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SOLPSVectorFunction2D, __pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  #else
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SOLPSVectorFunction2D, (PyObject *)&__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D = (PyTypeObject *)__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D;
  #else
  __pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D = &__pyx_type_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D;
  #endif
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_type___pyx_array = PyType_FromSpec(&__pyx_type___pyx_array_spec); if (unlikely(!__pyx_type___pyx_array)) __PYX_ERR(16, 109, __pyx_L1_error)
  #else
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(16, 109, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_SetVtable(__pyx_type___pyx_array, __pyx_vtabptr_array) < 0) __PYX_ERR(16, 109, __pyx_L1_error)
  #else
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(16, 109, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(16, 109, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_array_type = (PyTypeObject *)__pyx_type___pyx_array;
  #else
  __pyx_array_type = &__pyx_type___pyx_array;
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_type___pyx_MemviewEnum = PyType_FromSpec(&__pyx_type___pyx_MemviewEnum_spec); if (unlikely(!__pyx_type___pyx_MemviewEnum)) __PYX_ERR(16, 290, __pyx_L1_error)
  #else
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(16, 290, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(16, 290, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_MemviewEnum_type = (PyTypeObject *)__pyx_type___pyx_MemviewEnum;
  #else
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  #endif
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  __pyx_vtable_memoryview._get_base = (PyObject *(*)(struct __pyx_memoryview_obj *))__pyx_memoryview__get_base;
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_type___pyx_memoryview = PyType_FromSpec(&__pyx_type___pyx_memoryview_spec); if (unlikely(!__pyx_type___pyx_memoryview)) __PYX_ERR(16, 341, __pyx_L1_error)
  #else
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(16, 341, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(16, 341, __pyx_L1_error)
  #else
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(16, 341, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(16, 341, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_memoryview_type = (PyTypeObject *)__pyx_type___pyx_memoryview;
  #else
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  #endif
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_vtable__memoryviewslice.__pyx_base._get_base = (PyObject *(*)(struct __pyx_memoryview_obj *))__pyx_memoryviewslice__get_base;
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_t_6 = PyTuple_Pack(1, (PyObject *)__pyx_memoryview_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(16, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_type___pyx_memoryviewslice = PyType_FromSpecWithBases(&__pyx_type___pyx_memoryviewslice_spec, __pyx_t_6); if (unlikely(!__pyx_type___pyx_memoryviewslice)) __PYX_ERR(16, 975, __pyx_L1_error)
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  #else
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(16, 975, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(16, 975, __pyx_L1_error)
  #else
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(16, 975, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(16, 975, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_memoryviewslice_type = (PyTypeObject *)__pyx_type___pyx_memoryviewslice;
  #else
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  #endif
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(17, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 203, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 226, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 230, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 239, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 846, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math._vec3"); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_5_vec3__Vec3 = __Pyx_ImportType(__pyx_t_1, "raysect.core.math._vec3", "_Vec3", sizeof(struct __pyx_obj_7raysect_4core_4math_5_vec3__Vec3), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_5_vec3__Vec3) __PYX_ERR(18, 32, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_5_vec3__Vec3 = (struct __pyx_vtabstruct_7raysect_4core_4math_5_vec3__Vec3*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_5_vec3__Vec3); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_5_vec3__Vec3)) __PYX_ERR(18, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math._mat4"); if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_5_mat4__Mat4 = __Pyx_ImportType(__pyx_t_1, "raysect.core.math._mat4", "_Mat4", sizeof(struct __pyx_obj_7raysect_4core_4math_5_mat4__Mat4), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_5_mat4__Mat4) __PYX_ERR(19, 32, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_5_mat4__Mat4 = (struct __pyx_vtabstruct_7raysect_4core_4math_5_mat4__Mat4*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_5_mat4__Mat4); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_5_mat4__Mat4)) __PYX_ERR(19, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.affinematrix"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_12affinematrix_AffineMatrix3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.affinematrix", "AffineMatrix3D", sizeof(struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_12affinematrix_AffineMatrix3D) __PYX_ERR(2, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_12affinematrix_AffineMatrix3D = (struct __pyx_vtabstruct_7raysect_4core_4math_12affinematrix_AffineMatrix3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_12affinematrix_AffineMatrix3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_12affinematrix_AffineMatrix3D)) __PYX_ERR(2, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.vector"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_6vector_Vector3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.vector", "Vector3D", sizeof(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_6vector_Vector3D) __PYX_ERR(3, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_6vector_Vector3D = (struct __pyx_vtabstruct_7raysect_4core_4math_6vector_Vector3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_6vector_Vector3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_6vector_Vector3D)) __PYX_ERR(3, 35, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_6vector_Vector2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.vector", "Vector2D", sizeof(struct __pyx_obj_7raysect_4core_4math_6vector_Vector2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_6vector_Vector2D) __PYX_ERR(3, 78, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_6vector_Vector2D = (struct __pyx_vtabstruct_7raysect_4core_4math_6vector_Vector2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_6vector_Vector2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_6vector_Vector2D)) __PYX_ERR(3, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.point"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_5point_Point3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.point", "Point3D", sizeof(struct __pyx_obj_7raysect_4core_4math_5point_Point3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_5point_Point3D) __PYX_ERR(4, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_5point_Point3D = (struct __pyx_vtabstruct_7raysect_4core_4math_5point_Point3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_5point_Point3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_5point_Point3D)) __PYX_ERR(4, 36, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_5point_Point2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.point", "Point2D", sizeof(struct __pyx_obj_7raysect_4core_4math_5point_Point2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_5point_Point2D) __PYX_ERR(4, 73, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_5point_Point2D = (struct __pyx_vtabstruct_7raysect_4core_4math_5point_Point2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_5point_Point2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_5point_Point2D)) __PYX_ERR(4, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.normal"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_6normal_Normal3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.normal", "Normal3D", sizeof(struct __pyx_obj_7raysect_4core_4math_6normal_Normal3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_6normal_Normal3D) __PYX_ERR(5, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_6normal_Normal3D = (struct __pyx_vtabstruct_7raysect_4core_4math_6normal_Normal3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_6normal_Normal3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_6normal_Normal3D)) __PYX_ERR(5, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.quaternion"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_10quaternion_Quaternion = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.quaternion", "Quaternion", sizeof(struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_10quaternion_Quaternion) __PYX_ERR(6, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_10quaternion_Quaternion = (struct __pyx_vtabstruct_7raysect_4core_4math_10quaternion_Quaternion*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_10quaternion_Quaternion); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_10quaternion_Quaternion)) __PYX_ERR(6, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.statsarray"); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_10statsarray_StatsBin = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.statsarray", "StatsBin", sizeof(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsBin), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_10statsarray_StatsBin) __PYX_ERR(20, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsBin = (struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsBin*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_10statsarray_StatsBin); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsBin)) __PYX_ERR(20, 35, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.statsarray", "StatsArray1D", sizeof(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray1D) __PYX_ERR(20, 53, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsArray1D = (struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsArray1D)) __PYX_ERR(20, 53, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.statsarray", "StatsArray2D", sizeof(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray2D) __PYX_ERR(20, 81, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsArray2D = (struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsArray2D)) __PYX_ERR(20, 81, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.statsarray", "StatsArray3D", sizeof(struct __pyx_obj_7raysect_4core_4math_10statsarray_StatsArray3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray3D) __PYX_ERR(20, 109, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsArray3D = (struct __pyx_vtabstruct_7raysect_4core_4math_10statsarray_StatsArray3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_10statsarray_StatsArray3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_10statsarray_StatsArray3D)) __PYX_ERR(20, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.sampler.solidangle"); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.solidangle", "SolidAngleSampler", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler) __PYX_ERR(21, 41, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_SolidAngleSampler)) __PYX_ERR(21, 41, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SphereSampler = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.solidangle", "SphereSampler", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_SphereSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SphereSampler) __PYX_ERR(21, 54, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_SphereSampler = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_SphereSampler*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_SphereSampler); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_SphereSampler)) __PYX_ERR(21, 54, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.solidangle", "HemisphereUniformSampler", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler) __PYX_ERR(21, 58, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_HemisphereUniformSampler)) __PYX_ERR(21, 58, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.solidangle", "HemisphereCosineSampler", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler) __PYX_ERR(21, 62, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_HemisphereCosineSampler)) __PYX_ERR(21, 62, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.solidangle", "ConeUniformSampler", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler) __PYX_ERR(21, 66, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_10solidangle_ConeUniformSampler)) __PYX_ERR(21, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.sampler.surface3d"); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.surface3d", "SurfaceSampler3D", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D) __PYX_ERR(22, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_SurfaceSampler3D)) __PYX_ERR(22, 35, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.surface3d", "DiskSampler3D", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D) __PYX_ERR(22, 46, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_DiskSampler3D)) __PYX_ERR(22, 46, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.surface3d", "RectangleSampler3D", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D) __PYX_ERR(22, 53, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_RectangleSampler3D)) __PYX_ERR(22, 53, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.surface3d", "TriangleSampler3D", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D) __PYX_ERR(22, 60, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_9surface3d_TriangleSampler3D)) __PYX_ERR(22, 60, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.sampler.targetted"); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_7sampler_9targetted__TargettedSampler = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.targetted", "_TargettedSampler", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted__TargettedSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_9targetted__TargettedSampler) __PYX_ERR(23, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_9targetted__TargettedSampler = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted__TargettedSampler*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_9targetted__TargettedSampler); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_9targetted__TargettedSampler)) __PYX_ERR(23, 35, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.targetted", "TargettedHemisphereSampler", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler) __PYX_ERR(23, 60, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_9targetted_TargettedHemisphereSampler)) __PYX_ERR(23, 60, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.sampler.targetted", "TargettedSphereSampler", sizeof(struct __pyx_obj_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler) __PYX_ERR(23, 63, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler = (struct __pyx_vtabstruct_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7sampler_9targetted_TargettedSphereSampler)) __PYX_ERR(23, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.ray"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_3ray_Ray = __Pyx_ImportType(__pyx_t_1, "raysect.core.ray", "Ray", sizeof(struct __pyx_obj_7raysect_4core_3ray_Ray), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_3ray_Ray) __PYX_ERR(7, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_3ray_Ray = (struct __pyx_vtabstruct_7raysect_4core_3ray_Ray*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_3ray_Ray); if (unlikely(!__pyx_vtabptr_7raysect_4core_3ray_Ray)) __PYX_ERR(7, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.material"); if (unlikely(!__pyx_t_1)) __PYX_ERR(24, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_8material_Material = __Pyx_ImportType(__pyx_t_1, "raysect.core.material", "Material", sizeof(struct __pyx_obj_7raysect_4core_8material_Material), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_8material_Material) __PYX_ERR(24, 32, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_8material_Material = (struct __pyx_vtabstruct_7raysect_4core_8material_Material*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_8material_Material); if (unlikely(!__pyx_vtabptr_7raysect_4core_8material_Material)) __PYX_ERR(24, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.scenegraph._nodebase"); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_10scenegraph_9_nodebase__NodeBase = __Pyx_ImportType(__pyx_t_1, "raysect.core.scenegraph._nodebase", "_NodeBase", sizeof(struct __pyx_obj_7raysect_4core_10scenegraph_9_nodebase__NodeBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_10scenegraph_9_nodebase__NodeBase) __PYX_ERR(25, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.scenegraph.node"); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_10scenegraph_4node_Node = __Pyx_ImportType(__pyx_t_1, "raysect.core.scenegraph.node", "Node", sizeof(struct __pyx_obj_7raysect_4core_10scenegraph_4node_Node), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_10scenegraph_4node_Node) __PYX_ERR(26, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_10scenegraph_4node_Node = (struct __pyx_vtabstruct_7raysect_4core_10scenegraph_4node_Node*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_10scenegraph_4node_Node); if (unlikely(!__pyx_vtabptr_7raysect_4core_10scenegraph_4node_Node)) __PYX_ERR(26, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.scenegraph.signal"); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_10scenegraph_6signal_ChangeSignal = __Pyx_ImportType(__pyx_t_1, "raysect.core.scenegraph.signal", "ChangeSignal", sizeof(struct __pyx_obj_7raysect_4core_10scenegraph_6signal_ChangeSignal), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_10scenegraph_6signal_ChangeSignal) __PYX_ERR(27, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.scenegraph.primitive"); if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_10scenegraph_9primitive_Primitive = __Pyx_ImportType(__pyx_t_1, "raysect.core.scenegraph.primitive", "Primitive", sizeof(struct __pyx_obj_7raysect_4core_10scenegraph_9primitive_Primitive), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_10scenegraph_9primitive_Primitive) __PYX_ERR(28, 42, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_10scenegraph_9primitive_Primitive = (struct __pyx_vtabstruct_7raysect_4core_10scenegraph_9primitive_Primitive*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_10scenegraph_9primitive_Primitive); if (unlikely(!__pyx_vtabptr_7raysect_4core_10scenegraph_9primitive_Primitive)) __PYX_ERR(28, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.intersection"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_12intersection_Intersection = __Pyx_ImportType(__pyx_t_1, "raysect.core.intersection", "Intersection", sizeof(struct __pyx_obj_7raysect_4core_12intersection_Intersection), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_12intersection_Intersection) __PYX_ERR(8, 37, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_12intersection_Intersection = (struct __pyx_vtabstruct_7raysect_4core_12intersection_Intersection*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_12intersection_Intersection); if (unlikely(!__pyx_vtabptr_7raysect_4core_12intersection_Intersection)) __PYX_ERR(8, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.acceleration.accelerator"); if (unlikely(!__pyx_t_1)) __PYX_ERR(29, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_12acceleration_11accelerator_Accelerator = __Pyx_ImportType(__pyx_t_1, "raysect.core.acceleration.accelerator", "Accelerator", sizeof(struct __pyx_obj_7raysect_4core_12acceleration_11accelerator_Accelerator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_12acceleration_11accelerator_Accelerator) __PYX_ERR(29, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_12acceleration_11accelerator_Accelerator = (struct __pyx_vtabstruct_7raysect_4core_12acceleration_11accelerator_Accelerator*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_12acceleration_11accelerator_Accelerator); if (unlikely(!__pyx_vtabptr_7raysect_4core_12acceleration_11accelerator_Accelerator)) __PYX_ERR(29, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.scenegraph.world"); if (unlikely(!__pyx_t_1)) __PYX_ERR(30, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_10scenegraph_5world_World = __Pyx_ImportType(__pyx_t_1, "raysect.core.scenegraph.world", "World", sizeof(struct __pyx_obj_7raysect_4core_10scenegraph_5world_World), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_10scenegraph_5world_World) __PYX_ERR(30, 38, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_10scenegraph_5world_World = (struct __pyx_vtabstruct_7raysect_4core_10scenegraph_5world_World*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_10scenegraph_5world_World); if (unlikely(!__pyx_vtabptr_7raysect_4core_10scenegraph_5world_World)) __PYX_ERR(30, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.scenegraph.observer"); if (unlikely(!__pyx_t_1)) __PYX_ERR(31, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_10scenegraph_8observer_Observer = __Pyx_ImportType(__pyx_t_1, "raysect.core.scenegraph.observer", "Observer", sizeof(struct __pyx_obj_7raysect_4core_10scenegraph_8observer_Observer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_10scenegraph_8observer_Observer) __PYX_ERR(31, 34, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_10scenegraph_8observer_Observer = (struct __pyx_vtabstruct_7raysect_4core_10scenegraph_8observer_Observer*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_10scenegraph_8observer_Observer); if (unlikely(!__pyx_vtabptr_7raysect_4core_10scenegraph_8observer_Observer)) __PYX_ERR(31, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.scenegraph.utility"); if (unlikely(!__pyx_t_1)) __PYX_ERR(32, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_10scenegraph_7utility_BridgeNode = __Pyx_ImportType(__pyx_t_1, "raysect.core.scenegraph.utility", "BridgeNode", sizeof(struct __pyx_obj_7raysect_4core_10scenegraph_7utility_BridgeNode), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_10scenegraph_7utility_BridgeNode) __PYX_ERR(32, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_10scenegraph_7utility_BridgeNode = (struct __pyx_vtabstruct_7raysect_4core_10scenegraph_7utility_BridgeNode*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_10scenegraph_7utility_BridgeNode); if (unlikely(!__pyx_vtabptr_7raysect_4core_10scenegraph_7utility_BridgeNode)) __PYX_ERR(32, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.containers"); if (unlikely(!__pyx_t_1)) __PYX_ERR(33, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_10containers__Item = __Pyx_ImportType(__pyx_t_1, "raysect.core.containers", "_Item", sizeof(struct __pyx_obj_7raysect_4core_10containers__Item), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_10containers__Item) __PYX_ERR(33, 33, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_10containers_LinkedList = __Pyx_ImportType(__pyx_t_1, "raysect.core.containers", "LinkedList", sizeof(struct __pyx_obj_7raysect_4core_10containers_LinkedList), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_10containers_LinkedList) __PYX_ERR(33, 41, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_10containers_LinkedList = (struct __pyx_vtabstruct_7raysect_4core_10containers_LinkedList*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_10containers_LinkedList); if (unlikely(!__pyx_vtabptr_7raysect_4core_10containers_LinkedList)) __PYX_ERR(33, 41, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_10containers_Stack = __Pyx_ImportType(__pyx_t_1, "raysect.core.containers", "Stack", sizeof(struct __pyx_obj_7raysect_4core_10containers_Stack), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_10containers_Stack) __PYX_ERR(33, 61, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_10containers_Stack = (struct __pyx_vtabstruct_7raysect_4core_10containers_Stack*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_10containers_Stack); if (unlikely(!__pyx_vtabptr_7raysect_4core_10containers_Stack)) __PYX_ERR(33, 61, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_10containers_Queue = __Pyx_ImportType(__pyx_t_1, "raysect.core.containers", "Queue", sizeof(struct __pyx_obj_7raysect_4core_10containers_Queue), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_10containers_Queue) __PYX_ERR(33, 68, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_10containers_Queue = (struct __pyx_vtabstruct_7raysect_4core_10containers_Queue*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_10containers_Queue); if (unlikely(!__pyx_vtabptr_7raysect_4core_10containers_Queue)) __PYX_ERR(33, 68, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.boundingsphere"); if (unlikely(!__pyx_t_1)) __PYX_ERR(34, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_14boundingsphere_BoundingSphere3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.boundingsphere", "BoundingSphere3D", sizeof(struct __pyx_obj_7raysect_4core_14boundingsphere_BoundingSphere3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_14boundingsphere_BoundingSphere3D) __PYX_ERR(34, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_14boundingsphere_BoundingSphere3D = (struct __pyx_vtabstruct_7raysect_4core_14boundingsphere_BoundingSphere3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_14boundingsphere_BoundingSphere3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_14boundingsphere_BoundingSphere3D)) __PYX_ERR(34, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.boundingbox"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_11boundingbox_BoundingBox3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.boundingbox", "BoundingBox3D", sizeof(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox3D) __PYX_ERR(9, 37, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_11boundingbox_BoundingBox3D = (struct __pyx_vtabstruct_7raysect_4core_11boundingbox_BoundingBox3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_11boundingbox_BoundingBox3D)) __PYX_ERR(9, 37, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_11boundingbox_BoundingBox2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.boundingbox", "BoundingBox2D", sizeof(struct __pyx_obj_7raysect_4core_11boundingbox_BoundingBox2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox2D) __PYX_ERR(9, 92, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_11boundingbox_BoundingBox2D = (struct __pyx_vtabstruct_7raysect_4core_11boundingbox_BoundingBox2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_11boundingbox_BoundingBox2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_11boundingbox_BoundingBox2D)) __PYX_ERR(9, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.spatial.kdtree2d"); if (unlikely(!__pyx_t_1)) __PYX_ERR(35, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_Item2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.spatial.kdtree2d", "Item2D", sizeof(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_Item2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_Item2D) __PYX_ERR(35, 52, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.spatial.kdtree2d", "KDTree2DCore", sizeof(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore) __PYX_ERR(35, 59, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore = (struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2DCore)) __PYX_ERR(35, 59, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.spatial.kdtree2d", "KDTree2D", sizeof(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D) __PYX_ERR(35, 116, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D = (struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree2d_KDTree2D)) __PYX_ERR(35, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function2d.interpolate.common"); if (unlikely(!__pyx_t_1)) __PYX_ERR(36, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.common", "MeshKDTree2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D) __PYX_ERR(36, 37, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_6common_MeshKDTree2D)) __PYX_ERR(36, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(37, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_4base_Function = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.base", "Function", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_4base_Function), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_4base_Function) __PYX_ERR(37, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(38, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_4base_FloatFunction = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.base", "FloatFunction", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_4base_FloatFunction), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_4base_FloatFunction) __PYX_ERR(38, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function1d.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "Function1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D) __PYX_ERR(10, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D)) __PYX_ERR(10, 36, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "AddFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D) __PYX_ERR(10, 40, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_AddFunction1D)) __PYX_ERR(10, 40, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "SubtractFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D) __PYX_ERR(10, 44, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractFunction1D)) __PYX_ERR(10, 44, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "MultiplyFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D) __PYX_ERR(10, 48, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyFunction1D)) __PYX_ERR(10, 48, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "DivideFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D) __PYX_ERR(10, 52, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_DivideFunction1D)) __PYX_ERR(10, 52, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "PowFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D) __PYX_ERR(10, 56, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunction1D)) __PYX_ERR(10, 56, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "ModuloFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D) __PYX_ERR(10, 60, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunction1D)) __PYX_ERR(10, 60, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "AbsFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D) __PYX_ERR(10, 64, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_AbsFunction1D)) __PYX_ERR(10, 64, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "EqualsFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D) __PYX_ERR(10, 68, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsFunction1D)) __PYX_ERR(10, 68, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "NotEqualsFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D) __PYX_ERR(10, 72, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsFunction1D)) __PYX_ERR(10, 72, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "LessThanFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D) __PYX_ERR(10, 76, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanFunction1D)) __PYX_ERR(10, 76, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "GreaterThanFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D) __PYX_ERR(10, 80, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanFunction1D)) __PYX_ERR(10, 80, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "LessEqualsFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D) __PYX_ERR(10, 84, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsFunction1D)) __PYX_ERR(10, 84, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "GreaterEqualsFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D) __PYX_ERR(10, 88, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsFunction1D)) __PYX_ERR(10, 88, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "AddScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D) __PYX_ERR(10, 92, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_AddScalar1D)) __PYX_ERR(10, 92, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "SubtractScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D) __PYX_ERR(10, 97, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_SubtractScalar1D)) __PYX_ERR(10, 97, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "MultiplyScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D) __PYX_ERR(10, 102, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_MultiplyScalar1D)) __PYX_ERR(10, 102, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "DivideScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D) __PYX_ERR(10, 107, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_DivideScalar1D)) __PYX_ERR(10, 107, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "ModuloScalarFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D) __PYX_ERR(10, 112, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloScalarFunction1D)) __PYX_ERR(10, 112, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "ModuloFunctionScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D) __PYX_ERR(10, 117, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_ModuloFunctionScalar1D)) __PYX_ERR(10, 117, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "PowScalarFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D) __PYX_ERR(10, 122, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_PowScalarFunction1D)) __PYX_ERR(10, 122, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "PowFunctionScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D) __PYX_ERR(10, 127, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_PowFunctionScalar1D)) __PYX_ERR(10, 127, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "EqualsScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D) __PYX_ERR(10, 132, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_EqualsScalar1D)) __PYX_ERR(10, 132, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "NotEqualsScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D) __PYX_ERR(10, 137, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_NotEqualsScalar1D)) __PYX_ERR(10, 137, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "LessThanScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D) __PYX_ERR(10, 142, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessThanScalar1D)) __PYX_ERR(10, 142, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "GreaterThanScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D) __PYX_ERR(10, 147, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterThanScalar1D)) __PYX_ERR(10, 147, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "LessEqualsScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D) __PYX_ERR(10, 152, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_LessEqualsScalar1D)) __PYX_ERR(10, 152, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.base", "GreaterEqualsScalar1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D) __PYX_ERR(10, 157, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_4base_GreaterEqualsScalar1D)) __PYX_ERR(10, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function1d.constant"); if (unlikely(!__pyx_t_1)) __PYX_ERR(39, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.constant", "Constant1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D) __PYX_ERR(39, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_8constant_Constant1D)) __PYX_ERR(39, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function1d.blend"); if (unlikely(!__pyx_t_1)) __PYX_ERR(40, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.blend", "Blend1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D) __PYX_ERR(40, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5blend_Blend1D)) __PYX_ERR(40, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function1d.autowrap"); if (unlikely(!__pyx_t_1)) __PYX_ERR(41, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.autowrap", "PythonFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D) __PYX_ERR(41, 34, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_8autowrap_PythonFunction1D)) __PYX_ERR(41, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function1d.arg"); if (unlikely(!__pyx_t_1)) __PYX_ERR(42, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.arg", "Arg1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D) __PYX_ERR(42, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_3arg_Arg1D)) __PYX_ERR(42, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function1d.cmath"); if (unlikely(!__pyx_t_1)) __PYX_ERR(43, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.cmath", "Exp1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D) __PYX_ERR(43, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Exp1D)) __PYX_ERR(43, 35, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.cmath", "Sin1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D) __PYX_ERR(43, 39, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sin1D)) __PYX_ERR(43, 39, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.cmath", "Cos1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D) __PYX_ERR(43, 43, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Cos1D)) __PYX_ERR(43, 43, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.cmath", "Tan1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D) __PYX_ERR(43, 47, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Tan1D)) __PYX_ERR(43, 47, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.cmath", "Asin1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D) __PYX_ERR(43, 51, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Asin1D)) __PYX_ERR(43, 51, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.cmath", "Acos1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D) __PYX_ERR(43, 55, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Acos1D)) __PYX_ERR(43, 55, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.cmath", "Atan1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D) __PYX_ERR(43, 59, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan1D)) __PYX_ERR(43, 59, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.cmath", "Atan4Q1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D) __PYX_ERR(43, 63, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Atan4Q1D)) __PYX_ERR(43, 63, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.cmath", "Sqrt1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D) __PYX_ERR(43, 67, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Sqrt1D)) __PYX_ERR(43, 67, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.cmath", "Erf1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D) __PYX_ERR(43, 71, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_5cmath_Erf1D)) __PYX_ERR(43, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function1d.interpolate"); if (unlikely(!__pyx_t_1)) __PYX_ERR(44, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.interpolate", "Interpolator1DArray", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray) __PYX_ERR(44, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate_Interpolator1DArray)) __PYX_ERR(44, 36, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.interpolate", "_Interpolator1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D) __PYX_ERR(44, 47, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1D)) __PYX_ERR(44, 47, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.interpolate", "_Interpolator1DLinear", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear) __PYX_ERR(44, 57, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DLinear)) __PYX_ERR(44, 57, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.interpolate", "_Interpolator1DCubic", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic) __PYX_ERR(44, 61, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Interpolator1DCubic)) __PYX_ERR(44, 61, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.interpolate", "_Extrapolator1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D) __PYX_ERR(44, 69, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1D)) __PYX_ERR(44, 69, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.interpolate", "_Extrapolator1DNone", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone) __PYX_ERR(44, 78, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNone)) __PYX_ERR(44, 78, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.interpolate", "_Extrapolator1DNearest", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest) __PYX_ERR(44, 82, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DNearest)) __PYX_ERR(44, 82, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.interpolate", "_Extrapolator1DLinear", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear) __PYX_ERR(44, 86, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DLinear)) __PYX_ERR(44, 86, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.interpolate", "_Extrapolator1DQuadratic", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic) __PYX_ERR(44, 90, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__Extrapolator1DQuadratic)) __PYX_ERR(44, 90, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function1d.interpolate", "_ArrayDerivative1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D) __PYX_ERR(44, 98, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function1d_11interpolate__ArrayDerivative1D)) __PYX_ERR(44, 98, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function2d.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "AddFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D) __PYX_ERR(11, 41, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_AddFunction2D)) __PYX_ERR(11, 41, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "SubtractFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D) __PYX_ERR(11, 45, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractFunction2D)) __PYX_ERR(11, 45, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "MultiplyFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D) __PYX_ERR(11, 49, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyFunction2D)) __PYX_ERR(11, 49, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "DivideFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D) __PYX_ERR(11, 53, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_DivideFunction2D)) __PYX_ERR(11, 53, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "ModuloFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D) __PYX_ERR(11, 57, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunction2D)) __PYX_ERR(11, 57, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "PowFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D) __PYX_ERR(11, 61, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunction2D)) __PYX_ERR(11, 61, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "AbsFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D) __PYX_ERR(11, 65, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_AbsFunction2D)) __PYX_ERR(11, 65, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "EqualsFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D) __PYX_ERR(11, 69, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsFunction2D)) __PYX_ERR(11, 69, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "NotEqualsFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D) __PYX_ERR(11, 73, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsFunction2D)) __PYX_ERR(11, 73, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "LessThanFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D) __PYX_ERR(11, 77, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanFunction2D)) __PYX_ERR(11, 77, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "GreaterThanFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D) __PYX_ERR(11, 81, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanFunction2D)) __PYX_ERR(11, 81, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "LessEqualsFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D) __PYX_ERR(11, 85, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsFunction2D)) __PYX_ERR(11, 85, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "GreaterEqualsFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D) __PYX_ERR(11, 89, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsFunction2D)) __PYX_ERR(11, 89, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "AddScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D) __PYX_ERR(11, 93, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_AddScalar2D)) __PYX_ERR(11, 93, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "SubtractScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D) __PYX_ERR(11, 98, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_SubtractScalar2D)) __PYX_ERR(11, 98, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "MultiplyScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D) __PYX_ERR(11, 103, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_MultiplyScalar2D)) __PYX_ERR(11, 103, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "DivideScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D) __PYX_ERR(11, 108, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_DivideScalar2D)) __PYX_ERR(11, 108, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "ModuloScalarFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D) __PYX_ERR(11, 113, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloScalarFunction2D)) __PYX_ERR(11, 113, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "ModuloFunctionScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D) __PYX_ERR(11, 118, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_ModuloFunctionScalar2D)) __PYX_ERR(11, 118, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "PowScalarFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D) __PYX_ERR(11, 123, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_PowScalarFunction2D)) __PYX_ERR(11, 123, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "PowFunctionScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D) __PYX_ERR(11, 128, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_PowFunctionScalar2D)) __PYX_ERR(11, 128, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "EqualsScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D) __PYX_ERR(11, 133, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_EqualsScalar2D)) __PYX_ERR(11, 133, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "NotEqualsScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D) __PYX_ERR(11, 138, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_NotEqualsScalar2D)) __PYX_ERR(11, 138, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "LessThanScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D) __PYX_ERR(11, 143, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessThanScalar2D)) __PYX_ERR(11, 143, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "GreaterThanScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D) __PYX_ERR(11, 148, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterThanScalar2D)) __PYX_ERR(11, 148, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "LessEqualsScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D) __PYX_ERR(11, 153, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_LessEqualsScalar2D)) __PYX_ERR(11, 153, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.base", "GreaterEqualsScalar2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D) __PYX_ERR(11, 158, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_4base_GreaterEqualsScalar2D)) __PYX_ERR(11, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function2d.constant"); if (unlikely(!__pyx_t_1)) __PYX_ERR(45, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.constant", "Constant2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D) __PYX_ERR(45, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_8constant_Constant2D)) __PYX_ERR(45, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function2d.blend"); if (unlikely(!__pyx_t_1)) __PYX_ERR(46, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.blend", "Blend2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D) __PYX_ERR(46, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5blend_Blend2D)) __PYX_ERR(46, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function2d.autowrap"); if (unlikely(!__pyx_t_1)) __PYX_ERR(47, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.autowrap", "PythonFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D) __PYX_ERR(47, 34, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_8autowrap_PythonFunction2D)) __PYX_ERR(47, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function2d.interpolate.interpolator2dmesh"); if (unlikely(!__pyx_t_1)) __PYX_ERR(48, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.interpolator2dmesh", "Interpolator2DMesh", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh) __PYX_ERR(48, 37, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_18interpolator2dmesh_Interpolator2DMesh)) __PYX_ERR(48, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function2d.interpolate.discrete2dmesh"); if (unlikely(!__pyx_t_1)) __PYX_ERR(49, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.discrete2dmesh", "Discrete2DMesh", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh) __PYX_ERR(49, 37, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_14discrete2dmesh_Discrete2DMesh)) __PYX_ERR(49, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function2d.interpolate.interpolator2darray"); if (unlikely(!__pyx_t_1)) __PYX_ERR(50, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.interpolator2darray", "Interpolator2DArray", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray) __PYX_ERR(50, 42, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray_Interpolator2DArray)) __PYX_ERR(50, 42, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.interpolator2darray", "_Interpolator2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D) __PYX_ERR(50, 56, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2D)) __PYX_ERR(50, 56, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.interpolator2darray", "_Interpolator2DLinear", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear) __PYX_ERR(50, 67, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DLinear)) __PYX_ERR(50, 67, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.interpolator2darray", "_Interpolator2DCubic", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic) __PYX_ERR(50, 75, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Interpolator2DCubic)) __PYX_ERR(50, 75, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.interpolator2darray", "_Extrapolator2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D) __PYX_ERR(50, 85, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2D)) __PYX_ERR(50, 85, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.interpolator2darray", "_Extrapolator2DNone", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone) __PYX_ERR(50, 100, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNone)) __PYX_ERR(50, 100, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.interpolator2darray", "_Extrapolator2DNearest", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest) __PYX_ERR(50, 104, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DNearest)) __PYX_ERR(50, 104, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.interpolator2darray", "_Extrapolator2DLinear", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear) __PYX_ERR(50, 108, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__Extrapolator2DLinear)) __PYX_ERR(50, 108, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.interpolate.interpolator2darray", "_ArrayDerivative2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D) __PYX_ERR(50, 112, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_11interpolate_19interpolator2darray__ArrayDerivative2D)) __PYX_ERR(50, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function2d.arg"); if (unlikely(!__pyx_t_1)) __PYX_ERR(51, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.arg", "Arg2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D) __PYX_ERR(51, 37, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_3arg_Arg2D)) __PYX_ERR(51, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function2d.cmath"); if (unlikely(!__pyx_t_1)) __PYX_ERR(52, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.cmath", "Exp2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D) __PYX_ERR(52, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Exp2D)) __PYX_ERR(52, 35, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.cmath", "Sin2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D) __PYX_ERR(52, 39, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sin2D)) __PYX_ERR(52, 39, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.cmath", "Cos2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D) __PYX_ERR(52, 43, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Cos2D)) __PYX_ERR(52, 43, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.cmath", "Tan2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D) __PYX_ERR(52, 47, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Tan2D)) __PYX_ERR(52, 47, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.cmath", "Asin2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D) __PYX_ERR(52, 51, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Asin2D)) __PYX_ERR(52, 51, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.cmath", "Acos2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D) __PYX_ERR(52, 55, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Acos2D)) __PYX_ERR(52, 55, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.cmath", "Atan2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D) __PYX_ERR(52, 59, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan2D)) __PYX_ERR(52, 59, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.cmath", "Atan4Q2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D) __PYX_ERR(52, 63, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Atan4Q2D)) __PYX_ERR(52, 63, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.cmath", "Sqrt2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D) __PYX_ERR(52, 67, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Sqrt2D)) __PYX_ERR(52, 67, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function2d.cmath", "Erf2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D) __PYX_ERR(52, 71, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function2d_5cmath_Erf2D)) __PYX_ERR(52, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function3d.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "Function3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D) __PYX_ERR(12, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D)) __PYX_ERR(12, 36, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "AddFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D) __PYX_ERR(12, 40, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_AddFunction3D)) __PYX_ERR(12, 40, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "SubtractFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D) __PYX_ERR(12, 44, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractFunction3D)) __PYX_ERR(12, 44, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "MultiplyFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D) __PYX_ERR(12, 48, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyFunction3D)) __PYX_ERR(12, 48, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "DivideFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D) __PYX_ERR(12, 52, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_DivideFunction3D)) __PYX_ERR(12, 52, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "ModuloFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D) __PYX_ERR(12, 56, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunction3D)) __PYX_ERR(12, 56, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "PowFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D) __PYX_ERR(12, 60, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunction3D)) __PYX_ERR(12, 60, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "AbsFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D) __PYX_ERR(12, 64, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_AbsFunction3D)) __PYX_ERR(12, 64, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "EqualsFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D) __PYX_ERR(12, 68, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsFunction3D)) __PYX_ERR(12, 68, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "NotEqualsFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D) __PYX_ERR(12, 72, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsFunction3D)) __PYX_ERR(12, 72, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "LessThanFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D) __PYX_ERR(12, 76, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanFunction3D)) __PYX_ERR(12, 76, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "GreaterThanFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D) __PYX_ERR(12, 80, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanFunction3D)) __PYX_ERR(12, 80, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "LessEqualsFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D) __PYX_ERR(12, 84, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsFunction3D)) __PYX_ERR(12, 84, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "GreaterEqualsFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D) __PYX_ERR(12, 88, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsFunction3D)) __PYX_ERR(12, 88, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "AddScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D) __PYX_ERR(12, 92, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_AddScalar3D)) __PYX_ERR(12, 92, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "SubtractScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D) __PYX_ERR(12, 97, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_SubtractScalar3D)) __PYX_ERR(12, 97, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "MultiplyScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D) __PYX_ERR(12, 102, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_MultiplyScalar3D)) __PYX_ERR(12, 102, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "DivideScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D) __PYX_ERR(12, 107, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_DivideScalar3D)) __PYX_ERR(12, 107, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "ModuloScalarFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D) __PYX_ERR(12, 112, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloScalarFunction3D)) __PYX_ERR(12, 112, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "ModuloFunctionScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D) __PYX_ERR(12, 117, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_ModuloFunctionScalar3D)) __PYX_ERR(12, 117, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "PowScalarFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D) __PYX_ERR(12, 122, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_PowScalarFunction3D)) __PYX_ERR(12, 122, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "PowFunctionScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D) __PYX_ERR(12, 127, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_PowFunctionScalar3D)) __PYX_ERR(12, 127, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "EqualsScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D) __PYX_ERR(12, 132, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_EqualsScalar3D)) __PYX_ERR(12, 132, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "NotEqualsScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D) __PYX_ERR(12, 137, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_NotEqualsScalar3D)) __PYX_ERR(12, 137, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "LessThanScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D) __PYX_ERR(12, 142, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessThanScalar3D)) __PYX_ERR(12, 142, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "GreaterThanScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D) __PYX_ERR(12, 147, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterThanScalar3D)) __PYX_ERR(12, 147, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "LessEqualsScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D) __PYX_ERR(12, 152, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_LessEqualsScalar3D)) __PYX_ERR(12, 152, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.base", "GreaterEqualsScalar3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D) __PYX_ERR(12, 157, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_4base_GreaterEqualsScalar3D)) __PYX_ERR(12, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function3d.constant"); if (unlikely(!__pyx_t_1)) __PYX_ERR(53, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.constant", "Constant3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D) __PYX_ERR(53, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_8constant_Constant3D)) __PYX_ERR(53, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function3d.blend"); if (unlikely(!__pyx_t_1)) __PYX_ERR(54, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.blend", "Blend3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D) __PYX_ERR(54, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5blend_Blend3D)) __PYX_ERR(54, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function3d.autowrap"); if (unlikely(!__pyx_t_1)) __PYX_ERR(55, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.autowrap", "PythonFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D) __PYX_ERR(55, 34, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_8autowrap_PythonFunction3D)) __PYX_ERR(55, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.spatial.kdtree3d"); if (unlikely(!__pyx_t_1)) __PYX_ERR(56, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_Item3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.spatial.kdtree3d", "Item3D", sizeof(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_Item3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_Item3D) __PYX_ERR(56, 52, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.spatial.kdtree3d", "KDTree3DCore", sizeof(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore) __PYX_ERR(56, 59, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore = (struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3DCore)) __PYX_ERR(56, 59, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.spatial.kdtree3d", "KDTree3D", sizeof(struct __pyx_obj_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D) __PYX_ERR(56, 126, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D = (struct __pyx_vtabstruct_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_7spatial_8kdtree3d_KDTree3D)) __PYX_ERR(56, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function3d.interpolate.common"); if (unlikely(!__pyx_t_1)) __PYX_ERR(57, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.interpolate.common", "MeshKDTree3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D) __PYX_ERR(57, 37, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_6common_MeshKDTree3D)) __PYX_ERR(57, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function3d.interpolate.discrete3dmesh"); if (unlikely(!__pyx_t_1)) __PYX_ERR(58, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.interpolate.discrete3dmesh", "Discrete3DMesh", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh) __PYX_ERR(58, 37, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_14discrete3dmesh_Discrete3DMesh)) __PYX_ERR(58, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function3d.interpolate.interpolator3darray"); if (unlikely(!__pyx_t_1)) __PYX_ERR(59, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.interpolate.interpolator3darray", "Interpolator3DArray", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray) __PYX_ERR(59, 41, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray_Interpolator3DArray)) __PYX_ERR(59, 41, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.interpolate.interpolator3darray", "_Interpolator3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D) __PYX_ERR(59, 55, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3D)) __PYX_ERR(59, 55, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.interpolate.interpolator3darray", "_Interpolator3DLinear", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear) __PYX_ERR(59, 67, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DLinear)) __PYX_ERR(59, 67, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.interpolate.interpolator3darray", "_Interpolator3DCubic", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic) __PYX_ERR(59, 79, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Interpolator3DCubic)) __PYX_ERR(59, 79, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.interpolate.interpolator3darray", "_Extrapolator3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D) __PYX_ERR(59, 89, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3D)) __PYX_ERR(59, 89, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.interpolate.interpolator3darray", "_Extrapolator3DNone", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone) __PYX_ERR(59, 108, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DNone)) __PYX_ERR(59, 108, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.interpolate.interpolator3darray", "_Extrapolator3DLinear", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear) __PYX_ERR(59, 112, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__Extrapolator3DLinear)) __PYX_ERR(59, 112, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.interpolate.interpolator3darray", "_ArrayDerivative3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D) __PYX_ERR(59, 116, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_11interpolate_19interpolator3darray__ArrayDerivative3D)) __PYX_ERR(59, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function3d.arg"); if (unlikely(!__pyx_t_1)) __PYX_ERR(60, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.arg", "Arg3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D) __PYX_ERR(60, 37, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_3arg_Arg3D)) __PYX_ERR(60, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.float.function3d.cmath"); if (unlikely(!__pyx_t_1)) __PYX_ERR(61, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.cmath", "Exp3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D) __PYX_ERR(61, 35, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Exp3D)) __PYX_ERR(61, 35, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.cmath", "Sin3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D) __PYX_ERR(61, 39, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sin3D)) __PYX_ERR(61, 39, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.cmath", "Cos3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D) __PYX_ERR(61, 43, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Cos3D)) __PYX_ERR(61, 43, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.cmath", "Tan3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D) __PYX_ERR(61, 47, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Tan3D)) __PYX_ERR(61, 47, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.cmath", "Asin3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D) __PYX_ERR(61, 51, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Asin3D)) __PYX_ERR(61, 51, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.cmath", "Acos3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D) __PYX_ERR(61, 55, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Acos3D)) __PYX_ERR(61, 55, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.cmath", "Atan3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D) __PYX_ERR(61, 59, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan3D)) __PYX_ERR(61, 59, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.cmath", "Atan4Q3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D) __PYX_ERR(61, 63, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Atan4Q3D)) __PYX_ERR(61, 63, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.cmath", "Sqrt3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D) __PYX_ERR(61, 67, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Sqrt3D)) __PYX_ERR(61, 67, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.float.function3d.cmath", "Erf3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D) __PYX_ERR(61, 71, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_5float_10function3d_5cmath_Erf3D)) __PYX_ERR(61, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(62, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.base", "Vector3DFunction", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_4base_Vector3DFunction) __PYX_ERR(62, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function1d.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.base", "Function1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D) __PYX_ERR(13, 38, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D)) __PYX_ERR(13, 38, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.base", "AddFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D) __PYX_ERR(13, 42, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_AddFunction1D)) __PYX_ERR(13, 42, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.base", "SubtractFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D) __PYX_ERR(13, 46, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_SubtractFunction1D)) __PYX_ERR(13, 46, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.base", "MultiplyFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D) __PYX_ERR(13, 50, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_MultiplyFunction1D)) __PYX_ERR(13, 50, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.base", "DivideFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D) __PYX_ERR(13, 55, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_DivideFunction1D)) __PYX_ERR(13, 55, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.base", "NegFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D) __PYX_ERR(13, 60, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NegFunction1D)) __PYX_ERR(13, 60, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.base", "EqualsFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D) __PYX_ERR(13, 64, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_EqualsFunction1D)) __PYX_ERR(13, 64, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.base", "NotEqualsFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D) __PYX_ERR(13, 68, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_4base_NotEqualsFunction1D)) __PYX_ERR(13, 68, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function1d.constant"); if (unlikely(!__pyx_t_1)) __PYX_ERR(63, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.constant", "Constant1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D) __PYX_ERR(63, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_8constant_Constant1D)) __PYX_ERR(63, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function1d.blend"); if (unlikely(!__pyx_t_1)) __PYX_ERR(64, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.blend", "Blend1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D) __PYX_ERR(64, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_5blend_Blend1D)) __PYX_ERR(64, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function1d.autowrap"); if (unlikely(!__pyx_t_1)) __PYX_ERR(65, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.autowrap", "PythonFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D) __PYX_ERR(65, 34, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_PythonFunction1D)) __PYX_ERR(65, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function1d.utility"); if (unlikely(!__pyx_t_1)) __PYX_ERR(66, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function1d.utility", "FloatToVector3DFunction1D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D) __PYX_ERR(66, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function1d_7utility_FloatToVector3DFunction1D)) __PYX_ERR(66, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function2d.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function2d.base", "AddFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D) __PYX_ERR(14, 42, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_AddFunction2D)) __PYX_ERR(14, 42, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function2d.base", "SubtractFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D) __PYX_ERR(14, 46, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_SubtractFunction2D)) __PYX_ERR(14, 46, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function2d.base", "MultiplyFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D) __PYX_ERR(14, 50, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_MultiplyFunction2D)) __PYX_ERR(14, 50, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function2d.base", "DivideFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D) __PYX_ERR(14, 55, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_DivideFunction2D)) __PYX_ERR(14, 55, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function2d.base", "NegFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D) __PYX_ERR(14, 60, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NegFunction2D)) __PYX_ERR(14, 60, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function2d.base", "EqualsFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D) __PYX_ERR(14, 64, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_EqualsFunction2D)) __PYX_ERR(14, 64, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function2d.base", "NotEqualsFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D) __PYX_ERR(14, 68, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_4base_NotEqualsFunction2D)) __PYX_ERR(14, 68, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function2d.constant"); if (unlikely(!__pyx_t_1)) __PYX_ERR(67, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function2d.constant", "Constant2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D) __PYX_ERR(67, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_8constant_Constant2D)) __PYX_ERR(67, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function2d.blend"); if (unlikely(!__pyx_t_1)) __PYX_ERR(68, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function2d.blend", "Blend2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D) __PYX_ERR(68, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_5blend_Blend2D)) __PYX_ERR(68, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function2d.autowrap"); if (unlikely(!__pyx_t_1)) __PYX_ERR(69, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function2d.autowrap", "PythonFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D) __PYX_ERR(69, 34, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_PythonFunction2D)) __PYX_ERR(69, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function2d.utility"); if (unlikely(!__pyx_t_1)) __PYX_ERR(70, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function2d.utility", "FloatToVector3DFunction2D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D) __PYX_ERR(70, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function2d_7utility_FloatToVector3DFunction2D)) __PYX_ERR(70, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function3d.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.base", "Function3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D) __PYX_ERR(15, 38, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D)) __PYX_ERR(15, 38, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.base", "AddFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D) __PYX_ERR(15, 42, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_AddFunction3D)) __PYX_ERR(15, 42, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.base", "SubtractFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D) __PYX_ERR(15, 46, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_SubtractFunction3D)) __PYX_ERR(15, 46, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.base", "MultiplyFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D) __PYX_ERR(15, 50, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_MultiplyFunction3D)) __PYX_ERR(15, 50, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.base", "DivideFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D) __PYX_ERR(15, 55, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_DivideFunction3D)) __PYX_ERR(15, 55, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.base", "NegFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D) __PYX_ERR(15, 60, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NegFunction3D)) __PYX_ERR(15, 60, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.base", "EqualsFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D) __PYX_ERR(15, 64, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_EqualsFunction3D)) __PYX_ERR(15, 64, __pyx_L1_error)
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.base", "NotEqualsFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D) __PYX_ERR(15, 68, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_4base_NotEqualsFunction3D)) __PYX_ERR(15, 68, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function3d.constant"); if (unlikely(!__pyx_t_1)) __PYX_ERR(71, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.constant", "Constant3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D) __PYX_ERR(71, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_8constant_Constant3D)) __PYX_ERR(71, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function3d.blend"); if (unlikely(!__pyx_t_1)) __PYX_ERR(72, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.blend", "Blend3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D) __PYX_ERR(72, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_5blend_Blend3D)) __PYX_ERR(72, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function3d.autowrap"); if (unlikely(!__pyx_t_1)) __PYX_ERR(73, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.autowrap", "PythonFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D) __PYX_ERR(73, 34, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_PythonFunction3D)) __PYX_ERR(73, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.function.vector3d.function3d.utility"); if (unlikely(!__pyx_t_1)) __PYX_ERR(74, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D = __Pyx_ImportType(__pyx_t_1, "raysect.core.math.function.vector3d.function3d.utility", "FloatToVector3DFunction3D", sizeof(struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D) __PYX_ERR(74, 36, __pyx_L1_error)
  __pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D = (struct __pyx_vtabstruct_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D*)__Pyx_GetVtable(__pyx_ptype_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D); if (unlikely(!__pyx_vtabptr_7raysect_4core_4math_8function_8vector3d_10function3d_7utility_FloatToVector3DFunction3D)) __PYX_ERR(74, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("raysect.core.math.quaternion"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "new_quaternion_from_matrix", (void (**)(void))&__pyx_f_7raysect_4core_4math_10quaternion_new_quaternion_from_matrix, "struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "new_quaternion_from_axis_angle", (void (**)(void))&__pyx_f_7raysect_4core_4math_10quaternion_new_quaternion_from_axis_angle, "struct __pyx_obj_7raysect_4core_4math_10quaternion_Quaternion *(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = PyImport_ImportModule("raysect.core.math.transform"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "translate", (void (**)(void))&__pyx_f_7raysect_4core_4math_9transform_translate, "struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(double, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "rotate_x", (void (**)(void))&__pyx_f_7raysect_4core_4math_9transform_rotate_x, "struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "rotate_y", (void (**)(void))&__pyx_f_7raysect_4core_4math_9transform_rotate_y, "struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "rotate_z", (void (**)(void))&__pyx_f_7raysect_4core_4math_9transform_rotate_z, "struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "rotate_vector", (void (**)(void))&__pyx_f_7raysect_4core_4math_9transform_rotate_vector, "struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(double, struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "rotate", (void (**)(void))&__pyx_f_7raysect_4core_4math_9transform_rotate, "struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(double, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "rotate_basis", (void (**)(void))&__pyx_f_7raysect_4core_4math_9transform_rotate_basis, "struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *(struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, struct __pyx_obj_7raysect_4core_4math_6vector_Vector3D *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "to_cylindrical", (void (**)(void))&__pyx_f_7raysect_4core_4math_9transform_to_cylindrical, "PyObject *(struct __pyx_obj_7raysect_4core_4math_5point_Point3D *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "from_cylindrical", (void (**)(void))&__pyx_f_7raysect_4core_4math_9transform_from_cylindrical, "struct __pyx_obj_7raysect_4core_4math_5point_Point3D *(double, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "extract_rotation", (void (**)(void))&__pyx_f_7raysect_4core_4math_9transform_extract_rotation, "PyObject *(struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, int __pyx_skip_dispatch, struct __pyx_opt_args_7raysect_4core_4math_9transform_extract_rotation *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "extract_translation", (void (**)(void))&__pyx_f_7raysect_4core_4math_9transform_extract_translation, "PyObject *(struct __pyx_obj_7raysect_4core_4math_12affinematrix_AffineMatrix3D *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = PyImport_ImportModule("raysect.core.math.random"); if (!__pyx_t_3) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "uniform", (void (**)(void))&__pyx_f_7raysect_4core_4math_6random_uniform, "double (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = PyImport_ImportModule("raysect.core.math.cython.triangle"); if (!__pyx_t_4) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_4, "barycentric_coords", (void (**)(void))&__pyx_f_7raysect_4core_4math_6cython_8triangle_barycentric_coords, "void (double, double, double, double, double, double, double, double, double *, double *, double *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_4, "barycentric_interpolation", (void (**)(void))&__pyx_f_7raysect_4core_4math_6cython_8triangle_barycentric_interpolation, "double (double, double, double, double, double, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = PyImport_ImportModule("raysect.core.math.polygon"); if (!__pyx_t_5) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_5, "triangulate2d", (void (**)(void))&__pyx_f_7raysect_4core_4math_7polygon_triangulate2d, "PyArrayObject *(PyArrayObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = PyImport_ImportModule("raysect.core.math.function.float.function1d.autowrap"); if (!__pyx_t_6) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_6, "autowrap_function1d", (void (**)(void))&__pyx_f_7raysect_4core_4math_8function_5float_10function1d_8autowrap_autowrap_function1d, "struct __pyx_obj_7raysect_4core_4math_8function_5float_10function1d_4base_Function1D *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_7 = PyImport_ImportModule("raysect.core.math.function.float.function2d.autowrap"); if (!__pyx_t_7) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_7, "autowrap_function2d", (void (**)(void))&__pyx_f_7raysect_4core_4math_8function_5float_10function2d_8autowrap_autowrap_function2d, "struct __pyx_obj_7raysect_4core_4math_8function_5float_10function2d_4base_Function2D *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_8 = PyImport_ImportModule("raysect.core.math.function.float.function3d.autowrap"); if (!__pyx_t_8) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_8, "autowrap_function3d", (void (**)(void))&__pyx_f_7raysect_4core_4math_8function_5float_10function3d_8autowrap_autowrap_function3d, "struct __pyx_obj_7raysect_4core_4math_8function_5float_10function3d_4base_Function3D *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_9 = PyImport_ImportModule("raysect.core.math.function.vector3d.function1d.autowrap"); if (!__pyx_t_9) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_9, "autowrap_function1d", (void (**)(void))&__pyx_f_7raysect_4core_4math_8function_8vector3d_10function1d_8autowrap_autowrap_function1d, "struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function1d_4base_Function1D *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_10 = PyImport_ImportModule("raysect.core.math.function.vector3d.function2d.autowrap"); if (!__pyx_t_10) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_10, "autowrap_function2d", (void (**)(void))&__pyx_f_7raysect_4core_4math_8function_8vector3d_10function2d_8autowrap_autowrap_function2d, "struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function2d_4base_Function2D *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_11 = PyImport_ImportModule("raysect.core.math.function.vector3d.function3d.autowrap"); if (!__pyx_t_11) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_11, "autowrap_function3d", (void (**)(void))&__pyx_f_7raysect_4core_4math_8function_8vector3d_10function3d_8autowrap_autowrap_function3d, "struct __pyx_obj_7raysect_4core_4math_8function_8vector3d_10function3d_4base_Function3D *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_solps_2d_functions(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_solps_2d_functions},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "solps_2d_functions",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_COMPILING_IN_LIMITED_API
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_COMPILING_IN_LIMITED_API
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initsolps_2d_functions(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initsolps_2d_functions(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_solps_2d_functions(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_solps_2d_functions(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none) {
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
#endif
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


  static CYTHON_SMALL_CODE int __pyx_pymod_exec_solps_2d_functions(PyObject *__pyx_pyinit_module)
  #endif
  #endif
  {
    PyObject *__pyx_t_1 = NULL;
    PyObject *__pyx_t_2 = NULL;
    PyObject *__pyx_t_3 = NULL;
    static PyThread_type_lock __pyx_t_4[8];
    int __pyx_lineno = 0;
    const char *__pyx_filename = NULL;
    int __pyx_clineno = 0;
    __Pyx_RefNannyDeclarations
    #if CYTHON_PEP489_MULTI_PHASE_INIT
    if (__pyx_m) {
      if (__pyx_m == __pyx_pyinit_module) return 0;
      PyErr_SetString(PyExc_RuntimeError, "Module 'solps_2d_functions' has already been imported. Re-initialisation is not supported.");
      return -1;
    }
    #elif PY_MAJOR_VERSION >= 3
    if (__pyx_m) return __Pyx_NewRef(__pyx_m);
    #endif
    /*--- Module creation code ---*/
    #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_m = __pyx_pyinit_module;
    Py_INCREF(__pyx_m);
    #else
    #if PY_MAJOR_VERSION < 3
    __pyx_m = Py_InitModule4("solps_2d_functions", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
    if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
    #elif CYTHON_COMPILING_IN_LIMITED_API
    __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyState_AddModule(__pyx_t_1, &__pyx_moduledef) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    #else
    __pyx_m = PyModule_Create(&__pyx_moduledef);
    if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    #endif
    #if !CYTHON_COMPILING_IN_LIMITED_API
    __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
    Py_INCREF(__pyx_d);
    #endif
    __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
    Py_INCREF(__pyx_b);
    __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
    Py_INCREF(__pyx_cython_runtime);
    if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
    #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
    __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_solps_2d_functions(void)", 0);
    if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    #ifdef __Pxy_PyFrame_Initialize_Offsets
    __Pxy_PyFrame_Initialize_Offsets();
    #endif
    __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
    __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
    __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
    #ifdef __Pyx_CyFunction_USED
    if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    #ifdef __Pyx_FusedFunction_USED
    if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    #ifdef __Pyx_Generator_USED
    if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    #ifdef __Pyx_StopAsyncIteration_USED
    if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    /*--- Library function declarations ---*/
    /*--- Threads initialization code ---*/
    #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
    PyEval_InitThreads();
    #endif
    /*--- Initialize various global constants etc. ---*/
    if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    if (__pyx_module_is_main_cherab__solps__solps_2d_functions) {
      if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    }
    #if PY_MAJOR_VERSION >= 3
    {
      PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
      if (!PyDict_GetItemString(modules, "cherab.solps.solps_2d_functions")) {
        if (unlikely(PyDict_SetItemString(modules, "cherab.solps.solps_2d_functions", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #endif
    /*--- Builtin init code ---*/
    if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    /*--- Constants init code ---*/
    if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    /*--- Global type/function init code ---*/
    (void)__Pyx_modinit_global_init_code();
    (void)__Pyx_modinit_variable_export_code();
    (void)__Pyx_modinit_function_export_code();
    if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    (void)__Pyx_modinit_variable_import_code();
    if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    /*--- Execution code ---*/
    #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif

    /* "cherab/solps/solps_2d_functions.pyx":21
 * # under the Licence.
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * 
 */
    __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_numpy, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    {
      #if CYTHON_COMPILING_IN_LIMITED_API
      PyObject *__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 21, __pyx_L1_error)
      #endif
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "cherab/solps/solps_2d_functions.pyx":54
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self._grid_data, self._triangle_to_grid_map, self._kdtree
 * 
 */
    #if CYTHON_COMPILING_IN_LIMITED_API
    __pyx_t_2 = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
    #else
    __pyx_t_2 = __pyx_d;  if ((1)); else __PYX_ERR(0, 54, __pyx_L1_error);
    #endif
    __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_3__getstate__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SOLPSFunction2D___getstate, NULL, __pyx_n_s_cherab_solps_solps_2d_functions_2, __pyx_t_2, ((PyObject *)__pyx_codeobj__11)); __pyx_t_2 = NULL; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem((PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D->tp_dict, __pyx_n_s_getstate, __pyx_t_1) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    PyType_Modified(__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D);

    /* "cherab/solps/solps_2d_functions.pyx":57
 *         return self._grid_data, self._triangle_to_grid_map, self._kdtree
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._grid_data, self._triangle_to_grid_map, self._kdtree = state
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 */
    #if CYTHON_COMPILING_IN_LIMITED_API
    __pyx_t_2 = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 57, __pyx_L1_error)
    #else
    __pyx_t_2 = __pyx_d;  if ((1)); else __PYX_ERR(0, 57, __pyx_L1_error);
    #endif
    __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_5__setstate__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SOLPSFunction2D___setstate, NULL, __pyx_n_s_cherab_solps_solps_2d_functions_2, __pyx_t_2, ((PyObject *)__pyx_codeobj__13)); __pyx_t_2 = NULL; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem((PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D->tp_dict, __pyx_n_s_setstate, __pyx_t_1) < 0) __PYX_ERR(0, 57, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    PyType_Modified(__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D);

    /* "cherab/solps/solps_2d_functions.pyx":62
 *         self._grid_data_mv = self._grid_data
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 */
    #if CYTHON_COMPILING_IN_LIMITED_API
    __pyx_t_2 = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 62, __pyx_L1_error)
    #else
    __pyx_t_2 = __pyx_d;  if ((1)); else __PYX_ERR(0, 62, __pyx_L1_error);
    #endif
    __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_7__reduce__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SOLPSFunction2D___reduce, NULL, __pyx_n_s_cherab_solps_solps_2d_functions_2, __pyx_t_2, ((PyObject *)__pyx_codeobj__15)); __pyx_t_2 = NULL; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem((PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D->tp_dict, __pyx_n_s_reduce, __pyx_t_1) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    PyType_Modified(__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D);

    /* "cherab/solps/solps_2d_functions.pyx":65
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def instance(cls, SOLPSFunction2D instance not None, object grid_data=None):
 *         """
 */
    #if CYTHON_COMPILING_IN_LIMITED_API
    __pyx_t_2 = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
    #else
    __pyx_t_2 = __pyx_d;  if ((1)); else __PYX_ERR(0, 65, __pyx_L1_error);
    #endif
    __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_6cherab_5solps_18solps_2d_functions_15SOLPSFunction2D_9instance, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SOLPSFunction2D_instance, NULL, __pyx_n_s_cherab_solps_solps_2d_functions_2, __pyx_t_2, ((PyObject *)__pyx_codeobj__17)); __pyx_t_2 = NULL; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem((PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D->tp_dict, __pyx_n_s_instance, __pyx_t_1) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    PyType_Modified(__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D);
    __Pyx_GetNameInClass(__pyx_t_1, (PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D, __pyx_n_s_instance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_Method_ClassMethod(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (PyDict_SetItem((PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D->tp_dict, __pyx_n_s_instance, __pyx_t_3) < 0) __PYX_ERR(0, 65, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    PyType_Modified(__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSFunction2D);

    /* "cherab/solps/solps_2d_functions.pyx":138
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self._grid_vectors, self._triangle_to_grid_map, self._kdtree
 * 
 */
    #if CYTHON_COMPILING_IN_LIMITED_API
    __pyx_t_2 = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
    #else
    __pyx_t_2 = __pyx_d;  if ((1)); else __PYX_ERR(0, 138, __pyx_L1_error);
    #endif
    __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_3__getstate__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SOLPSVectorFunction2D___getstate, NULL, __pyx_n_s_cherab_solps_solps_2d_functions_2, __pyx_t_2, ((PyObject *)__pyx_codeobj__19)); __pyx_t_2 = NULL; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem((PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D->tp_dict, __pyx_n_s_getstate, __pyx_t_3) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    PyType_Modified(__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D);

    /* "cherab/solps/solps_2d_functions.pyx":141
 *         return self._grid_vectors, self._triangle_to_grid_map, self._kdtree
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self._grid_vectors, self._triangle_to_grid_map, self._kdtree = state
 *         self._grid_vectors_mv = self._grid_vectors
 */
    #if CYTHON_COMPILING_IN_LIMITED_API
    __pyx_t_2 = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
    #else
    __pyx_t_2 = __pyx_d;  if ((1)); else __PYX_ERR(0, 141, __pyx_L1_error);
    #endif
    __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_5__setstate__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SOLPSVectorFunction2D___setstate, NULL, __pyx_n_s_cherab_solps_solps_2d_functions_2, __pyx_t_2, ((PyObject *)__pyx_codeobj__21)); __pyx_t_2 = NULL; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem((PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D->tp_dict, __pyx_n_s_setstate, __pyx_t_3) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    PyType_Modified(__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D);

    /* "cherab/solps/solps_2d_functions.pyx":146
 *         self._triangle_to_grid_map_mv = self._triangle_to_grid_map
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 */
    #if CYTHON_COMPILING_IN_LIMITED_API
    __pyx_t_2 = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
    #else
    __pyx_t_2 = __pyx_d;  if ((1)); else __PYX_ERR(0, 146, __pyx_L1_error);
    #endif
    __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_7__reduce__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SOLPSVectorFunction2D___reduce, NULL, __pyx_n_s_cherab_solps_solps_2d_functions_2, __pyx_t_2, ((PyObject *)__pyx_codeobj__23)); __pyx_t_2 = NULL; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem((PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D->tp_dict, __pyx_n_s_reduce, __pyx_t_3) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    PyType_Modified(__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D);

    /* "cherab/solps/solps_2d_functions.pyx":149
 *         return self.__new__, (self.__class__, ), self.__getstate__()
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def instance(cls, SOLPSVectorFunction2D instance not None, object grid_vectors=None):
 *         """
 */
    #if CYTHON_COMPILING_IN_LIMITED_API
    __pyx_t_2 = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
    #else
    __pyx_t_2 = __pyx_d;  if ((1)); else __PYX_ERR(0, 149, __pyx_L1_error);
    #endif
    __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6cherab_5solps_18solps_2d_functions_21SOLPSVectorFunction2D_9instance, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SOLPSVectorFunction2D_instance, NULL, __pyx_n_s_cherab_solps_solps_2d_functions_2, __pyx_t_2, ((PyObject *)__pyx_codeobj__25)); __pyx_t_2 = NULL; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem((PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D->tp_dict, __pyx_n_s_instance, __pyx_t_3) < 0) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    PyType_Modified(__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D);
    __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D, __pyx_n_s_instance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem((PyObject *)__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D->tp_dict, __pyx_n_s_instance, __pyx_t_1) < 0) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    PyType_Modified(__pyx_ptype_6cherab_5solps_18solps_2d_functions_SOLPSVectorFunction2D);

    /* "cherab/solps/solps_2d_functions.pyx":1
 * # cython: language_level=3             # <<<<<<<<<<<<<<
 * 
 * # Copyright 2016-2018 Euratom
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    {
      #if CYTHON_COMPILING_IN_LIMITED_API
      PyObject *__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
      #endif
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "View.MemoryView":202
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
    __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(16, 202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    PyType_Modified(__pyx_array_type);

    /* "View.MemoryView":297
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XGOTREF(generic);
    __Pyx_DECREF_SET(generic, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "View.MemoryView":298
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XGOTREF(strided);
    __Pyx_DECREF_SET(strided, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "View.MemoryView":299
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XGOTREF(indirect);
    __Pyx_DECREF_SET(indirect, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "View.MemoryView":302
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XGOTREF(contiguous);
    __Pyx_DECREF_SET(contiguous, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "View.MemoryView":303
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XGOTREF(indirect_contiguous);
    __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "View.MemoryView":327
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
    __pyx_memoryview_thread_locks_used = 0;

    /* "View.MemoryView":328
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
    __pyx_t_4[0] = PyThread_allocate_lock();
    __pyx_t_4[1] = PyThread_allocate_lock();
    __pyx_t_4[2] = PyThread_allocate_lock();
    __pyx_t_4[3] = PyThread_allocate_lock();
    __pyx_t_4[4] = PyThread_allocate_lock();
    __pyx_t_4[5] = PyThread_allocate_lock();
    __pyx_t_4[6] = PyThread_allocate_lock();
    __pyx_t_4[7] = PyThread_allocate_lock();
    memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_4, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

    /* "View.MemoryView":560
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(16, 560, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    PyType_Modified(__pyx_memoryview_type);

    /* "View.MemoryView":1004
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_1 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 1004, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_1) < 0) __PYX_ERR(16, 1004, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    PyType_Modified(__pyx_memoryviewslice_type);

    /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
    __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    {
      #if CYTHON_COMPILING_IN_LIMITED_API
      PyObject *__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(16, 1, __pyx_L1_error)
      #endif
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_1) < 0) __PYX_ERR(16, 1, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

    /*--- Wrapped vars code ---*/

    goto __pyx_L0;
    __pyx_L1_error:;
    __Pyx_XDECREF(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_3);
    if (__pyx_m) {
      #if !CYTHON_COMPILING_IN_LIMITED_API
      if (__pyx_d) {
        #endif
        __Pyx_AddTraceback("init cherab.solps.solps_2d_functions", __pyx_clineno, __pyx_lineno, __pyx_filename);
        #if !CYTHON_COMPILING_IN_LIMITED_API
      }
      Py_CLEAR(__pyx_m);
      #endif
    } else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_ImportError, "init cherab.solps.solps_2d_functions");
    }
    __pyx_L0:;
    __Pyx_RefNannyFinishContext();
    #if CYTHON_PEP489_MULTI_PHASE_INIT
    return (__pyx_m != NULL) ? 0 : -1;
    #elif PY_MAJOR_VERSION >= 3
    return __pyx_m;
    #else
    return;
    #endif
  }
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;  // error
            return kwvalues[i];
        }
    }
    return NULL;  // not found (no exception set)
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        if (kwds_is_tuple) {
            if (pos >= PyTuple_GET_SIZE(kwds)) break;
            key = PyTuple_GET_ITEM(kwds, pos);
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result;
    Py_ssize_t i;
    argstuple = PyTuple_New(nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        PyTuple_SET_ITEM(argstuple, i, args[i]);
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
    Py_DECREF(argstuple);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
#ifdef __Pyx_CyFunction_USED
        if (PyCFunction_Check(func) || __Pyx_CyFunction_Check(func))
#else
        if (PyCFunction_Check(func))
#endif
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
                return __Pyx_PyObject_CallMethO(func, NULL);
            }
        }
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (PyCFunction_Check(func))
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
                return __Pyx_PyObject_CallMethO(func, args[0]);
            }
        }
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    #if CYTHON_VECTORCALL
    vectorcallfunc f = _PyVectorcall_Function(func);
    if (f) {
        return f(func, args, nargs, kwargs);
    }
    #elif __Pyx_CyFunction_USED && CYTHON_BACKPORT_VECTORCALL
    if (__Pyx_IS_TYPE(func, __pyx_CyFunctionType)) {
        __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
        if (f) return f(func, args, nargs, kwargs);
    }
    #endif
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    return __Pyx_PyObject_FastCall_fallback(func, args, nargs, kwargs);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY ||  CYTHON_COMPILING_IN_LIMITED_API
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define GCC_DIAGNOSTIC
#endif
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_8)[digit_pos];
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_10)[digit_pos];
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* CIntToPyUnicode */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define GCC_DIAGNOSTIC
#endif
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_8)[digit_pos];
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_10)[digit_pos];
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (meth) {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* None */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t a, Py_ssize_t b) {
    Py_ssize_t q = a / b;
    Py_ssize_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
#endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                #if CYTHON_COMPILING_IN_LIMITED_API
                module = PyImport_ImportModuleLevelObject(
                    name, empty_dict, empty_dict, from_list, 1);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                #endif
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            #if CYTHON_COMPILING_IN_LIMITED_API
            module = PyImport_ImportModuleLevelObject(
                name, empty_dict, empty_dict, from_list, level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        PyObject *sep;
        PyObject *slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, CYTHON_UNUSED PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__5;
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    Py_ssize_t i, nparts;
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (likely(module))
        return module;
    return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* ssize_strlen */
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_ass_subscript) {
            int r;
            PyObject *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return -1;
            r = mm->mp_ass_subscript(o, key, v);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return sm->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* None */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* CallNextTpDealloc */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_dealloc != current_tp_dealloc)
        type = type->tp_base;
    while (type && type->tp_dealloc == current_tp_dealloc)
        type = type->tp_base;
    if (type)
        type->tp_dealloc(obj);
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* SetVTable */
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_SetVtable(PyObject *type, void *vtable) {
#else
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#endif
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (!ob)
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (PyObject_SetAttr(type, __pyx_n_s_pyx_vtable, ob) < 0)
#else
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", __Pyx_PyType_Name(type_obj));
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* FetchCommonType */
  static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    PyObject *abi_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!abi_module) return NULL;
    Py_INCREF(abi_module);
    return abi_module;
}
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, type->tp_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              type->tp_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, type->tp_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *py_basicsize, *cached_type = NULL;
    Py_ssize_t basicsize;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, spec->name);
    if (cached_type) {
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (!py_basicsize) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred()) goto bad;
        if (__Pyx_VerifyCachedType(
              cached_type,
              spec->name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    cached_type = PyType_FromSpecWithBases(spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (PyObject_SetAttrString(abi_module, spec->name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
  #if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, Py_ssize_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (i = 0; i < nargs; i++) newargs[i] = args[i];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, Py_ssize_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
  #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if PY_VERSION_HEX < 0x030500A0
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), PyTuple_GET_SIZE(args), kw);
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         cyfunc->func.m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", cyfunc->func.m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = cyfunc->func.m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = cyfunc->func.m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = cyfunc->func.m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = cyfunc->func.m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = cyfunc->func.m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = cyfunc->func.m_self;
        break;
    default:
        return NULL;
    }
    return ((_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(__pyx_CyFunctionObject, func.vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#ifdef _Py_TPFLAGS_HAVE_VECTORCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(void) {
#if CYTHON_COMPILING_IN_LIMITED_API
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(&__pyx_CyFunctionType_spec, NULL);
#else
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
  static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* ClassMethod */
  static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYSTON || CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    static PyTypeObject *methoddescr_type = NULL;
    if (unlikely(methoddescr_type == NULL)) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (unlikely(!meth)) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else if (PyCFunction_Check(method)) {
        return PyClassMethod_New(method);
    }
#ifdef __Pyx_CyFunction_USED
    else if (__Pyx_CyFunction_Check(method)) {
        return PyClassMethod_New(method);
    }
#endif
    PyErr_SetString(PyExc_TypeError,
                   "Class-level classmethod() can only be called on "
                   "a method_descriptor or instance method.");
    return NULL;
}

/* GetNameInClass */
  static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    PyObject *dict;
    assert(PyType_Check(nmspace));
#if CYTHON_USE_TYPE_SLOTS
    dict = ((PyTypeObject*)nmspace)->tp_dict;
    Py_XINCREF(dict);
#else
    dict = PyObject_GetAttr(nmspace, __pyx_n_s_dict);
#endif
    if (likely(dict)) {
        result = PyObject_GetItem(dict, name);
        Py_DECREF(dict);
        if (result) {
            return result;
        }
    }
    PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if CYTHON_COMPILING_IN_LIMITED_API
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    if (c_line) {
        (void) __pyx_cfilenm;
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    _PyTraceback_Add(funcname, filename, c_line ? -c_line : py_line);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* MemviewSliceIsContig */
  static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
  static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
  static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, const char *sig)
{
    return PyCapsule_New(p, sig, NULL);
}

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* MemviewSliceCopyTemplate */
  static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* MemviewSliceInit */
  static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#ifdef HAVE_STDARG_PROTOTYPES
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    __pyx_atomic_int_type old_acquisition_count;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        return;
    }
    old_acquisition_count = __pyx_add_acquisition_count(memview);
    if (unlikely(old_acquisition_count <= 0)) {
        if (likely(old_acquisition_count == 0)) {
            if (have_gil) {
                Py_INCREF((PyObject *) memview);
            } else {
                PyGILState_STATE _gilstate = PyGILState_Ensure();
                Py_INCREF((PyObject *) memview);
                PyGILState_Release(_gilstate);
            }
        } else {
            __pyx_fatalerror("Acquisition count is %d (line %d)",
                             __pyx_get_slice_count(memview), lineno);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    __pyx_atomic_int_type old_acquisition_count;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    old_acquisition_count = __pyx_sub_acquisition_count(memview);
    memslice->data = NULL;
    if (likely(old_acquisition_count > 1)) {
        memslice->memview = NULL;
    } else if (likely(old_acquisition_count == 1)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    }
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
    const char neg_one = (char) ((char) 0 - (char) 1), const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* IsLittleEndian */
  static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
  static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* TypeInfoCompare */
    static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
    static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (unlikely(buf->strides[dim] != sizeof(void *))) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (unlikely(buf->strides[dim] != buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (unlikely(stride < buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (unlikely(spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(spec & (__Pyx_MEMVIEW_PTR))) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(buf->suboffsets)) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, CYTHON_UNUSED int ndim, int spec)
{
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (unlikely(buf->suboffsets && buf->suboffsets[dim] >= 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (unlikely(!buf->suboffsets || (buf->suboffsets[dim] < 0))) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (unlikely(buf->ndim != ndim)) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (unlikely(!__Pyx_BufFmt_CheckString(&ctx, buf->format))) goto fail;
    }
    if (unlikely((unsigned) buf->itemsize != dtype->size)) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    for (i = 0; i < ndim; i++) {
        spec = axes_specs[i];
        if (unlikely(!__pyx_check_strides(buf, i, ndim, spec)))
            goto fail;
        if (unlikely(!__pyx_check_suboffsets(buf, i, ndim, spec)))
            goto fail;
    }
    if (unlikely(buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag)))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
    static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
    static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_5numpy_int32_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
    static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 3,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CheckBinaryVersion */
    static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
    #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
    #if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}
#endif

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                __Pyx_PyType_Name(Py_TYPE(result)))) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, __Pyx_PyType_Name(Py_TYPE(result)));
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
